/*
 * SipO bootmem misc device
 *
 * Copyright 2001 SSH Communications Security Corp, All Rights Reserved.
 *                http://www.ssh.com
 *
 * 14.02.2001 Arto Vuori <avuori@ssh.com>
 *            Initial version
 *
 * Misc Character device interface for Bootmem access. Bootmem area is located
 * above end of all available memory at (void *)(__bd->memsize).
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/reboot.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include <asm/sipo.h>

#include <asm/immap_8260.h>
#include <asm/cpm_8260.h>
#include <asm/machdep.h>
#include <asm/mpc8260.h>

#define FLASH_BLOCKSIZE (128 * 1024)

static uint sipo_bootmem_size;
static void *sipo_bootmem_base;

static int bootmem_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int bootmem_release(struct inode *inode, struct file *file)
{
	return 0;
}

static ssize_t bootmem_write(struct file *file, const char *buf, size_t count,
			     loff_t *ppos)
{
	if (count < 0)
		return -EIO;

	if (*ppos + count > sipo_bootmem_size || *ppos < 0)
		return -EIO;

	if (copy_from_user((void *) (sipo_bootmem_base + *ppos), buf, count))
		return -EFAULT;

	*ppos += count;
	return count;
}

static ssize_t bootmem_read(struct file *file, char *buf, size_t count,
			    loff_t *ppos)
{
	if (*ppos > sipo_bootmem_size || *ppos < 0)
		return 0;

	if (count > sipo_bootmem_size - *ppos)
		count = sipo_bootmem_size - *ppos;

	if (copy_to_user(buf, (void *) (sipo_bootmem_base + *ppos), count))
		return -EFAULT;

	*ppos += count;
	return count;
}

static loff_t bootmem_lseek(struct file *file, loff_t offset, int orig)
{
	switch (orig) {
	case 0:
		file->f_pos = offset;
		return file->f_pos;
	case 1:
		file->f_pos += offset;
		return file->f_pos;
	case 2:
		file->f_pos = sipo_bootmem_size - offset;
		return file->f_pos;
	default:
		return -EINVAL;
	}
}

static int bootmem_ioctl(struct inode *node, struct file *file,
			 unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case IOCTL_FLASH_BLOCK_ERASE:
		/* File position must be inside bootmem and located at block
		 * boundary */
		if (file->f_pos >= sipo_bootmem_size ||
		    (file->f_pos & (FLASH_BLOCKSIZE - 1)))
			return -EIO;

		memset(sipo_bootmem_base + file->f_pos, 0xff, FLASH_BLOCKSIZE);

		return 0;
	default:
		return -EINVAL;
	}
}

static struct file_operations bootmem_fops=
{
	owner:          THIS_MODULE,
	llseek:         bootmem_lseek,
	read:           bootmem_read,
	write:          bootmem_write,
	open:           bootmem_open,
	ioctl:          bootmem_ioctl,
	release:        bootmem_release,
};

static struct miscdevice bootmem_miscdev=
{
	SIPOBOOTMEM_MINOR,
	"sipobootmem",
	&bootmem_fops
};

void __init sipobootmem_init(void)
{
	bd_t *__bd = (bd_t *) __res;

	misc_register(&bootmem_miscdev);
	printk("SIpo BootMem: 0.1\n");

	/* ReMap flash memory address range */
	sipo_bootmem_size = __bd->bi_memtotal - __bd->bi_memsize;
	sipo_bootmem_base = (void *) ioremap(__bd->bi_memsize,
					     sipo_bootmem_size);
	if (!sipo_bootmem_base) {
		printk("SIpo bootmem: Failed to ioremap\n");
		sipo_bootmem_size = 0;
		return;
	}
}
