/* Routines to read/write data to the Dallas RTC chip
 * Copyright 2000-2001 SSH Communications Security Corp
 * Marko Teiste <mte@ssh.com>
 *
 * 12-2-2001 Arto Vuori <avuori@ssh.com>
 * - nvram misc device interface
 * - /proc/sipo/bootstate interface
 *
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/rtc.h>
#include <linux/miscdevice.h>
#include <linux/ctype.h>
#include <linux/proc_fs.h>

#include <asm/immap_8260.h>
#include <asm/pgtable.h>
#include <asm/mpc8260.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>
#include <asm/cpm_8260.h>

/*
 *  OK. Where are trying to emulate the SPI by wiggling the pins directly
 *  Why? Because this code is only used to set/get data from rtc and going
 *  thru the interrupt hassle does not seem to be worth the trouble.
 *
 *  The thing is wired to port D pin
 *
 *      16      SPIMOSI (output to RTC)
 *      17      SPIMOSO (input from RTC)
 *      18      SPICLK
 *      19      SPISEL
 */
#define SIPO_RTC_OUT (0x80000000 >> 16)
#define SIPO_RTC_IN  (0x80000000 >> 17)
#define SIPO_RTC_CLK (0x80000000 >> 18)
#define SIPO_RTC_SEL (0x80000000 >> 19)
#define SIPO_RTC_ALL (SIPO_RTC_IN | SIPO_RTC_OUT | SIPO_RTC_CLK | SIPO_RTC_SEL)

/*  Help functions for toggling the I/O pins
 */
static int
pin_state(unsigned int pin)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;

	immap = (immap_t *) IMAP_ADDR;
	io = &immap->im_ioport;

	return (0 != (io->iop_pdatd & pin));
}

static void
pin_up(unsigned int pin)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;

	immap = (immap_t *) IMAP_ADDR;
	io = &immap->im_ioport;

	io->iop_pdatd |= pin;
}

static void
pin_down(unsigned int pin)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;

	immap = (immap_t *) IMAP_ADDR;
	io = &immap->im_ioport;

	io->iop_pdatd &= ~pin;
}

static void
pin_flip(unsigned int pin)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;

	immap = (immap_t *) IMAP_ADDR;
	io = &immap->im_ioport;

	if (io->iop_pdatd & pin)
		io->iop_pdatd &= ~pin;
	else
		io->iop_pdatd |= pin;
}


/*  Initialize the pins we use to communicate with the rtc
 *  e.g. mark them general and set the directions
 */
static void
pins_init(void)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;

	immap = (immap_t *) IMAP_ADDR;
	io = &immap->im_ioport;

	io->iop_ppard &= ~SIPO_RTC_ALL;
	io->iop_pdird |= SIPO_RTC_OUT | SIPO_RTC_CLK | SIPO_RTC_SEL;
	io->iop_pdird &= ~SIPO_RTC_IN;
}

/* Output one byte to rtc
 */
static void
rtc_out(unsigned char c)
{
	unsigned char mask = 0x80;
	int i;

	for (i = 0; i < 8; i++, mask >>= 1) {
		pin_flip(SIPO_RTC_CLK);
		if (c & mask)
			pin_up(SIPO_RTC_OUT);
		else
			pin_down(SIPO_RTC_OUT);
		udelay(1);
		pin_flip(SIPO_RTC_CLK);
		udelay(1);
	}
}

/* Get one byte from rtc
 */
static unsigned char
rtc_in(void)
{
	unsigned char mask = 0x80;
	unsigned char c = 0;
	int i;

	for (i = 0; i < 8; i++, mask >>= 1) {
		pin_flip(SIPO_RTC_CLK);
		udelay(1);
		pin_flip(SIPO_RTC_CLK);
		udelay(1);
		if (pin_state(SIPO_RTC_IN))
			c |= mask;
	}

	return c;
}

/* Write range of bytes to the RTC chip
 */
static void
sipo_rtc_write(uint address, size_t len, unsigned char *buffer)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;
	int i;

	immap = (immap_t *) IMAP_ADDR;
	io = &immap->im_ioport;

	pin_down(SIPO_RTC_CLK);
	pin_up(SIPO_RTC_SEL);
	udelay(1);
	rtc_out(address & 0xff);
	for (i = 0; i < len; i++)
		rtc_out(buffer[i]);
	pin_down(SIPO_RTC_SEL);
}

/* Read range of bytes from the RTC chip
 */
static void
sipo_rtc_read(uint address, size_t len, unsigned char *buffer)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;
	int i;

	immap = (immap_t *) IMAP_ADDR;
	io = &immap->im_ioport;

	pin_down(SIPO_RTC_CLK);
	pin_up(SIPO_RTC_SEL);
	udelay(1);
	rtc_out(address & 0xff);
	for (i = 0; i < len; i++)
		buffer[i] = rtc_in();
	pin_down(SIPO_RTC_SEL);
}

/* External interface functions
 */

/* SIPO clock is always running in UTZ, so all we do is init the
 * the communication routines and return an offset of zero.
 */
long
sipo_time_init(void)
{
	unsigned char buffer[18];

	printk("SIPO RTC version 0.3\n");

	pins_init();
	pin_down(SIPO_RTC_SEL);

	sipo_rtc_read(0x00, 18, buffer);

	if ((buffer[15] & 0x80)) {
		printk("Warning!!! Clock has stopped, restarting.");
		/* Enable write */
		buffer[0] = 0x00;
		sipo_rtc_write(0x8f, 1, buffer);

		buffer[0]  = 0x00;
		buffer[1]  = 0x00;
		buffer[2]  = 0x00;
		buffer[3]  = 0x06;
		buffer[4]  = 0x01;
		buffer[5]  = 0x09;
		buffer[6]  = 0x00;
		buffer[7]  = 0x00;
		buffer[8]  = 0x00;
		buffer[9]  = 0x00;
		buffer[10] = 0x00;
		buffer[11] = 0x00;
		buffer[12] = 0x00;
		buffer[13] = 0x00;
		buffer[14] = 0x00;
		buffer[15] = 0x00;
		buffer[16] = 0x00;
		buffer[17] = 0xa5;
		sipo_rtc_write(0x80, 18, buffer);

		/* Disable write */
		buffer[0] = 0x40;
		sipo_rtc_write(0x8f, 1, buffer);

		sipo_rtc_read(0x00, 18, buffer);
	}

	printk("Chip time: 20%02x-%02x-%02x %02x:%02x.%02x UTC\n",
		buffer[6], buffer[5], buffer[4],
		buffer[2], buffer[1], buffer[0]);

	return 0;
}

/*  Dallas 1305 stores the time as two BCD digits in 8bit registers 0 to 6.
 *      0       seconds         (00-59)
 *      1       minutes         (00-59)
 *      2       hours           (00-23)
 *      3       day             (01-07)   ignored
 *      4       day of month    (01-31)
 *      5       month           (01-12)
 *      6       year            (00-99)
 */
unsigned long
sipo_get_rtc_time(void)
{
	unsigned int year, mon, day, hour, min, sec;
	unsigned char buffer[7];

	sipo_rtc_read(0x00, 7, buffer);

	sec =  (((buffer[0] >> 4) & 0x0f) * 10) + (buffer[0] & 0x0f);
	min =  (((buffer[1] >> 4) & 0x0f) * 10) + (buffer[1] & 0x0f);
	hour = (((buffer[2] >> 4) & 0x0f) * 10) + (buffer[2] & 0x0f);
	day =  (((buffer[4] >> 4) & 0x0f) * 10) + (buffer[4] & 0x0f);
	mon =  (((buffer[5] >> 4) & 0x0f) * 10) + (buffer[5] & 0x0f);
	year = (((buffer[6] >> 4) & 0x0f) * 10) + (buffer[6] & 0x0f) + 2000;


	return mktime(year, mon, day, hour, min, sec);
}

int
sipo_set_rtc_time(unsigned long nowtime)
{
	unsigned char buffer[18];
	struct rtc_time tm;

	extern void to_tm(int, struct rtc_time *);

	to_tm(nowtime, &tm);

	/* Enable write */
	buffer[0] = 0x00;
	sipo_rtc_write(0x8f, 1, buffer);

	/* Create and write the time on chip */
	buffer[0] = (((tm.tm_sec / 10) << 4) & 0xf0)
		    | ((tm.tm_sec % 10) & 0x0f);
	buffer[1] = (((tm.tm_min / 10) << 4) & 0xf0)
		    | ((tm.tm_min % 10) & 0x0f);
	buffer[2] = (((tm.tm_hour / 10) << 4) & 0xf0)
		    | ((tm.tm_hour % 10) & 0x0f);
	buffer[3] = 1;
	buffer[4] = (((tm.tm_mday / 10) << 4) & 0xf0)
		    | ((tm.tm_mday % 10) & 0x0f);
	buffer[5] = ((((tm.tm_mon) / 10) << 4) & 0xf0)
		    | (((tm.tm_mon) % 10) & 0x0f);
	buffer[6] = ((((tm.tm_year % 100) / 10) << 4) & 0xf0)
		    | (((tm.tm_year % 100) % 10) & 0x0f);

	sipo_rtc_write(0x80, 7, buffer);

	/* Disable write */
	buffer[0] = 0x40;
	sipo_rtc_write(0x8f, 1, buffer);

	return 0;
}

/***********************************************************************

  Misc Device Driver for reading from & writing to NVRAM

***********************************************************************/
#ifdef CONFIG_SIPO_NVRAM

#define NVRAM_SIZE 0x80

static int nvram_open(struct inode *inode, struct file *file)
{
	return 0;
}


static int nvram_release(struct inode *inode, struct file *file)
{
	return 0;
}

static ssize_t nvram_write(struct file *file, const char *buf, size_t count,
			   loff_t *ppos)
{
	unsigned char buffer[NVRAM_SIZE], c[1];

	if (count < 0)
		return -EIO;

	if (*ppos > NVRAM_SIZE)
		return -EIO;

	if (*ppos + count > NVRAM_SIZE)
		count = NVRAM_SIZE - *ppos;

	if (copy_from_user(buffer, buf, count))
		return -EFAULT;

	/* Enable write */
	c[0] = 0x00;
	sipo_rtc_write(0x8f, 1, c);

	sipo_rtc_write((unsigned) *ppos + 0x80, count, buffer);

	/* Disable write */
	c[0] = 0x40;
	sipo_rtc_write(0x8f, 1, c);

	*ppos += count;
	return count;
}

static ssize_t nvram_read(struct file * file, char * buf, size_t count,
			  loff_t *ppos)
{
	unsigned char buffer[NVRAM_SIZE];

	if (count < 0)
		return -EIO;

	if (*ppos > NVRAM_SIZE)
		return 0;

	if (count > NVRAM_SIZE - *ppos)
		count = NVRAM_SIZE - *ppos;

	sipo_rtc_read((unsigned) *ppos, count, buffer);

	if (copy_to_user(buf, buffer, count))
		return -EFAULT;

	*ppos += count;
	return count;
}


#ifdef CONFIG_PROC_FS

/*
 * /proc/sipo/bootstate read & write routines. (/arch/ppc/sipo/proc.c)
 */

static int proc_sipo_bootstate_write(struct file *file, const char *buffer,
				     unsigned long len, void *data)
{
	unsigned char c[1];
	int state, i;
	sipo_nv_info nv;

	for (state = 0, i = 0; i < len; i++) {
		if (buffer[i] <= '9' && buffer[i] >= '0')
			state = state * 10 + (buffer[i] - '0');
		else
			break;
	}
	if (i < len && buffer[i] != 0 && !isspace(buffer[i]))
		return -EIO;

	sipo_rtc_read(0x20, sizeof(sipo_nv_info), (char *) &nv);
	nv.state = state;

	/* Enable write */
	c[0] = 0x00;
	sipo_rtc_write(0x8f, 1, c);

	sipo_rtc_write(0x20 + 0x80, sizeof(sipo_nv_info), (char *)&nv);

	/* Disable write */
	c[0] = 0x40;
	sipo_rtc_write(0x8f, 1, c);

	return len;
}

static int proc_sipo_bootstate_read(char *page, char **start, off_t off,
				    int count, int *eof, void *data)
{
	char *p = page;
	sipo_nv_info nv;
	int len;

	sipo_rtc_read(0x20, sizeof(sipo_nv_info), (char *) &nv);
	p+=sprintf(p, "%d\n", nv.state);

	len = p - page;

	if (len <= off + count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;

}
#endif /* CONFIG_PROC_FS */

static struct file_operations nvram_fops =
{
	owner:          THIS_MODULE,
	read:           nvram_read,
	write:          nvram_write,
	open:           nvram_open,
	release:        nvram_release,
};

static struct miscdevice nvram_miscdev =
{
	SIPONVRAM_MINOR,
	"siponvram",
	&nvram_fops
};

void __init siponvram_init(void)
{
#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *entry;
#endif
	misc_register(&nvram_miscdev);
	printk("SIpo NVRAM: 0.1\n");

#ifdef CONFIG_PROC_FS
	entry = create_proc_entry("bootstate", 0400, sipo_proc_root_dir);
	entry->nlink = 1;
	entry->read_proc = proc_sipo_bootstate_read;
	entry->write_proc = proc_sipo_bootstate_write;
#endif
}

#endif CONFIG_SIPO_NVRAM
