
/*
 *      sIPo audio 0.01:        An audio driver for sIPo box
 *
 *      Copyright 2000 SSH Communications Security Corp, All Rights Reserved.
 *                              http://www.ssh.com
 *
 *      14.1.2001 Arto Vuori <avuori@ssh.com>
 */


/*
  Theory of operation

  Two serial controllers is used - one for data and another for a
  left/right signal.  SCC1 is used for serial audio output. Data is
  sent MSB first, and is syncronous to BRGO3, which is data clock out
  (for both SCCs). SCC3 sends the left/right signal, which is
  syncronous to datastream. It starts with a pattern of 16 zeros, 16
  ones and then continues looping the pattern.

  Both CTS signals should be tied together and driven throuh PC0 GPIO
  pin, so that they start synchronously. Once started, data should be feeded
  continously without any pauses to avoid loosing sync.

  used signals:

  MPC8260                               Philips DAC ???

  PC24 - TxD3   >-------------------> Left /Right
  PC30 - TxD1   >-------------------> DataOut
  PC27 - BRG03  >-------------------> Clock
  PC15 - CTS1  <-----,
  PC8  - CTS3  <-----|
  PC30 CTSout >---- */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/reboot.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/poll.h>
#include <asm/delay.h>
#include <asm/uaccess.h>
#include <asm/immap_8260.h>
#include <asm/mpc8260.h>
#include <asm/cpm_8260.h>
#include <asm/irq.h>

/* Data buffers for audio operations */
#define RING_SIZE 32
#define BUFFER_SIZE 4096

/* Data buffers & descriptor rings */
static cbd_t    *tx_bdt;
static void     *tx_buff[RING_SIZE];
static cbd_t    *tx2_bdt;
static void     *tx2_buff[RING_SIZE];

/* Data buffers & descriptor rings */
static cbd_t    *rx_bdt;
static void     *rx_buff[RING_SIZE];
static cbd_t    *rx2_bdt;
static void     *rx2_buff[RING_SIZE];


static volatile int head = 0;                         /* Next to be written */
static volatile int tail = 0;                         /* Current pointer */
static volatile int full = 0, empty = 1, restart = 1; /* State flags */;

static wait_queue_head_t audio_wait;
static spinlock_t audio_lock = SPIN_LOCK_UNLOCKED;

#define AUDIO_BAUDRATE (44100 * 2 * 16)

static void set_cts(int i)
{
	volatile iop8260_t *io;

	io = &immr->im_ioport;

	io->iop_pparc &=~(0x80000000 >> 30);      /* SpareIO -> CTS1/3 */
	io->iop_podrc &=~(0x80000000 >> 30);
	io->iop_psorc &=~(0x80000000 >> 30);
	io->iop_pdirc |= (0x80000000 >> 30);
	if (i)
		io->iop_pdatc |= (0x80000000 >> 30);
	else
		io->iop_pdatc &=~(0x80000000 >> 30);
}

static void audio_interrupt(int irq, void *not_used, struct pt_regs *regs)
{
	volatile immap_t *immap;
	volatile cpm8260_t *cp;
	volatile scc_t *scp;
	unsigned long flags;
	uint events;

	spin_lock_irqsave(&audio_lock, flags);

	scp = &immr->im_scc[0];
	events = scp->scc_scce;
	if (events & SCCM_TX) {
	    while ((full || tail != head) &&
		   !(tx_bdt[tail].cbd_sc & BD_SC_READY)) {
		if (full) {
		    wake_up_interruptible(&audio_wait);
		    full = 0;
		}
		tail = (tail + 1) % RING_SIZE;

		if (head == tail) {
			wake_up_interruptible(&audio_wait);
			empty = 1;
		}

	    }
	}
	if (events & SCCM_TXE) {
		immap = immr;
		scp = &immap->im_scc[0];
		scp->scc_gsmrl &= ~(SCC_GSMRL_ENR | SCC_GSMRL_ENT);

		cp = &immap->im_cpm;
		cp->cp_cpcr = mk_cr_cmd(CPM_CR_SCC1_PAGE, CPM_CR_SCC1_SBLOCK,
					0, CPM_CR_RESTART_TX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);

		scp->scc_gsmrl |= (SCC_GSMRL_ENR | SCC_GSMRL_ENT);

		restart=1;
		set_cts(1);

		wake_up_interruptible(&audio_wait);
	}
	scp->scc_scce = events;

	/* SCC3 errors handler */
	scp = &immr->im_scc[2];
	events = scp->scc_scce;
	if (events & SCCM_TXE) {
		immap = immr;
		scp->scc_gsmrl &= ~(SCC_GSMRL_ENR | SCC_GSMRL_ENT);

		cp = &immap->im_cpm;
		cp->cp_cpcr = mk_cr_cmd(CPM_CR_SCC3_PAGE, CPM_CR_SCC3_SBLOCK,
					0, CPM_CR_RESTART_TX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);

		scp->scc_gsmrl |= (SCC_GSMRL_ENR | SCC_GSMRL_ENT);

		restart=1;
		set_cts(1);

		wake_up_interruptible(&audio_wait);
	}
	scp->scc_scce = events;

	spin_unlock_irqrestore(&audio_lock, flags);
}



/*
 *      Initialize audio HW
 */
static int audio_open(struct inode *inode, struct file *file)
{
	volatile immap_t *immap;
	volatile cpm8260_t *cp;
	volatile scc_t *scp;

	immap = immr;
	cp = &immap->im_cpm;

	cp->cp_cpcr = mk_cr_cmd(CPM_CR_SCC1_PAGE, CPM_CR_SCC1_SBLOCK, 0,
				CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);

	scp = &immap->im_scc[0];
	scp->scc_gsmrl |= (SCC_GSMRL_ENR | SCC_GSMRL_ENT);

	cp->cp_cpcr = mk_cr_cmd(CPM_CR_SCC3_PAGE, CPM_CR_SCC3_SBLOCK, 0,
				CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);

	scp = &immap->im_scc[2];
	scp->scc_gsmrl |= (SCC_GSMRL_ENR | SCC_GSMRL_ENT);

	return 0;
}


/*
 *      Release audio HW
 */
static int audio_release(struct inode *inode, struct file *file)
{
	unsigned long flags;
	volatile immap_t *immap;
	volatile scc_t *scp;

	DECLARE_WAITQUEUE(wait, current);

	printk("R");

	/*
	 * Wait until there is free buffers available
	 */
	add_wait_queue(&audio_wait, &wait);
	current->state = TASK_INTERRUPTIBLE;

	while (!empty) {
#if 0
		if (file->f_flags & O_NDELAY) {
			remove_wait_queue(&audio_wait, &wait);
			current->state = TASK_RUNNING;
			return -EWOULDBLOCK;
		}
#endif
		if (signal_pending(current)) {
			remove_wait_queue(&audio_wait, &wait);
			current->state = TASK_RUNNING;
			return -ERESTARTSYS;
		}
		schedule();
		current->state = TASK_INTERRUPTIBLE;
	}
	remove_wait_queue(&audio_wait, &wait);
	current->state = TASK_RUNNING;

	spin_lock_irqsave(&audio_lock, flags);

	immap = immr;
	scp = &immap->im_scc[0];
	scp->scc_gsmrl &= ~(SCC_GSMRL_ENR | SCC_GSMRL_ENT);
	scp = &immap->im_scc[2];
	scp->scc_gsmrl &= ~(SCC_GSMRL_ENR | SCC_GSMRL_ENT);

	head = tail = full = 0;
	empty = 1;
	restart = 1;
	set_cts(1);

	spin_unlock_irqrestore(&audio_lock, flags);

	return 0;
}

/*
 *      Send Frame to audio HW
 */
static ssize_t audio_write(struct file *file, const char *buf,
			   size_t buf_count, loff_t *ppos)
{
	unsigned long flags;
	size_t done, count;

	DECLARE_WAITQUEUE(wait, current);

	/*
	 * Wait until there is free buffers available
	 */
	add_wait_queue(&audio_wait, &wait);
	current->state = TASK_INTERRUPTIBLE;

	while (full) {
		if (file->f_flags & O_NDELAY) {
			remove_wait_queue(&audio_wait, &wait);
			current->state = TASK_RUNNING;
			return -EWOULDBLOCK;
		}
		if (signal_pending(current)) {
			remove_wait_queue(&audio_wait, &wait);
			current->state = TASK_RUNNING;
			return -ERESTARTSYS;
		}
		schedule();
		current->state = TASK_INTERRUPTIBLE;
	}
	remove_wait_queue(&audio_wait, &wait);
	current->state = TASK_RUNNING;

	/* Lock'n load */

	spin_lock_irqsave(&audio_lock, flags);

	/* Loop through buffers as long as we got more data */

	/* Buffers Size should be multiple of 2 x 16bits */
	buf_count &= ~3;

	for (done = 0; done < buf_count && !full; done += count) {
		count = buf_count - done;

		if (count > BUFFER_SIZE)
			count = BUFFER_SIZE;

		if (tx_bdt[head].cbd_sc & BD_SC_READY)
			panic("siPO audio: TX Buffers should be free before "
			      "write!! */");

		if (copy_from_user(tx_buff[head], buf, count)) {
			spin_unlock_irqrestore(&audio_lock, flags);
			return -EFAULT;
		}

		tx_bdt[head].cbd_datlen = count;
		tx_bdt[head].cbd_sc |= BD_SC_READY;

		tx2_bdt[head].cbd_datlen = count;
		tx2_bdt[head].cbd_sc |= BD_SC_READY;

		head = (head + 1) % RING_SIZE;
		if (head == tail)
			full = 1;

		if (empty)
			empty = 0;

		*ppos += count;
	}

	if (restart) {
		udelay(1000);
		set_cts(0);
		restart = 0;
	}

	spin_unlock_irqrestore(&audio_lock, flags);

	return done;
}

static ssize_t audio_read(struct file * file, char * buf, size_t buf_count,
			  loff_t *ppos)
{
	return 0;
}

static struct file_operations audio_fops =
{
	owner:          THIS_MODULE,
	read:           audio_read,
	write:          audio_write,
	open:           audio_open,
	release:        audio_release,
};

static struct miscdevice audio_miscdev =
{
	SIPOAUDIO_MINOR,
	"sipoaudio",
	&audio_fops
};

/*
 * Initialize Audio HW
 */
void __init sipoaudio_init(void)
{
	volatile immap_t *immap;
	volatile cpm8260_t *cp;
	volatile iop8260_t *io;
	volatile scc_t *scp;
	volatile scc_uart_t *sup;
	int i, j;

	misc_register(&audio_miscdev);
	printk("SIpo audio: 0.01b\n");

	init_waitqueue_head(&audio_wait);

	if (request_8xxirq(SIU_INT_SCC1, audio_interrupt, 0, "audio", 0) < 0) {
		printk("SIpo audio: Can't get Audio IRQ %d\n", SIU_INT_SCC1);
		return;
	}

	if (request_8xxirq(SIU_INT_SCC3, audio_interrupt, 0, "audio", 0) < 0) {
		printk("SIpo audio: Can't get Audio IRQ %d\n", SIU_INT_SCC3);
		return;
	}

	immap = immr;
	cp = &immap->im_cpm;
	io = &immap->im_ioport;

/*
 * Initialize SCC1
 */

	/* SCC1 pins */
	io->iop_ppard |= (0x80000000 >> 30);      /* TxD1 */
	io->iop_podrd &=~(0x80000000 >> 30);
	io->iop_psord |= (0x80000000 >> 30);
	io->iop_pdird |= (0x80000000 >> 30);

	io->iop_ppard |= (0x80000000 >> 29);      /* ~RTS1 */
	io->iop_podrd &=~(0x80000000 >> 29);
	io->iop_psord &=~(0x80000000 >> 29);
	io->iop_pdird |= (0x80000000 >> 29);

	io->iop_pparc |= (0x80000000 >> 15);      /* ~CTS1 */
	io->iop_podrc &=~(0x80000000 >> 15);
	io->iop_psorc &=~(0x80000000 >> 15);
	io->iop_pdirc &=~(0x80000000 >> 15);

	io->iop_pparc |= (0x80000000 >> 27);      /* BRG03 */
	io->iop_podrc &=~(0x80000000 >> 27);
	io->iop_psorc |= (0x80000000 >> 27);
	io->iop_pdirc |= (0x80000000 >> 27);

	/* Connect SCC1 to NMSI.  Connect BRG3 to SCC1.
	 */
	immap->im_cpmux.cmx_scr &= ~0xff000000;
	immap->im_cpmux.cmx_scr |= 0x12000000;

	/* Set up the baud rate generator. */
	m8260_cpm_fastbrg(2, AUDIO_BAUDRATE, 0);

	scp = &immap->im_scc[0];
	sup = (scc_uart_t *) &immap->im_dprambase[PROFF_SCC1];

	/* We need to allocate a transmit buffer
	 * descriptors from dual port ram, and a character
	 * buffer area from host mem.
	 */
	sup->scc_genscc.scc_rbase =
		i = m8260_cpm_dpalloc(sizeof(cbd_t) * RING_SIZE, 16);
	rx_bdt = (cbd_t *) &immap->im_dprambase[i];

	for (j = 0; j < RING_SIZE; j++) {
		rx_bdt[j].cbd_datlen = BUFFER_SIZE;
		if (!(rx_buff[j] = kmalloc(BUFFER_SIZE, GFP_KERNEL)))
			panic("SiPo audio: Unable to allocate RX buffers\n");
		rx_bdt[j].cbd_bufaddr = __pa(rx_buff[j]);
		rx_bdt[j].cbd_sc = BD_SC_INTRPT;
	}
	rx_bdt[--j].cbd_sc |= BD_SC_WRAP;

	sup->scc_genscc.scc_tbase =
		i = m8260_cpm_dpalloc(sizeof(cbd_t) * RING_SIZE, 16);
	tx_bdt = (cbd_t *) &immap->im_dprambase[i];

	for (j = 0; j < RING_SIZE; j++) {
		tx_bdt[j].cbd_datlen = BUFFER_SIZE;
		if (!(tx_buff[j] = kmalloc(BUFFER_SIZE, GFP_KERNEL)))
			panic("SiPo audio: Unable to allocate TX buffers\n");
		tx_bdt[j].cbd_bufaddr = __pa(tx_buff[j]);
		tx_bdt[j].cbd_sc = BD_SC_INTRPT;
	}
	tx_bdt[--j].cbd_sc |= BD_SC_WRAP;

	/* Send the CPM an initialize command.
	 */
	cp->cp_cpcr = mk_cr_cmd(CPM_CR_SCC1_PAGE, CPM_CR_SCC1_SBLOCK, 0,
				CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);

	/* Set up the uart parameters in the
	 * parameter ram.
	 */
	sup->scc_genscc.scc_rfcr = CPMFCR_GBL | CPMFCR_EB;
	sup->scc_genscc.scc_tfcr = CPMFCR_GBL | CPMFCR_EB;
	sup->scc_genscc.scc_mrblr = BUFFER_SIZE;

	scp->scc_scce = 0xffff; /* Clear prev. events */
	/* TX transmit & error interrupts enabled */
	scp->scc_sccm = (SCCM_TX|SCCM_TXE);
	scp->scc_dsr = 0x0000;

	/* Transparent mode, msb first */
	scp->scc_gsmrh = SCC_GSMRH_TRX | SCC_GSMRH_TTX | SCC_GSMRH_REVD;
	scp->scc_gsmrl = 0;
	/* (SCC_GSMRL_MODE_UART | SCC_GSMRL_TDCR_16 | SCC_GSMRL_RDCR_16); */

/*
 * Initialize SCC3
 */

	/* SCC3 pins */
	io->iop_ppard |= (0x80000000 >> 24);      /* TxD3 */
	io->iop_podrd &=~(0x80000000 >> 24);
	io->iop_psord &=~(0x80000000 >> 24);
	io->iop_pdird |= (0x80000000 >> 24);

	io->iop_pparc |= (0x80000000 >> 8);       /* ~CTS3 */
	io->iop_podrc &=~(0x80000000 >> 8);
	io->iop_psorc |= (0x80000000 >> 8);
	io->iop_pdirc &=~(0x80000000 >> 8);

	/* Connect SCC3 to NMSI.  Connect BRG3 to SCC3.
	 */
	immap->im_cpmux.cmx_scr &= ~0x0000ff00;
	immap->im_cpmux.cmx_scr |= 0x00001200;

	/* Set up the baud rate generator. (Already set) */
	/* m8260_cpm_fastbrg(2, AUDIO_BAUDRATE, 0); */

	scp = &immap->im_scc[2];
	sup = (scc_uart_t *)&immap->im_dprambase[PROFF_SCC3];

	/* We need to allocate a transmit buffer
	 * descriptors from dual port ram, and a character
	 * buffer area from host mem.
	 */
	sup->scc_genscc.scc_rbase =
		i = m8260_cpm_dpalloc(sizeof(cbd_t) * RING_SIZE, 16);
	rx2_bdt = (cbd_t *)&immap->im_dprambase[i];

	for (j=0; j<RING_SIZE; j++) {
		rx2_bdt[j].cbd_datlen=BUFFER_SIZE;
		if (!(rx2_buff[j]=kmalloc(BUFFER_SIZE, GFP_KERNEL)))
			panic("SiPo audio: Unable to allocate RX buffers\n");
		rx2_bdt[j].cbd_bufaddr = __pa(rx2_buff[j]);
		rx2_bdt[j].cbd_sc = BD_SC_INTRPT;
	}
	rx2_bdt[--j].cbd_sc |= BD_SC_WRAP;

	sup->scc_genscc.scc_tbase =
		i = m8260_cpm_dpalloc(sizeof(cbd_t) * 2, 16);
	tx2_bdt = (cbd_t *)&immap->im_dprambase[i];

	for (j=0; j<RING_SIZE; j++) {
		tx2_bdt[j].cbd_datlen=BUFFER_SIZE;
		if (!(tx2_buff[j]=kmalloc(BUFFER_SIZE, GFP_KERNEL)))
			panic("SiPo audio: Unable to allocate TX2 buffers\n");
		tx2_bdt[j].cbd_bufaddr = __pa(tx2_buff[j]);
		tx2_bdt[j].cbd_sc = BD_SC_INTRPT;
		/* Fill tx2 buffer with Left/Right bits */
		for (i = 0; i < BUFFER_SIZE / 2; i++) {
			if (i & 1)
				((short *)(tx2_buff[j]))[i]=0xfffe;
			else
				((short *)(tx2_buff[j]))[i]=0x0001;
		}

	}
	tx2_bdt[--j].cbd_sc |= BD_SC_WRAP;

	/* Send the CPM an initialize command.
	 */
	cp->cp_cpcr = mk_cr_cmd(CPM_CR_SCC3_PAGE, CPM_CR_SCC3_SBLOCK, 0,
				CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);

	/* Set up the uart parameters in the
	 * parameter ram.
	 */
	sup->scc_genscc.scc_rfcr = CPMFCR_GBL | CPMFCR_EB;
	sup->scc_genscc.scc_tfcr = CPMFCR_GBL | CPMFCR_EB;
	sup->scc_genscc.scc_mrblr = BUFFER_SIZE;

	scp->scc_scce = 0xffff; /* Clear prev. events */
	/* TX transmit & error interrupts enabled */
	scp->scc_sccm = (SCCM_TX|SCCM_TXE);
	scp->scc_dsr = 0x0000;

	/* Transparent mode, msb first */
	scp->scc_gsmrh =  SCC_GSMRH_TRX | SCC_GSMRH_TTX | SCC_GSMRH_REVD;
	scp->scc_gsmrl = 0;

/*
 *      unassert CTS for both SCC's
 */

	set_cts(1);
}

#ifdef MODULE
int init_module(void)
{
	sipoaudio_init();
	return 0;
}

void cleanup_module(void)
{
	misc_deregister(&audio_miscdev);
}
#endif
