/*
 *      sIPo flash 0.6:  Lowlevel flash I/O routines + Misc Character device
 *
 *      Copyright 2001 SSH Communications Security Corp, All Rights Reserved.
 *                              http://www.ssh.com
 *
 *      Initial implementation 16.1.2001 by Arto Vuori <avuori@ssh.com>
 *
 *      Misc device is not a block device nor a filesystem. (see devrom)
 *      You can read and write as you wish without any size & alignment
 *      restrictions. If you write a non-aligned block, write routine
 *      copies surrounding data to temporary buffer before erase and
 *      writes it back when re-programming the block.
 *
 *      8.2.2001 av
 *      - Also nonaligned writes work
 *
 *      19.2.2001 av
 *      - a rwlock flash_lock added
 *      - flash_read() acquires the read lock
 *      - flash_write_block() acquires the write lock
 *      - All reads & writes must go through these routines.
 *
 *      22.2.2001 av
 *      - rwlock sucks... implemented own mutex with wait_queue.
 *      - All operations are ordered by using acquire_ & release_flash().
 *      - Operations wait until previous operation finnish.
 *      - Flash status is read from IO-pin instead of reading status register
 *      - Flash writes done using faster write buffer mode.
 *
 *      23.2.2001 av
 *      - flash_program_area() is used to write into flash.
 *      - flash_write(), flash_erase() implemented.
 *
 *      16.3.2001 Jouni Malinen <jkm@ssh.com>
 *      - use a led to indicate writing/erasing
 *      - use blocking version acquire_flash() when accessing flash through MTD
 *
 *      18.5.2001 av
 *      - Misc device always does access true flash memory
 *      - other interfaces use bootmem area when in flash emulation mode
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/locks.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/reboot.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/delay.h>
#include <asm/uaccess.h>
#include <asm/immap_8260.h>
#include <asm/mpc8260.h>
#include <asm/cpm_8260.h>
#include <asm/irq.h>
#include <asm/sipo.h>

/* #define DEBUG */
/* #define EXTRA_DEBUG */

#define FLASH_MAX_RETRY 5
#define FLASH_MAX_SINGLE_WRITE 4096
#define FLASH_BLOCKSIZE (128*1024)
#define MADE_BY_INTEL 0x89
#define STRATA_FLASH_16MB 0x18
#define STRATA_FLASH_8MB 0x17

static uint  flash_lanes_swapped;
static uint  flash_A7_missing;
static void *flash_base;
static void *false_flash_base;
static uint  flash_size;

#define FLASHWORD(index) (*(volatile unsigned short *) (flash_base + (index)))

/*
 * Acquire_flash() & release_flash() are used to serialize flash access.
 */

static int flash_state = 0;
static wait_queue_head_t flash_wait;
static spinlock_t flash_lock = SPIN_LOCK_UNLOCKED;


static int acquire_flash(int block)
{
	DECLARE_WAITQUEUE(wait, current);

	add_wait_queue(&flash_wait, &wait);
	for (;;) {
		current->state = TASK_INTERRUPTIBLE;
		if (!block && signal_pending(current)) {
			remove_wait_queue(&flash_wait, &wait);
			current->state = TASK_RUNNING;
			return -ERESTARTSYS;
		}
		spin_lock_irq(&flash_lock);
		if (!flash_state) {
			flash_state++;
			spin_unlock_irq(&flash_lock);
			break;
		}
		spin_unlock_irq(&flash_lock);
		schedule();
	}
	remove_wait_queue(&flash_wait, &wait);
	current->state = TASK_RUNNING;

	return 0;
}

static void release_flash(void)
{
	spin_lock_irq(&flash_lock);
	flash_state--;
	spin_unlock_irq(&flash_lock);

	wake_up_interruptible(&flash_wait);
}

/******************************************************************************

 Flash low-level access routines.

******************************************************************************/

/*
 * flash_sts() reads flash status from IO pins.
 * Returns zero if either flash is busy
 */

#define PIN_STS_LO (0x80000000>>5)
#define PIN_STS_HI (0x80000000>>4)

static inline int flash_sts(void)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;

	immap = (immap_t *) IMAP_ADDR;
	io = &immap->im_ioport;

	if ((io->iop_pdatc & PIN_STS_HI) && (io->iop_pdatc & PIN_STS_LO))
		return 1; /* Both are ready */
	else
		return 0; /* One or both flash is not ready */
}

/*
  Erase single flash block
*/
static int
flash_erase_block(unsigned int blocknum)
{
	int i;
	unsigned short dummy;
	unsigned int blockaddr;

#ifdef DEBUG
	printk("sIPoflash: erase block %d starting...\n", blocknum);
#endif

	sipoled_power_blink(1);

	udelay(100);
	blockaddr = blocknum * FLASH_BLOCKSIZE;

	FLASHWORD(blockaddr) = 0x2020;
	mb();
	FLASHWORD(blockaddr) = 0xd0d0;
	mb();

	udelay(100);

	for (i = 0; i <= HZ * 5 && !flash_sts(); i++) {
	    current->state = TASK_UNINTERRUPTIBLE;
	    schedule_timeout(10);
	}
	if (i > HZ * 5) {
		printk("flash_erase_block(%u) - timeout? i=%d\n", blocknum, i);
	}

	current->state = TASK_RUNNING;

	FLASHWORD(blockaddr) = 0xffff;
	mb();
	dummy = FLASHWORD(blockaddr);
	mb();

#ifdef EXTRA_DEBUG
	/* check whether the block was erased successfully */
	{
		int j;
		unsigned char *fpos = (unsigned char *) flash_base;

		fpos += (blocknum * FLASH_BLOCKSIZE);
		for (j = 0; j < FLASH_BLOCKSIZE; j++) {
			if (*fpos != 0xff) {
				printk("flash_erase_block(%u) - at %d: 0x%02X"
				       " != 0xFF\n", blocknum, j, *fpos);
				break;
			}
			fpos++;
		}
	}
#endif

	sipoled_power_blink(0);

#ifdef DEBUG
	printk("sIPoflash: erase %d done\n", blocknum);
#endif

	return 0;
}

/*
 * Write single buffer. (32 bytes)
 */
static void __inline
flash_write_buffer(char *addr, uint i)
{
	uint k, blockaddr;
	unsigned short dummy;

	blockaddr = (i & ~(FLASH_BLOCKSIZE - 1));

	sipoled_power_blink(1);

	/* Buffer Write command */
	FLASHWORD(blockaddr) = 0xe8e8;
	mb();
	for (k = 100; k; k--) {
		udelay(10);
		if ((FLASHWORD(blockaddr) & 0x8080) != 0)
			break;
	}
	if (!k) {
		printk("sIPoflash: flash_write_buffer(%p, %u) - buffer not "
		       "available\n", addr, i);
		sipoled_power_blink(0);
		return; /* Buffer not available */
	}

	/* Write count & address registers*/
	mb();
	if (flash_lanes_swapped)
	    FLASHWORD(blockaddr) = 0x0f00;
	else
	    FLASHWORD(blockaddr) = 0x000f;
	mb();

	/* Write to buffer (with some loop unrolling) */
	for (k = 0; k < 32; k += 2 * 4) {
		FLASHWORD(i+k+2*0) = *(unsigned short *)(addr+k+2*0); mb();
		FLASHWORD(i+k+2*1) = *(unsigned short *)(addr+k+2*1); mb();
		FLASHWORD(i+k+2*2) = *(unsigned short *)(addr+k+2*2); mb();
		FLASHWORD(i+k+2*3) = *(unsigned short *)(addr+k+2*3); mb();
	}

	/* confirm write */
	FLASHWORD(blockaddr) = 0xd0d0;
	mb();

	/* Wait until status clear */
	dummy = FLASHWORD(blockaddr);
	mb();
	do { udelay(10); } while (!flash_sts());

	sipoled_power_blink(0);
}


/*
  Program any flash area.
*/

static int
flash_program_area(uint ppos, void *addr, uint count)
{
	unsigned short dummy;
	unsigned int i, j;
	unsigned char buffer[32];

#ifdef DEBUG
	unsigned int blocknum;
	blocknum = ppos / FLASH_BLOCKSIZE;
	printk("sIPoflash: program (block=%d) starting...\n", blocknum);
#endif

	sipoled_power_blink(1);
	udelay(100);
#ifdef EXTRA_DEBUG
	for (i = 0; i < count; i++) {
		unsigned char old, new;

		old = *((unsigned char *) flash_base + ppos + i);
		new = ((unsigned char *) addr)[i];
		/* check that it is okay to flash the new value without erasing
		 * flash (i.e., no need to set any bits from zero to one) */
		if (~old & new) {
			printk("flash_program_area(%u, %p, %u) - not empty at "
			       "%d (old=0x%02X new=0x%02X)\n", ppos, addr,
			       count, i, old, new);
			break;
		}
	}
#endif

	/* Write first partial buffer */
	if (ppos & 31) {
		j = 32 - (ppos & 31);
		if (j > count)
			j = count;
		memcpy(buffer, flash_base + (ppos & ~31), 32);
		memcpy(buffer + (ppos & 31), addr, j);

		flash_write_buffer(buffer, ppos&~31);

		/* Read array command */
		FLASHWORD(ppos & ~(FLASH_BLOCKSIZE - 1)) = 0xffff;
		mb();
		dummy = FLASHWORD(ppos & ~(FLASH_BLOCKSIZE - 1));
		mb();

		ppos += j; addr += j; count -= j;
	}

	/* Write middle buffers */
	while (count & ~31) {
		j = FLASH_BLOCKSIZE - (ppos & (FLASH_BLOCKSIZE - 1));

		if (j > (count & ~31))
			j = (count & ~31);

		if (j > FLASH_MAX_SINGLE_WRITE)
			j = FLASH_MAX_SINGLE_WRITE;

		for (i = ppos; i < ppos + j; i += 32, addr += 32)
			flash_write_buffer(addr, i);

		/* Read array command */
		FLASHWORD(ppos & ~(FLASH_BLOCKSIZE - 1)) = 0xffff;
		mb();
		dummy = FLASHWORD(ppos & ~(FLASH_BLOCKSIZE - 1));
		mb();
		schedule();

		ppos += j; count -= j;
	}

	/* Write last partial buffer */
	if (count) {
		memcpy(buffer, flash_base + (ppos & ~31), 32);
		memcpy(buffer + (ppos & 31), addr, count);

		flash_write_buffer(buffer, ppos & ~31);

		/* Read array command */
		FLASHWORD(ppos & ~(FLASH_BLOCKSIZE - 1)) = 0xffff;
		mb();
		dummy = FLASHWORD(ppos & ~(FLASH_BLOCKSIZE - 1));
		mb();

		ppos += j; addr += j; count -= j;
	}

	sipoled_power_blink(0);

#ifdef DEBUG
	printk("sIPoflash: program %d done\n", blocknum);
#endif
	return 0;
}

/*
  Read serial number of the box
  If the code should work on early boards also (below serial no 11),
  you have to call this routine at least once befor doing a write.
  This will check for how the chip was attached to the bus
  and set the static flash_lanes_swapped variable.
  Most of the commands don't care if the lanes are swapped or not.
  But the burst-write mode has to know it.
*/
static long long
flash_read_serial_number(unsigned char *Manuf_ID, unsigned char *Device_Type,
			 unsigned char *Lock, char *SerialLock)
{
	long long SerialNr;
	volatile unsigned short *flash;
	int lock_offset;

	flash = (unsigned short *) (flash_base);
	*flash = 0x9090;
	mb();
	flash_lanes_swapped = ((*flash & 0xff00) != 0);
	flash_A7_missing = (flash[64] != 0);
	if (flash_A7_missing)
		lock_offset = 64;
	else
		lock_offset = 128;
	if (flash_lanes_swapped) {
		*Manuf_ID = (flash[0]) >> 8;
		*Device_Type = (flash[1]) >> 8;
		*Lock = (flash[2]) >> 8;
		*SerialLock = (flash[lock_offset]) >> 8;
	} else {
		*Manuf_ID = flash[0];
		*Device_Type = flash[1];
		*Lock = flash[2];
		*SerialLock = flash[lock_offset];
	}
	((unsigned short *) &SerialNr)[0] = flash[lock_offset + 5];
	((unsigned short *) &SerialNr)[1] = flash[lock_offset + 6];
	((unsigned short *) &SerialNr)[2] = flash[lock_offset + 7];
	((unsigned short *) &SerialNr)[3] = flash[lock_offset + 8];
	mb();
	*flash = 0xffff;
	mb();
	return SerialNr;
}


static void
flash_clear_lock_bits(void)
{
	unsigned short dummy;
	int i;

	FLASHWORD(0)=0x6060;
	mb();
	FLASHWORD(0)=0xd0d0;
	mb();

	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(HZ / 2);

	dummy = FLASHWORD(0);
	mb();

	for (i = 0; i <= HZ; i++) {
	    current->state = TASK_INTERRUPTIBLE;
	    schedule_timeout(HZ / 10);
	    if ((FLASHWORD(0) & 0x8080) != 0)
		break;
	}

	FLASHWORD(0) = 0xffff;
	mb();
}

/*
 * Write single block. Caller must have called acquire_flash() earlier.
 */
static int flash_write_single_block(int blocknum, void *addr)
{
	unsigned int j;

	for (j = 0; j < FLASH_MAX_RETRY; j++) {
		if (!flash_erase_block(blocknum) &&
		    !flash_program_area(blocknum * FLASH_BLOCKSIZE, addr,
					FLASH_BLOCKSIZE) &&
		    !memcmp((void *) (flash_base +
				      (blocknum) * FLASH_BLOCKSIZE), addr,
			    FLASH_BLOCKSIZE))
			return 0;

		printk("sIPoflash: Programming failure on block %d. "
		       "Retrying... \n", blocknum);
	}

	printk("sIPoFlash: Unable to program flash block %d\n", blocknum);
	return -1;
}

/*
   Erase, program and verify multiple flash blocks.
 */
static int flash_write_blocks(int blocknum, void *addr, unsigned int count)
{
	unsigned int i, j, r;

	for (i = 0; i < count; i++) {
		for (j = 0; j < FLASH_MAX_RETRY; j++) {
			if ((r = acquire_flash(0)))
				return r;

			if (!flash_erase_block(blocknum + i) &&
			    !flash_program_area((blocknum + i) *
						FLASH_BLOCKSIZE,
						addr + i * FLASH_BLOCKSIZE,
						FLASH_BLOCKSIZE) &&
			    !memcmp((void *)
				    (flash_base +
				     (blocknum + i) * FLASH_BLOCKSIZE),
				    addr + i * FLASH_BLOCKSIZE,
				    FLASH_BLOCKSIZE)) {
				/* Programming successful */
				release_flash();
				break;
			}

			release_flash();
			printk("sIPoflash: Programming failure on block %d. "
			       "Retrying...\n", blocknum + i);
			schedule();
		}

		if (j == FLASH_MAX_RETRY) {
			printk("sIPoFlash: Unable to program flash block %d\n",
			       blocknum + i);
			return -1;
		}

		schedule();
	}

	return 0;
}

/*
 * Write non-aligned and not block sized memory areas.
 * Cut & Copy & Paste original data to the areas not to be programmed.
 */
static int flash_cutcopypaste(unsigned int offset, void *addr,
			      unsigned int count)
{
	static char *temp = NULL;
	unsigned int i, j, ret = 0;

#ifdef DEBUG
	printk("sIPoflash: Cutcopypaste (off=0x%x, addr=0x%x, count=0x%x)\n",
	       offset, (uint) addr, count);
#endif

	if (!temp) {
		temp = kmalloc(FLASH_BLOCKSIZE, GFP_KERNEL);
		if (!temp) {
			printk("sIPoflash: Unable to allocate temporary "
			       "buffer\n");
			return -EIO;
		}
	}

	i = (offset & (FLASH_BLOCKSIZE - 1));
	j = FLASH_BLOCKSIZE - i;
	if (j >= count) {
		if ((ret = acquire_flash(0)))
			return ret;
		/* Write single block */
		memcpy(temp, (void *) (flash_base + offset - i), i);
		if (copy_from_user(temp + i, addr, count)) {
			release_flash();
			return -EFAULT;
		}
		if (FLASH_BLOCKSIZE - i - count >= 0)
			memcpy(temp + i + count,
			       (void *) (flash_base + offset + count),
			       FLASH_BLOCKSIZE - i - count);
		ret = flash_write_single_block(offset / FLASH_BLOCKSIZE, temp);
		release_flash();
	} else {
		if ((ret = acquire_flash(0)))
			return ret;
		memcpy(temp, (void *) (flash_base + offset - i), i);
		if (copy_from_user(temp+i, addr, j)) {
			release_flash();
			return -EFAULT;
		}
		/* Write first block */
		if (flash_write_single_block(offset / FLASH_BLOCKSIZE, temp)) {
			release_flash();
			return -EIO;
		}
		addr += j; count -= j; offset += j;
		release_flash();

		/* Writing middle blocks */
		for (i = 0; i < count / FLASH_BLOCKSIZE; i++) {
			if (copy_from_user(temp, addr, FLASH_BLOCKSIZE))
				return -EFAULT;
			if (flash_write_blocks(offset / FLASH_BLOCKSIZE, temp,
					       1))
				return -EIO;
			addr += FLASH_BLOCKSIZE;
			offset += FLASH_BLOCKSIZE;
		}
		count -= count / FLASH_BLOCKSIZE * FLASH_BLOCKSIZE;

		if (count) {
			if ((ret = acquire_flash(0)))
				return ret;
			if (copy_from_user(temp, addr, count)) {
				release_flash();
				return -EFAULT;
			}
			memcpy(temp + count,
			       (void *) (flash_base + offset + count),
			       FLASH_BLOCKSIZE - count);

			/* Write last block */
			ret = flash_write_single_block(offset/FLASH_BLOCKSIZE,
						       temp);
			release_flash();
		}
	}

	return 0;
}

/*
 * Initialize flash device. Ioremap() flash address-space
 */

int flash_init(void)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;
	bd_t *__bd = (bd_t *) __res;
	unsigned char Manuf_ID;
	unsigned char Device_Type;
	unsigned char Lock;
	char SerialLock;
	long long SerialNr;

	/* ReMap flash memory address range */
	flash_base = (void *) ioremap(__bd->true_flash_base, __bd->flash_size);
	if (!flash_base) {
		printk("SIpo flash: Failed to ioremap\n");
		return 1;
	} else
		flash_size = __bd->flash_size;

	if (__bd->true_flash_base != __bd->flash_base) {
		false_flash_base = (void *)
			ioremap(__bd->bi_memsize,
				__bd->bi_memtotal - __bd->bi_memsize);
		if (!false_flash_base) {
			printk("SIpo flash: Failed to ioremap\n");
			return 1;
		}
	}

	/* Initialize flash status pins */
	immap = (immap_t *) IMAP_ADDR;
	io = &immap->im_ioport;
	io->iop_pparc &= ~PIN_STS_LO;
	io->iop_pdirc &= ~PIN_STS_LO;
	io->iop_podrc |=  PIN_STS_LO;
	io->iop_psorc &= ~PIN_STS_LO;

	io->iop_pparc &= ~PIN_STS_HI;
	io->iop_pdirc &= ~PIN_STS_HI;
	io->iop_podrc |=  PIN_STS_HI;
	io->iop_psorc &=  PIN_STS_HI;

	SerialNr = flash_read_serial_number(&Manuf_ID, &Device_Type, &Lock,
					    &SerialLock);

	if (Lock & 0x01) {
		printk("Flash requires unlocking\n");
		/* If first Block is locked, clear lock bits */
		flash_clear_lock_bits();
		SerialNr = flash_read_serial_number(&Manuf_ID, &Device_Type,
						    &Lock, &SerialLock);

		if (Lock & 0x01) {
			printk("Unlocking failed\n");
			return -1;
		}
	}

	init_waitqueue_head(&flash_wait);

	return 0;
}

/******************************************************************************

 Flash interface routines.

******************************************************************************/

/*
 * Read from flash. Hold flash busy while reading from device. This function
 * is only called from MTD code so use blocking acquire_flash().
 */

int flash_read(char * buf, uint ppos, size_t count)
{
	int r;

	if (ppos > flash_size)
		return -1;

	if (count > flash_size - ppos)
		return -1;

	if ((r = acquire_flash(1)))
		return r;

#ifdef DEBUG
	printk("sIPoflash: read (off=0x%x, count=0x%x)\n", ppos, count);
#endif

	if (false_flash_base)
		memcpy(buf, false_flash_base + ppos, count);
	else
		memcpy(buf, flash_base + ppos, count);


	release_flash();

	return 0;
}

/*
 * Write to flash. This function is only called from MTD code so use blocking
 * acquire_flash().
 */

int flash_write(char *buf, uint ppos, size_t count)
{
	int r;

	if (ppos > flash_size)
		return -1;

	if (count > flash_size - ppos)
		return -1;

	if ((r = acquire_flash(1)))
		return r;

#ifdef DEBUG
	printk("sIPoflash: write (off=0x%x, count=0x%x)\n", ppos, count);
#endif

	if (false_flash_base) {
		memcpy(false_flash_base + ppos, buf, count);
	} else {
		flash_program_area(ppos, buf, count);

		if (memcmp(buf, flash_base+ppos, count)) {
			printk("sIPoflash: write error (off=0x%x, "
			       "count=0x%x)\n", ppos, count);

			release_flash();
			return -EIO;
		}
	}

	release_flash();

	return 0;
}

/*
  Erase flash area. Area must be block-aligned. This function is only called
  from MTD code so use blocking acquire_flash().

*/
int flash_erase(uint ppos, size_t count)
{
	int i, r;

	if (ppos > flash_size || (ppos & (FLASH_BLOCKSIZE - 1))) {
		printk("sIPoflash: flash_erase(%u, %u) - invalid ppos\n",
		       ppos, count);
		return -1;
	}

	if (count > flash_size - ppos || (count & (FLASH_BLOCKSIZE - 1))) {
		printk("sIPoflash: flash_erase(%u, %u) - invalid count\n",
		       ppos, count);
		return -1;
	}

	if ((r = acquire_flash(1))) {
		printk("flash_erase(%u,%u) - could not acquire flash\n",
		       ppos, count);
		return r;
	}

#ifdef DEBUG
	printk("sIPoflash: erase (off=0x%x, count=0x%x)\n", ppos, count);
#endif

	if (false_flash_base) {
		memset(false_flash_base + ppos, 0xff, count);
	} else {
		for (i = 0;i < count;i += FLASH_BLOCKSIZE)
			flash_erase_block((ppos + i) / FLASH_BLOCKSIZE);
	}

	release_flash();

	return 0;
}


/******************************************************************************

  Misc Character device interface for FLASH

******************************************************************************/

static int d_flash_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int d_flash_release(struct inode *inode, struct file *file)
{
	return 0;
}

static ssize_t d_flash_write(struct file *file, const char *buf, size_t count,
			     loff_t *ppos)
{
	if (count < 0)
		return -EIO;

	if (*ppos + count > flash_size)
		return -EIO;

	if (flash_cutcopypaste((unsigned) *ppos, (void *) buf, count))
		return -EIO;

	*ppos += count;
	return count;
}

static ssize_t d_flash_read(struct file *file, char *buf, size_t count,
			    loff_t *ppos)
{
	int r;

	if (*ppos > flash_size)
		return 0;

	if (count > flash_size - *ppos)
		count = flash_size - *ppos;

	if ((r = acquire_flash(0)))
		return r;

#ifdef DEBUG
	printk("sIPoflash: d_read (off=0x%x, count=0x%x)\n",
	       (uint) *ppos, count);
#endif

	if (copy_to_user(buf, (void *) (flash_base + *ppos), count))
		return -EFAULT;

	release_flash();

	*ppos += count;
	return count;
}

static loff_t d_flash_lseek(struct file * file, loff_t offset, int orig)
{
	switch (orig) {
	case 0:
		file->f_pos = offset;
		return file->f_pos;
	case 1:
		file->f_pos += offset;
		return file->f_pos;
	default:
		return -EINVAL;
	}
}

static int d_flash_ioctl(struct inode *node, struct file *file,
			 unsigned int cmd, unsigned long arg)
{
	unsigned int r;

	switch (cmd) {
	case IOCTL_FLASH_BLOCK_ERASE:
		/* File position must be inside flash and located at block
		 * boundary */
		if (file->f_pos >= flash_size ||
		    (file->f_pos & (FLASH_BLOCKSIZE - 1)))
			return -EIO;

		if ((r = acquire_flash(0)))
			return r;

		flash_erase_block(file->f_pos / FLASH_BLOCKSIZE);

		release_flash();

		return 0;
	default:
		return -EINVAL;
	}
}


static struct file_operations d_flash_fops =
{
	owner:          THIS_MODULE,
	llseek:         d_flash_lseek,
	read:           d_flash_read,
	write:          d_flash_write,
	open:           d_flash_open,
	ioctl:          d_flash_ioctl,
	release:        d_flash_release,
};

static struct miscdevice d_flash_miscdev =
{
	SIPOFLASH_MINOR,
	"sipoflash",
	&d_flash_fops
};

/*
 * Initialize Flash HW
 */
void __init sipoflash_init(void)
{
	misc_register(&d_flash_miscdev);
	printk("SIpo flash: 0.6\n");
}
