/*
 *      sIPoLeds:  A led "device driver" for sIPo box
 *
 *      Copyright 2000-2002 SSH Communications Security Corp,
 *      All Rights Reserved. http://www.ssh.com
 *
 *      18.3.2002 Jouni Malinen <jkm@ssh.com>
 *              Added special "blink all leds" states
 *
 *      3.8.2001 Marko Teiste <mte@ssh.com>
 *              Added support for fatal/non-fatal status coding
 *
 *      24.8.2000 Marko Teiste <mte@ssh.com>
 *              Initial version
 */
#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/mpc8260.h>

#define LM_LED_POWER    0
#define LM_LED_STATUS   1
#define LM_LED_BLUE_0   3
#define LM_LED_BLUE_1   4
#define LM_LED_BLUE_2   2
#define LM_LED_BLUE_3   5
#define LM_LED_COMMAND  6

#define NUM_NF          6
#define SEQ_LENGTH      20

static struct semaphore led_mutex;

#define LED_LOCK_INIT() sema_init(&led_mutex, 1);
#define LED_LOCK()      down(&led_mutex);
#define LED_UNLOCK()    up(&led_mutex);

static int blinkingpower, blink_all;

static char ledmode[6];
static struct timer_list led_timer;
static int nfptr;
static int nfmax;
static int nfadd;
static char nfcodes[NUM_NF][6];

static int led_count;

static char buffer[7];
static enum { start, collect, wait } state;
static int bptr;

/*
 *      There is actually nothing to open
 */
static int ledshow_open(struct inode *inode, struct file *file)
{
	state = start;
	bptr = 0;

	return 0;
}

/*
 *      Or to release either
 */
static int ledshow_release(struct inode *inode, struct file *file)
{
	return 0;
}

/*
 *  Read led code from user
 */

static inline void ledshow_write_newline(void)
{
	int j, k;

	LED_LOCK();
	switch (buffer[LM_LED_COMMAND]) {
	case '!':
		/* Set default and erase non fatal codes */
		nfptr = NUM_NF;
		nfmax = 0;
		nfadd = 0;
		/* fall through */
	case '=':       /* Set default led mode */
		memcpy(ledmode, buffer, 6);
		break;
	case '+':       /* Add to non fatal codes */
		/* Check if code already exists */
		for (j = 0; j < nfmax && (0 != memcmp(buffer, nfcodes[j], 6));
		     j++)
			; /* VOID */

		if (j == nfmax) { /* nope, OK to add */
			memcpy(nfcodes[nfadd], buffer, 6);

			if (nfadd < (NUM_NF - 1)) {
				nfadd++;
				if (nfmax < NUM_NF)
					nfmax = nfadd;
			} else {
				nfadd = 0;
				nfmax = NUM_NF;
			}
		}
		break;
	case '-':     /* remove from non fatal codes */
		for (j = 0; j < nfmax && (0 != memcmp(buffer, nfcodes[j], 6));
		     j++)
			; /* VOID */
		if (j != nfmax) {
			/* Found the code? */
			/* removing resets the sequence */
			nfptr = NUM_NF;
			led_count = SEQ_LENGTH;

			/* Remove entry from table */
			for (k = j; k < (nfmax - 1); k++)
				memcpy(nfcodes[k], nfcodes[k + 1], 6);
			nfmax--;
			if (nfadd == 0)
				nfadd = nfmax;
			else
				nfadd--;
		}
		break;
	}
	LED_UNLOCK();
}


static ssize_t ledshow_write(struct file *file, const char *buf, size_t len,
			     loff_t *ppos)
{
	int i;
	char data[8];

	if (len < 0)
		return -EIO;

	if ((len - *ppos) > 8)
		len = 8;

	if (copy_from_user(data, buf, len))
		return -EFAULT;

	*ppos += len;

	for (i = 0; i < len; i++) {
		switch (data[i])
		{
		case 'Q': blink_all = 0; break;
		case 'W': blink_all = 1; break;
		case 'E': blink_all = 2; break;
		case '-':
		case '+':
		case '=':
		case '!':
			bptr = 0;
			state = collect;
			buffer[LM_LED_COMMAND] = data[i];
			break;
		case '\n':
			if (wait == state && (6 == bptr))
				ledshow_write_newline();
			state = start;
			bptr = 0;
			buffer[LM_LED_COMMAND] = ' ';
			break;
		default:
			if ((collect == state) &&
			    ((data[i] >= '0') && (data[i] <= '9'))
			    && (bptr < 6))
				buffer[bptr++] = data[i];
			if (bptr >= 6)
				state = wait;
			break;
		}
	}

	return len;
}

static ssize_t ledshow_read(struct file * file, char * buf, size_t len,
			    loff_t *ppos)
{
	/*
	 *      Tell user what the led status is
	 */
	char status[80];
	int i, k;

	if (len < 0)
		return -EIO;

	status[0] = '=';
	memcpy(&status[1], ledmode, 6);
	status[7] = '\n';
	for (k = 8, i = 0; i < nfmax; i++) {
		status[k++] = '+';
		memcpy(&status[k], nfcodes[i], 6);
		k += 6;
		status[k++] = '\n';
	}

	if (*ppos > k)
		return 0;

	if (len > (k - *ppos))
		len = k - *ppos;

	if (copy_to_user(buf, status, len))
		return -EFAULT;

	*ppos += len;

	return len;
}

static void
led_activity(unsigned long data)
{
	char mode[6];
	int alternate;

	LED_LOCK();
	led_count--;
	if (led_count < 0) {  /* switch code ? */
		if ((NUM_NF == nfptr) && nfmax)
			nfptr = 0;
		else if (nfptr < (nfmax - 1))
			nfptr++;
		else
			nfptr = NUM_NF;

		led_count = SEQ_LENGTH - 1;
	}

	if (nfptr < NUM_NF)
		memcpy(mode, nfcodes[nfptr], 6);
	else
		memcpy(mode, ledmode, 6);

	/* if any of the blue leds is affected, take led control away
	   from the fcc driver */
	if ((mode[LM_LED_BLUE_0] != '0') ||
	    (mode[LM_LED_BLUE_1] != '0') ||
	    (mode[LM_LED_BLUE_2] != '0') ||
	    (mode[LM_LED_BLUE_3] != '0') ||
	    blink_all)
		sipo_fcc_led_show = 0;
	else
		sipo_fcc_led_show = 1;

	alternate = led_count % 2;

	if (blink_all) {
		sipo_power_led_set(alternate ? 0 : 1);
		sipo_status_led_set(alternate ? 0 : 3);
		if (alternate) {
			sipo_blue_led_off(0);
			sipo_blue_led_off(1);
			sipo_blue_led_off(2);
			sipo_blue_led_off(3);
		} else {
			sipo_blue_led_on(0);
			sipo_blue_led_on(1);
			sipo_blue_led_on(2);
			sipo_blue_led_on(3);
		}

		goto done;
	}

	/* Power led 0 = off, 1 = on, 2 = blinking, unless overridden with
	 * blinking power
	 */
	if (blinkingpower)
		sipo_power_led_set(alternate ? 0 : 1);
	else {
		switch (mode[LM_LED_POWER]) {
		case '0':
			sipo_power_led_set(0);
			break;
		case '1':
			sipo_power_led_set(1);
			break;
		case '2':
			sipo_power_led_set(alternate ? 0 : 1);
			break;
		default:
			sipo_power_led_set(1);
			break;
		}
	}

	/* Status led 0 = off, 1 = green, 2 = yellow, 3 = red,
	   4 = blinking green, 5 =  blinking yellow, 6 = blinking red,
	   7 = green-yellow, 8 = green-red, 9 = yellow-red */
	switch (mode[LM_LED_STATUS]) {
	case '0':
		sipo_status_led_set(0);
		break;
	case '1':
		sipo_status_led_set(2);
		break;
	case '2':
		sipo_status_led_set(3);
		break;
	case '3':
		sipo_status_led_set(1);
		break;
	case '4':
		sipo_status_led_set(alternate ? 0 : 2);
		break;
	case '5':
		sipo_status_led_set(alternate ? 0 : 3);
		break;
	case '6':
		sipo_status_led_set(alternate ? 0 : 1);
		break;
	case '7':
		sipo_status_led_set(alternate ? 2 : 3);
		break;
	case '8':
		sipo_status_led_set(alternate ? 2 : 1);
		break;
	case '9':
		sipo_status_led_set(alternate ? 3 : 1);
		break;
	default:
		sipo_status_led_set(1);
	}

	/* Blue leds, 0 = off, 1 = on, 2 = blinking */
	switch (mode[LM_LED_BLUE_0]) {
	case '0':
		if (!sipo_fcc_led_show)
			sipo_blue_led_off(0);
		break;
	case '1':
		sipo_blue_led_on(0);
		break;
	case '2':
		if (alternate)
			sipo_blue_led_off(0);
		else
			sipo_blue_led_on(0);
		break;
	default:
		sipo_blue_led_on(0);
	}

	switch (mode[LM_LED_BLUE_1]) {
	case '0':
		if (!sipo_fcc_led_show)
			sipo_blue_led_off(1);
		break;
	case '1':
		sipo_blue_led_on(1);
		break;
	case '2':
		if (alternate)
			sipo_blue_led_off(1);
		else
			sipo_blue_led_on(1);
		break;
	default:
		sipo_blue_led_on(1);
	}

	switch (mode[LM_LED_BLUE_2]) {
	case '0':
		if (!sipo_fcc_led_show)
			sipo_blue_led_off(2);
		break;
	case '1':
		sipo_blue_led_on(2);
		break;
	case '2':
		if (alternate)
			sipo_blue_led_off(2);
		else
			sipo_blue_led_on(2);
		break;
	default:
		sipo_blue_led_on(2);
	}

	switch (mode[LM_LED_BLUE_3]) {
	case '0':
		sipo_blue_led_off(3);
		break;
	case '1':
		sipo_blue_led_on(3);
		break;
	case '2':
		if (alternate)
			sipo_blue_led_off(3);
		else
			sipo_blue_led_on(3);
		break;
	default:
		sipo_blue_led_on(3);
	}

 done:
	LED_UNLOCK();

	mod_timer(&led_timer, jiffies + HZ / (blink_all == 2 ? 4 : 2));
}

static struct file_operations ledshow_fops =
{
	owner:          THIS_MODULE,
	write:          ledshow_write,
	read:           ledshow_read,
	open:           ledshow_open,
	release:        ledshow_release,
};

static struct miscdevice ledshow_miscdev=
{
	SIPOLED_MINOR,
	"sipoled",
	&ledshow_fops
};

void sipoled_power_blink(int n)
{
	blinkingpower = n;
}

void __init sipoled_init(void)
{
	bd_t *__bd = (bd_t *) __res;

	memset(ledmode, '0', 6);
	ledmode[0] = '1';
	blinkingpower = 0;
	if (__bd->boot_state == nv_prepare || __bd->boot_state == nv_test ||
	    __bd->boot_state == nv_prepare_part2)
		blink_all = 1;
	else
		blink_all = 0;
	nfptr = NUM_NF;
	nfmax = 0;
	nfadd = 0;
	led_count = SEQ_LENGTH;
	LED_LOCK_INIT();

	misc_register(&ledshow_miscdev);
	printk("sIPo LedShow: 1.02\n");
	/*
	 * setup timer for led activity
	 */
	init_timer(&led_timer);
	led_timer.expires = jiffies + HZ / 2;
	led_timer.function = led_activity;
	add_timer(&led_timer);
}

#ifdef MODULE
int init_module(void)
{
	sipoled_init();
	return 0;
}

void cleanup_module(void)
{
	misc_deregister(&ledshow_miscdev);
}
#endif
