/*
 *      SipO FLASH ROM block device
 *
 *      Copyright 2001 SSH Communications Security Corp, All Rights Reserved.
 *                              http://www.ssh.com
 *
 *      15.2.2001 Arto Vuori <avuori@ssh.com>
 *              Initial version
 *
 *      Block device driver for read-only access on SipO FLASH partitions.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/reboot.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/fd.h>
#include <linux/blkpg.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/sipo.h>
#include <asm/immap_8260.h>
#include <asm/cpm_8260.h>
#include <asm/machdep.h>
#include <asm/mpc8260.h>
#include <asm/byteorder.h>

#define MAJOR_NR SIPO_ROMDISK_MAJOR

#define DEVICE_NAME "romdisk"
#define DEVICE_REQUEST d_make_request
#define DEVICE_NR(device) (MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)
#define DEVICE_NO_RANDOM
#define TIMEOUT_VALUE (HZ)
#include <linux/blk.h>

#define NUM_ROMDISKS pt_size

/* Various static variables go here.  Most are used only in the RAM disk code.
 */

static unsigned long d_length[NUM_ROMDISKS]; /* Size of RAM disks in bytes   */
static int d_hardsec[NUM_ROMDISKS];          /* Size of real blocks in bytes */
static int d_blocksizes[NUM_ROMDISKS];       /* Size of 1024 byte blocks :)  */
static int d_kbsize[NUM_ROMDISKS];           /* Size in blocks of 1024 bytes */
static uint d_base[NUM_ROMDISKS];
static devfs_handle_t devfs_handle;

static int d_blocksize = BLOCK_SIZE;         /* blocksize of the RAM disks */

static int __init romdisk_blocksize(char *str)
{
	d_blocksize = simple_strtol(str, NULL, 0);
	return 1;
}

__setup("romdisk_blocksize=", romdisk_blocksize);

static void d_make_request(request_queue_t *q)
{
	int nr;
	unsigned long offset, len;

	while (1) {
		INIT_REQUEST;

		nr = DEVICE_NR(CURRENT->rq_dev);

		if (nr >= NUM_ROMDISKS) {
		    printk(KERN_ERR DEVICE_NAME ": bad device\n");
		    end_request(0);
		    continue;
		}

		offset = CURRENT->sector << 9;
		len = CURRENT->current_nr_sectors << 9;

		if (offset + len > d_length[nr]) {
			printk(KERN_ERR DEVICE_NAME ": bad access: block=%ld, "
			       "count=%ld\n",
			       CURRENT->sector,
			       CURRENT->current_nr_sectors);
			end_request(0);
			continue;
		}

		if (CURRENT->cmd != READ) {
			printk(KERN_ERR DEVICE_NAME ": bad command: %d\n",
			       CURRENT->cmd);
			end_request(0);
			continue;
		}

		if (flash_read(CURRENT->buffer, d_base[nr] + offset, len)) {
			printk(KERN_ERR DEVICE_NAME ": Read failure: %d\n",
			       CURRENT->cmd );
			end_request(0);
			continue;
		}

		end_request(1);
	}
}

static int d_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		   unsigned long arg)
{
	unsigned int intval;
	unsigned int nr;

	if (!inode || !inode->i_rdev)
		return -EINVAL;

	nr = DEVICE_NR(inode->i_rdev);

	switch (cmd) {
	case BLKGETSIZE:   /* Return device size */
		if (!arg)
			return -EINVAL;
		return put_user(d_kbsize[nr] << 1, (long *) arg);

	/* ROMDISK is read only */
	case BLKROGET:
		intval = 1;
		return put_user(intval, (int *) arg);

	case BLKROSET:
		if (!arg)
			return -EINVAL;
		return 0;

	case BLKSSZGET:
		return blk_ioctl(inode->i_rdev, cmd, arg);

	default:
		return -EINVAL;
	};

	return 0;
}

static int d_open(struct inode * inode, struct file * filp)
{
	uint nr;
	bd_t *__bd = (bd_t *) __res;

	nr = DEVICE_NR(inode->i_rdev);

	set_device_ro(inode->i_rdev, 1);

	if (nr >= NUM_ROMDISKS)
		return -ENODEV;

	/* Check if partition exist */
	if (__bd->pt[nr].type == 0)
		return -ENODEV;

	/* Re-read size & offset */
	d_length[nr] = __bd->pt[nr].size;
	d_kbsize[nr] = (__bd->pt[nr].size >> 10);
	d_base[nr] = __bd->pt[nr].offset;

	MOD_INC_USE_COUNT;

	return 0;
}

static int d_release(struct inode *inode, struct file *filp)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

static struct block_device_operations fd_fops =
{
	open:           d_open,
	release:        d_release,
	ioctl:          d_ioctl,
};

/*
 * This is the registration and initialization section of the ROM disk driver
 */

int __init sipo_romdisk_init(void)
{
	bd_t *__bd = (bd_t *) __res;
	int i;

	if (d_blocksize > PAGE_SIZE || d_blocksize < 512 ||
	    (d_blocksize & (d_blocksize - 1))) {
		printk("SipO ROMDISK: wrong blocksize %d, reverting to "
		       "defaults\n",
		       d_blocksize);
		d_blocksize = BLOCK_SIZE;
	}

	if (register_blkdev(MAJOR_NR, "romdisk", &fd_fops)) {
		printk("SipO ROMDISK: Could not get major %d", MAJOR_NR);
		return -EIO;
	}

	devfs_handle = devfs_mk_dir (NULL, "romd", NULL);
	devfs_register_series (devfs_handle, "%u", NUM_ROMDISKS,
			       DEVFS_FL_DEFAULT, SIPO_ROMDISK_MAJOR, 0,
			       S_IFBLK | S_IRUSR,
			       &fd_fops, NULL);

	blk_init_queue(BLK_DEFAULT_QUEUE(MAJOR_NR), DEVICE_REQUEST);

	for (i = 0; i < NUM_ROMDISKS; i++) {
		d_length[i] = __bd->pt[i].size;
		d_hardsec[i] = d_blocksize;
		d_blocksizes[i] = d_blocksize;
		d_kbsize[i] = __bd->pt[i].size >> 10;
		d_base[i] = __bd->pt[i].offset;
	}
	hardsect_size[MAJOR_NR] = d_hardsec; /* Size of the RAM disk blocks */
	blksize_size[MAJOR_NR] = d_blocksizes;/* Avoid set_blocksize() check */
	blk_size[MAJOR_NR] = d_kbsize;	 /* Size of the RAM disk in kB  */


	for (i = 0; i < NUM_ROMDISKS; i++)
		register_disk(NULL, MKDEV(MAJOR_NR, i), 1, &fd_fops,
			      d_kbsize[i] << 1);

	printk("SipO ROMDISK driver initialized\n");

	return 0;
}
