/*
 *      sIPoScard 0.01: A smartcard "device driver" for sIPo box
 *
 *      Copyright 2000 SSH Communications Security Corp, All Rights Reserved.
 *                              http://www.ssh.com
 *
 *      4.9.2000 Karri Kaksonen <karrik@ssh.com>
 *              Interruptable version
 */
#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/reboot.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/immap_8260.h>
#include <asm/mpc8260.h>
#include <asm/cpm_8260.h>
#include <asm/irq.h>

#ifdef MODULE
MODULE_PARM(soft_margin,"i");
#endif

typedef struct serial_info {
	/* CPM Buffer Descriptor pointers.
	 */
	cbd_t                   *rx_bd_base;
	cbd_t                   *rx_cur;
	cbd_t                   *tx_bd_base;
	cbd_t                   *tx_cur;
	/* Wait queues needed for interruptable driver */
	wait_queue_head_t       scard_wait;
	/* Number of users. Should be 0 or 1 */
	int                     scard_users;
	/* A shared variable for signalling the waiting driver */
	int                     scard_event;
	uint                    scard_clk_freq;
	uint                    scard_baudrate;
} ser_info_t;

static ser_info_t *info;

/*
  Set Baudrate generator
*/
#define BRG_INT_CLK     (133333333/2)
#define BRG_UART_CLK    (BRG_INT_CLK/16)
#define CPM_BRG_EN      ((uint)0x00010000)

/*
  Code for providing the clock to a SmartCard.
  The clock has to be turned off when the SmartCard
  is not in use. Otherwise it gets enough current to
  continue operation even if the Vcc is off.
*/
static void
setscardclock(int rate, uint active)
{
	volatile uint *bp;

	bp = &immr->im_brgc8;
	*bp = ((BRG_INT_CLK / rate) << 1) | active;
}

/*
  The baudrate generator used for data communications
*/
static void
setbrg(int rate)
{
	volatile uint *bp;

	bp = &immr->im_brgc4;
	*bp = ((BRG_UART_CLK / rate) << 1) | CPM_BRG_EN;
}

#define RX_NUM_FIFO     16
#define RX_BUF_SIZE     32
#define TX_NUM_FIFO     16
#define TX_BUF_SIZE     32

/*
  The interrupt routine will just inform the driver that something has
  happened.
*/
static void scard_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	u_char events;
	volatile scc_t *sccp;
	volatile cbd_t *bpd;

	sccp = &immr->im_scc[3];
	events = sccp->scc_scce;
	if (events & SCCM_RX) {
		bpd = info->rx_cur;
		if (bpd->cbd_sc & BD_SC_EMPTY) {
			sccp->scc_scce = events;
			return;
		}
		info->scard_event = 1;
	}
	sccp->scc_scce = events;
	/*
	  If it was not a receive interrupt then wake up the sleeping task.
	*/
	wake_up_interruptible(&info->scard_wait);
}

/*
  Check for SmartCard in reader.
*/
static int scard_missing(void)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;

	immap = immr;
	io = &immap->im_ioport;

	if (!(io->iop_pdatc & 0x00100000))
		return -ENOMEDIUM;       /* No SmartCard in the reader */
	return 0;
}

/*
 *      Reset the card and setup everything.
 */
static int scard_open(struct inode *inode, struct file *file)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;
	volatile scc_t *scp;

	/* Allow only one user at a time. The scard has a concept of
	   "current file" so multiple simultaneous users won't work
	*/
	if (info->scard_users)
		return -EBUSY;

	info->scard_users = 1;

	/* All cards start up with a default speed of 9600 */
	info->scard_baudrate = 9600;
	info->scard_clk_freq = 372 * 9600;

	MOD_INC_USE_COUNT;

	immap = immr;
	io = &immap->im_ioport;

	if (scard_missing()) {
		info->scard_users = 0;
		MOD_DEC_USE_COUNT;
		return -ENOMEDIUM;       /* No SmartCard in the reader */
	}

	/* Set up the SmartCard clock. */
	setscardclock(info->scard_clk_freq, CPM_BRG_EN);
	/* Set up the baud rate generator. */
	setbrg(info->scard_baudrate);

	io->iop_ppard &= ~0x08000000; /* smartc_clc to io-pin */
	io->iop_pdatd &= ~0x00000800; /* smartc_vcc = 1 */
	io->iop_ppard |= 0x08000000;  /* smartc_clc to brg8 */
	scp = &immap->im_scc[3];

	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(HZ / 100);
	io->iop_pdatd &= ~0x00000100; /* smartc_rst  = 1 */
	scp->scc_gsmrl = (SCC_GSMRL_MODE_UART | SCC_GSMRL_TDCR_16 |
			  SCC_GSMRL_RDCR_16 | SCC_GSMRL_ENR | SCC_GSMRL_ENT);

	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(60 * HZ/100); /* About 60ms delay for welcome msg */
	return 0;
}

/*
  Close the SmartCard.
*/
static int scard_release(struct inode *inode, struct file *file)
{
	volatile cbd_t *bpd;
	volatile immap_t *immap;
	volatile iop8260_t *io;

	/*
	  Wait until the transmit buffer is empty before closing down the
	  device.
	*/
	immap = immr;
	io = &immap->im_ioport;
	bpd = info->tx_bd_base;
	while (1) {
		while (bpd->cbd_sc & BD_SC_READY) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(HZ / 100);
		}
		if (bpd->cbd_sc & BD_SC_WRAP)
			break;
		bpd++;
	}
	/*
	  Free device
	*/
	info->scard_users = 0;
	MOD_DEC_USE_COUNT;
	/*
	  Turn power off
	*/
	io->iop_ppard &= ~0x08000000; /* smartc_clc to brg8 */
	io->iop_pdatd &= ~0x08000000; /* smartc_clc = 0 */
	io->iop_pdatd |= 0x00000100; /* smartc_rst  = 0 */
	io->iop_pdatd |= 0x00000800; /* smartc_vcc = 0 */
	return 0;
}

static ssize_t scard_write(struct file *file, const char *data, size_t len,
			   loff_t *ppos)
{
	volatile cbd_t *bpd;
	size_t c, i;
	ssize_t ret;
	volatile immap_t *immap;
	volatile iop8260_t *io;
	char *cp;

	/*
	  Has someone removed the card?
	*/
	if (scard_missing())
		return -ENOMEDIUM;

	immap = immr;
	io = &immap->im_ioport;

	bpd = info->tx_cur;
	ret = 0;
	while (1) {
		c = (len < TX_BUF_SIZE) ? len : TX_BUF_SIZE;
		if (c <= 0)
			break;
		while (bpd->cbd_sc & BD_SC_READY) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(HZ/100);
		}
		cp = __va(bpd->cbd_bufaddr);
		for (i = 0; i < c; i++) {
			*cp++ = *data++;
		}
		bpd->cbd_datlen = c;
		bpd->cbd_sc |= BD_SC_READY;
		len -= c;
		ret += c;
		if (bpd->cbd_sc & BD_SC_WRAP)
			bpd = info->tx_bd_base;
		else
			bpd++;
		info->tx_cur = (cbd_t *)bpd;
	}
	return ret;
}

static ssize_t scard_read(struct file * file, char * buf, size_t count,
			  loff_t *ppos)
{
	volatile cbd_t *bpd;
	size_t ret, i;
	DECLARE_WAITQUEUE(wait, current);

	/*
	  Tag us to the wait list. Don't sleep yet.
	*/
	add_wait_queue(&info->scard_wait, &wait);
	/*
	  Tell kernel that we don't want to be rescheduled for a while.
	*/
	current->state = TASK_INTERRUPTIBLE;

	/*
	  Check if there is old data in the buffers
	*/
	bpd = info->rx_cur;
	if (!(bpd->cbd_sc & BD_SC_EMPTY))
		info->scard_event = 1;
	/*
	  Has an event occurred?
	*/
	while (!info->scard_event) {
		/*
		  Has someone removed the card?
		*/
		if (scard_missing()) {
			remove_wait_queue(&info->scard_wait, &wait);
			current->state = TASK_RUNNING;
			return -ENOMEDIUM;
		}
		/*
		  Should we return immediately if the event is not here?
		*/
		if (file->f_flags & O_NDELAY) {
			remove_wait_queue(&info->scard_wait, &wait);
			current->state = TASK_RUNNING;
			return -EWOULDBLOCK;
		}
		/*
		  Did someone kill us? Then DIE!
		*/
		if (signal_pending(current)) {
			remove_wait_queue(&info->scard_wait, &wait);
			current->state = TASK_RUNNING;
			return -ERESTARTSYS;
		}
		/*
		  Good grief. Nothing is happening. Time to take a nap.
		*/
		schedule();
		current->state = TASK_INTERRUPTIBLE;
	}
	/*
	  We got an event!
	*/
	remove_wait_queue(&info->scard_wait, &wait);
	info->scard_event = 0;
	current->state = TASK_RUNNING;

	/*
	  Has someone removed the card?
	*/
	if (scard_missing())
		return -ENOMEDIUM;

	/* Just loop through the closed BDs and copy the characters into
	 * the buffer.
	 */
	bpd = info->rx_cur;
	ret = 0;
	for (;;) {
		if ((bpd->cbd_sc & BD_SC_EMPTY) || !count)
			break;
		/* Get the number of characters and the buffer pointer. */
		i = bpd->cbd_datlen;
		if (count < i) {
			/*
			  Special case. We want to read only part of
			  the buffer.  As this is never the case in
			  Sipo I am not going to go through the
			  trouble of implementing buffer offsets
			  etc. So a partial read will lose eventual
			  remainding characters in the buffer.  I
			  will, however implement the count so that
			  data structures in tha calling program won't
			  be corrupted.
			*/
			i = count;
		}
		memcpy(buf, __va(bpd->cbd_bufaddr), i);
		buf += i;
		ret += i;
		count -= i;
		bpd->cbd_sc |= BD_SC_EMPTY;
		bpd->cbd_sc &= ~(BD_SC_BR | BD_SC_FR | BD_SC_PR | BD_SC_OV);
		if (bpd->cbd_sc & BD_SC_WRAP)
			bpd = info->rx_bd_base;
		else
			bpd++;
		info->rx_cur = (cbd_t *)bpd;
	}
	return ret;
}

static unsigned int scard_poll(struct file *file, poll_table *wait)
{
	poll_wait(file, &(info->scard_wait), wait);
	/*
	  Has someone removed the card?
	*/
	if (scard_missing()) {
		return -ENOMEDIUM;
	}

	if (info->scard_event) {
		info->scard_event = 0; /* !!! FIX */
		return POLLIN | POLLRDNORM;
	}
	return 0;
}

#define IOCTL_SCARD_BAUDRATE 0x8800

static int scard_ioctl(struct inode *inode, struct file *file,
		       unsigned int cmd, unsigned long arg)
{
	uint retval = 0;
	uint payload[2];

	switch(cmd) {
	case IOCTL_SCARD_BAUDRATE:
		if (copy_from_user(&payload, (uint *) arg, 2 * (sizeof(uint))))
			return -EFAULT;
		info->scard_baudrate = payload[0];
		info->scard_clk_freq = payload[1];
#ifdef DEBUG
		printk("SIpo scard ioctl: Set baudrate to ");
		printk("%d frequency to %d\n", info->scard_baudrate,
		       info->scard_clk_freq);
#endif
		/* Set up the SmartCard clock. */
		setscardclock(info->scard_clk_freq, CPM_BRG_EN);
		/* Set up the baud rate generator. */
		setbrg(info->scard_baudrate);
		break;

	default:
		return -EINVAL;
	}

	return retval;
}

static struct file_operations scard_fops=
{
	owner:          THIS_MODULE,
	read:           scard_read,
	write:          scard_write,
	poll:           scard_poll,
	open:           scard_open,
	release:        scard_release,
	ioctl:          scard_ioctl
};

static struct miscdevice scard_miscdev=
{
	SIPOSCARD_MINOR,
	"siposcard",
	&scard_fops
};

int __init scard_init(void)
{
	volatile cpm8260_t *cp;
	volatile immap_t *immap;
	volatile iop8260_t *io;
	volatile cbd_t *bdp;
	volatile scc_t *scp;
	volatile scc_uart_t *sup;
	uint mem_addr, dp_addr, page, sblock;
	int j;

	immap = immr;
	cp = &immap->im_cpm;
	io = &immap->im_ioport;

	io->iop_pdirc &= ~0x00100000; /* smartc_present input */
	io->iop_pdird |= 0x00000800; /* smartc_vcc_on output */
	io->iop_pdird |= 0x00000100; /* smartc_rst output */
	io->iop_pdird |= 0x08000000; /* smartc_clc output */
	io->iop_ppard |= 0x00000600; /* smartc_rxd smartc_txd to scc */
	io->iop_podrd &= ~0x00000600; /* smartc_rxd smartc_txd no open drain */
	io->iop_psord &= ~0x00000600; /* smartc_rxd smartc_txd psords = 0 */
	io->iop_pdird |= 0x00000400; /* smartc_txd output */
	io->iop_pdird &= ~0x00000200; /* smartc_rxd input */

	io->iop_ppard |= 0x08000000; /* smartc_clc to brg8 */

	/* Connect SCC4 to NMSI.  Connect BRG4 to SCC4.
	 */
	immap->im_cpmux.cmx_scr &= ~0x000000ff;
	immap->im_cpmux.cmx_scr |= 0x0000001b;

	misc_register(&scard_miscdev);
	printk("sIPO SmartCard: 0.01\n");

	info = kmalloc(sizeof(ser_info_t), GFP_KERNEL);
	if (!info) {
		printk("Cannot kmalloc info\n");
		return -1;
	}
	__clear_user(info,sizeof(ser_info_t));

	/* We need to allocate a transmit and receive buffer
	 * descriptors from dual port ram, and a character
	 * buffer area from host mem.
	 */
	dp_addr = m8260_cpm_dpalloc(sizeof(cbd_t) * RX_NUM_FIFO, 8);

	/* Allocate space for FIFOs in the host memory. */
	mem_addr = m8260_cpm_hostalloc(RX_NUM_FIFO * RX_BUF_SIZE, 1);

	/* Set the physical address of the host memory
	 * buffers in the buffer descriptors, and the
	 * virtual address for us to work with.
	 */
	bdp = (cbd_t *)&immap->im_dprambase[dp_addr];
	info->rx_cur = info->rx_bd_base = (cbd_t *)bdp;

	for (j = 0; j < (RX_NUM_FIFO - 1); j++) {
		bdp->cbd_bufaddr = __pa(mem_addr);
		bdp->cbd_sc = BD_SC_EMPTY | BD_SC_INTRPT;
		mem_addr += RX_BUF_SIZE;
		bdp++;
	}
	bdp->cbd_bufaddr = __pa(mem_addr);
	bdp->cbd_sc = BD_SC_WRAP | BD_SC_EMPTY | BD_SC_INTRPT;

	scp = &immap->im_scc[3];
	sup = (scc_uart_t *)&immap->im_dprambase[PROFF_SCC4];
	sup->scc_genscc.scc_rbase = dp_addr;

	dp_addr = m8260_cpm_dpalloc(sizeof(cbd_t) * TX_NUM_FIFO, 8);

	/* Allocate space for FIFOs in the host memory. */
	mem_addr = m8260_cpm_hostalloc(TX_NUM_FIFO * TX_BUF_SIZE, 1);

	/* Set the physical address of the host memory
	 * buffers in the buffer descriptors, and the
	 * virtual address for us to work with.
	 */
	bdp = (cbd_t *)&immap->im_dprambase[dp_addr];
	info->tx_cur = info->tx_bd_base = (cbd_t *)bdp;

	for (j = 0; j < (TX_NUM_FIFO - 1); j++) {
		bdp->cbd_bufaddr = __pa(mem_addr);
		bdp->cbd_sc = BD_SC_INTRPT;
		mem_addr += TX_BUF_SIZE;
		bdp++;
	}
	bdp->cbd_bufaddr = __pa(mem_addr);
	bdp->cbd_sc = (BD_SC_WRAP | BD_SC_INTRPT);

	sup->scc_genscc.scc_tbase = dp_addr;

	/* Send the CPM an initialize command.
	 */
	page = CPM_CR_SCC4_PAGE;
	sblock = CPM_CR_SCC4_SBLOCK;

	cp->cp_cpcr = mk_cr_cmd(page, sblock, 0, CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);

	/* Set up the uart parameters in the
	 * parameter ram.
	 */
	sup->scc_genscc.scc_rfcr = CPMFCR_GBL | CPMFCR_EB;
	sup->scc_genscc.scc_tfcr = CPMFCR_GBL | CPMFCR_EB;

	/* Set this to 1 for now, so we get single
	 * character interrupts.  Using idle charater
	 * time requires some additional tuning.
	 */
	sup->scc_genscc.scc_mrblr = RX_BUF_SIZE;
	sup->scc_maxidl = 100;
	sup->scc_brkcr = 1;
	sup->scc_parec = 0;
	sup->scc_frmec = 0;
	sup->scc_nosec = 0;
	sup->scc_brkec = 0;
	sup->scc_uaddr1 = 0;
	sup->scc_uaddr2 = 0;
	sup->scc_toseq = 0;
	sup->scc_char1 = 0x8000;
	sup->scc_char2 = 0x8000;
	sup->scc_char3 = 0x8000;
	sup->scc_char4 = 0x8000;
	sup->scc_char5 = 0x8000;
	sup->scc_char6 = 0x8000;
	sup->scc_char7 = 0x8000;
	sup->scc_char8 = 0x8000;
	sup->scc_rccm = 0xc0ff;

	/* !!! step 17 & 18 missing */

	scp->scc_scce = 0xffff; /* Clear prev. events */

	request_8xxirq(SIU_INT_SCC4, scard_interrupt, 0, "scard", info);

	scp->scc_sccm = 3;      /* 3 = Allow Tx and Rx interrupts */

	/* !!! step 21 missing */

	scp->scc_gsmrh = SCC_GSMRH_RFW;
	scp->scc_gsmrl = (SCC_GSMRL_MODE_UART | SCC_GSMRL_TDCR_16 |
			  SCC_GSMRL_RDCR_16);

	scp->scc_dsr = 0x7e7e;
	/* PSMR (Protocol-Specific Mode Register):
	 * 0111000000011010 = 0x701a:
	 * FLC=0 (normal flow control)
	 * SL=1 (two stop bits)
	 * CL=11 (8 data bits)
	 * UM=00 (normal UART operation)
	 * FRZ=0 (normal freeze transmission operation)
	 * RZS=0 (normally, but at least one stop bit between characters)
	 * SYN=0 (normal asynch.)
	 * DRT=0 (normal, do not disable receiver while transmitting)
	 * RES=0 (reserved)
	 * PEN=1 (enable parity)
	 * RPM=10 (even receiver parity)
	 * TPM=10 (even transmitter parity)
	 */
	scp->scc_pmsr = 0x701a;

	io->iop_ppard &= ~0x08000000; /* smartc_clc to io-pin */
	io->iop_pdatd &= ~0x08000000; /* smartc_clc = 0 */
	io->iop_pdatd |= 0x00000100; /* smartc_rst  = 0 */
	io->iop_pdatd |= 0x00000800; /* smartc_vcc = 0 */
	init_waitqueue_head(&info->scard_wait);
	return 0;
}

#ifdef MODULE
int init_module(void)
{
	if (scard_init() < 0)
		return -ENODEV;
	return 0;
}

void cleanup_module(void)
{
	misc_deregister(&scard_miscdev);
}
#endif
