/*
 * Fast Ethernet Controller (FCC) driver for Motorola MPC8260.
 * Copyright (c) 2000 MontaVista Software, Inc.   Dan Malek (dmalek@jlc.net)
 *
 * This version of the driver is a combination of the 8xx fec and
 * 8260 SCC Ethernet drivers.  People seem to be choosing common I/O
 * configurations, so this driver will work on the EST8260 boards and
 * others yet to be announced.
 *
 *
 * 16-Oct-2000 - Jouni Malinen <jkm@ssh.com>:
 *      - removed skb memory leaks caused by full TX ring
 *      - receive data directly to skb (i.e., avoid copying when packet
 *        length >= RX_COPY_THRESHOLD)
 *      - align skb data and add extra space to optimize  encapsulation header
 *        addition
 *      - use only one pointer for Rx/Tx buffer (cbd_t) and skb
 *      - added statistics counters for development use
 *      - use spinlock earlier in TX interrupt
 *      - optimized error checking in interrupt handler
 *
 * 19-Oct-2000 - Arto Vuori <avuori@ssh.com>
 *      - Disable FCC RX interrupts.
 *      - TX interrupts are enabled for middle and last BD
 *      - Use PIT interrupts to poll FCC events.
 *      - spinlock moved to fcc_enet_interrupt from fcc_enet_tx
 *      - Link status interrupt handler
 *      - Restart on link change, failure or timout
 *      - Send flow control frames on receive buffer overruns
 *      - Fixed restart on TX errors
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>

#include <asm/immap_8260.h>
#include <asm/pgtable.h>
#include <asm/mpc8260.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>
#include <asm/cpm_8260.h>
#include <linux/ethtool.h>
#include <linux/mii.h>

#define DRV_NAME        "FCC ENET"
#define DRV_VERSION     "0.3"

/* The transmitter timeout
 */
#define TX_TIMEOUT      (HZ/2)

#ifdef CONFIG_SIPO
#define SIPO_ACTIVITY_CHECK_INTERVAL            (HZ/20)
#define SIPO_LINK_CHECK_INIT_COUNT              50
#define DLINK_KLUDGE
#endif

/* Instead of using reqular receive and transmit interrupts, use
 * a periodic timer interrupt to poll FCC status. This should
 * decrease amount of interrupts under heavy load. */
#define USE_PIT_INTR_POLL
#define PIT_RATE_HZ     2048            /* PIT frequency in HZ */

#define PHY_INTR_MASK 0x0f00

/* Send flow-control frames on receive buffer overruns */
#define USE_FLOW_CONTROL
/* Backoff time in slots (5.12us) */
#define FC_BACKOFF_TIME_MSB 0x00
#define FC_BACKOFF_TIME_LSB 0x80

/* Check that all the skb's get freed. This causes a bit more load on
 * processing, but should prevent memory leaks that occured with the previous
 * version of this driver. */
/* #define CHECK_SKB_FREEING */

/* Add statistical information collection for debug/optimization purposes.
 * This makes the driver a bit slower and is designed only for development
 * purposes */
/* #define DO_STATISTICS */

/* Remove additional checks from the final version if everything is working
 * properly. */
#define final_version

/* Reserved buffer head for each received frame. This can be used to align the
 * IP header and to optimize packet encapsulation.
 * E.g. 2 results in 16 byte aligned IP header */
#define RESERVE_BUFFER_HEAD (2+96)
/* Reserved buffer tail for each received frame. This can be used to optimize
 * e.g., ESP trailer addition */
#define RESERVE_BUFFER_TAIL 32

#if 0
#define DPRINTK(fmt, args...) \
printk(KERN_DEBUG "%s: " fmt, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif


/* The number of Tx and Rx buffers.  These are allocated from the page
 * pool.  The code may assume these are power of two, so it it best
 * to keep them that size.
 * We don't need to allocate pages for the transmitter.  We just use
 * the skbuffer directly.
 */
#define TX_RING_SIZE            64      /* Must be power of two */
#define TX_RING_MOD_MASK        63      /*   for this to work */
#define RX_RING_SIZE            64      /* Must be power of two */
#define RX_RING_MOD_MASK        63      /*   for this to work */

/* Packets smaller than this get copied into a newly allocated skbuf on
 * reception, to avoid having large and mostly empty skbuf's hanging around
 * in the TCP stack unnecessarily.
 */
/* #define RX_COPY_THRESHOLD    128 */
#define RX_COPY_THRESHOLD       0

/* The FCC stores dest/src/type, data, and checksum for receive packets.
 */
#define PKT_MAXBUF_SIZE         1518
#define PKT_MINBUF_SIZE         64

/* Maximum input DMA size.  Must be a should(?) be a multiple of 4.
*/
#define PKT_MAXDMA_SIZE         1520

/* Maximum input buffer size.  Must be a multiple of 32.
*/
#define PKT_MAXBLR_SIZE         1536

static int fcc_enet_open(struct net_device *dev);
static int fcc_enet_start_xmit(struct sk_buff *skb, struct net_device *dev);
static void fcc_enet_restart(struct net_device *dev);
static void inline fcc_enet_tx(struct net_device *dev);
static void inline fcc_enet_rx(struct net_device *dev);
static void fcc_enet_backoff(struct net_device *dev);
static void fcc_enet_interrupt(int irq, void * dev_id, struct pt_regs * regs);
static int fcc_enet_close(struct net_device *dev);
static struct net_device_stats *fcc_enet_get_stats(struct net_device *dev);
static void set_multicast_list(struct net_device *dev);

/* These will be configurable for the FCC choice.
 * Multiple ports can be configured.  There is little choice among the
 * I/O pins to the PHY, except the clocks.  We will need some board
 * dependent clock selection.
 * Why in the hell did I put these inside #ifdef's?  I dunno, maybe to
 * help show what pins are used for each device.
 */

/* I/O Pin assignment for FCC1.  I don't yet know the best way to do this,
 * but there is little variation among the choices.
 */
#define PA1_COL         ((uint)0x00000001)
#define PA1_CRS         ((uint)0x00000002)
#define PA1_TXER        ((uint)0x00000004)
#define PA1_TXEN        ((uint)0x00000008)
#define PA1_RXDV        ((uint)0x00000010)
#define PA1_RXER        ((uint)0x00000020)
#define PA1_TXDAT       ((uint)0x00003c00)
#define PA1_RXDAT       ((uint)0x0003c000)
#define PA1_PSORA0      (PA1_RXDAT | PA1_TXDAT)
#define PA1_PSORA1      (PA1_COL | PA1_CRS | PA1_TXER | PA1_TXEN | \
                                PA1_RXDV | PA1_RXER)
#define PA1_DIRA0       (PA1_RXDAT | PA1_CRS | PA1_COL | PA1_RXER | PA1_RXDV)
#define PA1_DIRA1       (PA1_TXDAT | PA1_TXEN | PA1_TXER)

#ifdef CONFIG_SIPO
/* CLK12 is receive, CLK11 is transmit.  These are board specific.
*/
#define PC_F1RXCLK      ((uint)0x00000400)
#define PC_F1TXCLK      ((uint)0x00000800)
#define CMX1_CLK_ROUTE  ((uint)0x37000000)
#define CMX1_CLK_MASK   ((uint)0xff000000)
#else /* SIPO */
#error
/* CLK12 is receive, CLK11 is transmit.  These are board specific.
*/
#define PC_F1RXCLK      ((uint)0x00000800)
#define PC_F1TXCLK      ((uint)0x00000400)
#define CMX1_CLK_ROUTE  ((uint)0x3e000000)
#define CMX1_CLK_MASK   ((uint)0xff000000)
#endif /* SIPO */

/* I/O Pin assignment for FCC2.  I don't yet know the best way to do this,
 * but there is little variation among the choices.
 */
#define PB2_TXER        ((uint)0x00000001)
#define PB2_RXDV        ((uint)0x00000002)
#define PB2_TXEN        ((uint)0x00000004)
#define PB2_RXER        ((uint)0x00000008)
#define PB2_COL         ((uint)0x00000010)
#define PB2_CRS         ((uint)0x00000020)
#define PB2_TXDAT       ((uint)0x000003c0)
#define PB2_RXDAT       ((uint)0x00003c00)
#define PB2_PSORB0      (PB2_RXDAT | PB2_TXDAT | PB2_CRS | PB2_COL | \
                                PB2_RXER | PB2_RXDV | PB2_TXER)
#define PB2_PSORB1      (PB2_TXEN)
#define PB2_DIRB0       (PB2_RXDAT | PB2_CRS | PB2_COL | PB2_RXER | PB2_RXDV)
#define PB2_DIRB1       (PB2_TXDAT | PB2_TXEN | PB2_TXER)

/* CLK13 is receive, CLK14 is transmit.  These are board dependent.
*/
#define PC_F2RXCLK      ((uint)0x00001000)
#define PC_F2TXCLK      ((uint)0x00002000)
#define CMX2_CLK_ROUTE  ((uint)0x00250000)
#define CMX2_CLK_MASK   ((uint)0x00ff0000)

/* I/O Pin assignment for FCC3.  I don't yet know the best way to do this,
 * but there is little variation among the choices.
 */
#define PB3_RXDV        ((uint)0x00004000)
#define PB3_RXER        ((uint)0x00008000)
#define PB3_TXER        ((uint)0x00010000)
#define PB3_TXEN        ((uint)0x00020000)
#define PB3_COL         ((uint)0x00040000)
#define PB3_CRS         ((uint)0x00080000)
#define PB3_TXDAT       ((uint)0x0f000000)
#define PB3_RXDAT       ((uint)0x00f00000)
#define PB3_PSORB0      (PB3_RXDAT | PB3_TXDAT | PB3_CRS | PB3_COL | \
                                PB3_RXER | PB3_RXDV | PB3_TXER | PB3_TXEN)
#define PB3_PSORB1      (0)
#define PB3_DIRB0       (PB3_RXDAT | PB3_CRS | PB3_COL | PB3_RXER | PB3_RXDV)
#define PB3_DIRB1       (PB3_TXDAT | PB3_TXEN | PB3_TXER)

/* CLK15 is receive, CLK16 is transmit.  These are board dependent.
*/
#define PC_F3RXCLK      ((uint)0x00004000)
#define PC_F3TXCLK      ((uint)0x00008000)
#define CMX3_CLK_ROUTE  ((uint)0x00003700)
#define CMX3_CLK_MASK   ((uint)0x0000ff00)

/* MII status/control serial interface.
*/
#define PC_MDIO         ((uint)0x00400000)
#define PC_MDCK         ((uint)0x00200000)

/* A table of information for supporting FCCs.  This does two things.
 * First, we know how many FCCs we have and they are always externally
 * numbered from zero.  Second, it holds control register and I/O
 * information that could be different among board designs.
 */

#define FCC_FORCE_100_FULL              1
#define FCC_FORCE_10_FULL               2
#define FCC_FORCE_100_HALF              3
#define FCC_FORCE_10_HALF               4

typedef struct fcc_info {
	uint    fc_fccnum;
	uint    fc_cpmblock;
	uint    fc_cpmpage;
	uint    fc_proff;
	uint    fc_interrupt;
	uint    fc_trxclocks;
	uint    fc_clockroute;
	uint    fc_clockmask;
	uint    fc_mdio;
	uint    fc_mdck;
	uint    fc_phyid;
	uint    fc_link_intr;
	uint    fc_led;
	uint    mode;
	uint    link_up;
	uint    link_full_duplex;
	uint    link_100;
	uint    fcc_full_duplex;
} fcc_info_t;

static fcc_info_t fcc_ports[] = {
#ifdef CONFIG_FCC1_ENET
	{ 0, CPM_CR_FCC1_SBLOCK, CPM_CR_FCC1_PAGE, PROFF_FCC1, SIU_INT_FCC1,
		(PC_F1RXCLK | PC_F1TXCLK), CMX1_CLK_ROUTE, CMX1_CLK_MASK,
		PC_MDIO, PC_MDCK, 31, SIU_INT_IRQ1, 1, 0},
#endif
#ifdef CONFIG_FCC2_ENET
	{ 1, CPM_CR_FCC2_SBLOCK, CPM_CR_FCC2_PAGE, PROFF_FCC2, SIU_INT_FCC2,
		(PC_F2RXCLK | PC_F2TXCLK), CMX2_CLK_ROUTE, CMX2_CLK_MASK,
		PC_MDIO, PC_MDCK, 30, SIU_INT_IRQ2, 0, 0},
#endif
#ifdef CONFIG_FCC3_ENET
	{ 2, CPM_CR_FCC3_SBLOCK, CPM_CR_FCC3_PAGE, PROFF_FCC3, SIU_INT_FCC3,
		(PC_F3RXCLK | PC_F3TXCLK), CMX3_CLK_ROUTE, CMX3_CLK_MASK,
		PC_MDIO, PC_MDCK, 29, SIU_INT_IRQ4, 2, 0},
#endif
};

/* maxmimum number of interfaces */
#define NUM_FCC (sizeof(fcc_ports) / sizeof(fcc_info_t))
/* real number of interfaces (based on bd hwver) */
static int real_num_fcc = NUM_FCC;

/* The FCC buffer descriptors track the ring buffers.  The rx_bd_base and
 * tx_bd_base always point to the base of the buffer descriptors.  The
 * cur_rx and cur_tx point to the currently available buffer.
 * The dirty_tx tracks the current buffer that is being sent by the
 * controller.  The cur_tx and dirty_tx are equal under both completely
 * empty and completely full conditions.  The empty/ready indicator in
 * the buffer descriptor determines the actual condition.
 */
struct fcc_enet_private {
	/* The saved address of a sent-in-place packet/buffer, for skfree(). */
	struct  sk_buff* tx_skbuff[TX_RING_SIZE];
	struct  sk_buff* rx_skbuff[RX_RING_SIZE];

	/* CPM dual port RAM relative addresses.
	*/
	cbd_t   *rx_bd_base;	    /* Address of Rx and Tx buffers. */
	cbd_t   *tx_bd_base;
	ushort  cur_tx;         /* The next free ring entry */
	ushort  dirty_tx;       /* The ring entries to be free()ed. */
	ushort  cur_rx;
	volatile fcc_t  *fccp;
	volatile fcc_enet_t     *ep;
	struct  net_device_stats stats;
	uint    tx_full;
	spinlock_t lock;
	uint    phy_address;
	uint    phy_type;
	uint    phy_duplex;
	fcc_info_t      *fip;
#ifdef CONFIG_SIPO
	/* Stuff to make the leds blink on the SIPO box */
	struct timer_list eth_timer;
	uint led_tx_packets;
	uint led_rx_packets;
	unsigned long link_check;
	ushort last_problem_skb_cur;
#ifdef DLINK_KLUDGE
	int dlink_updown_level;
	int link_down_last_time;
	uint dlink_found;
	uint dlink_prev_mode;
#endif
#endif
#ifdef DO_STATISTICS
#define MAX_RX_PER_INTERRUPT RX_RING_SIZE
	uint rx_per_interrupt[MAX_RX_PER_INTERRUPT + 1];
#define MAX_TX_PER_INTERRUPT TX_RING_SIZE
	uint tx_per_interrupt[MAX_TX_PER_INTERRUPT + 1];
	uint tx_missed_skb_frees;
	uint num_interrupts;
#endif
#ifdef USE_FLOW_CONTROL
	int no_flow_control;
#endif
};

#ifdef CONFIG_SIPO
unsigned int sipo_fcc_led_show = 1;
static struct net_device *fcc_devs[NUM_FCC];
#endif

static void init_fcc_shutdown(fcc_info_t *fip, struct fcc_enet_private *cep,
			      volatile immap_t *immap);
static void init_fcc_startup(fcc_info_t *fip, struct net_device *dev);
static void init_fcc_ioports(fcc_info_t *fip, volatile iop8260_t *io,
			     volatile immap_t *immap);
static void init_fcc_param(fcc_info_t *fip, struct net_device *dev,
			   volatile immap_t *immap);

#ifdef USE_FLOW_CONTROL
#define ETH_FC_LEN      60
static char flow_control_frame[ETH_FC_LEN]={
#if 1
	0x01,0x80,0xc2,0x00,0x00,0x01,  /* destination = Reserved Multicast address */
#else
	0xff,0xff,0xff,0xff,0xff,0xff,
#endif
	0x00,0x00,0x00,0x00,0x00,0x00,  /* source */
        0x88,0x08,                      /* Len/Type */
	0x00,0x01,                      /* Pause command */
	FC_BACKOFF_TIME_MSB,
	FC_BACKOFF_TIME_LSB,            /* Pause period in slot times (5.12us) MSB first */
	/* 42 reserved bytes followed by FCS (CRC) */
};
#endif

#ifndef CONFIG_SIPO

/* MII processing.  We keep this as simple as possible.  Requests are
 * placed on the list (if there is room).  When the request is finished
 * by the MII, an optional function may be called.
 */
typedef struct mii_list {
	uint    mii_regval;
	void    (*mii_func)(uint val, struct net_device *dev);
	struct  mii_list *mii_next;
} mii_list_t;

#define         NMII    20
mii_list_t      mii_cmds[NMII];
mii_list_t      *mii_free;
mii_list_t      *mii_head;
mii_list_t      *mii_tail;

static  int     phyaddr;
static  uint    phytype;

static int      mii_queue(int request, void (*func)(uint, struct net_device *));
static void     mii_startup_cmds(void);
static uint     mii_send_receive(fcc_info_t *fip, uint cmd);

/* Make MII read/write commands for the FCC.
*/

#define mk_mii_phyaddr(ADDR)    (0x60020000 | ((ADDR) << 23) | (2 << 18))

#define mk_mii_read(REG)        (0x60020000 | ((phyaddr << 23) | \
                                                (REG & 0x1f) << 18))

#define mk_mii_write(REG, VAL)  (0x50020000 | ((phyaddr << 23) | \
                                                (REG & 0x1f) << 18) | \
                                                (VAL & 0xffff))


/* The MII is simulated from the 8xx FEC implementation.  The FCC
 * is not responsible for the MII control/status interface.
 */
static void
fcc_enet_mii(struct net_device *dev)
{
	struct  fcc_enet_private *fep;
	mii_list_t      *mip;
	uint            mii_reg;

	fep = (struct fcc_enet_private *)dev->priv;
#if 0
	ep = &(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec);
	mii_reg = ep->fec_mii_data;
#endif

	if ((mip = mii_head) == NULL) {
		printk("MII and no head!\n");
		return;
	}

	if (mip->mii_func != NULL)
		(*(mip->mii_func))(mii_reg, dev);

	mii_head = mip->mii_next;
	mip->mii_next = mii_free;
	mii_free = mip;

#if 0
	if ((mip = mii_head) != NULL)
		ep->fec_mii_data = mip->mii_regval;
#endif
}

static int
mii_queue(int regval, void (*func)(uint, struct net_device *))
{
	unsigned long   flags;
	mii_list_t      *mip;
	int             retval;

	retval = 0;

	save_flags(flags);
	cli();

	if ((mip = mii_free) != NULL) {
		mii_free = mip->mii_next;
		mip->mii_regval = regval;
		mip->mii_func = func;
		mip->mii_next = NULL;
		if (mii_head) {
			mii_tail->mii_next = mip;
			mii_tail = mip;
		}
		else {
			mii_head = mii_tail = mip;
#if 0
			(&(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec))->fec_mii_data = regval;
#endif
		}
	}
	else {
		retval = 1;
	}

	restore_flags(flags);

	return(retval);
}

static  volatile uint   full_duplex;

static void
mii_status(uint mii_reg, struct net_device *dev)
{
	volatile uint   prev_duplex;

	if (((mii_reg >> 18) & 0x1f) == 1) {
		/* status register.
		*/
		printk("fec: ");
		if (mii_reg & 0x0004)
			printk("link up");
		else
			printk("link down");

		if (mii_reg & 0x0010)
			printk(",remote fault");
		if (mii_reg & 0x0020)
			printk(",auto complete");
		printk("\n");
	}
	if (((mii_reg >> 18) & 0x1f) == 0x14) {
		/* Extended chip status register.
		*/
		prev_duplex = full_duplex;
		printk("fec: ");
		if (mii_reg & 0x0800)
			printk("100 Mbps");
		else
			printk("10 Mbps");

		if (mii_reg & 0x1000) {
			printk(", Full-Duplex\n");
			full_duplex = 1;
		}
		else {
			printk(", Half-Duplex\n");
			full_duplex = 0;
		}
#if 0
		if (prev_duplex != full_duplex)
			restart_fec(dev);
#endif
	}
	if (((mii_reg >> 18) & 0x1f) == 31) {
		/* QS6612 PHY Control/Status.
		 * OK, now we have it all, so figure out what is going on.
		 */
		prev_duplex = full_duplex;
		printk("fec: ");

		mii_reg = (mii_reg >> 2) & 7;

		if (mii_reg & 1)
			printk("10 Mbps");
		else
			printk("100 Mbps");

		if (mii_reg > 4) {
			printk(", Full-Duplex\n");
			full_duplex = 1;
		}
		else {
			printk(", Half-Duplex\n");
			full_duplex = 0;
		}

#if 0
		if (prev_duplex != full_duplex)
			restart_fec(dev);
#endif
	}
}

static  uint    phyno;

static void
mii_discover_phy3(uint mii_reg, struct net_device *dev)
{
	phytype <<= 16;
	phytype |= (mii_reg & 0xffff);
	printk("fec: Phy @ 0x%x, type 0x%08x\n", phyno, phytype);
	mii_startup_cmds();
}

static void
mii_discover_phy(uint mii_reg, struct net_device *dev)
{
	if (phyno < 32) {
		if ((phytype = (mii_reg & 0xffff)) != 0xffff) {
			phyaddr = phyno;
			mii_queue(mk_mii_read(3), mii_discover_phy3);
		}
		else {
			phyno++;
			mii_queue(mk_mii_phyaddr(phyno), mii_discover_phy);
		}
	}
	else {
		printk("FEC: No PHY device found.\n");
	}
}

static void
mii_discover_phy_poll(fcc_info_t *fip)
{
	uint    rv;
	int     i;

	for (i=0; i<32; i++) {
		rv = mii_send_receive(fip, mk_mii_phyaddr(i));
		if ((phytype = (rv & 0xffff)) != 0xffff) {
			phyaddr = i;
			rv = mii_send_receive(fip, mk_mii_read(3));
			phytype <<= 16;
			phytype |= (rv & 0xffff);
			printk("fec: Phy @ 0x%x, type 0x%08x\n", phyaddr, phytype);
		}
	}
}


static  void
mii_startup_cmds(void)
{
#if 1
	/* Level One PHY.
	*/

	/* Read status registers to clear any pending interrupt.
	*/
	mii_queue(mk_mii_read(1), mii_status);
	mii_queue(mk_mii_read(18), mii_status);

	/* Read extended chip status register.
	*/
	mii_queue(mk_mii_read(0x14), mii_status);

	/* Set default operation of 100-TX....for some reason
	 * some of these bits are set on power up, which is wrong.
	 */
	mii_queue(mk_mii_write(0x13, 0), NULL);

	/* Enable Link status change interrupts.
	*/
	mii_queue(mk_mii_write(0x11, 0x0002), NULL);

	/* Don't advertize Full duplex.
	mii_queue(mk_mii_write(0x04, 0x0021), NULL);
	*/
#endif

}

/* This supports the mii_link interrupt below.
 * We should get called three times.  Once for register 1, once for
 * register 18, and once for register 20.
 */
static  uint mii_saved_reg1;

static void
mii_relink(uint mii_reg, struct net_device *dev)
{
	volatile uint   prev_duplex;
	unsigned long   flags;

	if (((mii_reg >> 18) & 0x1f) == 1) {
		/* Just save the status register and get out.
		*/
		mii_saved_reg1 = mii_reg;
		return;
	}
	if (((mii_reg >> 18) & 0x1f) == 18) {
		/* Not much here, but has to be read to clear the
		 * interrupt condition.
		 */
		if ((mii_reg & 0x8000) == 0)
			printk("fec: re-link and no IRQ?\n");
		if ((mii_reg & 0x4000) == 0)
			printk("fec: no PHY power?\n");
	}
	if (((mii_reg >> 18) & 0x1f) == 20) {
		/* Extended chip status register.
		 * OK, now we have it all, so figure out what is going on.
		 */
		prev_duplex = full_duplex;
		printk("fec: ");
		if (mii_saved_reg1 & 0x0004)
			printk("link up");
		else
			printk("link down");

		if (mii_saved_reg1 & 0x0010)
			printk(", remote fault");
		if (mii_saved_reg1 & 0x0020)
			printk(", auto complete");

		if (mii_reg & 0x0800)
			printk(", 100 Mbps");
		else
			printk(", 10 Mbps");

		if (mii_reg & 0x1000) {
			printk(", Full-Duplex\n");
			full_duplex = 1;
		}
		else {
			printk(", Half-Duplex\n");
			full_duplex = 0;
		}
		if (prev_duplex != full_duplex) {
			save_flags(flags);
			cli();
#if 0
			restart_fec(dev);
#endif
			restore_flags(flags);
		}
	}
	if (((mii_reg >> 18) & 0x1f) == 31) {
		/* QS6612 PHY Control/Status.
		 * OK, now we have it all, so figure out what is going on.
		 */
		prev_duplex = full_duplex;
		printk("fec: ");
		if (mii_saved_reg1 & 0x0004)
			printk("link up");
		else
			printk("link down");

		if (mii_saved_reg1 & 0x0010)
			printk(", remote fault");
		if (mii_saved_reg1 & 0x0020)
			printk(", auto complete");

		mii_reg = (mii_reg >> 2) & 7;

		if (mii_reg & 1)
			printk(", 10 Mbps");
		else
			printk(", 100 Mbps");

		if (mii_reg > 4) {
			printk(", Full-Duplex\n");
			full_duplex = 1;
		}
		else {
			printk(", Half-Duplex\n");
			full_duplex = 0;
		}

#if 0
		if (prev_duplex != full_duplex) {
			save_flags(flags);
			cli();
			restart_fec(dev);
			restore_flags(flags);
		}
#endif
	}
}

/* MII command/status interface.
 * I'm not going to describe all of the details.  You can find the
 * protocol definition in many other places, including the data sheet
 * of most PHY parts.
 * I wonder what "they" were thinking (maybe weren't) when they leave
 * the I2C in the CPM but I have to toggle these bits......
 */
static uint
mii_send_receive(fcc_info_t *fip, uint cmd)
{
	unsigned long   flags;
	uint            retval;
	int             read_op, i;
	volatile        immap_t         *immap;
	volatile        iop8260_t       *io;

	immap = (immap_t *)IMAP_ADDR;
	io = &immap->im_ioport;

	/* When we get here, both clock and data are high, outputs.
	 * Output is open drain.
	 * Data transitions on high->low clock, is valid on low->high clock.
	 * Spec says edge transitions no closer than 160 nSec, minimum clock
	 * cycle 400 nSec.  I could only manage about 500 nSec edges with
	 * an XOR loop, so I won't worry about delays yet.
	 * I disable interrupts during bit flipping to ensure atomic
	 * updates of the registers.  I do lots of interrupt disable/enable
	 * to ensure we don't hang out too long with interrupts disabled.
	 */

	/* First, crank out 32 1-bits as preamble.
	 * This is 64 transitions to clock the bits, with clock/data
	 * left high.
	 */
	save_flags(flags);
	cli();
	for (i=0; i<64; i++) {
		io->iop_pdatc ^= fip->fc_mdck;
		udelay(0);
	}
	restore_flags(flags);

	read_op = ((cmd & 0xf0000000) == 0x60000000);

	/* We return the command word on a write op, or the command portion
	 * plus the new data on a read op.  This is what the 8xx FEC does,
	 * and it allows the functions to simply look at the returned value
	 * and know the PHY/register as well.
	 */
	if (read_op)
		retval = cmd;
	else
		retval = (cmd >> 16);

	/* Clock out the first 16 MS bits of the command.
	*/
	save_flags(flags);
	cli();
	for (i=0; i<16; i++) {
		io->iop_pdatc &= ~(fip->fc_mdck);
		if (cmd & 0x80000000)
			io->iop_pdatc |= fip->fc_mdio;
		else
			io->iop_pdatc &= ~(fip->fc_mdio);
		cmd <<= 1;
		io->iop_pdatc |= fip->fc_mdck;
		udelay(0);
	}

	/* Do the turn-around.  If read op, we make the IO and input.
	 * If write op, do the 1/0 thing.
	 */
	io->iop_pdatc &= ~(fip->fc_mdck);
	if (read_op)
		io->iop_pdirc &= ~(fip->fc_mdio);
	else
		io->iop_pdatc |= fip->fc_mdio;
	io->iop_pdatc |= fip->fc_mdck;

	/* I do this mainly to get just a little delay.
	*/
	restore_flags(flags);
	save_flags(flags);
	cli();
	io->iop_pdatc &= ~(fip->fc_mdck);
	io->iop_pdirc &= ~(fip->fc_mdio);
	io->iop_pdatc |= fip->fc_mdck;

	restore_flags(flags);
	save_flags(flags);
	cli();

	/* For read, clock in 16 bits.  For write, clock out
	 * rest of command.
	 */
	if (read_op) {
		io->iop_pdatc &= ~(fip->fc_mdck);
		udelay(0);
		for (i=0; i<16; i++) {
			io->iop_pdatc |= fip->fc_mdck;
			udelay(0);
			retval <<= 1;
			if (io->iop_pdatc & fip->fc_mdio)
				retval |= 1;
			io->iop_pdatc &= ~(fip->fc_mdck);
			udelay(0);
		}
	}
	else {
		for (i=0; i<16; i++) {
			io->iop_pdatc &= ~(fip->fc_mdck);
			if (cmd & 0x80000000)
				io->iop_pdatc |= fip->fc_mdio;
			else
				io->iop_pdatc &= ~(fip->fc_mdio);
			cmd <<= 1;
			io->iop_pdatc |= fip->fc_mdck;
			udelay(0);
		}
		io->iop_pdatc &= ~(fip->fc_mdck);
	}
	restore_flags(flags);

	/* Some diagrams show two 1 bits for "idle".  I don't know if
	 * this is really necessary or if it was just to indicate nothing
	 * is going to happen for a while.
	 * Make the data pin an output, set the data high, and clock it.
	 */
	save_flags(flags);
	cli();
	io->iop_pdatc |= fip->fc_mdio;
	io->iop_pdirc |= fip->fc_mdio;
	for (i=0; i<3; i++)
		io->iop_pdatc ^= fip->fc_mdck;
	restore_flags(flags);

	/* We exit with the same conditions as entry.
	*/
	return(retval);
}

#else /* SIPO */

#define MDIO_DATA       0x00400000                      /* Data bit */
#define MDIO_CLK        0x00200000                      /* Clock bit */
#define MDIO_DIR        ((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdirc        /* PIN Direction register */
#define MDIO_IO         ((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdatc        /* PIN Data (Read/Write) register */

static void inline mdio_delay(void)
{
	mb();
	/* delay(0); */
}

/* SYNC with 32bits of '1' */
static void mdio_sync(void)
{
	int i;

	MDIO_DIR |= MDIO_DATA;
	mdio_delay();
	for(i = 0; i < 32 ; i++) {
		mdio_delay();
		MDIO_IO = (MDIO_IO & ~MDIO_CLK) | MDIO_DATA;
		mdio_delay();
		MDIO_IO = MDIO_IO | MDIO_CLK | MDIO_DATA;
	}
	mdio_delay();
}



static void mdio_write(int phy_id, int location, int value)
{
	int mii_cmd = (0x5002 << 16) | (phy_id << 23) | (location<<18) | value;
	int i;

	mdio_sync();

	/* Shift command bits out. */
	for (i = 31; i >= 0; i--) {
		int dataval = (mii_cmd & (1 << i)) ? MDIO_DATA : 0;
		MDIO_IO = (MDIO_IO & ~(MDIO_DATA | MDIO_CLK)) | dataval;
		mdio_delay();
		MDIO_IO |= MDIO_CLK;
		mdio_delay();
	}

	/* Switch iopin direction */
	MDIO_DIR &= ~MDIO_DATA;

	/* Clear out extra bits. */
	for (i = 2; i > 0; i--) {
		MDIO_IO &= ~(MDIO_DATA | MDIO_CLK);
		mdio_delay();
		MDIO_IO |= MDIO_CLK;
		mdio_delay();
	}
	MDIO_IO &= ~(MDIO_DATA | MDIO_CLK);

	return;

}

static int mdio_read(int phy_id, int location)
{
	int mii_cmd = (0x6000 << 16) | (phy_id << 23) | (location<<18);
	int retval = 0;
	int i;

	mdio_sync();

	/* Shift command bits out. */
	for (i = 31; i > 17; i--) {
		int dataval = (mii_cmd & (1 << i)) ? MDIO_DATA : 0;
		MDIO_IO = (MDIO_IO & ~(MDIO_DATA | MDIO_CLK)) | dataval;
		mdio_delay();
		MDIO_IO |= MDIO_CLK;
		mdio_delay();
	}

	/* Switch iopin direction */
	MDIO_DIR &= ~MDIO_DATA;

	/* Read two transitions, 16 data, and wire-idle bits. */
	for (i = 19; i > 0; i--) {
		MDIO_IO &= ~(MDIO_CLK | MDIO_DATA);
		mdio_delay();
		retval = (retval << 1) | ((MDIO_IO & MDIO_DATA) ? 1 : 0);
		MDIO_IO |= MDIO_CLK;
		mdio_delay();
	}
	MDIO_IO &= ~(MDIO_DATA | MDIO_CLK);

	return (retval>>1) & 0xffff;
}

/* Poll link state */

static void mdio_check(fcc_info_t *fip)
{
	uint x;

	if(!fip->mode) {
		x=mdio_read(fip->fc_phyid,5);
		fip->link_full_duplex = (x&((1<<8) | (1<<6)))?1:0;
		fip->link_100 = (x&((1<<8) | (1<<7)))?1:0;
	} else {

		if(fip->mode==FCC_FORCE_100_FULL || fip->mode==FCC_FORCE_10_FULL)
			fip->link_full_duplex = 1;
		else
			fip->link_full_duplex = 0;

		if(fip->mode==FCC_FORCE_100_FULL || fip->mode==FCC_FORCE_100_HALF)
			fip->link_100 = 1;
		else
			fip->link_100 = 0;
	}

	fip->link_up = (mdio_read(fip->fc_phyid,1) & 0x0004)?1:0;
}

/* Reset and set link mode */

static void mdio_reset(fcc_info_t *fip)
{
	mdio_write(fip->fc_phyid,0,0x8000);    /* Reset */
	mdio_write(fip->fc_phyid,19,0x0000);   /* 1:1 transformer, no low-power mode */
	mdio_write(fip->fc_phyid,21,0x0004);   /* Scrambling off */
	mdio_write(fip->fc_phyid,7, 0);	 /* Next page adv. null */


	mdio_write(fip->fc_phyid,17,PHY_INTR_MASK);

	if(fip->mode) {
		mdio_write(fip->fc_phyid,16,0x0020);   /* Disable auto polarity correction */
		mdio_write(fip->fc_phyid,4,0);  /* Advertise nothing */

		switch(fip->mode) {
		case FCC_FORCE_100_FULL:
			mdio_write(fip->fc_phyid,0,0x2100);
			break;
		case FCC_FORCE_100_HALF:
			mdio_write(fip->fc_phyid,0,0x2000);
			break;
		case FCC_FORCE_10_FULL:
			mdio_write(fip->fc_phyid,0,0x0100);
			break;
		case FCC_FORCE_10_HALF:
			mdio_write(fip->fc_phyid,0,0x0000);
			break;
		}
	} else
		mdio_write(fip->fc_phyid,0,0x1000);    /* Autonegotiation */
}

static void
mdio_phy_init(fcc_info_t *fip, struct fcc_enet_private *cep, struct net_device *dev)
{
	int type;

	type=mdio_read(fip->fc_phyid,3);

	if ((type & 0xffff) != 0xffff) {
		mdio_check(fip);

		mdio_reset(fip);

		udelay(200000);
#ifdef DEBUG
		printk("%s: Phy @ 0x%x, type 0x%04x, %s, %s, %s\n", dev->name, fip->fc_phyid, type,
		       (fip->link_up)?"link up":"link down",
		       (fip->link_full_duplex)?"full duplex":"half duplex",
		       (fip->link_100)?"100Mbps":"10Mbps");
#endif

	} else
		printk("%s: No PHY found\n",dev->name);
}

/* PHY management interface interrupt */
static void
mdio_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct  net_device *dev = dev_id;
	struct fcc_enet_private *cep = (struct fcc_enet_private *)dev->priv;
	ushort          events;
	fcc_info_t      *fip;

	fip=cep->fip;

	events=mdio_read(fip->fc_phyid,17);

#ifdef DEBUG
	printk("%s: MDIO Interrupt", dev->name);
	if(events&0x0001)
		printk(", Autonegotion completed");
	if(events&0x0002)
		printk(", Remote fault");
	if(events&0x0004)
		printk(", Link Down");
	if(events&0x0008)
		printk(", Link partner ACK");
	if(events&0x0080)
		printk(", Jabber");
	if(events&0x0040)
		printk(", Rx error");
	if(events&0x0020)
		printk(", New page");
	if(events&0x0010)
		printk(", PD Fault");
	printk("\n");
#endif

	/* Link gone down */
	if(events&0x0004) {
		printk("%s: Link down\n", dev->name);

		mdio_phy_init(fip,cep,dev);

#ifdef DLINK_KLUDGE
		if(cep->dlink_found) /* DLink was just disconnected, reset settings */
		{
			cep->dlink_found=0;
			cep->dlink_updown_level=0;
			fip->mode = cep->dlink_prev_mode;
			mdio_phy_init(fip,cep,dev);
		} else {
			cep->dlink_updown_level += HZ*10 - (jiffies - cep->link_down_last_time);
			if(cep->dlink_updown_level<0)
				cep->dlink_updown_level=0;
			else if(cep->dlink_updown_level>HZ*100) /* Dlink found */
			{
				printk("%s: Link state changed too often. Forcing 10mb half, duplex\n",
				       dev->name);
				cep->dlink_found = 1;
				cep->dlink_prev_mode = fip->mode;
				fip->mode = FCC_FORCE_10_HALF;
				mdio_phy_init(fip,cep,dev);
			}

		}
		cep->link_down_last_time=jiffies;
#endif
	}

	/* Auto negotion completed or Link partner ACK */
	if(events&0x0001) {
		mdio_check(fip);

		printk("%s: %s,  %s, %s\n", dev->name,
		       (fip->link_up)?"Link up":"Link down",
		       (fip->link_full_duplex)?"full duplex":"half duplex",
		       (fip->link_100)?"100Mbps":"10Mbps");

	}

	/* Reflect new settings to FCC on duplex change */
	if(fip->link_up==1 &&
	   fip->fcc_full_duplex != fip->link_full_duplex)
		fcc_enet_restart(dev);
}


static void
sipo_eth_led_show(unsigned long data)
{
	struct net_device *dev=(struct net_device *)data;
	struct fcc_enet_private *cep = (struct fcc_enet_private *)dev->priv;
	fcc_info_t      *fip=cep->fip;

	/* Poll link state, if mode is forced (no autoneg completed interrupts)*/
	if(fip->mode)
		mdio_check(fip);

	if ((cep->led_tx_packets != cep->stats.tx_packets) ||
	    (cep->led_rx_packets != cep->stats.rx_packets)) {
		cep->led_tx_packets = cep->stats.tx_packets;
		cep->led_rx_packets = cep->stats.rx_packets;
		if (sipo_fcc_led_show)
			sipo_blue_led_flip(cep->fip->fc_led);
		if(!cep->fip->link_up) {
			cep->fip->link_up=1;
			printk("%s: Packets flowing while link down(?) "
			       "Forcing link up.\n", dev->name);
		}
	} else if (sipo_fcc_led_show) {
		if (cep->fip->link_up)
			sipo_blue_led_on(cep->fip->fc_led);
		else
			sipo_blue_led_off(cep->fip->fc_led);
	}

	mod_timer(&(cep->eth_timer), jiffies + SIPO_ACTIVITY_CHECK_INTERVAL);
}

#endif /* SIPO */

static int netdev_ethtool_ioctl(struct net_device *dev, void *useraddr)
{
	struct ethtool_cmd ecmd;
	struct fcc_enet_private *cep = (struct fcc_enet_private *) dev->priv;
	fcc_info_t *fip = cep->fip;

	if (copy_from_user(&ecmd, useraddr, sizeof(ecmd)))
		return -EFAULT;

	switch (ecmd.cmd) {
	case ETHTOOL_GSET:
		ecmd.supported =
			SUPPORTED_10baseT_Half |
			SUPPORTED_10baseT_Full |
			SUPPORTED_100baseT_Half |
			SUPPORTED_100baseT_Full |
			SUPPORTED_Autoneg |
			SUPPORTED_MII;

		ecmd.advertising = ADVERTISED_MII
			| ADVERTISED_10baseT_Half
			| ADVERTISED_10baseT_Full
			| ADVERTISED_100baseT_Half
			| ADVERTISED_100baseT_Full;

		if (!fip->mode) {
			ecmd.advertising |= ADVERTISED_Autoneg;
			ecmd.autoneg = AUTONEG_ENABLE;
		} else
			ecmd.autoneg = AUTONEG_DISABLE;

		ecmd.port = PORT_MII;
		ecmd.transceiver = XCVR_INTERNAL;
		ecmd.phy_address = fip->fc_phyid;
		ecmd.speed = fip->link_100 ? SPEED_100 : SPEED_10;
		ecmd.duplex = fip->link_full_duplex ? DUPLEX_FULL :
			DUPLEX_HALF;
		ecmd.maxtxpkt = TX_RING_SIZE;
		ecmd.maxrxpkt = 0;

		if (copy_to_user(useraddr, &ecmd, sizeof(ecmd)))
			return -EFAULT;
		return 0;

	case ETHTOOL_SSET: {
		uint autoneg, speed100, full_duplex, old_mode=fip->mode;

		autoneg = (ecmd.autoneg == AUTONEG_ENABLE);
		speed100 = (ecmd.speed == SPEED_100);
		full_duplex = (ecmd.duplex == DUPLEX_FULL);

		if(autoneg) {
			fip->mode=0;
		} else {
			if(speed100) {
				if(full_duplex)
					fip->mode=FCC_FORCE_100_FULL;
				else
					fip->mode=FCC_FORCE_100_HALF;
			} else {
				if(full_duplex)
					fip->mode=FCC_FORCE_10_FULL;
				else
					fip->mode=FCC_FORCE_10_HALF;
			}
		}

		if (old_mode != fip->mode) {
			mdio_reset(fip);
			fcc_enet_restart(dev);
		}

		return 0;
	}

	case ETHTOOL_GDRVINFO: {
		struct ethtool_drvinfo info;

		memset(&info, 0, sizeof(info));
		info.cmd = ecmd.cmd;
		strcpy(info.driver, DRV_NAME);
		strcpy(info.version, DRV_VERSION);
		*info.fw_version = 0;
		if (copy_to_user(useraddr, &info, sizeof(info)))
		       return -EFAULT;
		return 0;
	}
	default:
		return -EOPNOTSUPP;
	}

}

static int fcc_enet_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	struct mii_ioctl_data *data = (struct mii_ioctl_data *)&rq->ifr_data;
	struct fcc_enet_private *cep = (struct fcc_enet_private *)dev->priv;

#ifdef DO_STATISTICS
	int i;
	unsigned long flags;
#endif

	switch(cmd) {
	case SIOCETHTOOL:
		return netdev_ethtool_ioctl(dev, (void *) rq->ifr_data);

	case SIOCGMIIPHY:	       /* Get address of MII PHY in use. */
	case SIOCDEVPRIVATE:	    /* for binary compat, remove in 2.5 */
		data->phy_id = cep->fip->fc_phyid;
		/* Fall Through */

	case SIOCGMIIREG:               /* Read MII PHY register. */
	case SIOCDEVPRIVATE+1:          /* for binary compat, remove in 2.5 */
		data->val_out = mdio_read(cep->fip->fc_phyid, data->reg_num & 0x1f);
		return 0;

	case SIOCSMIIREG:               /* Write MII PHY register. */
	case SIOCDEVPRIVATE+2:          /* for binary compat, remove in 2.5 */
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		mdio_write(cep->fip->fc_phyid, data->reg_num & 0x1f, data->val_in);
		return 0;

#ifdef DO_STATISTICS
	case SIOCDEVPRIVATE:
		spin_lock_irqsave(&cep->lock, flags);
		printk("%s: statistics\n", dev->name);
		printk("   tx missed skb frees: %d, interrupts: %d\n",
		       cep->tx_missed_skb_frees, cep->num_interrupts);
		printk("   rx per interrupt:");
		for (i = 0; i <= MAX_RX_PER_INTERRUPT; i++)
			printk(" %d:%d", i, cep->rx_per_interrupt[i]);
		printk("\n");
		printk("   tx per interrupt:");
		for (i = 0; i <= MAX_TX_PER_INTERRUPT; i++)
			printk(" %d:%d", i, cep->tx_per_interrupt[i]);
		printk("\n");
		spin_unlock_irqrestore(&cep->lock, flags);
		return 0;
	case SIOCDEVPRIVATE + 1:
		spin_lock_irqsave(&cep->lock, flags);
		cep->tx_missed_skb_frees = 0;
		cep->num_interrupts = 0;
		memset(&cep->rx_per_interrupt, 0,
		       sizeof(cep->rx_per_interrupt));
		memset(&cep->tx_per_interrupt, 0,
		       sizeof(cep->tx_per_interrupt));
		spin_unlock_irqrestore(&cep->lock, flags);
		return 0;
#endif

#ifdef USE_FLOW_CONTROL
	case SIOCDEVPRIVATE + 3:
		if (cep->no_flow_control) {
			cep->no_flow_control = 0;
			printk("%s: Enabling flow control\n", dev->name);
		} else {
			cep->no_flow_control = 1;
			printk("%s: Disabling flow control\n", dev->name);
		}
		return 0;
#endif
	default:
		return -EOPNOTSUPP;
	}
}

static int
fcc_enet_open(struct net_device *dev)
{
	netif_start_queue(dev);
	return 0;                                       /* Always succeed */
}

static int
fcc_enet_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct fcc_enet_private *cep = (struct fcc_enet_private *)dev->priv;
	volatile cbd_t  *bdp;
	unsigned long flags;

	/* Fill in a Tx ring entry */
	spin_lock_irqsave(&cep->lock, flags);

	if (cep->cur_tx == cep->dirty_tx && cep->tx_full)
	{
		spin_unlock_irqrestore(&cep->lock, flags);
		printk("%s: tx queue full!", dev->name);
		return 1;
	}

	bdp = &cep->tx_bd_base[cep->cur_tx];

#ifndef final_version
	if (bdp->cbd_sc & BD_ENET_TX_READY) {
		/* Ooops.  All transmit buffers are full.  Bail out.
		 * This should not happen, since cep->tx_full should be set.
		 */
		spin_unlock_irqrestore(&cep->lock, flags);
		printk("%s: tx queue full!\n", dev->name);
		return 1;
	}
#endif

#ifdef CHECK_SKB_FREEING
	if (cep->tx_skbuff[cep->cur_tx] != NULL) {
		if (bdp->cbd_sc & BD_ENET_TX_READY) {
			printk("%s: could not get free tx_skbuff in "
			       "fcc_enet_start_xmit()\n", dev->name);
			spin_unlock_irqrestore(&cep->lock, flags);
			return 1;
		} else {
#if 0
			printk("%s: freeing tx_skbuff[%d] in "
			       "fcc_enet_start_xmit()\n",
			       dev->name, cep->cur_tx);
#endif
#ifdef DO_STATISTICS
			cep->tx_missed_skb_frees++;
#endif
			dev_kfree_skb_irq(cep->tx_skbuff[cep->cur_tx]);
			cep->tx_skbuff[cep->cur_tx] = NULL;
		}
	}
#endif

	/* Clear all of the status flags.
	 */
	bdp->cbd_sc &= ~BD_ENET_TX_STATS;

	/* If the frame is short, tell CPM to pad it.
	*/
	if (skb->len <= ETH_ZLEN)
		bdp->cbd_sc |= BD_ENET_TX_PAD;
	else
		bdp->cbd_sc &= ~BD_ENET_TX_PAD;

	/* Set buffer length and buffer pointer.
	*/
	bdp->cbd_datlen = skb->len;
	bdp->cbd_bufaddr = __pa(skb->data);

	/* Save skb pointer.
	*/
	cep->tx_skbuff[cep->cur_tx] = skb;

	cep->stats.tx_bytes += skb->len;
	cep->cur_tx = (cep->cur_tx + 1) & TX_RING_MOD_MASK;

	/* Send it on its way.  Tell CPM its ready,
	 * its the last BD of the frame, and to put the CRC on the end.
	 */
	bdp->cbd_sc |= (BD_ENET_TX_READY | BD_ENET_TX_LAST |
			BD_ENET_TX_TC);

#if 0
	/* Errata says don't do this.
	 * FIX: which errata??
	 * (MPC8260 user manual errata did not seem to mention this..)
	*/
	cep->fccp->fcc_ftodr = 0x8000;
#endif
	dev->trans_start = jiffies;

	/* Check whether the TX ring is full */
	if (cep->dirty_tx == cep->cur_tx) {
		cep->tx_full = 1;
		netif_stop_queue(dev);
#ifdef USE_PIT_INTR_POLL
		/* We might want to poll */
		fcc_enet_tx(dev);
#endif
	}

	spin_unlock_irqrestore(&cep->lock, flags);

	return 0;
}


static void
fcc_enet_timeout(struct net_device *dev)
{
	struct fcc_enet_private *cep = (struct fcc_enet_private *)dev->priv;

	DPRINTK("%s: transmit timed out.\n", dev->name);
	cep->stats.tx_errors++;
#ifndef final_version
	{
		int     i;
		cbd_t   *bdp;
		printk(" Ring data dump: cur_tx %d%s cur_rx %d.\n",
		       cep->cur_tx, cep->tx_full ? " (full)" : "",
		       cep->cur_rx);
		bdp = cep->tx_bd_base;
		printk(" Tx @base %p :\n", bdp);
		for (i = 0 ; i < TX_RING_SIZE; i++, bdp++)
			printk("%04x %04x %08x\n",
			       bdp->cbd_sc,
			       bdp->cbd_datlen,
			       bdp->cbd_bufaddr);
		bdp = cep->rx_bd_base;
		printk(" Rx @base %p :\n", bdp);
		for (i = 0 ; i < RX_RING_SIZE; i++, bdp++)
			printk("%04x %04x %08x\n",
			       bdp->cbd_sc,
			       bdp->cbd_datlen,
			       bdp->cbd_bufaddr);
	}
#endif
	fcc_enet_restart(dev);

	if (!cep->tx_full)
		netif_wake_queue(dev);
}

static void
fcc_enet_restart(struct net_device *dev)
{
	volatile struct fcc_enet_private *cep = (struct fcc_enet_private *)dev->priv;
	cbd_t                   *bdp;
	volatile fcc_t          *fccp;
	unsigned long flags;
	int                     i;
	fcc_info_t              *fip;

	fip=cep->fip;

	printk("%s: Restart\n", dev->name);

	spin_lock_irqsave(&cep->lock, flags);

	fip->fcc_full_duplex = fip->link_full_duplex;

	fccp = cep->fccp;

	/* Restart rx & tx */
	cpmp->cp_cpcr =
	    mk_cr_cmd(cep->fip->fc_cpmpage, cep->fip->fc_cpmblock, 0x0c,
			  CPM_CR_GRACEFULLSTOP_TX) | CPM_CR_FLG;
	do { udelay(10); } while (cpmp->cp_cpcr & CPM_CR_FLG);

	fccp->fcc_gfmr &= ~(FCC_GFMR_ENR | FCC_GFMR_ENT);

	/* Clear transmit BD's and release skb's */
	for(i=cep->dirty_tx; i!=cep->cur_tx || cep->tx_full; i=((i+1)&TX_RING_MOD_MASK), cep->tx_full=0)
	{
		bdp = &cep->tx_bd_base[i];

		bdp->cbd_sc &= ~BD_ENET_TX_READY;
		dev_kfree_skb_irq(cep->tx_skbuff[i]);
		cep->tx_skbuff[i]=NULL;
	}

	/* Clear Receive BD's */
	for(bdp=cep->rx_bd_base, i=0; i<RX_RING_SIZE; i++) {
		bdp->cbd_sc |= BD_ENET_RX_EMPTY;
		bdp++;
	}

	/* Reset buffer pointers */
	cep->dirty_tx = cep->cur_tx = cep->cur_rx = 0;
	cep->ep->fen_genfcc.fcc_rbase = __pa(cep->rx_bd_base);
	cep->ep->fen_genfcc.fcc_tbase = __pa(cep->tx_bd_base);
	cep->ep->fen_genfcc.fcc_rbptr = 0;
	cep->ep->fen_genfcc.fcc_tbptr = 0;

	/* Reset state */
	cep->ep->fen_genfcc.fcc_rstate &= 0xff000000;
	cep->ep->fen_genfcc.fcc_tstate &= 0xff000000;

	/* Clear the Out-of-sequence TxBD.
	*/
	cep->ep->fen_tfcstat = 0;
	cep->ep->fen_tfclen = 0;
	cep->ep->fen_tfcptr = 0;

	/* Change mode register settings */
	if(cep->fip->link_full_duplex)
	    fccp->fcc_fpsmr = FCC_PSMR_ENCRC | FCC_PSMR_FDE |
		    FCC_PSMR_FCE |  FCC_PSMR_LPB;
	else
	    fccp->fcc_fpsmr = FCC_PSMR_ENCRC | FCC_PSMR_FCE;

	/* Restart rx & tx */
	cpmp->cp_cpcr =
	    mk_cr_cmd(cep->fip->fc_cpmpage, cep->fip->fc_cpmblock, 0x0c,
			  CPM_CR_INIT_TRX) | CPM_CR_FLG;
	do { udelay(10); } while (cpmp->cp_cpcr & CPM_CR_FLG);

	fccp->fcc_gfmr |= (FCC_GFMR_ENR | FCC_GFMR_ENT);
	spin_unlock_irqrestore(&cep->lock, flags);
}


#ifdef USE_PIT_INTR_POLL

/* Periodic Timer interrupt */
static void
pit_interrupt(int irq, void *dev, struct pt_regs * regs)
{
	int i;

	volatile	immap_t	 *immap;
	immap = (immap_t *)IMAP_ADDR;   /* and to internal registers */

	/* Ack 82xx PIC interrupt */
	immap->im_sit.sit_piscr|=PISCR_PS;

	for(i = 0; i < real_num_fcc; i++)
	    fcc_enet_interrupt(0, fcc_devs[i], regs);

	return;
}
#endif /* USE_PIT_INTR_POLL */

/* The interrupt handler.
 */
static void
fcc_enet_interrupt(int irq, void * dev_id, struct pt_regs * regs)
{
	struct  net_device *dev = dev_id;
	volatile struct fcc_enet_private *cep;
	ushort  int_events;
	unsigned long flags;

	cep = (struct fcc_enet_private *)dev->priv;

	/* Fill in a Tx ring entry */
	spin_lock_irqsave(&cep->lock, flags);

#ifdef DO_STATISTICS
	cep->num_interrupts++;
#endif
	/* Get the interrupt events that caused us to be here and clear the
	 * FCCE bits.
	*/
	int_events = cep->fccp->fcc_fcce;
	cep->fccp->fcc_fcce = int_events;

	/* Check for a transmit error.  The manual is a little unclear
	 * about this, so the debug code until I get it figured out.  It
	 * appears that if TXE is set, then TXB is not set.  However,
	 * if carrier sense is lost during frame transmission, the TXE
	 * bit is set, "and continues the buffer transmission normally."
	 * I don't know if "normally" implies TXB is set when the buffer
	 * descriptor is closed.....trial and error :-).
	 */

	/* Transmit OK, or non-fatal error.  Update the buffer descriptors.
	*/
#ifdef USE_PIT_INTR_POLL
	if (int_events & (FCC_ENET_TXE | FCC_ENET_TXB))
#endif
		fcc_enet_tx(dev);

	/* Handle receive event in its own function.
	*/
#ifdef USE_PIT_INTR_POLL
	if (int_events & FCC_ENET_RXF)
#endif
		fcc_enet_rx(dev);

	/* Check for receive busy, i.e. packets coming but no place to
	 * put them.
	 */
	if (int_events & FCC_ENET_BSY) {
		cep->stats.rx_dropped++;
#ifdef USE_FLOW_CONTROL
		fcc_enet_backoff(dev);
#endif
	}

	/* FIX: Motorola manual (Sect. 28.10) says that FCCE should be cleared
	 * here; should we now clear the bits again?
	 */
#if 0
	cep->fccp->fcc_fcce = 0xffff;
#endif
	spin_unlock_irqrestore(&cep->lock, flags);

	return;
}

/* Part of the interrupt handler for cleaning the sent packets.
 */
static void inline
fcc_enet_tx(struct net_device *dev)
{
	volatile struct fcc_enet_private *cep;
	int must_restart;
	ushort cbd_sc;
	volatile cbd_t *bdp;
#ifdef DO_STATISTICS
	int tx_pkt_count = 0;
#endif

	cep = (struct fcc_enet_private *)dev->priv;
	must_restart = 0;

	bdp = &cep->tx_bd_base[cep->dirty_tx];
	while (((cbd_sc = bdp->cbd_sc) & BD_ENET_TX_READY) == 0) {
		if (cep->dirty_tx == cep->cur_tx && cep->tx_full == 0)
			break;

#ifdef DO_STATISTICS
		tx_pkt_count++;
#endif

		if (cbd_sc &
		    (BD_ENET_TX_HB | BD_ENET_TX_LC | BD_ENET_TX_RL |
		     BD_ENET_TX_UN | BD_ENET_TX_CSL | BD_ENET_TX_DEF)) {
			if (cbd_sc & BD_ENET_TX_HB)     /* No heartbeat */
				cep->stats.tx_heartbeat_errors++;
			if (cbd_sc & BD_ENET_TX_LC)     /* Late collision */
				cep->stats.tx_window_errors++;
			if (cbd_sc & BD_ENET_TX_RL)     /* Retrans limit */
				cep->stats.tx_aborted_errors++;
			if (cbd_sc & BD_ENET_TX_UN)     /* Underrun */
				cep->stats.tx_fifo_errors++;
			if (cbd_sc & BD_ENET_TX_CSL)    /* Carrier lost */
				cep->stats.tx_carrier_errors++;

			/* No heartbeat or Lost carrier are not really bad
			 * errors.
			 * The others require a restart transmit command.
			 */
			if (cbd_sc &
			    (BD_ENET_TX_LC | BD_ENET_TX_RL | BD_ENET_TX_UN)) {
				must_restart = 1;
				cep->stats.tx_errors++;
			}

			/* Deferred means some collisions occurred during
			 * transmit, but we eventually sent the packet OK.
			 */
			if (cbd_sc & BD_ENET_TX_DEF) {
				cep->stats.collisions++;
				cep->stats.tx_packets++;
			}
		} else
			cep->stats.tx_packets++;

		/* Free the sk buffer associated with this last transmit.
		*/
#ifdef CHECK_SKB_FREEING
		if (cep->tx_skbuff[cep->dirty_tx] == NULL)
			printk("%s: tx_skbuff[%d] == NULL when freeing in "
			       "TX interrupt\n", dev->name, cep->dirty_tx);
		else
			dev_kfree_skb_irq(cep->tx_skbuff[cep->dirty_tx]);
		cep->tx_skbuff[cep->dirty_tx] = NULL;
#else
		dev_kfree_skb_irq(cep->tx_skbuff[cep->dirty_tx]);
#endif

		/* Update pointer to next buffer descriptor to be transmitted.
		*/
		cep->dirty_tx = (cep->dirty_tx + 1) & TX_RING_MOD_MASK;
		bdp = &cep->tx_bd_base[cep->dirty_tx];

		/* I don't know if we can be held off from processing these
		 * interrupts for more than one frame time.  I really hope
		 * not.  In such a case, we would now want to check the
		 * currently available BD (cur_tx) and determine if any
		 * buffers between the dirty_tx and cur_tx have also been
		 * sent.  We would want to process anything in between that
		 * does not have BD_ENET_TX_READY set.
		 */

		/* Since we have freed up a buffer, the ring is no longer
		 * full.
		 */
		if (cep->tx_full) {
			cep->tx_full = 0;
			if (netif_queue_stopped(dev))
				netif_wake_queue(dev);
		}
	}

	if (must_restart) {
		volatile cpm8260_t *cp;

		/* Some transmit errors cause the transmitter to shut
		 * down.  We now issue a restart transmit.  Since the
		 * errors close the BD and update the pointers, the restart
		 * _should_ pick up without having to reset any of our
		 * pointers either.
		 */

		cep->fccp->fcc_gfmr &= ~(FCC_GFMR_ENT);

		cp = cpmp;
		cp->cp_cpcr =
			mk_cr_cmd(cep->fip->fc_cpmpage, cep->fip->fc_cpmblock,
				  0x0c, CPM_CR_RESTART_TX) | CPM_CR_FLG;
		do { udelay(10); } while (cp->cp_cpcr & CPM_CR_FLG);

		cep->fccp->fcc_gfmr |= (FCC_GFMR_ENT);
	}

#ifdef DO_STATISTICS
	if (tx_pkt_count > MAX_TX_PER_INTERRUPT)
		tx_pkt_count = MAX_TX_PER_INTERRUPT;
	cep->tx_per_interrupt[tx_pkt_count]++;
#endif
}

#ifdef CONFIG_SIPO

#define CRC_POLYNOMIAL 0xedb88320UL
#define CRC_MASK 0xdebb20e3UL

/* len includes 4 bytes for FCS */
static int ethernet_fcs_ok(unsigned char *data, int len)
{
	u32 crc, poly = CRC_POLYNOMIAL;
	int byte, bit, j;

	if (len < 4)
		return 0;

	crc = 0xffffffff;
	for (byte = 0; byte < len; byte++) {
		for (bit = data[byte], j = 0; j < 8; j++, bit >>= 1) {
			int test = ((bit ^ crc) & 0x01);
			crc >>= 1;
			if (test)
				crc = crc ^ poly;
		}
	}

	return crc == CRC_MASK;
}

#endif /* CONFIG_SIPO */

/* Part of the interrupt handler for receiving packets.
 * During a receive, the cur_rx points to the current incoming buffer.
 * When we update through the ring, if the next incoming buffer has
 * not been given to the system, we just set the empty indicator,
 * effectively tossing the packet.
 */
static void inline
fcc_enet_rx(struct net_device *dev)
{
	struct  fcc_enet_private *cep;
	volatile cbd_t  *bdp;
	struct  sk_buff *new_skb, *skb;
	ushort  pkt_len;
	ushort  cbd_sc;
#if RX_COPY_TRESHOLD!=0
	int     do_copy;
#endif
#ifdef DO_STATISTICS
	int     rx_pkt_count = 0;
#endif

	cep = (struct fcc_enet_private *)dev->priv;

	for (;;) {
		/* First, grab all of the stats for the incoming packet.
		 * These get messed up if we get called due to a busy
		 * condition.
		 */
		bdp = &cep->rx_bd_base[cep->cur_rx];
		cbd_sc = bdp->cbd_sc;
		if (cbd_sc & BD_ENET_RX_EMPTY)
			break;
#ifdef DO_STATISTICS
		if (++rx_pkt_count > MAX_RX_PER_INTERRUPT) {
			DPRINTK("rx_pkt_count > max\n");
			break;
		}
#endif
#ifndef final_version
		/* Since we have allocated space to hold a complete frame, both
		 * the first and last indicators should be set.
		 */
		if ((cbd_sc & (BD_ENET_RX_FIRST | BD_ENET_RX_LAST)) !=
		    (BD_ENET_RX_FIRST | BD_ENET_RX_LAST))
			printk("CPM ENET: rcv is not first+last\n");
#endif
		if (cbd_sc & (BD_ENET_RX_LG | BD_ENET_RX_SH | BD_ENET_RX_NO |
			      BD_ENET_RX_CR | BD_ENET_RX_OV | BD_ENET_RX_CL)) {
			/* Frame too long or too short.
			 */
			if (cbd_sc & (BD_ENET_RX_LG | BD_ENET_RX_SH))
				cep->stats.rx_length_errors++;
			if (cbd_sc & BD_ENET_RX_NO)     /* Frame alignment */
				cep->stats.rx_frame_errors++;
			if (cbd_sc & BD_ENET_RX_CR)     /* CRC Error */
				cep->stats.rx_crc_errors++;
			if (cbd_sc & BD_ENET_RX_OV)     /* FIFO overrun */
				cep->stats.rx_fifo_errors++;

#ifdef CONFIG_SIPO
			/* http://e-www.motorola.com/brdata/PDFDB/docs/
			 * MPC8260CE.pdf;
			 * MPC8260CE Rev. 2.1, 01/2002 Device Errata:
			 * CPM 21. False indication of collision in fast
			 * Ethernet (3927) */

			/* FCC has known problem with at least some Motorola
			 * versions. It gives late RX collisions for frames
			 * that are actually okay. Accept these frames if
			 * Ethernet FCS is correct. */
			if (cbd_sc & BD_ENET_RX_CL) {
				unsigned char *data =
					(unsigned char *)
					__va(bdp->cbd_bufaddr);

				if (ethernet_fcs_ok(data, bdp->cbd_datlen)) {
					/* RX late coll, but FCS OK; accept */
				} else {
					/* RX late coll and FCS !OK; drop */
					cep->stats.rx_errors++;
					cep->stats.rx_frame_errors++;
					goto drop;
				}
			}
#else /* CONFIG_SIPO */
			/* Report late collisions as a frame error.
			 * On this error, the BD is closed, but we don't know
			 * what we have in the buffer.  So, just drop this
			 * frame on the floor.
			 */
			if (cbd_sc & BD_ENET_RX_CL) {
				cep->stats.rx_errors++;
				cep->stats.rx_frame_errors++;
				goto drop;
			}
#endif /* CONFIG_SIPO */
		}

		/* Process the incoming frame.
		*/
		pkt_len = bdp->cbd_datlen;
#ifndef final_version
		if (pkt_len > PKT_MAXBLR_SIZE) {
			printk("%s: Oversized Ethernet frame, %d bytes\n",
			       dev->name, pkt_len);
			pkt_len = PKT_MAXBLR_SIZE;
		} else if (pkt_len < 4)
			printk("%s: Undersized Ethernet frame, %d bytes\n",
			       dev->name, pkt_len);
		else
			pkt_len -= 4; /* omit four octet CRC */
#else
		pkt_len -= 4; /* omit four octer CRC */
#endif

		cep->stats.rx_bytes += pkt_len;

#if RX_COPY_TRESHOLD!=0
		/* If the packet is small, we will do a copy. */
		do_copy = pkt_len < RX_COPY_THRESHOLD;

		/* Allocate the new skbuff _before_ we hand the old one up
		 * to netif_rx, to guarantee that we never get left without
		 * one when memory runs low.  If the alloc fails, we drop the
		 * incoming packet, and re-use the previous buffer next time.
		 */
		new_skb = dev_alloc_skb((do_copy ? pkt_len : PKT_MAXBLR_SIZE) +
					RESERVE_BUFFER_HEAD +
					RESERVE_BUFFER_TAIL);

		if (new_skb == NULL) {
			printk("%s: Memory squeeze, dropping packet.\n",
				dev->name);
			cep->stats.rx_dropped++;
			goto drop;
		}

		/* byte align the IP header */
		skb_reserve(new_skb, RESERVE_BUFFER_HEAD);

		if (do_copy) {
			skb = new_skb;
			eth_copy_and_sum(skb,
				(unsigned char *)__va(bdp->cbd_bufaddr),
				pkt_len, 0);
		} else {
			skb = cep->rx_skbuff[cep->cur_rx];
			/* Replace buffer just passed off with new one */
			cep->rx_skbuff[cep->cur_rx] = new_skb;
			/* Set buffer descriptor to point to new buffer */
			bdp->cbd_bufaddr = __pa(new_skb->data);
		}
#else
		new_skb = dev_alloc_skb(PKT_MAXBLR_SIZE + RESERVE_BUFFER_HEAD +
			RESERVE_BUFFER_TAIL);

		if (new_skb == NULL) {
			printk("%s: Memory squeeze, dropping packet.\n",
				dev->name);
			cep->stats.rx_dropped++;
			goto drop;
		}

		/* byte align the IP header */
		skb_reserve(new_skb, RESERVE_BUFFER_HEAD);

		skb = cep->rx_skbuff[cep->cur_rx];
		/* Replace buffer just passed off with new one */
		cep->rx_skbuff[cep->cur_rx] = new_skb;
		/* Set buffer descriptor to point to new buffer */
		bdp->cbd_bufaddr = __pa(new_skb->data);
#endif
		skb_put(skb, pkt_len);
		skb->dev = dev;
		skb->protocol = eth_type_trans(skb,dev);
		cep->stats.rx_packets++;
		netif_rx(skb);

	drop:
		/* Clear the status flags for this buffer.
		 */
		bdp->cbd_sc &= ~BD_ENET_RX_STATS;

		/* Mark the buffer empty.
		 */
		bdp->cbd_sc |= BD_ENET_RX_EMPTY;

		/* Update BD pointer to next entry.
		 */
		cep->cur_rx = (cep->cur_rx + 1) & RX_RING_MOD_MASK;
	}

#ifdef DO_STATISTICS
	cep->rx_per_interrupt[rx_pkt_count]++;
#endif
}

#ifdef USE_FLOW_CONTROL
/* Send Ethernet Flow-control backoff frame to reduce link partner transmit rate.
 */
static void
fcc_enet_backoff(struct net_device *dev)
{
	volatile struct fcc_enet_private *cep;

	cep = (struct fcc_enet_private *)dev->priv;

	if (cep->no_flow_control)
		return;

	/* FIX: this might cause some problems if multiple interfaces would
	 * like to send flow-control frames simultaneously.. only one buffer
	 * used at the moment and src addr should be diffrent */
	memcpy(&flow_control_frame[6], dev->dev_addr, 6); /* src addr */

	cep->ep->fen_tfclen = ETH_FC_LEN;
	cep->ep->fen_tfcptr = __pa(flow_control_frame);
	cep->ep->fen_tfcstat = (BD_ENET_TX_READY | BD_ENET_TX_LAST |
				BD_ENET_TX_TC);
}
#endif

static int
fcc_enet_close(struct net_device *dev)
{
	/* Don't know what to do yet.
	*/
	netif_stop_queue(dev);

	return 0;
}

static struct net_device_stats *fcc_enet_get_stats(struct net_device *dev)
{
	struct fcc_enet_private *cep = (struct fcc_enet_private *)dev->priv;

	return &cep->stats;
}

/* Set or clear the multicast filter for this adaptor.
 * Skeleton taken from sunlance driver.
 * The CPM Ethernet implementation allows Multicast as well as individual
 * MAC address filtering.  Some of the drivers check to make sure it is
 * a group multicast address, and discard those that are not.  I guess I
 * will do the same for now, but just remove the test if you want
 * individual filtering as well (do the upper net layers want or support
 * this kind of feature?).
 */
static void
set_multicast_list(struct net_device *dev)
{
	struct  fcc_enet_private *cep;
	struct  dev_mc_list *dmi;
	u_char  *mcptr, *tdptr;
	volatile fcc_enet_t *ep;
	int     i, j;

	cep = (struct fcc_enet_private *)dev->priv;

	/* Get pointer to FCC area in parameter RAM.
	*/
	ep = (fcc_enet_t *)dev->base_addr;

	if (dev->flags&IFF_PROMISC) {

		/* Log any net taps. */
		printk("%s: Promiscuous mode enabled.\n", dev->name);
		cep->fccp->fcc_fpsmr |= FCC_PSMR_PRO;
	} else {

		cep->fccp->fcc_fpsmr &= ~FCC_PSMR_PRO;

		if (dev->flags & IFF_ALLMULTI) {
			/* Catch all multicast addresses, so set the
			 * filter to all 1's.
			 */
			ep->fen_gaddrh = 0xffffffff;
			ep->fen_gaddrl = 0xffffffff;
		}
		else {
			/* Clear filter and add the addresses in the list.
			*/
			ep->fen_gaddrh = 0;
			ep->fen_gaddrl = 0;

			dmi = dev->mc_list;

			for (i=0; i<dev->mc_count; i++) {

				/* Only support group multicast for now.
				*/
				if (!(dmi->dmi_addr[0] & 1))
					continue;

				/* The address in dmi_addr is LSB first,
				 * and taddr is MSB first.  We have to
				 * copy bytes MSB first from dmi_addr.
				 */
				mcptr = (u_char *)dmi->dmi_addr + 5;
				tdptr = (u_char *)&ep->fen_taddrh;
				for (j=0; j<6; j++)
					*tdptr++ = *mcptr--;

				/* Ask CPM to run CRC and set bit in
				 * filter mask.
				 */
				cpmp->cp_cpcr = mk_cr_cmd(cep->fip->fc_cpmpage,
						cep->fip->fc_cpmblock, 0x0c,
						CPM_CR_SET_GADDR) | CPM_CR_FLG;
				udelay(10);
				while (cpmp->cp_cpcr & CPM_CR_FLG);
			}
		}
	}
}

/* Initialize the CPM Ethernet on FCC.
 */
int __init fec_enet_init(void)
{
	struct net_device *dev;
	struct fcc_enet_private *cep;
	fcc_info_t      *fip;
	int     i, np;
	volatile        immap_t         *immap;
	volatile        iop8260_t       *io;
	bd_t *__bd=(bd_t *)__res;

	if (__bd->hw_version == 6) {
		real_num_fcc = 2;
		/* swap DMZ and Private leds, since two interface model has
		 * different back panel, but same led locations */
		fcc_ports[0].fc_led = 2;
		fcc_ports[2].fc_led = 1;
	}

	immap = (immap_t *)IMAP_ADDR;   /* and to internal registers */
	io = &immap->im_ioport;

	fip = fcc_ports;

	for (np = 0; np < real_num_fcc; np++)
	{
		/* Allocate some private information.
		*/
		cep = (struct fcc_enet_private *)
					kmalloc(sizeof(*cep), GFP_KERNEL);
		__clear_user(cep,sizeof(*cep));
		spin_lock_init(&cep->lock);
		cep->fip = fip;

		/* Create an Ethernet device instance.
		*/
		dev = init_etherdev(0, 0);
		dev->priv = cep;

		init_fcc_shutdown(fip, cep, immap);
		init_fcc_ioports(fip, io, immap);
		init_fcc_param(fip, dev, immap);

		dev->base_addr = (unsigned long)(cep->ep);

		/* The CPM Ethernet specific entries in the device
		 * structure.
		 */
		dev->open = fcc_enet_open;
		dev->hard_start_xmit = fcc_enet_start_xmit;
		dev->tx_timeout = fcc_enet_timeout;
		dev->watchdog_timeo = TX_TIMEOUT;
		dev->stop = fcc_enet_close;
		dev->get_stats = fcc_enet_get_stats;
		dev->set_multicast_list = set_multicast_list;
		dev->do_ioctl = fcc_enet_ioctl;

		printk("%s: " DRV_NAME " Version " DRV_VERSION ". @", dev->name);
		for (i=0; i<5; i++)
			printk("%02x:", dev->dev_addr[i]);
		printk("%02x\n", dev->dev_addr[5]);

		mdio_phy_init(fip, cep, dev);

		/* Install link interrupt handler */
		if (request_8xxirq(fip->fc_link_intr, mdio_interrupt, 0,
				   "fccphy", dev) < 0)
			printk("%s: Can't get PHY irq %d\n", dev->name, fip->fc_interrupt);

		init_fcc_startup(fip, dev);

#ifdef CONFIG_SIPO
		init_timer(&(cep->eth_timer));
		cep->eth_timer.expires = jiffies +
			SIPO_ACTIVITY_CHECK_INTERVAL;
		cep->eth_timer.data = (unsigned long)dev;
		cep->eth_timer.function = sipo_eth_led_show;
		cep->link_check = SIPO_LINK_CHECK_INIT_COUNT;
		add_timer(&(cep->eth_timer));
#endif

#ifdef DLINK_KLUDGE
		cep->dlink_updown_level=0;
		cep->link_down_last_time=jiffies;
#endif
		fcc_devs[np]=dev;
		fip++;
	}

#ifdef USE_PIT_INTR_POLL
	/* Install PIT interrupt handler.
	*/

	m8260_cpm_fastbrg(0, 8192*4, 0);		/* Set BRG1 */

	immap->im_sit.sit_piscr=0x0000;

	if (request_8xxirq(SIU_INT_PIT, pit_interrupt, 0, "fenet PIT", dev) < 0)
		printk("Can't get PIT IRQ %d\n", SIU_INT_PIT);

	immap->im_sit.sit_pitc=(32768/PIT_RATE_HZ/4 << 16);
	immap->im_sit.sit_piscr=PISCR_PIE|PISCR_PTF|PISCR_PTE;		  /* Enable counter, interrupt, clock =32Khz */
#endif
	return 0;
}

/* Make sure the device is shut down during initialization.
*/
static void __init
init_fcc_shutdown(fcc_info_t *fip, struct fcc_enet_private *cep,
						volatile immap_t *immap)
{
	volatile        fcc_enet_t      *ep;
	volatile        fcc_t           *fccp;

	/* Get pointer to FCC area in parameter RAM.
	*/
	ep = (fcc_enet_t *)(&immap->im_dprambase[fip->fc_proff]);

	/* And another to the FCC register area.
	*/
	fccp = (volatile fcc_t *)(&immap->im_fcc[fip->fc_fccnum]);
	cep->fccp = fccp;	       /* Keep the pointers handy */
	cep->ep = ep;

	/* Disable receive and transmit in case someone left it running.
	*/
	fccp->fcc_gfmr &= ~(FCC_GFMR_ENR | FCC_GFMR_ENT);
}

/* Initialize the I/O pins for the FCC Ethernet.
*/
static void __init
init_fcc_ioports(fcc_info_t *fip, volatile iop8260_t *io,
						volatile immap_t *immap)
{

	/* FCC1 pins are on port A/C.  FCC2/3 are port B/C.
	*/
	if (fip->fc_proff == PROFF_FCC1) {
		/* Configure port A and C pins for FCC1 Ethernet.
		 */
		io->iop_pdira &= ~PA1_DIRA0;
		io->iop_pdira |= PA1_DIRA1;
		io->iop_psora &= ~PA1_PSORA0;
		io->iop_psora |= PA1_PSORA1;
		io->iop_ppara |= (PA1_DIRA0 | PA1_DIRA1);
	}
	if (fip->fc_proff == PROFF_FCC2) {
		/* Configure port B and C pins for FCC Ethernet.
		 */
		io->iop_pdirb &= ~PB2_DIRB0;
		io->iop_pdirb |= PB2_DIRB1;
		io->iop_psorb &= ~PB2_PSORB0;
		io->iop_psorb |= PB2_PSORB1;
		io->iop_pparb |= (PB2_DIRB0 | PB2_DIRB1);
	}
	if (fip->fc_proff == PROFF_FCC3) {
		/* Configure port B and C pins for FCC Ethernet.
		 */
		io->iop_pdirb &= ~PB3_DIRB0;
		io->iop_pdirb |= PB3_DIRB1;
		io->iop_psorb &= ~PB3_PSORB0;
		io->iop_psorb |= PB3_PSORB1;
		io->iop_pparb |= (PB3_DIRB0 | PB3_DIRB1);
	}

	/* Port C has clocks......
	*/
	io->iop_psorc &= ~(fip->fc_trxclocks);
	io->iop_pdirc &= ~(fip->fc_trxclocks);
	io->iop_pparc |= fip->fc_trxclocks;

	/* ....and the MII serial clock/data.
	*/
	io->iop_pdatc |= (fip->fc_mdio | fip->fc_mdck);
	io->iop_podrc |= fip->fc_mdio;
	io->iop_pdirc |= (fip->fc_mdio | fip->fc_mdck);
	io->iop_pparc &= ~(fip->fc_mdio | fip->fc_mdck);

	/* Configure Serial Interface clock routing.
	 * First, clear all FCC bits to zero,
	 * then set the ones we want.
	 */
	immap->im_cpmux.cmx_fcr &= ~(fip->fc_clockmask);
	immap->im_cpmux.cmx_fcr |= fip->fc_clockroute;
}

static void __init
init_fcc_param(fcc_info_t *fip, struct net_device *dev,
						volatile immap_t *immap)
{
	unsigned char   *eap;
	unsigned long   mem_addr;
	bd_t            *bd;
	int             i;
	struct          fcc_enet_private *cep;
	volatile        fcc_enet_t      *ep;
	volatile        cbd_t           *bdp;
	volatile        cpm8260_t       *cp;

	cep = (struct fcc_enet_private *)(dev->priv);
	ep = cep->ep;
	cp = cpmp;

	bd = (bd_t *)__res;

	/* Zero the whole thing.....I must have missed some individually.
	 * It works when I do this.
	 */
	memset((char *)ep, 0, sizeof(fcc_enet_t));

	/* Allocate space for the buffer descriptors in the DP ram.
	 * These are relative offsets in the DP ram address space.
	 * Initialize base addresses for the buffer descriptors.
	 */
#if 0
	/* I really want to do this, but for some reason it doesn't
	 * work with the data cache enabled, so I allocate from the
	 * main memory instead.
	 */
	i = m8260_cpm_dpalloc(sizeof(cbd_t) * RX_RING_SIZE, 8);
	ep->fen_genfcc.fcc_rbase = (uint)&immap->im_dprambase[i];
	cep->rx_bd_base = (cbd_t *)&immap->im_dprambase[i];

	i = m8260_cpm_dpalloc(sizeof(cbd_t) * TX_RING_SIZE, 8);
	ep->fen_genfcc.fcc_tbase = (uint)&immap->im_dprambase[i];
	cep->tx_bd_base = (cbd_t *)&immap->im_dprambase[i];
#else
	cep->rx_bd_base = (cbd_t *)
		m8260_cpm_hostalloc(sizeof(cbd_t) * RX_RING_SIZE, 8);
	ep->fen_genfcc.fcc_rbase = __pa(cep->rx_bd_base);
	cep->tx_bd_base = (cbd_t *)
		m8260_cpm_hostalloc(sizeof(cbd_t) * TX_RING_SIZE, 8);
	ep->fen_genfcc.fcc_tbase = __pa(cep->tx_bd_base);
#endif

	cep->dirty_tx = cep->cur_tx = cep->cur_rx = 0;

	ep->fen_genfcc.fcc_rstate = (CPMFCR_GBL | CPMFCR_EB) << 24;
	ep->fen_genfcc.fcc_tstate = (CPMFCR_GBL | CPMFCR_EB) << 24;

	/* Set maximum bytes per receive buffer.
	 * It must be a multiple of 32.
	 */
	ep->fen_genfcc.fcc_mrblr = PKT_MAXBLR_SIZE;

	/* Allocate space in the reserved FCC area of DPRAM for the
	 * internal buffers.  No one uses this space (yet), so we
	 * can do this.  Later, we will add resource management for
	 * this area.
	 */
	mem_addr = CPM_FCC_SPECIAL_BASE + (fip->fc_fccnum * 128);
	ep->fen_genfcc.fcc_riptr = mem_addr;
	ep->fen_genfcc.fcc_tiptr = mem_addr+32;
	ep->fen_padptr = mem_addr+64;
	memset((char *)(&(immap->im_dprambase[(mem_addr+64)])), 0x88, 32);

	ep->fen_genfcc.fcc_rbptr = 0;
	ep->fen_genfcc.fcc_tbptr = 0;
	ep->fen_genfcc.fcc_rcrc = 0;
	ep->fen_genfcc.fcc_tcrc = 0;
	ep->fen_genfcc.fcc_res1 = 0;
	ep->fen_genfcc.fcc_res2 = 0;

	ep->fen_camptr = 0;     /* CAM isn't used in this driver */

	/* Set CRC preset and mask.
	*/
	ep->fen_cmask = 0xdebb20e3;
	ep->fen_cpres = 0xffffffff;

	ep->fen_crcec = 0;      /* CRC Error counter */
	ep->fen_alec = 0;       /* alignment error counter */
	ep->fen_disfc = 0;      /* discard frame counter */
	ep->fen_retlim = 15;    /* Retry limit threshold */
	ep->fen_pper = 0;       /* Normal persistence */

	/* Clear hash filter tables.
	*/
	ep->fen_gaddrh = 0;
	ep->fen_gaddrl = 0;
	ep->fen_iaddrh = 0;
	ep->fen_iaddrl = 0;

	/* Clear the Out-of-sequence TxBD.
	*/
	ep->fen_tfcstat = 0;
	ep->fen_tfclen = 0;
	ep->fen_tfcptr = 0;

	ep->fen_mflr = PKT_MAXBUF_SIZE;   /* maximum frame length register */
	ep->fen_minflr = PKT_MINBUF_SIZE;  /* minimum frame length register */

	/* Set Ethernet station address.
	 *
	 * This is supplied in the board information structure, so we
	 * copy that into the controller.
	 * So, far we have only been given one Ethernet address. We make
	 * it unique by setting a few bits in the upper byte of the
	 * non-static part of the address.
	 */
	eap = (unsigned char *)&(ep->fen_paddrh);
#ifdef CONFIG_SIPO
	/*
	 * In SSH SIPO the given address is the first ethernet, others are just
	 * simply added from there
	 */
	{
		uint laddr = (((bd->bi_enetaddr[3] << 16) & 0xff0000)
			     |((bd->bi_enetaddr[4] << 8) & 0xff00)
			     | (bd->bi_enetaddr[5] & 0xff)) + fip->fc_fccnum;

		eap[5] = dev->dev_addr[0] = bd->bi_enetaddr[0];
		eap[4] = dev->dev_addr[1] = bd->bi_enetaddr[1];
		eap[3] = dev->dev_addr[2] = bd->bi_enetaddr[2];
		eap[2] = dev->dev_addr[3] = (laddr >> 16) & 0xff;
		eap[1] = dev->dev_addr[4] = (laddr >> 8) & 0xff;
		eap[0] = dev->dev_addr[5] = laddr & 0xff;
	}
#else
	for (i=5; i>=0; i--) {
		if (i == 3) {
			dev->dev_addr[i] = bd->bi_enetaddr[i];
			dev->dev_addr[i] |= (1 << (7 - fip->fc_fccnum));
			*eap++ = dev->dev_addr[i];
		}
		else {
			*eap++ = dev->dev_addr[i] = bd->bi_enetaddr[i];
		}
	}
#endif

	ep->fen_taddrh = 0;
	ep->fen_taddrm = 0;
	ep->fen_taddrl = 0;

	ep->fen_maxd1 = PKT_MAXDMA_SIZE;	/* maximum DMA1 length */
	ep->fen_maxd2 = PKT_MAXDMA_SIZE;	/* maximum DMA2 length */

	/* Clear stat counters, in case we ever enable RMON.
	*/
	ep->fen_octc = 0;
	ep->fen_colc = 0;
	ep->fen_broc = 0;
	ep->fen_mulc = 0;
	ep->fen_uspc = 0;
	ep->fen_frgc = 0;
	ep->fen_ospc = 0;
	ep->fen_jbrc = 0;
	ep->fen_p64c = 0;
	ep->fen_p65c = 0;
	ep->fen_p128c = 0;
	ep->fen_p256c = 0;
	ep->fen_p512c = 0;
	ep->fen_p1024c = 0;

	ep->fen_rfthr = 0;      /* Suggested by manual */
	ep->fen_rfcnt = 0;
	ep->fen_cftype = 0;

	/* Now allocate the host memory pages and initialize the
	 * buffer descriptors.
	 */
	bdp = cep->tx_bd_base;
	for (i=0; i<TX_RING_SIZE; i++) {

		/* Initialize the BD for every fragment in the page.
		*/
#ifdef USE_PIT_INTR_POLL
		bdp->cbd_sc = 0;
#else
		bdp->cbd_sc = BD_ENET_TX_INTR;
#endif
		bdp->cbd_datlen = 0;
		bdp->cbd_bufaddr = 0;
		bdp++;
	}

#ifdef USE_PIT_INTR_POLL
	/* Only last and middle BD's generate interrupt */
	cep->tx_bd_base[TX_RING_SIZE/2].cbd_sc |= BD_ENET_TX_INTR;
	cep->tx_bd_base[TX_RING_SIZE-1].cbd_sc |= BD_ENET_TX_INTR;
#endif
	/* Set the last buffer to wrap.
	*/
	bdp--;
	bdp->cbd_sc |= BD_SC_WRAP;

	bdp = cep->rx_bd_base;
	for (i=0; i<RX_RING_SIZE; i++) {
		struct sk_buff *skb;

		skb = dev_alloc_skb(PKT_MAXBLR_SIZE + RESERVE_BUFFER_HEAD +
			RESERVE_BUFFER_TAIL);
		if (skb == NULL) {
			printk("%s: could not allocate all the skb\n",
			       dev->name);
			break;
		}

		/* 16 byte align the IP header. */
		skb_reserve(skb, RESERVE_BUFFER_HEAD);

		/* Set buffer descriptor to point to new buffer */
#ifdef USE_PIT_INTR_POLL
		bdp->cbd_sc = BD_ENET_RX_EMPTY;
#else
		bdp->cbd_sc = BD_ENET_RX_EMPTY | BD_ENET_RX_INTR;
#endif
		bdp->cbd_bufaddr = __pa(skb->data);
		cep->rx_skbuff[i] = skb;
		bdp++;
	}

	/* Set the last buffer to wrap.
	*/
	bdp--;
	bdp->cbd_sc |= BD_SC_WRAP;

	/* Let's re-initialize the channel now.  We have to do it later
	 * than the manual describes because we have just now finished
	 * the BD initialization.
	 */
	cp->cp_cpcr = mk_cr_cmd(fip->fc_cpmpage, fip->fc_cpmblock, 0x0c,
			CPM_CR_INIT_TRX) | CPM_CR_FLG;
	while (cp->cp_cpcr & CPM_CR_FLG);
}

/* Let 'er rip.
*/
static void __init
init_fcc_startup(fcc_info_t *fip, struct net_device *dev)
{
	volatile fcc_t  *fccp;
	struct fcc_enet_private *cep;

	cep = (struct fcc_enet_private *)(dev->priv);
	fccp = cep->fccp;

	/* Set GFMR to enable Ethernet operating mode.
	 */
	fccp->fcc_gfmr = (FCC_GFMR_TCI | FCC_GFMR_MODE_ENET);

	/* Set protocol specific processing mode for Ethernet.
	 * This has to be adjusted for Full Duplex operation after we can
	 * determine how to detect that.
	 */
	if(cep->fip->link_full_duplex)
	    fccp->fcc_fpsmr = FCC_PSMR_ENCRC | FCC_PSMR_FDE |
		    FCC_PSMR_FCE | FCC_PSMR_LPB;
	else
	    fccp->fcc_fpsmr = FCC_PSMR_ENCRC | FCC_PSMR_FCE;

	cep->fip->fcc_full_duplex=cep->fip->link_full_duplex;

	/* Set sync/delimiters.
	*/
	fccp->fcc_fdsr = 0xd555;

	/* Clear any pending events */
	fccp->fcc_fcce = 0xffff;

	/* Enable interrupts for transmit error, complete frame
	 * received, and any transmit buffer we have also set the
	 * interrupt flag.
	 */

#ifdef USE_PIT_INTR_POLL
	fccp->fcc_fccm = FCC_ENET_BSY | FCC_ENET_TXB | FCC_ENET_TXE;
#else
	fccp->fcc_fccm = (FCC_ENET_TXE | FCC_ENET_RXF | FCC_ENET_TXB |
			  FCC_ENET_BSY);
#endif

	/* Install our interrupt handler.
	*/
	if (request_8xxirq(fip->fc_interrupt, fcc_enet_interrupt, 0,
							"fenet", dev) < 0)
		printk("Can't get FCC IRQ %d\n", fip->fc_interrupt);

	/* FIX: should these steps mentioned in the manual be used? */
	/* write SCPRR_H to configure the FCC interrupt priority */
	/* clear out any current interrupts in the SIPNR_L, if preferred */
	/* write the SIMR_L to enable interrupts to the CP interrupt
	 * controller */
	/* issue an INIT TX AND RX PARAMETERS command */

	/* And last, enable the transmit and receive processing.
	*/
	fccp->fcc_gfmr |= (FCC_GFMR_ENR | FCC_GFMR_ENT);
}

static int __init fcc_setup(char *str)
{
	int n;

	if ((n = *str - '0') < 0 || n >= NUM_FCC) {
		printk(KERN_WARNING "fcc_setup: Invalid fcc num, %s "
		       "ignored.\n", str);
		return 1;
	}

	if (str[1] != '=') {
		printk("fcc_setup: Syntax Error while expecting '='. %s "
		       "ignored\n", str);
		return 1;
	}

	if (!strncmp(str + 2, "100,full", 8)) {
		fcc_ports[n].mode = FCC_FORCE_100_FULL;
		printk("fcc%d is now forced to 100mbit/s full duplex\n", n);
		str += 8;
	} else if (!strncmp(str + 2, "10,full", 8)) {
		fcc_ports[n].mode = FCC_FORCE_10_FULL;
		printk("fcc%d is now forced to 10mbit/s full duplex\n", n);
		str += 7;
	} else if (!strncmp(str + 2, "100,half", 8)) {
		fcc_ports[n].mode = FCC_FORCE_100_HALF;
		printk("fcc%d is now forced to 100mbit/s half duplex\n", n);
		str += 8;
	} else if (!strncmp(str + 2, "10,half", 8)) {
		fcc_ports[n].mode = FCC_FORCE_10_HALF;
		printk("fcc%d is now forced to 10mbit/s half duplex\n", n);
		str += 7;
	} else
		printk("fcc_setup: Invalid string '%s'. Expecting "
		       "'fcc%d=<10|100>,<half|full>'\n", str, n);

	return 1;
}

__setup("fcc", fcc_setup);
