/*
 *      SipO FPGA 0.0001b: FPGA programming interface driver
 *
 *      Copyright 2000 SSH Communications Security Corp, All Rights Reserved.
 *                              http://www.ssh.com
 *
 *      23.9.2000 Arto Vuori <avuori@ssh.com>
 *              Initial version
 *
 *      4.12.2001 Arto Vuori <avuori@ssh.com>
 *              Pin re-mapping for sipo2 hardware
 */
#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/reboot.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include <asm/sipo.h>

#include <asm/immap_8260.h>
#include <asm/cpm_8260.h>
#include <asm/mpc8260.h>

typedef struct {
	uint *init_dir;
	uint *init_dat;
	uint  init_pin;

	uint *prog_dir;
	uint *prog_dat;
	uint  prog_pin;

	uint *done_dir;
	uint *done_dat;
	uint  done_pin;

	uint *cclk_dir;
	uint *cclk_dat;
	uint  cclk_pin;

	uint *data_dir;
	uint *data_dat;
	uint  data_pin;

	uint *pwdn_dir;
	uint *pwdn_dat;
	uint  pwdn_pin;
} pin_table;

pin_table sipo_pins =
	{
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdirc,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdatc,
		(0x80000000>>23),

		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdira,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdata,
		(0x80000000>>2),

		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdira,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdata,
		(0x80000000>>3),

		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdirc,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdatc,
		(0x80000000>>31),

		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdira,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdata,
		(0x80000000>>13),

		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdird,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdatd,
		(0x80000000>>10)
	};

pin_table sipo2_pins =
	{
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdird,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdatd,
		(0x80000000>>11),

		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdira,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdata,
		(0x80000000>>1),

		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdira,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdata,
		(0x80000000>>2),

		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdirc,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdatc,
		(0x80000000>>31),

		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdira,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdata,
		(0x80000000>>13),

		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdird,
		&((immap_t *)(IMAP_ADDR))->im_ioport.iop_pdatd,
		(0x80000000>>10)
	};

static pin_table * fpga_pins=0;

#define FPGA_INIT_DIR (*(fpga_pins->init_dir))
#define FPGA_INIT_DAT (*(fpga_pins->init_dat))
#define FPGA_INIT_PIN (fpga_pins->init_pin)

#define FPGA_PROG_DIR (*(fpga_pins->prog_dir))
#define FPGA_PROG_DAT (*(fpga_pins->prog_dat))
#define FPGA_PROG_PIN (fpga_pins->prog_pin)

#define FPGA_DONE_DIR (*(fpga_pins->done_dir))
#define FPGA_DONE_DAT (*(fpga_pins->done_dat))
#define FPGA_DONE_PIN (fpga_pins->done_pin)

#define FPGA_CCLK_DIR (*(fpga_pins->cclk_dir))
#define FPGA_CCLK_DAT (*(fpga_pins->cclk_dat))
#define FPGA_CCLK_PIN (fpga_pins->cclk_pin)

#define FPGA_DATA_DIR (*(fpga_pins->data_dir))
#define FPGA_DATA_DAT (*(fpga_pins->data_dat))
#define FPGA_DATA_PIN (fpga_pins->data_pin)

#define FPGA_PWDN_DIR (*(fpga_pins->pwdn_dir))
#define FPGA_PWDN_DAT (*(fpga_pins->pwdn_dat))
#define FPGA_PWDN_PIN (fpga_pins->pwdn_pin)


static uint bitlen;
typedef enum {uninitialized, config, done, initialized, error} fpga_prg_state;

static fpga_prg_state fpga_state=uninitialized;

#ifdef MODULE
MODULE_PARM(soft_margin,"i");
#endif

/*
 *      Xilinx FPGA configuration start sequence.
 */
static int fpga_open(struct inode *inode, struct file *file)
{
	int i;

	if (fpga_state != uninitialized && fpga_state != initialized) {
		printk("SipO FPGA: not ready for configuration\n");
		return -EBUSY;
	}

	FPGA_PROG_DAT &= ~FPGA_PROG_PIN;
	FPGA_CCLK_DAT &= ~FPGA_CCLK_PIN;
	FPGA_DATA_DAT &= ~FPGA_DATA_PIN;

	/* PROG down, wait for INIT to go down */
	for (i = 0; i <= HZ; i++) {
		if (!(FPGA_INIT_DAT & FPGA_INIT_PIN))
			break;
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(1);
	}
	if (FPGA_INIT_DAT & FPGA_INIT_PIN) {
		printk("SipO FPGA:: INIT should not remain high\n");
		return -EIO;
	}

	/* PROG UP, wait fo INIT to go up */
	FPGA_PROG_DAT |= FPGA_PROG_PIN;
	for (i = 0; i < HZ; i++) {
		if (FPGA_INIT_DAT & FPGA_INIT_PIN)
			break;
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(1);
	}
	if (!(FPGA_INIT_DAT & FPGA_INIT_PIN)) {
		printk("SipO FPGA:: INIT should rise after PROG\n");
		return -EIO;
	}

	fpga_state = config;
	bitlen = 0;
	return 0;
}
/*
 *      Configuration end sequence. Check that chip configuration is OK.
 */
static int fpga_release(struct inode *inode, struct file *file)
{
	int i;

	if (fpga_state != done) {
		printk("SipO FPGA: Not initialized\n");
		fpga_state = uninitialized;
		return -EIO;
	}
	fpga_state = uninitialized;

	if (!(FPGA_INIT_DAT & FPGA_INIT_PIN)) {
		printk("SipO FPGA: INIT gone low after configuration\n");
		return -EIO;
	}

	/* Wait for DONE */
	for(i = 0; i < HZ; i++) {
		if (FPGA_DONE_DAT & FPGA_DONE_PIN)
			break;
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(1);
	}

	if (!(FPGA_DONE_DAT & FPGA_DONE_PIN)) {
		printk("SipO FPGA: DONE should rise after configuration\n");
		return -EIO;
	}

	printk("SipO FPGA: Configuration done\n");
	fpga_state = initialized;
	return 0;
}

/*
 * Write data block to serial configuration interface
 */

static ssize_t fpga_write(struct file *file, const char *data, size_t len,
			  loff_t *ppos)
{
	int l = len, b, c;
	const char *cp = data;

	if (fpga_state != config) {
		printk("SipO FPGA: Not in configuration mode\n");
		return -EIO;
	}

	/* first write after open */
	if (bitlen==0) {
		/* Check MAGIC */
		if (len < 8 ||
		    data[0] != (char) 0xf0 || data[1] != (char) 0x0b ||
		    data[2] != (char) 0xa3 || data[3] != (char) 0x7e) {
			fpga_state = error;
			printk("SipO FPGA: Bitstream header error\n");
			return -EIO;
		}

		/* Read length */
		bitlen = 0;
		for (b = 0; b < 4; b++)
			bitlen = (bitlen << 8) + ((uint) data[b + 4] & 0xff);
		cp += 8;
		l -= 8;
	}

	if (!(FPGA_INIT_DAT & FPGA_INIT_PIN)) {
		fpga_state = error;
		printk("SipO FPGA: INIT gone low while doing configuration\n");
		return -EIO;
	}

	cli();
	udelay(100);

	/* Shift bitstream out */
	for (b = 0 ; l > 0 && bitlen > 0; bitlen--) {
		c = *cp;
		if (c & (0x80 >> b))
			FPGA_DATA_DAT |= FPGA_DATA_PIN;
		else
			FPGA_DATA_DAT &= ~FPGA_DATA_PIN;

		udelay(1);
		FPGA_CCLK_DAT |= FPGA_CCLK_PIN;

		udelay(1);
		FPGA_CCLK_DAT &= ~FPGA_CCLK_PIN;

		b = b + 1;
		if ( b > 7) {
			b = 0;
			l--;
			cp++;

			if (l % 1024 == 0) {
				udelay(100);
				sti();
				udelay(1);
				cli();
				udelay(100);
			}
		}
	}
	udelay(100);
	sti();

	if (!(FPGA_INIT_DAT & FPGA_INIT_PIN)) {
		fpga_state = error;
		printk("SipO FPGA: INIT gone low while doing configuration\n");
		return -EIO;
	}

	if (l > 1) {
		fpga_state = error;
		printk("SipO FPGA: Bitstream too long\n");
		return -EIO;
	}

	if (bitlen == 0)
		fpga_state = done;

	return len;
}

#if 0
static ssize_t fpga_read(struct file * file, char * buf, size_t count,
			 loff_t *ppos)
{
	/* TODO: */
	return 0;
}
#endif

static struct file_operations fpga_fops=
{
	owner:          THIS_MODULE,
	write:          fpga_write,
	open:           fpga_open,
	release:        fpga_release,
};

static struct miscdevice fpga_miscdev=
{
	SIPOFPGA_MINOR,
	"sipofpga",
	&fpga_fops
};

/*
 * Initialize FPGA configuration IO pins
 */

void __init sipofpga_init(void)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;

	bd_t *__bd = (bd_t *) __res;

	misc_register(&fpga_miscdev);
	printk("SipO FPGA: 0.0001b\n");

	immap = (immap_t *)IMAP_ADDR;
	io = &immap->im_ioport;

	if (__bd->hw_version == 4) {
		fpga_pins = &sipo2_pins;

		io->iop_podra &= ~(FPGA_PROG_PIN|FPGA_DATA_PIN|FPGA_DONE_PIN);
		io->iop_podrc &= ~(FPGA_CCLK_PIN);
		io->iop_podrd &= ~(FPGA_PWDN_PIN|FPGA_INIT_PIN);

		io->iop_ppara &= ~(FPGA_PROG_PIN|FPGA_DATA_PIN|FPGA_DONE_PIN);
		io->iop_pparc &= ~(FPGA_CCLK_PIN);
		io->iop_ppard &= ~(FPGA_PWDN_PIN|FPGA_INIT_PIN);
	} else {
		fpga_pins = &sipo_pins;

		io->iop_podra &= ~(FPGA_PROG_PIN|FPGA_DATA_PIN|FPGA_DONE_PIN);
		io->iop_podrc &= ~(FPGA_CCLK_PIN|FPGA_INIT_PIN);
		io->iop_podrd &= ~(FPGA_PWDN_PIN);

		io->iop_ppara &= ~(FPGA_PROG_PIN|FPGA_DATA_PIN|FPGA_DONE_PIN);
		io->iop_pparc &= ~(FPGA_CCLK_PIN|FPGA_INIT_PIN);
		io->iop_ppard &= ~(FPGA_PWDN_PIN);
	}


	FPGA_PROG_DIR |= FPGA_PROG_PIN;
	FPGA_INIT_DIR &= ~FPGA_INIT_PIN;
	FPGA_DONE_DIR &= ~FPGA_DONE_PIN;
	FPGA_CCLK_DIR |= FPGA_CCLK_PIN;
	FPGA_DATA_DIR |= FPGA_DATA_PIN;
	FPGA_PWDN_DIR |= FPGA_PWDN_PIN;

	FPGA_PROG_DAT &= ~FPGA_PROG_PIN;
	FPGA_CCLK_DAT &= ~FPGA_CCLK_PIN;
	FPGA_DATA_DAT &= ~FPGA_DATA_PIN;
	FPGA_PWDN_DAT |= FPGA_PWDN_PIN;
}

#ifdef MODULE
int init_module(void)
{
	sipofpga_init();
	return 0;
}

void cleanup_module(void)
{
	misc_deregister(&fpga_miscdev);
}
#endif
