/*

  fpgauart.c - UART (scard)

  Author: Santeri Paavolainen <santtu@ssh.com>

  Copyright 2002 SSH Communications Security Corp,
  All Rights Reserved. http://www.ssh.com

*/

/* For documentation see vpn/fpga/README.uart in sipo_uart branch. */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/timer.h>

#include <asm/delay.h>
#include <asm/io.h>
#include <asm/immap_8260.h>
#include <asm/cpm_8260.h>

#include <asm/uaccess.h>
#include <asm/mpc8260.h>

#include "fpgaio.h"

#if 0
/* uart ready? */
static enum {
  NOTHERE,
  PRESENT
} uart_state = NOTHERE;
#endif

#define HERE(...)                                               \
do                                                              \
{                                                               \
  printk("%s:%d:%s: ", __FILE__, __LINE__, __FUNCTION__);        \
  printk(__VA_ARGS__);                                          \
  printk("\n");                                                 \
}                                                               \
while (0)

#if 0
static int clk_values_set, serial_clk_value, bit_clk_value;
static unsigned int *uart_reg_clk, *uart_reg_fifo, *uart_reg_status;
#endif

#define DEFAULT_SERIAL_CLK_DIV  20
#define DEFAULT_BIT_CLK_DIV 372
#define RECEIVE_BUFFER_SIZE 1024
#define TRANSMIT_BUFFER_SIZE 1024
#define UART_FIFO_SIZE 511

/* Status word masks */
#define UART_STATUS_TX (1 << 14)
#define UART_STATUS_RX (1 << 13)
#define UART_STATUS_TX_ERR (1 << 12)
#define UART_STATUS_RX_ERR (1 << 11)
#define UART_STATUS_CARD_READY (1 << 10)
#define UART_STATUS_CARD_PRESENT (1 << 9)
#define UART_STATUS_ELEMS 0x1ff

typedef struct fpga_uart_rec {
  /* Spin lock to protect against concurrent access to FIFO and this
     structure (this structure is allocated and deallocated durring
     single-threaded init/uninit) */
  spinlock_t lock;

  /* Whether we're open or closed. Only a single open is allowed. BTW,
     during open we actually check whether the device is present or
     not.. */
  enum {
    FPGA_UART_CLOSED,
    FPGA_UART_CLOSING,
    FPGA_UART_OPEN
  } state;

  /* UART register pointers */
  unsigned int *reg_clk, *reg_fifo, *reg_status;

  /* Serial and bit clock divisors (set to default on open) */
  unsigned int serial_clk_div, bit_clk_div;

  /* Notice: Buffers always start from offset 0 -- data that has been
     already either transmitted to UART FIFO or read by user is
     removed, and the rest is memmove'd to the beginning of the
     buffer. Yes, it is not O(n^2). */

  /* Receive buffer */
  unsigned char receive_buf[RECEIVE_BUFFER_SIZE];
  int receive_buf_len, receive_buf_overflow;

  /* Transmit buffer */
  unsigned char transmit_buf[TRANSMIT_BUFFER_SIZE];
  int transmit_buf_len;

  /* Error counters */
  unsigned int receive_errors, transmit_errors;

  /* Periodic timer -- checks for various things on the UART, like
     card removal, queue filling etc. (yeah, this is a polling
     interface on the UART, wait until we get a real IRQ support
     in)*/
  struct timer_list timer;

  /* Wait queue for polls */
  wait_queue_head_t wait_queue;

  /* Proc dir entry for this UART instance */
  struct proc_dir_entry *entry;

} fpga_uart_t;

static fpga_uart_t *uart = NULL;

#if 0
static void bitstring(unsigned int n, char *buf, int len)
{
  int i;

  for (i = 0; i < len; i++)
    if (n & ((1 << (len - 1 - i))))
      buf[i] = '1';
    else
      buf[i] = '0';

  buf[i] = '\0';
}
#endif

#if 0
static void reg_dump(void)
{
  unsigned int val, *vaddr;
  unsigned char *addr;
  int i, k;
  char bit1[33], bit2[33];

  for (i = 0; i < 32; i += 4)
    {
      /*for (k = 0; k < 5; k++)*/
      for (k = 0; k < 1; k++)
        {
          addr = sipofpgaio_ptr;
          addr = addr + i;

          vaddr = addr;
          val = *vaddr;

          bitstring((unsigned int) addr, bit1, 32);
          bitstring(val, bit2, 32);

          printk("FPGA UART: i=%3d addr=%08p [%s] "
                 "val=%010lu 0x%08lx [%s] (%d)\n",
                 i, addr, bit1, val, val, bit2, k);

          udelay(100000);
        }
    }
}
#endif

static int check_errors(fpga_uart_t *uart, unsigned int *val)
{
  int ret = 0;

  /* counter errors */
  if (*val & UART_STATUS_TX_ERR)
    {
      ret = 1;
      uart->transmit_errors++;
    }

  if (*val & UART_STATUS_RX_ERR)
    {
      ret = 1;
      uart->receive_errors++;
    }

  /* cause any errors to be cleared */
  *val &= ~(UART_STATUS_TX_ERR | UART_STATUS_RX_ERR);

  return ret;
}

static void uart_info(fpga_uart_t *uart)
{
  unsigned int val;

  val = *(uart->reg_status);
  if (check_errors(uart, &val))
    *(uart->reg_status) = val;

  printk("UART: %s%s%s%s%s%sElems=%d",
         (val & UART_STATUS_TX) ? "Transmitting " : "",
         (val & UART_STATUS_RX) ? "Receiving " : "",
         (val & UART_STATUS_TX_ERR) ? "TxErr " : "",
         (val & UART_STATUS_RX_ERR) ? "RxErr " : "",
         (val & UART_STATUS_CARD_READY) ? "CardReady " : "",
         (val & UART_STATUS_CARD_PRESENT) ? "CardPresent " : "",
         val & UART_STATUS_ELEMS);

  if (uart)
    printk(" %s %sRxBufLen=%d TxBufLen=%d RxErrs=%d TxErrs=%d",
           uart->state == FPGA_UART_CLOSED ? "Closed" :
           uart->state == FPGA_UART_CLOSING ? "Closing" : "Open",
           uart->receive_buf_overflow ? "ReceiveOverflow " : "",
           uart->receive_buf_len, uart->transmit_buf_len,
           uart->receive_errors, uart->transmit_errors);

  printk("\n");
}

#if 0
static int uart_init(void)
{
  unsigned int val;

  printk("FPGA UART: Initialize.\n");

  uart_reg_clk = (char*)sipofpgaio_ptr + 8;
  uart_reg_status = (char*)sipofpgaio_ptr + 16;
  uart_reg_fifo = (char*)sipofpgaio_ptr + 24;

  /* check magic constant, eg. version info */
  val = *uart_reg_clk;

  if (val != 0x02000101)
    {
      printk(KERN_ERR "UART: Version information (0x%08x) "
             "is not expected. FPGA probably not initialised?\n", val);
      return 1;
    }

  clk_values_set = 0;

  return 0;
}

static int uart_uninit(void)
{
  printk("FPGA UART: Uninitialize.\n");

  return 0;
}
#endif

static int proc_sipo_uart_read (char *buf, char **start, off_t off,
                                int len, int *eof, void *data)
{
  buf[0] = '\0';
  return 0;
}

static int proc_sipo_uart_write (struct file *file, const char *buffer,
                                 unsigned long len, void *data)
{
  return len;
#if 0
  unsigned char buf[512], *end, tmp[33];
  unsigned long val;


  if (len > sizeof(buf) - 1)
    return -EINVAL;

  if (copy_from_user(buf, buffer, len))
    return -EFAULT;

  buf[len] = '\0';

  /*printk("state=%d buf=%s\n", uart_state, buf);*/

  /* read argument value, if present */
  if (len > 1 && buf[1] >= '0' && buf[1] <= '9')
    {
      val = simple_strtoul(buf + 1, &end, 0);

      if (buf + 1 == end)
        return -EINVAL;
    }
  else
    val = 0;

  switch (*buf)
    {
      /* uninit */
    case '0':
      if (uart_state != PRESENT || uart_uninit())
        return -EINVAL;
      uart_state = NOTHERE;
      break;
      /* init */
    case '1':
      if (uart_state != NOTHERE || uart_init())
        return -EINVAL;
      uart_state = PRESENT;
      uart_info(NULL);
      break;
      /* info */
    case 'i':
      uart_info(NULL);
      break;
      /* reg dump */
    case 'd':
      reg_dump();
      break;
      /* write to status */
    case 's':
      uart_info(NULL);
      *uart_reg_status = val;
      uart_info(NULL);
      break;
      /* write to clock */
    case 'c':
      if (!clk_values_set)
        printk("UART: Clock not set.\n");
      else
        printk("UART: Old clock divisors: serial=%d bit=%d\n",
               serial_clk_value, bit_clk_value);

      serial_clk_value = (val >> 0) & 0x1ff;
      bit_clk_value = (val >> 16) & 0x1ff;
      clk_values_set = 1;

      *uart_reg_clk = val;

      printk("UART: New clock divisors: serial=%d bit=%d\n",
             serial_clk_value, bit_clk_value);
      break;
      /* write to FIFO */
    case 'w':
      uart_info(NULL);
      *uart_reg_fifo = val & 0xff;
      *uart_reg_fifo = val & 0xff;
      *uart_reg_fifo = val & 0xff;
      uart_info(NULL);
      break;
      /* read from FIFO */
    case 'r':
      uart_info(NULL);
      val = *uart_reg_fifo;
      uart_info(NULL);
      printk("UART: Read byte = %d 0x%02x\n", val & 0xff, val & 0xff);
      break;
    default:
      printk(KERN_ERR "UART: Unrecognized command '%c'.\n", *buf);
    }

  return len;
#endif
}

/************************************************************************/

static void uart_set_clk_div(fpga_uart_t *uart,
                             unsigned int serial, unsigned int bit)
{
  unsigned int val;

  /* mask unused bits to zero and combine */
  uart->serial_clk_div = serial & 0x1ff;
  uart->bit_clk_div = bit & 0x1ff;

  val = uart->bit_clk_div << 16 | uart->serial_clk_div;

  *(uart->reg_clk) = val;
}

static void uart_set_card_ready(fpga_uart_t *uart, int ready)
{
  unsigned int val;

  val = *(uart->reg_status);

  if (ready)
    val |= UART_STATUS_CARD_READY;
  else
    val &= ~UART_STATUS_CARD_READY;

  check_errors(uart, &val);

  *(uart->reg_status) = val;

  /* small delay before we actually start to do anything.. */
  udelay(1000);
}

/************************************************************************/

static void uart_timer(unsigned long data)
{
  fpga_uart_t *uart = (fpga_uart_t*) data;
  unsigned int val, i, wakeup, max;
  static int cnt = 0;

  wakeup = 0;

  spin_lock_irq(&uart->lock);

#if 1
  /* give status output 100 times every second */
  if ((cnt % 100) == 0)
    uart_info(uart);

  cnt++;
#endif

 redo:

  /* Get current status, update error counters */
  val = *(uart->reg_status);

  /* XXX: probably should change error bits to be clear-on-read
     semantics, there's a race here which potentially causes us to
     drop errors .. */
  if (check_errors(uart, &val))
    *(uart->reg_status) = val;

  /* Check if card has been removed. Wake up anyone. */
  if ((val & UART_STATUS_CARD_PRESENT) == 0)
    {
      uart->state = FPGA_UART_CLOSING;
      goto unlock_and_wakeup;
    }

  /* If there is data in the receive side, read it off */
  while (val & UART_STATUS_RX && (val & UART_STATUS_ELEMS) > 0)
    {
      wakeup = 1;

      for (i = val & UART_STATUS_ELEMS; i > 0; i--)
        {
          val = *(uart->reg_fifo);

          /* catch overflows */
          if (uart->receive_buf_len == sizeof(uart->receive_buf))
            uart->receive_buf_overflow = 1;
          else
            {
              uart->receive_buf[uart->receive_buf_len++] =
                (unsigned char) val & 0xff;
            }
        }

      /* go back again, and redo all */
      goto redo;
    }

  /* check if we can transmit data */
  if ((val & UART_STATUS_RX) == 0 && uart->transmit_buf_len > 0)
    {
      /* how many bytes can we fit into the UART FIFO? */
      max = UART_FIFO_SIZE - (val & UART_STATUS_ELEMS);

      for (i = 0; i < max && i < uart->transmit_buf_len; i++)
        *(uart->reg_fifo) = uart->transmit_buf[i];

      memmove(uart->transmit_buf, uart->transmit_buf + i,
              uart->transmit_buf_len - i);

      uart->transmit_buf_len -= i;

      /* No use of waking up the waiting writer if we just wrote a few
         bytes... so wake them up only when we've emptied the whole
         write buffer (to the uart's fifo) */

      if (uart->transmit_buf_len == 0)
        wakeup = 1;
    }

  if (wakeup)
    goto unlock_and_wakeup;

  spin_unlock_irq(&uart->lock);
  goto reset_timer;

 unlock_and_wakeup:
  spin_unlock_irq(&uart->lock);
  wake_up_interruptible(&uart->wait_queue);

 reset_timer:
  mod_timer(&uart->timer, jiffies + (HZ / 100));
}

/************************************************************************/

/* FPGA UART device operations */
static ssize_t uart_read(struct file * file, char * buf, size_t count,
                         loff_t *ppos)
{
  int i, len, ret;
  DECLARE_WAITQUEUE(wait, current);

  HERE("file=%p buf=%p count=%d ppos=%p", file, buf, count, ppos);

  if (!uart)
    return -ENODEV;

  spin_lock_irq(&uart->lock);

  uart_info(uart);

  ret = 0;

  add_wait_queue(&uart->wait_queue, &wait);

 redo:

  /* signal? */
  if (signal_pending(current))
    {
      printk("UART: Signal received while reading.\n");
      ret = -ERESTARTSYS;
      goto unlock_and_return;
    }

  /* card removed? */
  if (uart->state == FPGA_UART_CLOSING)
    {
      printk("UART: Card removed while/before reading.\n");
      ret = -ENOMEDIUM;
      goto unlock_and_return;
    }

  /* first check how much we have in the read buffer, and give that
     out if there is anything */

  if (uart->receive_buf_len > 0)
    {
      len = count > uart->receive_buf_len ? uart->receive_buf_len : count;

      memcpy(buf, uart->receive_buf, len);
      memmove(uart->receive_buf, uart->receive_buf + len,
              uart->receive_buf_len - len);
      uart->receive_buf_len -= len;

      ret = len;
      goto unlock_and_return;
    }

  /* Nonblocking read? */
  if (file->f_flags & O_NONBLOCK)
    {
      ret = 0;
      goto unlock_and_return;
    }

  printk("UART: Read waiting...\n");

  /* wait for something to happen */
  spin_unlock_irq(&uart->lock);
  interruptible_sleep_on(&uart->wait_queue);
  spin_lock_irq(&uart->lock);

  goto redo;

 unlock_and_return:
  remove_wait_queue(&uart->wait_queue, &wait);
  spin_unlock_irq(&uart->lock);
  return ret;
}

static ssize_t uart_write(struct file *file, const char *data, size_t len,
                          loff_t *ppos)
{
  int ret = 0;
  size_t left, max, pos;
  DECLARE_WAITQUEUE(wait, current);

  HERE("file=%p data=%p len=%d ppos=%p", file, data, len, ppos);

  if (!uart)
    return -ENODEV;

  spin_lock_irq(&uart->lock);

  uart_info(uart);

  /* Move data to the end of transmit buffer */
  left = len;
  ret = 0;

  add_wait_queue(&uart->wait_queue, &wait);

 redo:

  /* signal? */
  if (signal_pending(current))
    {
      printk("UART: Signal received while writing.\n");
      ret = -ERESTARTSYS;
      goto unlock_and_return;
    }

  /* card removed? */
  if (uart->state == FPGA_UART_CLOSING)
    {
      printk("UART: Card removed while/before writing.\n");
      ret = -ENOMEDIUM;
      goto unlock_and_return;
    }

  if (left == 0)
    goto unlock_and_return;

  max = sizeof(uart->transmit_buf) - uart->transmit_buf_len;
  max = max > left ? left : max;

  if (max > 0)
    {
      memcpy(uart->transmit_buf + uart->transmit_buf_len, data, max);
      uart->transmit_buf_len += max;
      left -= max;
      ret += max;
    }

  printk("UART: Wrote %d bytes, %d left (of total %d bytes), "
         "%d buffered.\n",
         max, left, len, uart->transmit_buf_len);

  /* If not all written, wait a bit -- unless this is nonblocking write */
  if (left > 0)
    {
      if (file->f_flags & O_NONBLOCK)
        {
          printk("UART: File in nonblocking mode, not waiting.\n");
          goto unlock_and_return;
        }

      printk("UART: Waiting...\n");

      spin_unlock_irq(&uart->lock);
      interruptible_sleep_on(&uart->wait_queue);
      spin_lock_irq(&uart->lock);
    }

  goto redo;

 unlock_and_return:
  uart_info(uart);
  remove_wait_queue(&uart->wait_queue, &wait);
  spin_unlock_irq(&uart->lock);
  return ret;
}

static unsigned int uart_poll(struct file *file, poll_table *wait)
{
  if (!uart)
    return -ENODEV;

  if (uart->state != FPGA_UART_OPEN)
    return -ENOMEDIUM;

  poll_wait(file, &(uart->wait_queue), wait);

  if (uart->state != FPGA_UART_OPEN)
    return -ENOMEDIUM;

  return POLLIN | POLLRDNORM;
}

static int uart_open(struct inode *inode, struct file *file)
{
  unsigned int val;

  HERE("inode=%p file=%p", inode, file);

  if (!uart)
    {
      printk("UART: Uart state not initialised.\n");
      return -ENODEV;
    }

  if (uart->state != FPGA_UART_CLOSED)
    {
      printk("UART: Already open.\n");
      return -EBUSY;
    }

  if (sipofpgaio_ptr == NULL)
    {
      printk("UART: No base FPGA pointer present.\n");
      return -ENODEV;
    }

  uart->reg_clk = (unsigned int*) ((char*)sipofpgaio_ptr + 8);
  uart->reg_status = (unsigned int*) ((char*)sipofpgaio_ptr + 16);
  uart->reg_fifo = (unsigned int*) ((char*)sipofpgaio_ptr + 24);

  val = *(uart->reg_clk);

  if (val != 0x02000101)
    {
      printk("UART: Magic version identifier invalid (0x%08x).\n", val);
      return -ENODEV;
    }

  val = *(uart->reg_status);

  if ((val & UART_STATUS_CARD_PRESENT) == 0)
    {
      printk("UART: Card not present.\n");
      return -ENOMEDIUM;
    }

  /* We have good FPGA, UART is present, card is present: okay to proceed */
  MOD_INC_USE_COUNT;

  uart->state = FPGA_UART_OPEN;
  uart->receive_buf_len = 0;
  uart->receive_buf_overflow = 0;
  uart->transmit_buf_len = 0;
  uart->transmit_errors = 0;
  uart->receive_errors = 0;

  uart_set_clk_div(uart, DEFAULT_SERIAL_CLK_DIV, DEFAULT_BIT_CLK_DIV);
  uart_set_card_ready(uart, 1);

  uart_info(uart);

  /* start timer periodically polling the UART */
  init_timer(&uart->timer);
  uart->timer.function = uart_timer;
  uart->timer.expires = jiffies + (HZ / 100);
  uart->timer.data = (unsigned long) uart;
  add_timer(&uart->timer);

  return 0;
}

static int uart_release(struct inode *inode, struct file *file)
{
  if (uart->state == FPGA_UART_CLOSED)
    return -ENODEV;

  uart_info(uart);

  /* remove timer */
  del_timer(&uart->timer);

  /* close it (card might have already been removed, we don't care) */
  uart_set_card_ready(uart, 0);

  uart->state = FPGA_UART_CLOSED;

  uart_info(uart);

  return 0;
}

static int uart_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
                      unsigned long arg)
{
  printk("UART: IOCTL cmd=0x%08x arg=0x%08lx\n", cmd, arg);
  return -EINVAL;
}

static struct file_operations uart_fops = {
  owner: THIS_MODULE,
  read: uart_read,
  write: uart_write,
  poll: uart_poll,
  open: uart_open,
  release: uart_release,
  ioctl: uart_ioctl
};

static struct miscdevice uart_miscdev = {
  SIPOFPGAUART_MINOR,
  "uart",
  &uart_fops
};

void __init sipouart_init(void)
{
  struct proc_dir_entry *entry;

  printk("FPGA UART: Starting in uninitialized state.\n");

  uart = kmalloc(sizeof(*uart), GFP_KERNEL);

  if (!uart)
    {
      printk(KERN_ERR "UART: Could not allocate state structure!\n");
      return;
    }

  uart->lock = SPIN_LOCK_UNLOCKED;
  uart->state = FPGA_UART_CLOSED;

  init_waitqueue_head(&uart->wait_queue);

  entry = create_proc_entry("uart", 0400, sipo_proc_root_dir);

  if (!entry)
    {
      printk(KERN_ERR "UART: Could not create proc entry!\n");
      kfree(uart);
      uart = NULL;
      return;
    }

  entry->nlink = 1;
  entry->read_proc = proc_sipo_uart_read;
  entry->write_proc = proc_sipo_uart_write;

  uart->entry = entry;

  misc_register(&uart_miscdev);
}
