/*
 * $Id: mtdflash.c,v 1.1 2002/09/15 08:38:14 jkm Exp $
 *
 * Flash mtd driver (uses kernel flash routines)
 *
 * Copyright 2001 SSH Communications Security Corp, All Rights Reserved.
 *                http://www.ssh.com
 *
 * 23.02.2001   Marko Teiste <marko.teiste@ssh.com>
 *              Initial version
 *
 * 16.03.2001   Jouni Malinen <jkm@ssh.com>
 *              Pass function return values to upper layers and save the
 *              original current->state.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>

#include <asm/sipo.h>
#include <asm/io.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>

#define FLASH_BLOCKSIZE (128*1024)
static int n_mtd = 0;

static int
sipo_mtd_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	unsigned long offset = (unsigned long)mtd->priv;
	long saved_state = current->state;
	int ret, tries = 0;

	if (instr->addr + instr->len > mtd->size) {
		printk("MTDFLASH: Erase goes beyond the partition\n");
		return -EINVAL;
	}

	if (instr->addr % FLASH_BLOCKSIZE) {
		printk("MTDFLASH: Erase does not start on block boundary\n");
		return -EINVAL;
	}

	if ((instr->addr + instr->len) % FLASH_BLOCKSIZE) {
		printk("MTDFLASH: Erase is not size of erase block\n");
		return -EINVAL;
	}

	do {
		ret = flash_erase(offset + instr->addr, instr->len);
		if (ret == 0)
			break;
		schedule();
		tries++;
	} while (tries < 10);

	if (ret != 0) {
		printk("flash_erase returned %d, tries=%d\n", ret, tries);
	}

	current->state = saved_state;

	if (ret)
		return ret;

	instr->state = MTD_ERASE_DONE;

	if (instr->callback)
		(*(instr->callback))(instr);

	return 0;
}

static int
sipo_mtd_read(struct mtd_info *mtd, loff_t from, size_t len,
	      size_t *retlen, u_char *buf)
{
	unsigned long offset = (unsigned long)mtd->priv;
	long saved_state = current->state;
	int ret;

	if (from + len > mtd->size) {
		printk("MTDFLASH: Trying to read outside the flash area\n");
		return -EINVAL;
	}

	ret = flash_read(buf, offset + from, len);
	current->state = saved_state;

	if (ret != 0) {
		printk("MTDFLASH: flash_read returned %d\n", ret);
	}

	*retlen = len;
	return ret;
}

static int
sipo_mtd_write(struct mtd_info *mtd, loff_t to, size_t len,
	       size_t *retlen, const u_char *buf)
{
	unsigned long offset = (unsigned long)mtd->priv;
	long saved_state = current->state;
	int ret;

	if (to + len > mtd->size) {
		printk("MTDFLASH: Trying to write outside the flash area\n");
		return -EINVAL;
	}

	ret = flash_write((char *) buf, offset + to, len);
	current->state = saved_state;

	if (ret != 0) {
		printk("MTDFLASH: flash_write returned %d\n", ret);
	}

	*retlen = len;
	return ret;
}

static struct mtd_info sipo_mtds[4] = {
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0,
		erasesize: FLASH_BLOCKSIZE,
		name: "SIPO MTD partition 0",
		erase: sipo_mtd_erase,
		read: sipo_mtd_read,
		write: sipo_mtd_write,
		priv: (void *)0
	},
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0,
		erasesize: FLASH_BLOCKSIZE,
		name: "SIPO MTD partition 1",
		erase: sipo_mtd_erase,
		read: sipo_mtd_read,
		write: sipo_mtd_write,
		priv: (void *)0
	},
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0,
		erasesize: FLASH_BLOCKSIZE,
		name: "SIPO MTD partition 2",
		erase: sipo_mtd_erase,
		read: sipo_mtd_read,
		write: sipo_mtd_write,
		priv: (void *)0
	},
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0,
		erasesize: FLASH_BLOCKSIZE,
		name: "SIPO MTD partition 3",
		erase: sipo_mtd_erase,
		read: sipo_mtd_read,
		write: sipo_mtd_write,
		priv: (void *)0
	}
};


#if LINUX_VERSION_CODE < 0x20212 && defined(MODULE)
#define init_sipomtd init_module
#define cleanup_sipomtd cleanup_module
#endif

int __init init_sipomtd(void)
{
	bd_t *__bd = (bd_t *) __res;
	int i;
	int mtd_i;

	printk(KERN_NOTICE "SIPO mtd flash device version 0.1\n");
	for (mtd_i = i = 0; i < pt_size; i++) {
		if (__bd->pt[i].type == pt_fs || __bd->pt[i].type == pt_jffs2)
		{
			printk(KERN_NOTICE "filesystem partition %d ", i);
			if (mtd_i < 4) {
				printk("mapped to mtd %d\n", mtd_i);
				sipo_mtds[mtd_i].size = __bd->pt[i].size;
				sipo_mtds[mtd_i].priv = (void *)
					__bd->pt[i].offset;
				if (((int) sipo_mtds[mtd_i].priv) %
				    FLASH_BLOCKSIZE)
					printk(KERN_NOTICE
					       "WARNING! partition does not "
					       "start at block boundary\n");
				if (sipo_mtds[mtd_i].size % FLASH_BLOCKSIZE)
					printk(KERN_NOTICE
					       "WARNING! partition size 0x%0x "
					       "is not multiple of "
					       "block size 0x%0x\n",
					       (unsigned int)
					       sipo_mtds[mtd_i].size,
					       FLASH_BLOCKSIZE);
				mtd_i++;
			} else {
				printk(KERN_NOTICE
				       "not mapped (number of supported mtd "
				       "devices (%d) exceeded)",
				       mtd_i);
			}
		}
	}

	n_mtd = mtd_i;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#ifdef MODULE
	for (i = 0; i < mtd_i; i++)
		sipo_mtds[i].module = THIS_MODULE;
#endif
#endif

	for (i = 0; i < mtd_i; i++)
		add_mtd_device(&sipo_mtds[i]);

  return 0;
}

static void __exit cleanup_sipomtd(void)
{
	int i;

	for (i = 0; i < n_mtd; i++)
		del_mtd_device(&sipo_mtds[i]);
}

module_init(init_sipomtd);
module_exit(cleanup_sipomtd);
