/*
 *      SipO Proc: /proc/sipo interface
 *
 *      Copyright 2001 SSH Communications Security Corp, All Rights Reserved.
 *                              http://www.ssh.com
 *
 *      12.2.2001 Arto Vuori <avuori@ssh.com>
 *              Initial version
 *
 *      This routines creates a /proc/sipo directory entry. In the directory
 *      is created file /proc/sipo/bd, which will can be used to view
 *      sipo board descriptor structure.
 */
#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/reboot.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include <asm/sipo.h>

#include <asm/immap_8260.h>
#include <asm/cpm_8260.h>
#include <asm/machdep.h>
#include <asm/mpc8260.h>

/*
 * /proc directory entry for sIPo specific stuff.
 */

struct proc_dir_entry *sipo_proc_root_dir;

static int proc_calc_metrics(char *page, char **start, off_t off,
			     int count, int *eof, int len)
{
	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

/*
 * Print contents of board descriptor in human readable format.
 */

static int proc_sipo_bd_read(char *page, char **start, off_t off,
			     int count, int *eof, void *data)
{
	int len;
	bd_t *__bd = (bd_t *) __res;
	char *p = page;

	/* Memory start address */
	p += sprintf(p, "memstart = 0x%08x\n", __bd->bi_memstart);
	/* Usable memory size in bytes */
	p += sprintf(p, "memsize  = 0x%08x\n", __bd->bi_memsize);
	/* Physical memory size in bytes */
	p += sprintf(p, "memtotal = 0x%08x\n", __bd->bi_memtotal);
	/* Internal Freq, in Hz */
	p += sprintf(p, "intfreq  = %d\n", __bd->bi_intfreq);
	/* Bus Freq, in MHz */
	p += sprintf(p, "busfreq  = %d\n", __bd->bi_busfreq);
	/* CPM Freq, in MHz */
	p += sprintf(p, "cpmfreq  = %d\n", __bd->bi_cpmfreq);
	/* BRG Freq, in MHz */
	p += sprintf(p, "brgfreq  = %d\n", __bd->bi_brgfreq);
	/* VCO Out from PLL */
	p += sprintf(p, "vco      = %d\n", __bd->bi_vco);
	/* Default console baud rate */
	p += sprintf(p, "baudrate = %d\n", __bd->bi_baudrate);
        /* IMMR when called from boot rom */
	p += sprintf(p, "immr     = 0x%08x\n", __bd->bi_immr);
	p += sprintf(p, "enetaddr = 0x%02x%02x%02x%02x%02x%02x\n",
		     __bd->bi_enetaddr[0], __bd->bi_enetaddr[1],
		     __bd->bi_enetaddr[2], __bd->bi_enetaddr[3],
		     __bd->bi_enetaddr[4], __bd->bi_enetaddr[5]);

	p += sprintf(p, "flashbase= 0x%08x\n", __bd->flash_base);
	p += sprintf(p, "flashsize= 0x%08x\n", __bd->flash_size);
	/* Board version descriptor */
	p += sprintf(p, "bd ver   = %d\n", __bd->bd_version);
	/* hardware version */
	p += sprintf(p, "hw ver   = %d\n", __bd->hw_version);
	/* software version */
	p += sprintf(p, "sw ver   = %d\n", __bd->sw_version);
	/* Boot state copied from NVRAM at initialization */
	p += sprintf(p, "bootstate= %d\n", __bd->boot_state);
	p += sprintf(p, "netboot  = %s\n",
		     __bd->enable_netboot ? "enabled" : "disabled");
	p += sprintf(p, "debugport= %s\n",
		     __bd->enable_debugport ? "enabled" : "disabled");
	p += sprintf(p, "passwd   = %s\n",
		     __bd->enable_passwd ? "enabled" : "disabled");
	p += sprintf(p, "watchdog = %s\n",
		     __bd->enable_wdt ? "enabled" : "disabled");

	len = p - page;

	return proc_calc_metrics(page, start, off, count, eof, len);
}

/*
 * Print contents of the partition table
 */

static int proc_sipo_pt_read(char *page, char **start, off_t off,
			     int count, int *eof, void *data)
{
	int len, i;
	bd_t *__bd = (bd_t *) __res;
	char *p = page;

	for(i = 0; i < pt_size; i++)
		if(__bd->pt[i].type) {
			p += sprintf(p, "index = %d, ", i);
			p += sprintf(p, "type = %d, ", __bd->pt[i].type);
			p += sprintf(p, "offset = 0x%08x, ",
				     __bd->pt[i].offset);
			p += sprintf(p, "size = 0x%x\n", __bd->pt[i].size);
		}

	len = p - page;

	return proc_calc_metrics(page, start, off, count, eof, len);
}


/*
 * Print status on smart card reader (1 = card inserted, 0 = no card)
 */

static int proc_sipo_scard_read(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	volatile immap_t *immap;
	volatile iop8260_t *io;
	int len, i;
	char *p = page;

	immap = immr;
	io = &immap->im_ioport;

	if (io->iop_pdatc & 0x00100000)
		i = 1;
	else
		i = 0;

	p += sprintf(p, "%d\n", i);
	len = p - page;

	return proc_calc_metrics(page, start, off, count, eof, len);
}


void m8260_restart(char *cmd); /* kernel/m8260_setup.c */

static int proc_sipo_restart_write(struct file *file, const char *buffer,
				   unsigned long len, void *data)
{
	m8260_restart(NULL);
	return 0;
}

void proc_sipo_init(void)
{
	struct proc_dir_entry *entry;

	printk("siPO: Directory entry '/proc/sipo'\n");

	/* create /proc/sipo */
	sipo_proc_root_dir = proc_mkdir("sipo", 0);

	/* create /proc/sipo/bd */
	entry = create_proc_entry("bd", 0400, sipo_proc_root_dir);
	entry->nlink = 1;
	entry->read_proc = proc_sipo_bd_read;

	/* create /proc/sipo/pt */
	entry = create_proc_entry("pt", 0400, sipo_proc_root_dir);
	entry->nlink = 1;
	entry->read_proc = proc_sipo_pt_read;

	/* create /proc/sipo/scard */
	entry = create_proc_entry("scard", 0400, sipo_proc_root_dir);
	entry->nlink = 1;
	entry->read_proc = proc_sipo_scard_read;

	/* create /proc/sipo/bd */
	entry = create_proc_entry("restart", 0400, sipo_proc_root_dir);
	entry->nlink = 1;
	entry->write_proc = proc_sipo_restart_write;
}
