/*
 * SIPO PC Card socket driver
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/vmalloc.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/bitops.h>
#include <asm/segment.h>
#include <asm/system.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/ss.h>
#include <pcmcia/cs.h>
#include <pcmcia/bus_ops.h>

#include <asm/uaccess.h>
#include <asm/delay.h>
#include <asm/immap_8260.h>


#define PCMCIA_DEBUG 3

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) \
do { if (pc_debug > (n)) printk(KERN_DEBUG args); } while (0)
#else
#define DEBUG(n, args...) do { } while (0)
#endif

/* These definitions set signals active for all four clock phases
 * setting individual clock phases is not covered here
 */
#define UPM_CS_0 0x00000000
#define UPM_CS_1 0xF0000000
#define UPM_BS_0 0x00000000
#define UPM_BS_1 0x0F000000
#define UPM_G0_0 0x00A00000
#define UPM_G0_1 0x00F00000
#define UPM_G1_0 0x00000000
#define UPM_G1_1 0x000C0000
#define UPM_G2_0 0x00000000
#define UPM_G2_1 0x00030000
#define UPM_G3_0 0x00000000
#define UPM_G3_1 0x0000C000
#define UPM_G3HL 0x00008000
#define UPM_G4_0 0x00000000
#define UPM_G4_1 0x00003000
#define UPM_G5_0 0x00000000
#define UPM_G5_1 0x00000C00
#define UPM_WAEN 0x00001000
#define UPM_WADI 0x00000000
#define UPM_REP1 0x00000000
#define UPM_REP2 0x00000100
#define UPM_REP3 0x00000200
#define UPM_REP4 0x00000300
#define UPM_LOOP 0x00000080
#define UPM_EXEN 0x00000040
#define UPM_ACOL 0x00000000
#define UPM_AROW 0x00000020
#define UPM_AMAR 0x00000030
#define UPM_NEXT 0x00000008
#define UPM_UTA  0x00000004
#define UPM_TODT 0x00000002
#define UPM_LAST 0x00000001

/* read single beat (RSS) offset 0x00  */
/* read burst (RBS) offset 0x08        */
/* write single beat (WSS) offset 0x18 */
/* write burst (WBS) offset 0x20       */
/* refresh timer (PTS) offset 0x30     */
/* exception condition (EXS) 0x3C      */

#define UPM_ARRAYSIZE 64
/* NOTES:
 * chip select and byte select are combined by external logic
 * because pc-card uses only byte select signals
 * For some yet unknown reason UPM_WAIT works only if the instruction
 * is repeated 4 times. I tried many combinations but it was only
 * one that worked
 * Processor also stops driving byte select signals after UTA is asserted
 * no matter what byte select values are in microinstructions
 */
static const unsigned long upm_array[UPM_ARRAYSIZE] = {
  /* RSS */
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_1 | UPM_G2_1 | UPM_G3_1 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_AROW,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_1 | UPM_G2_1 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_AROW,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_1 | UPM_G2_1 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_ACOL,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_1 | UPM_G2_0 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP2 | UPM_ACOL,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_0 | UPM_G1_1 | UPM_G2_0 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP3 | UPM_ACOL,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_0 | UPM_G1_1 | UPM_G2_0 | UPM_G3_0 | UPM_WAEN |
  UPM_G5_1 | UPM_REP4 | UPM_ACOL,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_0 | UPM_G1_1 | UPM_G2_0 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_ACOL | UPM_UTA | UPM_LAST,
  0,
  /* RBS */
  UPM_CS_1 | UPM_BS_1 | UPM_G0_1 | UPM_G1_1 | UPM_G2_1 | UPM_G3_1 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_LAST,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  /* WSS */
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_1 | UPM_G2_1 | UPM_G3_1 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_AROW,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_1 | UPM_G2_1 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_AROW,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_1 | UPM_G2_1 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_ACOL,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_1 | UPM_G2_0 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP2 | UPM_ACOL,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_0 | UPM_G2_0 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP3 | UPM_ACOL,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_0 | UPM_G2_0 | UPM_G3_0 | UPM_WAEN |
  UPM_G5_1 | UPM_REP4 | UPM_ACOL,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_1 | UPM_G2_0 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_ACOL,
  UPM_CS_0 | UPM_BS_0 | UPM_G0_1 | UPM_G1_1 | UPM_G2_0 | UPM_G3_0 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_ACOL | UPM_UTA | UPM_LAST,
  /* WBS */
  UPM_CS_1 | UPM_BS_1 | UPM_G0_1 | UPM_G1_1 | UPM_G2_1 | UPM_G3_1 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_LAST,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  /* PTS */
  UPM_CS_1 | UPM_BS_1 | UPM_G0_1 | UPM_G1_1 | UPM_G2_1 | UPM_G3_1 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_LAST,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  /* EXS */
  UPM_CS_1 | UPM_BS_1 | UPM_G0_1 | UPM_G1_1 | UPM_G2_1 | UPM_G3_1 | UPM_WADI |
  UPM_G5_1 | UPM_REP1 | UPM_LAST,
  0, 0, 0
};


#define MMR_B60X   0x00000000
#define MMR_BLOCAL 0x80000000
#define MMR_RFEN   0x40000000
#define MMR_NORMAL 0x00000000
#define MMR_WRITE  0x10000000
#define MMR_READ   0x20000000
#define MMR_RUN    0x30000000
#define MMR_DS1    0x00000000
#define MMR_DS2    0x00400000
#define MMR_DS3    0x00800000
#define MMR_DS4    0x00C00000
#define MMR_G4DI   0x00040000
#define MMR_MUX    0x04000000

#define BR_AMASK   0xFFFF8000
#define BR_PS8     0x00000800
#define BR_PS16    0x00001000
#define BR_PS32    0x00001800
#define BR_PS64    0x00000000
#define BR_MS_UPMA 0x00000080
#define BR_VALID   0x00000001

/* PC Card base address and offsets from base to io, attribute and memory space
 */
#define PCCARD_BASE 0xD0000000
#define PCCARD_SIZE 0x04000000
#define PCCARD_ATTR 0x00000000
#define PCCARD_IO   0x02000000
#define PCCARD_MEM  0x01000000
#define PCCARD_WINDOW_SIZE 0x800

#define PCCARD_IRQ_A 21
#define PCCARD_IRQ_B 23

/* these apply only to UPM mode */
/* #define OR_MEMSIZE 0xFFF00000 */
#define OR_MEMSIZE ~(PCCARD_SIZE-1)
#define OR_NOBURST 0x00000100

#define PIN22 0x00000200
#define PIN24 0x00000080
#define PIN26 0x00000020
#define PIN28 0x00000008
#define PIN30 0x00000002
#define PIN11 0x00100000

/* COR - Configuration Option Register (in attribute base) */
#define COR_SRESET 0x80
#define COR_LEVLREQ 0x40
#define COR_ENABLE_FUNC 0x01

#define CARD_POLL_INTERVAL (1 * HZ)
#define NUM_PCCARD_SLOTS 2

struct pccard_card_conf {
	char slot; /* A/B */
	int card_present;
	int card_ok;
	unsigned char *pccard_addr; /* ioremap'ed access to PC Card
				     * attr/mem/io */
	unsigned int cor_offset;
	u8 cor_index;
	u16 manfid1, manfid2;
	int only_5v;
	int io;
	int irq;

	void (*handler)(void *info, u_int events);
	void *info;
	socket_cap_t cap;
	u32 socket_flags;
	int power;
	u32 csc_mask;
	struct proc_dir_entry *proc_dir;
};

struct pccard_conf_data {
	struct pccard_card_conf slots[NUM_PCCARD_SLOTS];

	struct timer_list poll_timer;
};

static struct pccard_conf_data *pccard_conf;



int ssh_pccard_ready(struct pccard_card_conf *card);


void ssh_pccard_hard_reset(struct pccard_card_conf *card)
{
	volatile immap_t *immap = (immap_t *) IMAP_ADDR;
	int i;

	if (card->slot == 'A') {
		immap->im_ioport.iop_pdata &= ~PIN11;
		udelay(100);
		immap->im_ioport.iop_pdata |= PIN11;
	} else if (card->slot == 'B') {
		immap->im_ioport.iop_pdatc &= ~PIN30;
		udelay(100);
		immap->im_ioport.iop_pdatc |= PIN30;
	} else {
		printk("ssh_pccard_hard_reset: unknown slot\n");
		return;
	}

	udelay(100);
	i = 0;
	while (i < 10000) {
		if (ssh_pccard_ready(card)) {
			printk(KERN_DEBUG "ssh_pccard_hard_reset: pccard_ready"
			       " in %d*100 usec\n", i);
			return;
		}
		udelay(100);
		i++;
	}

	printk("ssh_pccard_hard_reset: timeout while waiting PC Card "
	       "initialization\n");
}


int ssh_pccard_present(struct pccard_card_conf *card)
{
	volatile immap_t *immap = (immap_t *) IMAP_ADDR;
	unsigned int pin;

	if (card->slot == 'A')
		pin = PIN26;
	else if (card->slot == 'B')
		pin = PIN28;
	else {
		printk("ssh_pccard_present: unknown slot\n");
		return 0;
	}

	if (immap->im_ioport.iop_pdatc & pin)
		return 0;

	return 1;
}


/* Note this routine clears pending interrupt bit
 * Call this function only after pc-card hard reset with
 * appropriate interrupt masked. After initialisation
 * same line acts as interrupt request and polling then 
 * may cause loss of interrupt requests.
 * We assume that interrupt lines are configured to
 * level sensitive mode.
 */
int ssh_pccard_ready(struct pccard_card_conf *card)
{
	volatile immap_t *immap = (immap_t *) IMAP_ADDR;
	unsigned int irq_mask;

	if (card->slot == 'A')
		irq_mask = 0x00001000; /* IRQ 3 */
	else if (card->slot == 'B')
		irq_mask = 0x00000400; /* IRQ 5 */
	else {
		printk("ssh_pccard_ready: unknown slot\n");
		return 0;
	}

	/* if pending interrupt is set it means that rdy/irq line
	 * is still low (=busy)
	 */
	if (immap->im_intctl.ic_sipnrh & irq_mask) {
		/* clear pending bit */
		immap->im_intctl.ic_sipnrh |= irq_mask;
		return 0;
	}

	return 1;
}


static void ssh_pccard_set_power(int sock, int enable)
{
	volatile immap_t       *immap;
	immap = (immap_t *) IMAP_ADDR;

	if (sock == 0 && enable)
		immap->im_ioport.iop_pdatc |= PIN22;
	else if (sock == 0 && !enable)
		immap->im_ioport.iop_pdatc &= ~PIN22;
	else if (sock == 1 && enable)
		immap->im_ioport.iop_pdatc |= PIN24;
	else if (sock == 1 && !enable)
		immap->im_ioport.iop_pdatc &= ~PIN24;

	pccard_conf->slots[sock].power = enable;
}

static void ssh_upm_setup(void)
{
	volatile unsigned long *pccard;
	volatile immap_t       *immap;
	volatile memctl8260_t  *memc;
	int i;

	immap = (immap_t *) IMAP_ADDR;

	/* set card A & B power enable pins */
	immap->im_ioport.iop_pparc &= ~(PIN24 | PIN22); /* gen.purp. I/O */
	immap->im_ioport.iop_pdirc |=  (PIN24 | PIN22); /* output */
	immap->im_ioport.iop_psorc &= ~(PIN24 | PIN22);
	immap->im_ioport.iop_podrc &= ~(PIN24 | PIN22); /* actively driven as
							 * output */
	/* set power off by default and enable it only when a card is present
	 */
	immap->im_ioport.iop_pdatc &= ~(PIN24 | PIN22); /* power off */

	/* set card A & B detection port to input mode */
	immap->im_ioport.iop_pparc &= ~(PIN26 | PIN28);
	immap->im_ioport.iop_pdirc &= ~(PIN26 | PIN28); /* input */
	immap->im_ioport.iop_psorc &= ~(PIN26 | PIN28);
	immap->im_ioport.iop_podrc &= ~(PIN26 | PIN28);

	/* Reset both cards until memory controller is configured */
	/* card A reset is controlled by port A, pin 11 */
	immap->im_ioport.iop_ppara &= ~PIN11;
	immap->im_ioport.iop_pdira |=  PIN11; /* output */
	immap->im_ioport.iop_psora &= ~PIN11;
	immap->im_ioport.iop_podra &= ~PIN11;
	immap->im_ioport.iop_pdata &= ~PIN11;

	/* card B reset is controlled by port B, pin 30 */
	immap->im_ioport.iop_pparc &= ~PIN30;
	immap->im_ioport.iop_pdirc |=  PIN30; /* output */
	immap->im_ioport.iop_psorc &= ~PIN30;
	immap->im_ioport.iop_podrc &= ~PIN30;
	immap->im_ioport.iop_pdatc &= ~PIN30;

	/* Setup 8260 Memory controller to map PC Card I/O, address and memory
	 * space */
	immap = (immap_t *) IMAP_ADDR;
	memc = &immap->im_memctl;
	memc->memc_or8 = OR_MEMSIZE | OR_NOBURST;
	memc->memc_br8 = (PCCARD_BASE & BR_AMASK) | BR_PS16 | BR_MS_UPMA |
		BR_VALID;

	memc->memc_or9 = OR_MEMSIZE | OR_NOBURST;
	memc->memc_br9 = ((PCCARD_BASE + PCCARD_SIZE) & BR_AMASK) | BR_PS16 |
		BR_MS_UPMA | BR_VALID;

	/* configure machine mode register for upm program load
	 * GPL4 must not be configured as upm_wait when program is loaded */
	memc->memc_mamr = MMR_BLOCAL | MMR_WRITE;
	printk(KERN_DEBUG "MAMR was set for write\n");

	/* load upm program */
	pccard = (volatile unsigned long *) pccard_conf->slots[0].pccard_addr;
	for (i = 0; i < UPM_ARRAYSIZE; i++) {
		/* load data into upm memory data register */
		memc->memc_mdr = upm_array[i];
		/* access to memory bank performs upm program write */
		*pccard = upm_array[i];
	}
	printk(KERN_DEBUG "UPM program loaded\n");

	/* configure machine mode register for normal operation */
	memc->memc_mamr = MMR_BLOCAL | MMR_NORMAL | MMR_DS1 | MMR_G4DI |
		MMR_MUX;
	printk(KERN_DEBUG "MAMR was set for normal\n");

	udelay(100);

	/* release resets */
	immap->im_ioport.iop_pdata |= PIN11;
	immap->im_ioport.iop_pdatc |= PIN30;
}


/* We do not have voltage detection pins available, so only way to determine
 * whether 3.3V or 5V is used is to parse CIS and figure out whether there are
 * cftable entries only for 3.3V */

/* FIX: CIS parsing should be removed from sipo_pcmcia.c...
 * CS already has code for this and it has probably even parsed CIS when this
 * info is needed, but it might be a bit ugly to modify CS for making this
 * easier.. */

#define CISTPL_CONFIG 0x1A
#define CISTPL_CFTABLE_ENTRY 0x1B
#define CISTPL_MANFID 0x20
#define CISTPL_END 0xFF

#define CIS_MAX_LEN 256

static inline int cftable_entry_3v(unsigned char v)
{
	int mant;
	DEBUG(4, "cftable_entry_3v(0x%x): exp=%d mant=%d\n",
	      v, (v & 0x7), ((v >> 3) & 0xf));
	if ((v & 0x7) == 5 /* 1 V */) {
		mant = (v >> 3) & 0xf;
		if (mant >= 5 && mant <= 7) /* 2.5V, 3V, 3.5V */
			return 1;
	}
	return 0;
}


static int ssh_pccard_read_cis(struct pccard_card_conf *card)
{
	unsigned char cis[CIS_MAX_LEN], *ptr, tpce_pd;
	int i;
	unsigned int pos;
	unsigned int rmsz, rasz;
	int tuple_config_found = 0, tuple_manfid_found = 0;
	int found_3v_cftable = 0;

	card->cor_offset = 0;
	card->cor_index = 0;
	card->manfid1 = card->manfid1 = 0;
	card->only_5v = 0;

	/* read CIS */
	for (i = 0; i < CIS_MAX_LEN; i++)
		cis[i] = *(card->pccard_addr + i * 2);

	pos = 0;
	while (pos < CIS_MAX_LEN - 1 && cis[pos] != CISTPL_END) {
		if (pos + cis[pos + 1] >= CIS_MAX_LEN) {
			printk(KERN_INFO "pccard: CIS overflow: pos=%d "
			       "cis[pos + 1]=%d\n", pos, cis[pos + 1]);
			if (tuple_manfid_found &&
			    (card->manfid1 == 0xc250 ||
			     card->manfid1 == 0xc240) &&
			    card->manfid2 == 2) {
				printk(KERN_INFO "pccard: Trying to cope with "
				       "invalid CIS on Emtac card\n");
				card->cor_index = 0x23;
				card->cor_offset = PCCARD_ATTR | 0x3e0;
				return 0;
			}

			return -1;
		}

		switch (cis[pos]) {
		case CISTPL_CONFIG:
			tuple_config_found = 1;
			rmsz = (cis[pos + 2] & 0x3C) >> 2;
			rasz = cis[pos + 2] & 0x03;
			if (4 + rasz + rmsz > cis[pos + 1])
				return -1;
			card->cor_index = cis[pos + 3] & 0x3F;
			card->cor_offset = PCCARD_ATTR;
			for (i = 0; i <= rasz; i++)
				card->cor_offset += cis[pos + 4 + i] <<
					(8 * i);
			DEBUG(4, "  cor_offset=0x%08x, cor_index=0x%02x\n",
			      card->cor_offset, card->cor_index);
			if (card->cor_offset > PCCARD_SIZE) {
				printk(KERN_INFO " CIS has too large COR "
				       "offset (0x%08x)\n", card->cor_offset);
				card->cor_offset = 0;
				return -1;
			}
			break;

		case CISTPL_MANFID:
			tuple_manfid_found = 1;
			if (cis[pos + 1] < 4)
				return -1;
			card->manfid1 = cis[pos + 2] + (cis[pos + 3] << 8);
			card->manfid2 = cis[pos + 4] + (cis[pos + 5] << 8);
			DEBUG(4, "  manfid= 0x%04x, 0x%04x\n", card->manfid1,
			      card->manfid2);
			break;

		case CISTPL_CFTABLE_ENTRY:
			/* FIX: check for cis overflow due to optional fields
			 */
			if (cis[pos + 1] < 2)
				return -1;
			ptr = &cis[pos + 1];
			DEBUG(4, "CFTABLE_ENTRY (TPL_LINK=%d)\n", *ptr);
			ptr++;
			DEBUG(4, "   TPCE_INDX=0x%02x\n", *ptr);
			if (*ptr & 0x80) {
				ptr++;
				DEBUG(4, "    TPCE_IF=0x%02x\n", *ptr);
			}
			ptr++;
			if (!(*ptr & 0x03)) {
				DEBUG(4, "    not power desc\n");
				break;
			}
			ptr++;
			tpce_pd = *ptr;
			ptr++;
			DEBUG(4, "   TPCE_PD=0x%02x\n", tpce_pd);
			if ((tpce_pd & 0x01 /* NormV */) &&
			    cftable_entry_3v(*ptr)) {
				DEBUG(4, "       NormV 3V\n");
				found_3v_cftable = 1;
				break;
			} else if ((tpce_pd & 0x02 /* MinV */) &&
				   cftable_entry_3v(*ptr)) {
				DEBUG(4, "       MinV 3V\n");
				found_3v_cftable = 1;
				break;
			} else if ((tpce_pd & 0x02 /* MaxV */) &&
				   cftable_entry_3v(*ptr)) {
				DEBUG(4, "       MaxV 3V\n");
				found_3v_cftable = 1;
				break;
			}

			break;
		}
		pos += cis[pos + 1] + 2;
	}

	if (pos >= CIS_MAX_LEN || cis[pos] != CISTPL_END) {
		printk(KERN_INFO "pccard: invalid CIS termination: cis[%d] = "
		       "0x%02x\n", pos, cis[pos]);
		return -1;
	}

	if (!tuple_config_found || !tuple_manfid_found) {
		printk(KERN_INFO "pccard: No CONFIG or MANFID tuple found in "
		       "CIS\n");
		return -1;
	}

	if (!found_3v_cftable)
		card->only_5v = 1;

	return 0;
}


static void ssh_pccard_set_io_mode(struct pccard_card_conf *card)
{
	printk(KERN_DEBUG "Setting IO mode (COR: 0x%02X => ",
	       card->pccard_addr[card->cor_offset]);
	card->pccard_addr[card->cor_offset] = card->cor_index |
		COR_LEVLREQ | COR_ENABLE_FUNC;
	printk("0x%02X)\n", card->pccard_addr[card->cor_offset]);
}


static int ssh_pccard_proc_read(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int sock;
	struct pccard_card_conf *slot;
	char *p = page;

	if (off != 0) {
		*eof = 1;
		return 0;
	}

	sock = (int) data;
	if (sock < 0 || sock >= NUM_PCCARD_SLOTS)
		return 0;
	slot = &pccard_conf->slots[sock];

	p += sprintf(p, "SSH PC Card v1\n");
	p += sprintf(p, "slot=%c\ncard=%d\nok=%d\nmanfid=0x%04x,0x%04x\n"
		     "irq=%d\nio=0x%08x\n5v=%d\ncor_offset=0x%x\n"
		     "cor_index=%d\nflags=0x%x\npower=%d\ncsc_mask=0x%x\n",
		     slot->slot, slot->card_present, slot->card_ok,
		     slot->manfid1, slot->manfid2, slot->irq, slot->io,
		     slot->only_5v, slot->cor_offset, slot->cor_index,
		     slot->socket_flags, slot->power, slot->csc_mask);
	p += sprintf(p, "cap: io_offset=0x%x map_size=0x%x "
		     "irq_mask=0x%x pci_irq=%d\n",
		     slot->cap.io_offset, slot->cap.map_size,
		     slot->cap.irq_mask, slot->cap.pci_irq);

	return (p - page);
}


static u_int pending_events[NUM_PCCARD_SLOTS];
static spinlock_t pending_event_lock = SPIN_LOCK_UNLOCKED;

static void sipo_pccard_bh(void *dummy)
{
	u_int events;
	int i;

	for (i = 0; i < NUM_PCCARD_SLOTS; i++) {
		spin_lock_irq(&pending_event_lock);
		events = pending_events[i];
		pending_events[i] = 0;
		spin_unlock_irq(&pending_event_lock);
		if (!pccard_conf->slots[i].handler)
			continue;

		DEBUG(4, "sipo_pccard_bh: calling handler\n");
		pccard_conf->slots[i].handler(pccard_conf->slots[i].info,
					      events);
		DEBUG(4, "sipo_pccard_bh: handler returned\n");
	}

	MOD_DEC_USE_COUNT;
}


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,44))

#define WORK_QUEUE struct tq_struct

#define SCHEDULE_WORK(q) \
MOD_INC_USE_COUNT; \
if (schedule_task((q)) == 0) \
	MOD_DEC_USE_COUNT;

#define FLUSH_SCHEDULED_WORK() flush_scheduled_tasks()

static inline void WORK_QUEUE_INIT(struct tq_struct *tq,
				   void (*routine)(void *), void *data)
{
	INIT_LIST_HEAD(&tq->list);
	tq->sync = 0;
	tq->routine = routine;
	tq->data = data;
}

#else /* kernel < 2.5.44 */

#define WORK_QUEUE struct work_struct

#define SCHEDULE_WORK(q) \
MOD_INC_USE_COUNT; \
if (schedule_work((q)) == 0) \
	MOD_DEC_USE_COUNT;

#define FLUSH_SCHEDULED_WORK() flush_scheduled_work()

static inline void WORK_QUEUE_INIT(struct work_struct *wq,
				   void (*routine)(void *), void *data)
{
	INIT_WORK(wq, routine, data);
}

#endif /* kernel < 2.5.44 */


static WORK_QUEUE sipo_pccard_task;


static void ssh_pccard_handle_poll(unsigned long data)
{
	struct pccard_conf_data *conf = (struct pccard_conf_data *) data;
	int i;

	for (i = 0; i < NUM_PCCARD_SLOTS; i++) {
		u_int events = 0;
		int present = ssh_pccard_present(&conf->slots[i]);

		if (present && !conf->slots[i].card_present) {
			printk("pccardupm: card inserted in slot %c\n",
			       conf->slots[i].slot);
			conf->slots[i].card_present = 1;
			conf->slots[i].card_ok = 1;
			events |= SS_DETECT;
			pccard_conf->slots[i].socket_flags = 0;
		} else if (!present && conf->slots[i].card_present) {
			printk("pccardupm: card removed from slot %c\n",
			       conf->slots[i].slot);
			conf->slots[i].card_present = 0;
			events |= SS_DETECT;
		}

		if (events) {
			DEBUG(4, "pccard: scheduling events 0x%x for slot "
			      "%d\n", events, i);
			spin_lock(&pending_event_lock);
			pending_events[i] |= events;
			spin_unlock(&pending_event_lock);
			SCHEDULE_WORK(&sipo_pccard_task);
		}
	}

	conf->poll_timer.expires = jiffies + CARD_POLL_INTERVAL;
	add_timer(&conf->poll_timer);
}


static int sipo_pccard_register_callback(unsigned int sock,
					 void (*handler)(void *, unsigned int),
					 void *info)
{
	DEBUG(4, "sipo_pccard_register_callback(%d, 0x%p, 0x%p)\n",
	      sock, handler, info);

	pccard_conf->slots[sock].handler = handler;
	pccard_conf->slots[sock].info = info;

	if (handler == NULL) {
		MOD_DEC_USE_COUNT;
	} else {
		MOD_INC_USE_COUNT;
	}

	return 0;
}


static int sipo_pccard_inquire_socket(unsigned int sock, socket_cap_t *cap)
{
	DEBUG(4, "sipo_pccard_inquire_socket\n");
	*cap = pccard_conf->slots[sock].cap;
	return 0;
}


#ifdef CONFIG_PROC_FS

static void sipo_pccard_proc_setup(unsigned int sock,
				   struct proc_dir_entry *base)
{
	DEBUG(4, "sipo_pccard_proc_setup\n");
	/* Create a proc entry (/proc/bus/pccard/...) */

	if (!create_proc_read_entry("info", 0, base,
				    ssh_pccard_proc_read, (void *) sock)) {
		printk(KERN_INFO "pccard: procfs making read entry failed\n");
		return;
	}
	pccard_conf->slots[sock].proc_dir = base;
}

static void sipo_pccard_proc_remove(u_short sock)
{
	struct proc_dir_entry *base = pccard_conf->slots[sock].proc_dir;
	if (base == NULL)
		return;
	remove_proc_entry("info", base);
}

#else /* CONFIG_PROC_FS */

#define sipo_pccard_proc_setup NULL

#endif /* CONFIG_PROC_FS */


static int sipo_pccard_get_status(unsigned int sock, u_int *value)
{
	DEBUG(4, "sipo_pccard_get_status(%d)\n", sock);

	*value = 0;

	if (pccard_conf->slots[sock].card_present) {
		*value |= SS_DETECT;
		if (pccard_conf->slots[sock].power)
			*value |= SS_POWERON | SS_READY;
	}

	if (!pccard_conf->slots[sock].only_5v)
		*value |= SS_3VCARD;

	/* FIX: should these be added?
	 * SS_READY - A memory card has raised its ready signal.
	 * SS_DETECT - A card is present.
	 * SS_POWERON - Power has been applied to the socket.
	 * SS_STSCHG - An IO card has raised the status change signal.
	 */

	DEBUG(4, "sipo_pccard_get_status(%d) = %#4.4x\n", sock, *value);
	return 0;
}


static int sipo_pccard_get_socket(unsigned int sock, socket_state_t *state)
{
	DEBUG(4, "sipo_pccard_get_socket(%d)\n", sock);

	state->flags = pccard_conf->slots[sock].socket_flags;

	if (pccard_conf->slots[sock].only_5v) {
		state->Vcc = 50;
		state->Vpp = 50;
	} else {
		state->Vcc = 33;
		state->Vpp = 33;
	}

	state->csc_mask = pccard_conf->slots[sock].csc_mask;

	return 0;
}

static int sipo_pccard_set_socket(unsigned int sock, socket_state_t *state)
{
	DEBUG(4, "sipo_pccard_set_socket(%d)\n", sock);

	printk("  flags=0x%x Vcc=%d Vpp=%d csc_mask=0x%x\n",
	       state->flags, state->Vcc, state->Vpp, state->csc_mask);

	pccard_conf->slots[sock].csc_mask = state->csc_mask;

	if (state->flags & SS_RESET) {
		printk("  SS_RESET\n");
		ssh_pccard_hard_reset(&pccard_conf->slots[sock]);
		if (ssh_pccard_read_cis(&pccard_conf->slots[sock]) < 0) {
			printk(KERN_INFO "pccardupm: CIS read failed "
			       "for slot %c\n",
			       pccard_conf->slots[sock].slot);
			ssh_pccard_hard_reset(&pccard_conf->slots[sock]);
			if (ssh_pccard_read_cis(&pccard_conf->slots[sock]) < 0)
			{
				pccard_conf->slots[sock].card_ok = 0;
				return -1;
			}
			printk(KERN_INFO "pccardupm: CIS read "
			       "succeeded after card "
			       "reset (slot %c)\n",
			       pccard_conf->slots[sock].slot);
			pccard_conf->slots[sock].card_ok = 1;
		}
	}
	if (state->flags & SS_IOCARD) {
		printk("  SS_IOCARD\n");
		if (!(pccard_conf->slots[sock].socket_flags & SS_IOCARD)) {
			printk("  setting card to IO mode\n");
			pccard_conf->slots[sock].socket_flags |= SS_IOCARD;
			ssh_pccard_set_io_mode(&pccard_conf->slots[sock]);
		}
	} else
		pccard_conf->slots[sock].socket_flags &= ~SS_IOCARD;

	if (state->flags & SS_OUTPUT_ENA) {
		printk("  SS_OUTPUT_ENA\n");
		pccard_conf->slots[sock].socket_flags |= SS_OUTPUT_ENA;
	} else
		pccard_conf->slots[sock].socket_flags &= ~SS_OUTPUT_ENA;

	if (state->flags & SS_PWR_AUTO) {
		printk("  SS_PWR_AUTO (NOT SUPPORTED)\n");
	}

	if (state->Vcc == 0)
		ssh_pccard_set_power(sock, 0);
	else
		ssh_pccard_set_power(sock, 1);

	return 0;
}

static int sipo_pccard_get_io_map(unsigned int sock, struct pccard_io_map *io)
{
	DEBUG(1, "sipo_pccard_get_io_map(%d)\n", sock);

	io->start = pccard_conf->slots[sock].io;
	io->stop = pccard_conf->slots[sock].io + 0xff /* FIX */;
	io->flags = MAP_ACTIVE; /* FIX */

	return 0;
}

static int sipo_pccard_set_io_map(unsigned int sock, struct pccard_io_map *io)
{
	DEBUG(1, "sipo_pccard_set_io_map(%d) map=%d flags=0x%x speed=%d ns "
	      "io=0x%x-0x%x\n", sock, io->map, io->flags, io->speed, io->start,
	      io->stop);

	/* FIX: this might need to be implemented instead of doing ioremap()
	 * in the actual card driver (?)
	 * io->start and stop are between 0 and 0xffff;
	 * in tests hardcode 0x0a00-0x0a3f */

	return 0;
}

static int sipo_pccard_get_mem_map(unsigned int sock,
				   struct pccard_mem_map *mem)
{
	DEBUG(1, "sipo_pccard_get_mem_map(%d)\n", sock);

	return 0;
}

static int sipo_pccard_set_mem_map(unsigned int sock,
				   struct pccard_mem_map *mem)
{
	DEBUG(1, "sipo_pccard_set_mem_map(%d)\n", sock);

	return 0;
}

static int sipo_pccard_init(unsigned int s)
{
#if 0
	int i;
	pccard_io_map io = { 0, 0, 0, 0, 1 };
	pccard_mem_map mem = { 0, 0, 0, 0, 0, 0 };
#endif

	DEBUG(1, "sipo_pccard_init(%d)\n", s);

#if 0
	mem.sys_stop = 0x1000;
	sipo_pccard_set_socket(s, &dead_socket);

	for (i = 0; i < 2; i++) {
		io.map = i;
		sipo_pccard_set_io_map(s, &io);
	}

	for (i = 0; i < 5; i++) {
		mem.map = i;
		sipo_pccard_set_mem_map(s, &mem);
	}
#endif

	return 0;
}

static int sipo_pccard_suspend(unsigned int sock)
{
	DEBUG(1, "sipo_pccard_suspend(%d)\n", sock);
	return sipo_pccard_set_socket(sock, &dead_socket);
}


static struct pccard_operations sipo_pccard_operations = {
	init: sipo_pccard_init,
	suspend: sipo_pccard_suspend,
	register_callback: sipo_pccard_register_callback,
	inquire_socket: sipo_pccard_inquire_socket,
	get_status: sipo_pccard_get_status,
	get_socket: sipo_pccard_get_socket,
	set_socket: sipo_pccard_set_socket,
	get_io_map: sipo_pccard_get_io_map,
	set_io_map: sipo_pccard_set_io_map,
	get_mem_map: sipo_pccard_get_mem_map,
	set_mem_map: sipo_pccard_set_mem_map,
	proc_setup: sipo_pccard_proc_setup
};


/* SIPO specific bus_ops
 * FIX: check whether these are needed at all? Most of these could be done with
 * default functions. ioremap() is at the moment only one that uses something
 * extra.. */

static u32 sipo_pccard_bus_b_in(void *bus, u32 port, s32 sz)
{
	printk("b_in(%p, %d, %d)\n", bus, port, sz);
	return 0;
}

static void sipo_pccard_bus_b_ins(void *bus, u32 port, void *buf, u32 count,
				  s32 sz)
{
	printk("b_ins(%p, %d, %p, %d, %d)\n", bus, port, buf, count, sz);
}

static void sipo_pccard_bus_b_out(void *bus, u32 val, u32 port, s32 sz)
{
	printk("b_out(%p, %d, %d, %d)\n", bus, val, port, sz);
}

static void sipo_pccard_bus_b_outs(void *bus, u32 port, void *buf, u32 count,
				   s32 sz)
{
	printk("b_outs(%p, %d, %p, %d, %d)\n", bus, port, buf, count, sz);
}

static void * sipo_pccard_bus_b_ioremap(void *bus, u_long ofs, u_long sz)
{
	void * res;
	struct bus_operations *bus_op = bus;
	printk(KERN_DEBUG "b_ioremap(%p, %ld, %ld) slot=%d\n", bus, ofs, sz,
	       (int) bus_op->priv);
	res = ioremap(PCCARD_BASE + (int) bus_op->priv * PCCARD_SIZE + ofs,
		      sz);
	printk(KERN_DEBUG "==> b_ioremap=%p\n", res);
	return res;
}

static void sipo_pccard_bus_b_iounmap(void *bus, void *addr)
{
	printk(KERN_DEBUG "b_iounmap(%p, %p)\n", bus, addr);
	iounmap(addr);
}

static u32 sipo_pccard_bus_b_read(void *bus, void *addr, s32 sz)
{
	u32 res;
	if (sz == 0)
		res = readb(addr);
	else if (sz == 1)
		res = readw(addr);
	else if (sz == 2)
		res = readl(addr);
	else {
		printk("sipo_pccard_bus_b_read: invalid sz=%d\n", sz);
		res = 0;
	}
	/* printk(KERN_DEBUG "b_read(%p, %p, %d): 0x%x\n", bus, addr, sz, res);
	 */

	return res;
}

static void sipo_pccard_bus_b_write(void *bus, u32 val, void *addr, s32 sz)
{
	printk(KERN_DEBUG "b_write(%p, %d, %p, %d)\n", bus, val, addr, sz);

	if (sz == 0)
		writeb(val, addr);
	else if (sz == 1)
		writew(val, addr);
	else if (sz == 2)
		writel(val, addr);
	else {
		printk("sipo_pccard_bus_b_write: unsupported sz=%d\n", sz);
	}
}

static void sipo_pccard_bus_b_copy_from(void *bus, void *d, void *s, u32 count)
{
	printk("b_copy_from(%p, %p, %p, %d)\n", bus, d, s, count);
}

static void sipo_pccard_bus_b_copy_to(void *bus, void *d, void *s, u32 count)
{
	printk("b_copy_to(%p, %p, %p, %d)\n", bus, d, s, count);
}

static int sipo_pccard_bus_b_request_irq(void *bus, u_int irq,
					 void (*handler)(int, void *,
							 struct pt_regs *),
					 u_long flags, const char *device,
					 void *dev_id)
{
	printk("b_request_irq(%p, %d, %p, %ld, %s, %p)\n",
	       bus, irq, handler, flags, device, dev_id);

	return request_8xxirq(irq, handler, flags, device, dev_id);
}

static void sipo_pccard_bus_b_free_irq(void *bus, u_int irq, void *dev_id)
{
	printk("b_free_irq(%p, %d, %p)\n", bus, irq, dev_id);
	free_irq(irq, dev_id);
}


/*====================================================================*/


static int __init init_sipo_pccard(void)
{
	servinfo_t serv;
	int i, j;
	struct bus_operations *bus;
	bd_t *__bd;

	DEBUG(4, "init_sipo_pccard\n");

	pcmcia_get_card_services_info(&serv);
	if (serv.Revision != CS_RELEASE_CODE) {
		printk(KERN_NOTICE "sipo_pcmcia: Card Services release "
		       "does not match!\n");
		return -1;
	}

	__bd = (bd_t *) __res;
	if (__bd->hw_version != 4)
		return -ENODEV;

	pccard_conf = (struct pccard_conf_data *)
		kmalloc(sizeof(struct pccard_conf_data), GFP_KERNEL);
	if (pccard_conf == NULL) {
		printk(KERN_INFO "init_sipo_pccard: kmalloc failed\n");
		return -1;
	}
	memset(pccard_conf, 0, sizeof(struct pccard_conf_data));

	for (i = 0; i < NUM_PCCARD_SLOTS; i++) {
		unsigned int phaddr;
		pccard_conf->slots[i].slot = 'A' + i;

		if (i == 0) {
			pccard_conf->slots[i].irq = PCCARD_IRQ_A;
			pccard_conf->slots[i].io = PCCARD_BASE | PCCARD_IO;
		} else if (i == 1) {
			pccard_conf->slots[i].irq = PCCARD_IRQ_B;
			pccard_conf->slots[i].io = PCCARD_BASE | PCCARD_SIZE |
				PCCARD_IO;
		}

		pccard_conf->slots[i].pccard_addr = (unsigned char *)
			ioremap(PCCARD_BASE + i * PCCARD_SIZE,
				PCCARD_WINDOW_SIZE);
		if (pccard_conf->slots[i].pccard_addr == NULL) {
			printk(KERN_INFO
			       "pccard: ioremap failed for Slot %c\n",
			       pccard_conf->slots[i].slot);
			for (j = 0; j < i; j++)
				iounmap(pccard_conf->slots[j].pccard_addr);
			kfree(pccard_conf);
			pccard_conf = NULL;
			return -1;
		}

		pccard_conf->slots[i].socket_flags = 0;

		pccard_conf->slots[i].cap.features = SS_CAP_PCCARD |
			SS_CAP_STATIC_MAP | SS_CAP_VIRTUAL_BUS;

		phaddr = PCCARD_BASE | PCCARD_IO;
		if (i != 0)
			phaddr |= PCCARD_SIZE;
		/* FIX: what is the reason for using 0xa00 here? 0x200 should
		 * be identical, but why that one bit is set?? */
#if 0
		phaddr |= 0xa00;
#else
		phaddr |= 0x200;
#endif

		pccard_conf->slots[i].cap.io_offset =
			(unsigned int) ioremap(phaddr, PCCARD_WINDOW_SIZE);
		printk("sipo_pcmcia: slot%d cap_io_offset=0x%x "
		       "(phaddr=0x%x)\n", i,
		       pccard_conf->slots[i].cap.io_offset, phaddr);
		pccard_conf->slots[i].cap.map_size = PCCARD_WINDOW_SIZE;
		pccard_conf->slots[i].cap.irq_mask = 0;
		pccard_conf->slots[i].cap.pci_irq =
			i == 0 ? PCCARD_IRQ_A : PCCARD_IRQ_B;
		bus = kmalloc(sizeof(*bus), GFP_KERNEL);
		if (bus == NULL) {
			printk("kmalloc failed for bus_operations table\n");
			/* FIX: free other possibly reserved allocations */
			kfree(pccard_conf);
			pccard_conf = NULL;
			return -1;
		}
		memset(bus, 0, sizeof(*bus));
		bus->priv = (void *) i;
		bus->b_in = sipo_pccard_bus_b_in;
		bus->b_ins = sipo_pccard_bus_b_ins;
		bus->b_out = sipo_pccard_bus_b_out;
		bus->b_outs = sipo_pccard_bus_b_outs;
		bus->b_ioremap = sipo_pccard_bus_b_ioremap;
		bus->b_iounmap = sipo_pccard_bus_b_iounmap;
		bus->b_read = sipo_pccard_bus_b_read;
		bus->b_write = sipo_pccard_bus_b_write;
		bus->b_copy_from = sipo_pccard_bus_b_copy_from;
		bus->b_copy_to = sipo_pccard_bus_b_copy_to;
		bus->b_request_irq = sipo_pccard_bus_b_request_irq;
		bus->b_free_irq = sipo_pccard_bus_b_free_irq;
		pccard_conf->slots[i].cap.bus = bus;
	}

	ssh_upm_setup();

	WORK_QUEUE_INIT(&sipo_pccard_task, sipo_pccard_bh, NULL);

	init_timer(&pccard_conf->poll_timer);
	pccard_conf->poll_timer.expires = jiffies + CARD_POLL_INTERVAL;
	pccard_conf->poll_timer.data = (unsigned long) pccard_conf;
	pccard_conf->poll_timer.function = ssh_pccard_handle_poll;
	add_timer(&pccard_conf->poll_timer);

	if (register_ss_entry(NUM_PCCARD_SLOTS, &sipo_pccard_operations) != 0)
		printk(KERN_NOTICE
		       "sipo_pcmcia: register_ss_entry() failed\n");

	DEBUG(4, "init_sipo_pccard: ss_entry registered\n");

	return 0;
}

static void __exit exit_sipo_pccard(void)
{
	int i;

	printk("exit_sipo_pccard\n");

	FLUSH_SCHEDULED_WORK();

	for (i = 0; i < NUM_PCCARD_SLOTS; i++) {
#ifdef CONFIG_PROC_FS
		sipo_pccard_proc_remove(i);
#endif
		ssh_pccard_set_power(i, 0);

		if (pccard_conf->slots[i].pccard_addr != NULL)
			iounmap(pccard_conf->slots[i].pccard_addr);

		if (pccard_conf->slots[i].cap.bus != NULL)
			kfree(pccard_conf->slots[i].cap.bus);
	}

	unregister_ss_entry(&sipo_pccard_operations);

	del_timer(&pccard_conf->poll_timer);

	kfree(pccard_conf);
	pccard_conf = NULL;
}

module_init(init_sipo_pccard);
module_exit(exit_sipo_pccard);
