/*******************************************************************************

  Simple command line interface for sIPo bootloader                                                                             

  Author: Arto Vuori <avuori@ssh.com>

  Copyright (C) 2000 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved. 

  Simple parser for bootloader's command line interface.

*******************************************************************************/

#include "sipo.h"
#include "stdio.h"

/* Parsing helper functions */
static int parse_int(char **cpp, int *val);
static int parse_str(char **cpp, char **val);

/* top level commands */
static int cli_help(char *cp);
static int cli_bootloader(char *cmd);
static int cli_pt(char *cp);
static int cli_load(char *cp);
static int cli_put(char *cp);
static int cli_erase(char *cp);
static int cli_set(char *cp);
static int cli_net(char *cp);
static int cli_write(char *cp);
static int cli_passwd(char *cp);
static int cli_show(char *cp);
static int cli_test(char *cp);
static int cli_boot(char *cp);
static int cli_mr(char *cp);
static int cli_mw(char *cp);
static int cli_nvr(char *cp);
static int cli_nvw(char *cp);
static int cli_fe(char *cp);
static int cli_fw(char *cp);
static int cli_reboot(char *cp);

/* sub commands */
static int cli_set_ipaddr(char *cmd, char *name, in_addr *var);
static int cli_set_serial(char *cmd);
static int cli_set_phyaddr(char *cmd);
static int cli_set_cmdline(char *cmd);
static int cli_set_str(char *cmd, char *name, char *var);
static int cli_set_date(char *cmd);
static int cli_set_bool(char *cmd, char *name, unsigned *var);
static int cli_set_int(char *cmd, char *name, unsigned *var);
static int cli_set_led(char *cmd);

typedef struct {
    int (* action)(char *cmd);                  /* Pointer to a function performing any actions */
    char        *command;                       /* First token of the command */
    char        *short_help;                    /* Short description of the command, printed as a help topic */
    char        *long_help;                     /* Long description about the command */
} t_cli_command;

/*
 * All CLI commands are listed in following command table.
 */
static t_cli_command commands[]={
    { 
        cli_help,
        "help",
        "help [cmd]                  Print help topics or command description.\n",
        "\nNAME\n"
        "        help - print help topics or command description.\n"
        "\nSYNOPSIS\n"
        "        help [cmd]\n"
        "\nDESCRIPTION\n"
        "        Without any arguments this command prints all commands and help\n"
        "        topics. If optional argument is given, full description on that\n"
        "        command or topic is printed.\n"
        "\n"
    },
    { 
        cli_bootloader,
        "bootloader",
        "bootloader <filename> [NOVERIFY]   Download and program new bootloader.\n",
        "\nNAME\n"
        "        bootloader - Download and program new bootloader.\n"
        "\nSYNOPSIS\n"
        "        bootloader <filename>\n"
        "\nDESCRIPTION\n"
        "        Download bootloader binary image and program it to flash memory\n"
        "        as a new bootloader code. Required argument is path and filename\n"
        "        for bootloader image. BOOTP protocol is used to find network\n"
        "        settings and a TFTP server. If BOOTP succeeds, TFTP protocol is\n"
        "        used to download the image from the server.\n"
        "\n"
        "        CAUTION! Incorrect image or other failure may cause inrecoverable\n"
        "        damage. Do not try this at home or without proper authorization.\n"
        "\n"
    },
    { 
        cli_boot,
        "boot",
        "boot [cmd]                  Boot kernel stored in flash\n",
        "\nNAME\n"
        "        boot - Boot kernel stored in flash\n"
        "\nSYNOPSIS\n"
        "        boot [cmd]\n"
        "\nDESCRIPTION\n"
        "        Boot using image stored in flash kernel partion. Optional\n"
        "        parameters are passed to kernel as boot parameters.\n"
        "\n"
    },
    { 
        net_boot,
        "netboot",
        "netboot [cmd]               Boot from network using BOOTP & TFTP\n",
        "\nNAME\n"
        "        netboot - Boot from network using BOOTP & TFTP\n"
        "\nSYNOPSIS\n"
        "        netboot [cmd]\n"
        "\nDESCRIPTION\n"
        "        Boot using image downloaded from network. Optional parameters are\n"
        "        passed to kernel as boot parameters. BOOTP protocol is used to\n"
        "        find network settings and TFTP server. If BOOTP succeeds, TFTP\n"
        "        protocol is used to download kernel image from the server.\n"
        "\n"
    },
    {
        cli_pt,
        "pt",
        "pt [...]                    Show partition table or modify it\n",
        "\nNAME\n"
        "        pt - Show partition table or modify it\n"
        "\nSYNOPSIS\n"
        "        pt [<index> <type> <offset> <size>]\n"
        "\nDESCRIPTION\n"
        "        Without any argumets this command shows current partition table.\n"
        "        With optional arguments new partition can be created or old one\n"
        "        can be adjusted. If partition table is uninitialized new default\n"
        "        partition table is initialized before operation.\n"
        "\n"
    },
    {
        cli_load,
        "load",
        "load <num> <filename> [NOVERIFY]   Download image to a partition\n",
        "\nNAME\n"
        "        load - Download image to a partition\n"
        "\nSYNOPSIS\n"
        "        load <num> <filename>\n"
        "\nDESCRIPTION\n"
        "        Download image to flash partition. First argument is partition\n"
        "        number. Second argument is a partition image filename. The image\n"
        "        is downloaded from TFTP server which is found by using BOOTP\n"
        "        protocol.\n"
        "\n"
    },
#ifdef DEBUG
    {
        cli_put,
        "put",
        "put <num> <filename>       Upload a partition to TFTP server\n",
        "\nNAME\n"
        "        put - Upload a partition to TFTP server\n"
        "\nSYNOPSIS\n"
        "        put <num> <filename>\n"
        "\nDESCRIPTION\n"
        "        Upload image of flash partition to TFTP server. First argument\n"
        "        is partition number. Second argument is a filename to be used on\n"
        "        TFTP server. The image is uploaded to TFTP server which is found\n"
        "        by using BOOTP protocol.\n"
        "\n"
    },
#endif /* DEBUG */
    {
        cli_erase,
        "erase",
        "erase <num>                 Erase partition\n",
        "\nNAME\n"
        "        erase - Erase partition\n"
        "\nSYNOPSIS\n"
        "        erase <num>\n"
        "\nDESCRIPTION\n"
        "        Erase partition. This command erases contents of the partition.\n"
        "        First argument is number of the partition to be erased.\n"
        "\n"
    },
    {
        cli_set,
        "set",
        "set [<variable> <ipaddrnetmask>] Set/show variable\n",
        "\nNAME\n"
        "        set - Set/show variable\n"
        "\nSYNOPSIS\n"
        "        set [[variable] [value(s)...]]\n"
        "\nDESCRIPTION\n"
        "        If no arguments is given, command shows all variables and their\n"
        "        current values. The first argument is a name of a variable. With\n"
        "        only one argument value of the variable is shown. When more\n"
        "        arguments is given, the variable is set to value(s) of remaining\n"
        "        argument(s).\n"
        "\n"
    },
    {
        cli_net,
        "net",
        "net [<variable> <ipaddr>] Set/show network settings\n",
        "\nNAME\n"
        "        net - Set/show network settings\n"
        "\nSYNOPSIS\n"
        "        net <[ipaddr|netmask|gateway|tftpserver|bootfile> <ipaddr>]\n"
        "\nDESCRIPTION\n"
        "        Without any arguments, this command shows all network related\n"
        "        settings. With optional arguments, network related settings can be\n"
        "        set. These settings only affect only bootloader and are not\n"
        "        storedto nonvolatile memory."
        "\n"
    },
    {
        cli_write,
        "write",
        "write                       Write settings to flash\n",
        "\nNAME\n"
        "        write - Write settings to flash\n"
        "\nSYNOPSIS\n"
        "        write\n"
        "\nDESCRIPTION\n"
        "        Writes all settings to board info block stored in flash. Normally\n"
        "        all settings will be restored on next reboot. By giving this\n"
        "        command you can set current settings as new default bootup\n"
        "        settings\n"
        "\n"
    },
    {
        cli_passwd,
        "passwd",
        "passwd                      Change authentication token\n",
        "\nNAME\n"
        "        passwd - Change authentication token\n"
        "\nSYNOPSIS\n"
        "        passwd\n"
        "\nDESCRIPTION\n"
        "        Passwd is used to change bootloaders authetication token. Command\n"
        "        first requests current password, which will be verified. Then you\n"
        "        can enter a new password.\n"
        "        NOTE! Password won't be stored before you give 'write' command.\n"
        "\n"
    },
    {
        cli_show,
        "show",
        "show                        Show board descriptor\n",
        "\nNAME\n"
        "        show - Show board descriptor\n"
        "\nSYNOPSIS\n"
        "        show\n"
        "\nDESCRIPTION\n"
        "        Prints contents of the board descriptor stored in flash.\n"
        "\n"
    },
    {
        cli_test,
        "test",
        "test [cmd]                  Run self diagnostic tests\n",
        "\nNAME\n"
        "        test - Run self diagnostic tests\n"
        "\nSYNOPSIS\n"
        "        test [flash|spd|sdram|...]\n"
        "\nDESCRIPTION\n"
        "        If no arguments is given, command runs all tests. Otherwise only\n"
        "        single test is run. Optional argument(s) indicates which test(s)\n"
        "        should be run.\n"
        "\n"
    },
    { 
        cli_reboot,
        "reboot",
        "reboot                      Do hard reset\n",
        "\nNAME\n"
        "        reboot - Do hardreset\n"
        "\nSYNOPSIS\n"
        "        reboot\n"
        "\nDESCRIPTION\n"
        "        Restart system by doing hard reset\n"
        "\n"
    },

#ifdef DEBUG
    {
        cli_nvr,
        "nvr",
        "nvr [addr [len]]            Read from NVRAM\n",
        "\nNAME\n"
        "        nvr - read from memory\n"
        "\nSYNOPSIS\n"
        "        nvr [addr [num]]\n"
        "\nDESCRIPTION\n"
        "        Reads num bytes from NVRAM address specified by first argument.\n"
        "        If addr is not specified, NVRAM will read from the beginning.\n"
        "        Also if num is not specified, 0x60 bytes bytes will be read.\n"
        "\n"
    },
    {
        cli_nvw,
        "nvw",
        "nvw <addr> <hex>            Write to NVRAM\n",
        "\nNAME\n"
        "        nvw - write to memory\n"
        "\nSYNOPSIS\n"
        "        nvw <addr> <hex>\n"
        "\nDESCRIPTION\n"
        "        Writes bytes to NVRAM address specified by first argument.\n"
        "        Data is a hex byte stream starting with '0x' prefix.\n"
        "\n"
    },
    {
        cli_mr,
        "mr",
        "mr <addr> [len]             Read from memory\n",
        "\nNAME\n"
        "        mr - read from memory\n"
        "\nSYNOPSIS\n"
        "        mr <addr> [num]\n"
        "\nDESCRIPTION\n"
        "        Reads num bytes from memory address specified by first argument.\n"
        "        If num is not specified, 32 bytes bytes will be read.\n"
        "\n"
    },
    {
        cli_mw,
        "mw",
        "mw <addr> <hex>             Write to memory\n",
        "\nNAME\n"
        "        mw - write to memory\n"
        "\nSYNOPSIS\n"
        "        mw <addr> <hex>\n"
        "\nDESCRIPTION\n"
        "        Writes bytes to memory address specified by first argument.\n"
        "        Data is a hex byte stream starting with '0x' prefix.\n"
        "\n"
    },
    {
        cli_fw,
        "fw",
        "fw <addr> <hex>             Write to flash\n",
        "\nNAME\n"
        "        fw - write to flash\n"
        "\nSYNOPSIS\n"
        "        fw <addr> <hex>\n"
        "\nDESCRIPTION\n"
        "        Writes bytes to flash address specified by first argument.\n"
        "        Data is a hex byte stream starting with '0x' prefix.\n"
        "\n"
    },
    {
        cli_fe,
        "fe",
        "fe <addr> [cnt]             Erase flash blocks\n",
        "\nNAME\n"
        "        fe - Erase flash blocks\n"
        "\nSYNOPSIS\n"
        "        fe <addr> [cnt]\n"
        "\nDESCRIPTION\n"
        "        Erase flash blocks starting from given address. If count\n"
        "        is not specifed, single flash block will be erased.\n"
        "\n"
    },
#endif /* DEBUG */
    {NULL, NULL, NULL, NULL}
};

/*******************************************************************************

  CLI parser helper functions

*******************************************************************************/


static char **eat_spaces(char **cp)
{
    while(**cp==' ') 
        (*cp)++;

    return cp;
}

/* Read single integer. cpp is incremented past read characters. 
   Returns nonzero on failure.
   Reads also hex numbers if input starts with "0x".
*/

static int parse_int(char **cpp, int *val) 
{
    unsigned int i=0;
    char *cp;

    cp=*eat_spaces(cpp);

    if(cp[0]=='0' && cp[1]=='x') /* Hex number */
    {
        cp+=2;
        while(1) {
            if(*cp>='0' && *cp<='9')
                i=i*16+*cp-'0';
            else if(*cp>='a' && *cp<='f')
                i=i*16+*cp-'a'+10;
            else
                break;
            cp++;
        }
    } else
    {   
        while(*cp>='0' && *cp<='9') {
            i=i*10+*cp-'0';
            cp++;
        }
    }

    if((*cp!=0 && *cp!=' ') || cp==*cpp)
        return -1;
    else {
        *cpp=cp;
        *val = i;
        return 0;
    }
}
/* Read single ip address. 0x prefix can be user for 32bit hex numbers.
*/

static int parse_ip(char **cpp, int *val) 
{
    unsigned int i,j,ip;
    char *cp;

    cp=*eat_spaces(cpp);

    if(cp[0]=='0' && cp[1]=='x') /* Hex number */
        return parse_int(cpp,val);      
    
    for(j=ip=0;j<4;j++) 
    {
        if(j!=0 && *cp!='.')
            return -1;

        if(j!=0) cp++;

        for(i=0;*cp>='0' && *cp<='9';cp++)
            i=i*10+*cp-'0';

        if(i>255) return -1;

        ip=ip*256+i;
    }

    if((*cp!=0 && *cp!=' ') || cp==*cpp)
        return -1;
    else {
        *cpp=cp;
        *val = ip;
        return 0;
    }
}

/* Read string. Leading whitespaces are ignored. *val is set to point at first non-space character.
   First space character after string is changed to 0 and cpp is incremented to point next string.
   returns nonzero on error.
*/

static int parse_str(char **cpp, char **val)
{
    char *cp=*cpp;

    eat_spaces(&cp);
    *val=cp;

    while(*cp!=' ' && *cp!=0) cp++;
    if(*cp!=0)
        *cp++=0;
    
    if(cp!=*val) {
        *cpp=cp;
        return 0;
    } else
        return -1;
}

/* Read hex string. Read max num bytes. Return actual amout of bytes read. 
   Correct string must start with "0x" prefix.
 */

static int parse_hex(char **cpp, char *val, int num) {
    int i,j;
    char *cp;

    eat_spaces(cpp);
    cp=*cpp;

    /* Check for "0x" prefix */
    if(cp[0]!='0' || cp[1]!='x')
        return -1;
    else
        cp+=2;

    for(i=0,j=0;i<num*2;i++,cp++) 
    {
        if(*cp>='0' && *cp<='9')
            j=(j<<4)+*cp-'0';
        else if(*cp>='a' && *cp<='f')
            j=(j<<4)+*cp-'a'+10;
        else {
            break;
        }

        if(i&1) {
            val[i/2]=j;
            j=0;
        }
    }
    if((*cp!=' ' && *cp!=0) || i<2 || i&1)
        return -1;

    *cpp=cp;

    return i/2;
}

static int match(char **var, char *cons)
{
    if(!strncmp(*var, cons, strlen(cons))) {
        *var+=strlen(cons);
        eat_spaces(var);
        return 1;
    }
    else
        return 0;
}

/*******************************************************************************

  CLI action functions

*******************************************************************************/

#ifdef DEBUG

static int cli_mr(char *cp)
{
    unsigned int addr, len, i;

    if(parse_int(&cp, &addr)) {
        printf("mr: Memory address must be specified.\n");
        return -1;
    }

    if(parse_int(&cp, &len))
        len=32;

    for(i=0;i<len;i++)
    {
        if((i&15)==0)
            printf("%08x: ", addr+i);
        
        printf("%02x ", ((char *)addr)[i]);

        if((i&15)==7 && i!=len-1)
            printf("  ");
        else if((i&15)==15 || i==len-1)
            printf("\n");
    }
}

static int cli_mw(char *cp)
{
    unsigned int addr;
    int         len,i;
    char        buf[0x100];

    if(parse_int(&cp, &addr)) {
        printf("mw: Memory address must be specified.\n");
        return -1;
    }

    if((len=parse_hex(&cp, buf, 0x100))<=0) {
        printf("mw: Invalid data\n");
    }

    for(i=0;i<len;i++)
        ((char *)addr)[i]=buf[i];

    return 0;
}

static int cli_nvr(char *cp)
{
    unsigned int addr, len, i;
    unsigned char c;

    if(parse_int(&cp, &addr))
        addr=0;

    if(parse_int(&cp, &len))
        len=0x60;

    for(i=0;i<len;i++)
    {
        if((i&15)==0)
            printf("%08x: ", addr+i);
        
        if(nv_read(addr+i,1,&c)==0) {
            printf("\n");
            return -1;         
        }

        printf("%02x ", c);

        if((i&15)==7 && i!=len-1)
            printf("  ");
        else if((i&15)==15 || i==len-1)
            printf("\n");
    }
    return 0;
}

static int cli_nvw(char *cp)
{
    unsigned int addr;
    int         len,i;
    char        buf[0x100];

    if(parse_int(&cp, &addr)) {
        printf("nvw: Memory address must be specified.\n");
        return -1;
    }

    if((len=parse_hex(&cp, buf, 0x100))<=0) {
        printf("nvw: Invalid data\n");
        return -1;
    }

    return (nv_write(addr,len,buf)==0);
}

static int cli_fw(char *cp)
{
    unsigned int addr;
    int         len,i;
    char        buf[0x100];

    if(parse_int(&cp, &addr)) {
        printf("mw: Flash address must be specified.\n");
        return -1;
    }

    if((addr<FLASH_START_ADDRESS) || (addr-FLASH_START_ADDRESS >= __bd->flash_size)) {
        printf("fw: Address must be in flash memory range\n");
        return -1;
    }

    if((len=parse_hex(&cp, buf, 0x100))<=0) {
        printf("fw: Invalid data\n");
        return -1;
    }

    return flash_cutcopypaste(addr-FLASH_START_ADDRESS, buf, len);
}

static int cli_fe(char *cp)
{
    unsigned int addr;
    int         len,i;

    if(parse_int(&cp, &addr)) {
        printf("fe: Flash address must be specified.\n");
        return -1;
    }

    if(parse_int(&cp, &len))
        len=1;

    if(addr&(FLASH_BLOCKSIZE-1))
    {
        printf("address must be block aligned\n");
        return -1;
    }
       
    for(i=0;i<len;i++)
        flash_erase_block(addr/FLASH_BLOCKSIZE+i);

    return 0;
}

#endif /* DEBUG */

static int cli_boot(char *cp)
{
    eat_spaces(&cp);

    boot(cp);

    printf("Unable to boot from kernel partition.\n");

    return -1;
}

static int cli_show(char *cp) {
    print_bd();
    return 0;
}

static int cli_test(char *cp)
{
    eat_spaces(&cp);

    if(*cp==0) {
        return 0;
    }

    if(match(&cp, "flash yes I know what I am doing"))
        flash_test();
    else if(match(&cp, "spd"))
            spd_test(cp);
    else {
            printf("Invalid test '%s'\n", cp);
            return -1;
    }

    return 0;
}

static int cli_reboot(char *cp)
{
    hardreset();
    return -1;
}

static int cli_set(char *cp)
{
    eat_spaces(&cp);

    if(*cp==0) {
        cli_set_serial(NULL);
        cli_set_phyaddr(NULL);
        cli_set_date(NULL);
        cli_set_cmdline(NULL);
        cli_set_bool(0, "netboot", &__bd->enable_netboot);
        cli_set_bool(0, "debugport", &__bd->enable_debugport);
        cli_set_bool(0, "passwd", &__bd->enable_passwd);
        cli_set_bool(0, "watchdog", &__bd->enable_wdt);
        return 0;
    }

    if(match(&cp, "serial"))
        return cli_set_serial(cp);
    else if(match(&cp, "phyaddr"))
        return cli_set_phyaddr(cp);
    else if(match(&cp, "date"))
        return cli_set_date(cp);
    else if(match(&cp, "cmdline"))
        return cli_set_cmdline(cp);
    else if(match(&cp, "netboot"))
        return cli_set_bool(cp, "netboot", &__bd->enable_netboot);
    else if(match(&cp, "debugport"))
        return cli_set_bool(cp, "debugport", &__bd->enable_debugport);
    else if(match(&cp, "watchdog"))
        return cli_set_bool(cp, "watchdog", &__bd->enable_wdt);
    else if(match(&cp, "led"))
        return cli_set_led(cp);
    else if(match(&cp, "passwd"))
        return cli_passwd(cp);
    else {
        printf("Invalid variable '%s'\n", cp);
        return -1;
    }
}

static int cli_set_cmdline(char *cp) {
    if(!cp) {
        printf("cmdline = \"%s\"\n", __bd->cmdline);
        return 0;
    }

    strcpy(__bd->cmdline, cp);
    return 0;
}

static int cli_set_date(char *cp) 
{
    unsigned int year, mon, day, hour, min, sec;

    if(!cp || **eat_spaces(&cp)==0) 
    {
        print_rtc_time();
        return 0;
    }

    if(parse_int(&cp, &year) ||
       parse_int(&cp, &mon) ||
       parse_int(&cp, &day) ||
       parse_int(&cp, &hour) ||
       parse_int(&cp, &min) ||
       parse_int(&cp, &sec))
    {
        printf("Invalid arguments. ");
        printf("Usage: set date <year> <month> <day> <hour> <min> <sec>\n");
        return -1;
    }

    return set_rtc_time(sec,min,hour,day,mon,year);
}

/*
  Writing a serial number is an irreversible process.
*/
static int cli_set_serial(char *cp) 
{
    unsigned char Manuf_ID;
    unsigned char Device_Type;
    unsigned char Lock;
    char SerialLock;
    long long SerialNr;
    int i;

    if(!cp || **eat_spaces(&cp)==0) 
    {
        SerialNr = flash_read_serial_number(&Manuf_ID, &Device_Type, &Lock, &SerialLock);
        printf("serial = %ld\n", SerialNr);
        return 0;
    }

    SerialNr = 0;
    while (*cp>='0' && *cp<='9') {
        SerialNr=SerialNr*10+*cp-'0';
        cp++;
    }
    
    flash_write_serial_number(SerialNr);

    return 0;
}

static int cli_set_str(char *cp, char *name, char *var)
{
    if(!cp || **eat_spaces(&cp)==0) 
    {
        printf("%s = '%s'\n", name, var);
        return 0;
    }

    strcpy(var, cp);

    return 0;
}

static int cli_set_bool(char *cp, char *name, unsigned *var)
{
    if(!cp || **eat_spaces(&cp)==0) 
    {
        printf("%s = %s\n", name, (*var!=0)?"true":"false");
        return 0;
    }

    if(match(&cp, "true"))
        *var=1;
    else if(match(&cp, "false"))
        *var=0;
    else {
        printf("cli_set_bool: Boolean variable '%s' can only be set to 'true' or 'false'\n", name);
        return -1;
    }
    return 0;
}

static int cli_set_int(char *cp, char *name, unsigned *var)
{
    int i;

    if(!cp || **eat_spaces(&cp)==0) 
    {
        printf("%s = %u\n", name, *var);
        return 0;
    }

    if(parse_int(&cp, &i))
    {
        printf("cli_set_int: Invalid value\n");
        return -1;
    }

    *var=i;

    return 0;
}

static int cli_set_led(char *cp)
{
    int i;

    if(parse_int(&cp, &i))
    {
        printf("cli_set_led: Invalid value\n");
        return -1;
    }

    set_leds(i);

    return 0;
}

static int cli_set_phyaddr(char *cp) 
{
    int i,j;

    /* If no arguments */
    if(!cp || **eat_spaces(&cp)==0) 
    {
        printf("phyaddr = 0x%02x%02x%02x%02x%02x%02x\n",
               __bd->bi_enetaddr[0],
               __bd->bi_enetaddr[1],
               __bd->bi_enetaddr[2],
               __bd->bi_enetaddr[3],
               __bd->bi_enetaddr[4],
               __bd->bi_enetaddr[5],
               __bd->bi_enetaddr[6]);

        return 0;
    }

    if(parse_hex(&cp, (char *)(__bd->bi_enetaddr), 6)!=6) 
    {
        printf("Invalid MAC address\n");
        return -1;
    }

    return;
}

static int cli_set_ipaddr(char *cp, char *name, in_addr *var)
{
    if(!cp || **eat_spaces(&cp)==0) 
    {
        printf("%s = %u.%u.%u.%u\n", name, 
               (*var>>24)&255,
               (*var>>16)&255,
               (*var>>8)&255,
               (*var>>0)&255);

        return 0;
    }

    if(parse_ip(&cp, var))
    {
        printf("cli_set_ip: Invalid value\n");
        return -1;
    }
}

static int cli_net(char *cp)
{
    eat_spaces(&cp);

    if(*cp==0) {
        cli_set_ipaddr(0,"ipaddr", &MY_IPADDR);
        cli_set_ipaddr(0,"gateway", &MY_GATEWAY);
        cli_set_ipaddr(0,"netmask", &MY_NETMASK);
        cli_set_ipaddr(0,"tftpserver", &SERVER_IPADDR);
        cli_set_str(0,"bootfile", bootfile);
        return 0;
    }

    if(match(&cp, "ipaddr"))
        return cli_set_ipaddr(cp,"ipaddr", &MY_IPADDR);
    else if(match(&cp, "gateway"))
        return cli_set_ipaddr(cp,"gateway", &MY_GATEWAY);
    else if(match(&cp, "netmask"))
        return cli_set_ipaddr(cp,"netmask", &MY_NETMASK);
    else if(match(&cp, "tftpserver"))
        return cli_set_ipaddr(cp,"tftpserver", &SERVER_IPADDR);
    else if(match(&cp, "bootfile"))
        cli_set_str(cp, "bootfile", bootfile);
    else {
        printf("Invalid variable '%s'\n", cp);
        return -1;
    }
}

static int cli_write(char *cp)
{
    return write_bd();
}

/* Create password salt */
static void salt(SHA1Digest dgst) {
    static unsigned int count=0;
    unsigned char buffer[0x20];

    memset(buffer,0,0x20);
    *(unsigned int *)buffer=utime();
    *(unsigned int *)(buffer+4)=count++;
    rtc_read(0x00, 7, buffer+4);

    sha1(buffer,0x20, dgst);

    return;
}

static unsigned char secret[]=SIPO_SECRET;

/* SIPO password routines */
static int passwd(char *old, char *new) {
    SHA1_CTX    ctx;
    SHA1Digest  dgst;

    /* Check old password */
    if(old) {
        SHA1Init(&ctx);
        SHA1Update(&ctx, secret, sizeof(secret));
        SHA1Update(&ctx, __bd->salt, sizeof(__bd->salt));
        SHA1Update(&ctx, old, strlen(old));
        SHA1Final(dgst, &ctx);
        if(memcmp(dgst, __bd->passwd, sizeof(__bd->passwd)))
           return -1;
    }

    /* Set new password */    
    if(new) {
        salt(__bd->salt);
        SHA1Init(&ctx);
        SHA1Update(&ctx, secret, sizeof(secret));
        SHA1Update(&ctx, __bd->salt, sizeof(__bd->salt));
        SHA1Update(&ctx, new, strlen(new));
        SHA1Final(dgst, &ctx);
        memcpy(__bd->passwd,dgst,sizeof(SHA1Digest));
    }

    return 0;
}

static int cli_passwd(char *cp)
{
    char        old[0x32], new[0x32], new2[0x32];

    if(**eat_spaces(&cp)!=0) {
        printf("No arguments expected\n");
        return -1;
    }
    
    if(__bd->enable_passwd && !hw_passwd_disable()) {
        printf("Current passwd: ");
        readpasswd(old, 0x32);
        if(passwd(old, NULL)) {
            printf("Invalid password\n");
            udelay(1000000);
            return -1;
        }
    }

    printf("New passwd: ");
    readpasswd(new, 0x32);
    printf("Again: ");
    readpasswd(new2, 0x32);
    if(strcmp(new,new2)) {
        printf("Passwords don't match\n");
        return -1;
    }

    if(new[0]!=0) {
        if(passwd(NULL,new)) {
            printf("Invalid passwd??? This should never happen\n");
            return -1;
        }
        __bd->enable_passwd=1;
    }
    else
        __bd->enable_passwd=0;

    return 0;
}

static int cli_load(char *cp)
{
    int i, skip_verify;
    char *fname, *extra;

    if(parse_int(&cp, &i) ||
       parse_str(&cp, &fname))
    {
        printf("load: Invalid arguments\n");
        return -1;
    }

    if (parse_str(&cp, &extra) == 0 &&
	memcmp(extra, "NOVERIFY", 8) == 0)
      skip_verify = 1;
    else
      skip_verify = 0;

    if(i>1 && i<pt_size)
        return flash_tftp(fname, i, skip_verify);
    else
        printf("Invalid partition\n");


}

#ifdef DEBUG
static int cli_put(char *cp)
{
  int i;
  char *fname;

  if (parse_int(&cp, &i) || parse_str(&cp, &fname))
    {
      printf("put: Invalid arguments\n");
      return -1;
    }

  if (i > 1 && i < pt_size)
    return flash_to_tftp(fname, i);

  printf("Invalid partition\n");
}
#endif /* DEBUG */

static int cli_erase(char *cp)
{
    int i;

    if(parse_int(&cp, &i))
    {
        printf("load: Invalid arguments\n");
        return -1;
    }

    if(i>1 && i<pt_size)
        partition_erase(i);
    else
        printf("Invalid partition\n");


}

static int cli_pt(char *cp) 
{
    int ix,type,off,sz;

    eat_spaces(&cp);
    if(*cp==0) 
    {
        /* No arguments -> print partition table */
        partition_print();
        return 0;
    }
    else
    {
        /* Parse argumets */
        if(parse_int(&cp,&ix) ||
           parse_int(&cp,&type) ||
           parse_int(&cp,&off) ||
           parse_int(&cp,&sz))
        {
            printf("pt: Invalid arguments\n");
            return -1;
        }

        return partition_set(ix, type, off, sz);
    }
}

static int cli_bootloader(char *cmd)
{
    int skip_verify;
    char *fname, *extra;

    if(parse_str(&cmd, &fname))
    {
        printf("Invalid arguments. Filename required as single argument");
        return -1;
    }

    if (parse_str(&cmd, &extra) == 0 &&
	memcmp(extra, "NOVERIFY", 8) == 0)
      skip_verify = 1;
    else
      skip_verify = 0;

    return flash_tftp(fname, pt_bootloader, skip_verify);
}

/*
 * CLI help command. 
 *
 * Print all topics, if no arguments. If command is given print it's description.
 */

static int cli_help(char *cmd) 
{
    t_cli_command *cp;

    if(*cmd==0) {
        /* If no arguments, print all commands and topics */

        printf("\nABOUT\n");
        printf("        %s\n",SIPO_VERSION_STRING);
        printf("        %s\n",SIPO_COPYRIGHT_STRING);
        printf("\nCOMMANDS\n");

        for(cp=commands;cp->action!=NULL;cp++)
            printf("        %s", cp->short_help);

        printf("\n");
        return 0;

    } else {
        /* Print long command description */
        for(cp=commands;cp->action!=NULL;cp++)
            if(cp->long_help && !strncmp(cmd, cp->command, strlen(cp->command))) 
            {
                printf("%s",cp->long_help);
                return 0;
            }

        printf("Description for command '%s' not found\n", cmd);
        return -1;
    }
}

/*
 * CLI command processor. Parse command table for matching first token.
 * Exceute action if command is found. 
 */

static int cli_command(char *cmd) 
{
    t_cli_command *cp;

    /* ignore if empty command is given */
    if(**eat_spaces(&cmd)==0)
        return 0;

    /* Search command table for a string match */
    for(cp=commands;cp->action!=NULL;cp++)
        if(!strncmp(cmd, cp->command, strlen(cp->command)))
        {
            /* Space or EOL must be after first token */
            if(cmd[strlen(cp->command)]!=0 && cmd[strlen(cp->command)]!=' ')
                continue;

            cmd+=strlen(cp->command);
            eat_spaces(&cmd);
            return cp->action(cmd);
        }

    printf("Unrecognized command\n");

    return -1;
}


/* Read single line of input from serial port.
 */
char * readline(char *cmdline, int n) {
    char        *cp;
    int         ch;

    cp=cmdline;
    printf("> ");
    while((ch = serial_getc()) != '\n' && ch != '\r') {
        if (ch == '\b') {
            if (cp != cmdline) {
                cp--;
                printf("\b \b");
            }
        } else {
            if(cp<cmdline+n-1) {
                *cp++ = ch;
                serial_putc(ch);
            }
        }
    }
    *cp=0;
    printf("\n");

    return cmdline;
}       

/* Read line, echo '*'s */
char * readpasswd(char *pwd, int n)
{
    char        *cp;
    int         ch;

    cp=pwd;
    while((ch = serial_getc()) != '\n' && ch != '\r') 
    {
        if (ch == '\b') {
            if (cp != pwd) {
                cp--;
                printf("\b \b");
            }   
        } else {        
            if(cp<pwd+n-1) {
                *cp++ = ch;
                serial_putc('*');
            }   
        }       
    }
    *cp=0;
    printf("\n");

    return pwd;
}


/*
 * CLI timeout & input processing loop
 */

void cli() 
{
    int i;
    char        cmd[0x100];

    if(RUNNING_IN_FLASH) {
        for(i=0;i<50;i++) {
            udelay(10000);

            if(serial_tstc())
                break;
        }

        if(!serial_tstc() || serial_peekc()!='Q')
            return;
    }
    else
        if(!serial_tstc() && !hw_passwd_disable())
            return;


    /* Password check loop */
    while(bd_valid() && __bd->enable_passwd && !hw_passwd_disable()) 
    {
        if(!RUNNING_IN_FLASH)
            wdt_countdown=SIPO_CLI_COUNTDOWN;

        printf("Password: ");
        readpasswd(cmd,0x100);
        if(!passwd(cmd, NULL))
            break;
        printf("Invalid password\n");
        udelay(1000000);
    }

    /* CLI command processing loop */
    do {
        if(!RUNNING_IN_FLASH)
            wdt_countdown=SIPO_CLI_COUNTDOWN;

        readline(cmd, 0x100);

        if(RUNNING_IN_FLASH && !strncmp(cmd, "cont", 4))
            return;
        else
            cli_command(cmd);
    } while(1);
}
