/*******************************************************************************

  FCC Ethernet interface
  -----------------------

  Author: Arto Vuori <avuori@ssh.com>

  Copyright (C) 2000 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved. 

  Simple polling ethernet interface for MPC8260 FCC Ethernet controllers
  Adapted from FCC ethernet example from motorola.

*******************************************************************************/

#include "sipo.h"
#include <stdio.h>
#include <string.h>

/* Ethernet interfaces */
eth_if eth_interface[] = {
    { 0, uninitialized, "eth0 @ FCC1", 31, {0xde,0xad,0,0,0,1}, 
      (bufferdesc *)(&(IMMR->dpram1[0x400])), (bufferdesc *)(&(IMMR->dpram1[0x500])) },
    { 1, uninitialized, "eth1 @ FCC2", 30, {0xde,0xad,0,0,0,2}, 
      (bufferdesc *)(&(IMMR->dpram1[0x200])), (bufferdesc *)(&(IMMR->dpram1[0x300])) },
    { 2, uninitialized, "eth2 @ FCC3", 29, {0xde,0xad,0,0,0,3}, 
      (bufferdesc *)(&(IMMR->dpram1[0x600])), (bufferdesc *)(&(IMMR->dpram1[0x700])) },
    { -1 }
};


/* Send one frame 
     ptr = pointer data frame
     len = frame length (must be less than BUFFER_SIZE-4)
*/
int eth_send(eth_if *eth, void *ptr, int len)
{
    /* Wait until buffer becomes free */
    while(eth->TxBD[eth->tx_current].bd_cstatus&0x8000);

     /* Panic on too long frames */
    if(len>=BUFFER_SIZE-4) 
        return -1;

    /* copy data to buffer & set buffer length*/
    memcpy(eth->TxBD[eth->tx_current].bd_addr, ptr, len);
    eth->TxBD[eth->tx_current].bd_length = len;

    /* Mark buffer ready */
    eth->TxBD[eth->tx_current].bd_cstatus|=0x8000;

    /* Advance current TxBD pointer */
    eth->tx_current=(eth->tx_current+1<NUM_TXBDS)?eth->tx_current+1:0;

    return 0;
}


/* Receive one frame
   
   parameters:
     ptr = pointer to data frame
     maxlen = maximum frame length

   returns:
     0  if no data frame available
     -1 on errors
     >0 received frame actual length
*/
int eth_recv(eth_if *eth, void *ptr, int maxlen) {
    int len,s;

    /* Check if buffer is empty */
    if(eth->RxBD[eth->rx_current].bd_cstatus&0x8000)
        return 0;

    /* Read frame status */
    s=eth->RxBD[eth->rx_current].bd_cstatus;

    /* copy data to buffer & read buffer length */
    len=eth->RxBD[eth->rx_current].bd_length;
    memcpy(ptr, eth->RxBD[eth->rx_current].bd_addr, (maxlen>=len)?len:maxlen);

    /* Mark buffer empty */
    eth->RxBD[eth->rx_current].bd_cstatus |= 0x8000;
    
    /* Advance current TxBD pointer */
    eth->rx_current=(eth->rx_current+1<NUM_TXBDS)?eth->rx_current+1:0;

    if(s&BD_RX_ERROR)
        return -1;

    return len;
}


/* Initialize buffer descriptor rings */

static void InitBDs(eth_if *eth)
{  
    UHWORD index;

    /* Disable FCC2 while we program the buffer  */
    /* descriptors and the parameter RAM.        */
    /* (Just good practice)                      */

    /* Clear the ENT/ENR bits in the GFMR -- disable Transmit/Receive */
    IMMR->fcc_regs[eth->num].gfmr &= !(GFMR_ENT | GFMR_ENR);

    /* Initialize RxBDs. */
    for (index = 0; index < NUM_RXBDS; index++)
    {
        /* Allocate Receive Buffers */
        eth->RxBD[index].bd_addr = eth->RxBuffers[index];

        eth->RxBD[index].bd_length = BUFFER_SIZE-4;

        if( index != (NUM_RXBDS-1) )
        {
            eth->RxBD[index].bd_cstatus = 0x8000;    /* Empty */
        }
        else
        {    /* Last RX BD. Set the Empty, Wrap, and Interrupt bits */
            eth->RxBD[index].bd_cstatus = 0xB000;  
        }
    }

    /* Initialize TxBDs. */
    for (index=0; index < NUM_TXBDS; index++)    
    {
        /* load the address of the data buffer in external memory & set length */
        eth->TxBD[index].bd_addr = eth->TxBuffers[index];

        eth->TxBD[index].bd_length = 0; 

        if( index != (NUM_TXBDS-1) )
        {
            /* Set NOTReady, PAD, Last, TC bits */
            eth->TxBD[index].bd_cstatus = 0x4C00;     
        }
        else
        {
            /* Set NOTReady, PAD, Wrap, Last, and TC bits */
            eth->TxBD[index].bd_cstatus = 0x6C00;     
        }
    }

    eth->rx_current=0;
    eth->tx_current=0;
}

void eth_status(eth_if *eth) {
    int x;

    printf("%s: ", eth->name);
    if(eth->status==uninitialized) {
        printf("Uninitialized\n");
        return;
    } else if(eth->status==ready) {
        printf("Ready, ");
    } else {
        printf("Error");
        return;
    }
    
    mdio_puke(eth->mdio_phyid);

    x=mdio_read(eth->mdio_phyid,18);
    if(x&(1<<11)) 
        printf("Full duplex, ");
    else
        printf("Half duplex, ");

    if(x&(1<<10)) 
        printf("100Mbs, ");
    else
        printf("10Mbs,");

    if(x&(1<<9)) 
        printf("Signal received, ");
    else
        printf("No signal, ");

    if(x&(1<<8)) 
        printf("PLL Lock\n");
    else
        printf("No PLL lock\n");

}


/* Initialize FCC ethernet controller */

void eth_init(eth_if *eth)
{
    unsigned char *eap;
    int i;
    t_EnetFcc_Pram* fcc;

    printf("Initializing %s ...\n", eth->name);

    eth->RxBuffers=malloc(BUFFER_SIZE*NUM_RXBDS);
    eth->TxBuffers=malloc(BUFFER_SIZE*NUM_TXBDS);

    /* Initialize management interface */
    mdio_init();
    mdio_write(eth->mdio_phyid,0,0x8000);    /* Reset */
#if 0
    mdio_write(eth->mdio_phyid,0,0x3000);    /* 100mbit mode, Autonegotiation */
    mdio_write(eth->mdio_phyid,19,0x0000);   
    mdio_write(eth->mdio_phyid,21,0x0004);   
#else
    mdio_write(eth->mdio_phyid,4,0x00a1);    /* Advertise: can do: 10/100mbit, half-duplex */
    mdio_write(eth->mdio_phyid,19,0x0000);   /* Disable power savings */
    mdio_write(eth->mdio_phyid,21,0x0004);   /* Disable scrambling */
    mdio_write(eth->mdio_phyid,0,0x3200);    /* 100Mbps, Half Duplex, restart autonegotiation */
#endif
    udelay(100000);

    /* initialize buffer descriptor rings */
    InitBDs(eth);
     
    switch(eth->num) {
    case 0: /* FCC1 */
        /* Initialize the parallel port I/O pins */
        IMMR->io_regs[PORTA].ppar = (IMMR->io_regs[PORTA].ppar & 0xfffc03c0) | 0x0003fc3f;
        IMMR->io_regs[PORTA].pdir = (IMMR->io_regs[PORTA].pdir & 0xfffc03c0) | 0x00003c0c;
        IMMR->io_regs[PORTA].psor = (IMMR->io_regs[PORTA].psor & 0xfffc03c0) | 0x0000003f;
        IMMR->io_regs[PORTA].podr = (IMMR->io_regs[PORTA].podr & 0xfffc03c0) | 0x00000000;

        /* Enable CLK 11 & 12 pins */
        IMMR->io_regs[PORTC].ppar |= 0x00000c00;
        IMMR->io_regs[PORTC].pdir &= ~0x00000c00;
        IMMR->io_regs[PORTC].psor &= ~0x00000c00;
        IMMR->io_regs[PORTC].podr &= ~0x00000c00;

        /* Connect FCC1's Tx and Rx clocks to CLK 11 & 12 */
        IMMR->cpm_mux_cmxfcr = (IMMR->cpm_mux_cmxfcr & 0x80ffffff) | 0x37000000;
        break;
    case 1: /* FCC2 */
        /* Initialize the parallel port I/O pins */
        IMMR->io_regs[PORTB].ppar = (IMMR->io_regs[PORTB].ppar & ~0x00003fff) | 0x00003fff;
        IMMR->io_regs[PORTB].pdir = (IMMR->io_regs[PORTB].pdir & ~0x00003fff) | 0x000003c5;
        IMMR->io_regs[PORTB].psor = (IMMR->io_regs[PORTB].psor & ~0x00003fff) | 0x00000004;
        IMMR->io_regs[PORTB].podr = (IMMR->io_regs[PORTB].podr & ~0x00003fff) | 0x00000000;

        /* Enable CLK 13 & 14 pins */
        IMMR->io_regs[PORTC].ppar |= 0x00003000;
        IMMR->io_regs[PORTC].pdir &= ~0x00003000;
        IMMR->io_regs[PORTC].psor &= ~0x00003000;
        IMMR->io_regs[PORTC].podr &= ~0x00003000;

        /* Connect FCC2's Tx and Rx clocks to CLK 13 & 14 */
        IMMR->cpm_mux_cmxfcr = (IMMR->cpm_mux_cmxfcr & 0xff80ffff) | 0x00250000;
        break; 
    case 2: /* FCC3 */
        /* Initialize the parallel port I/O pins */
        IMMR->io_regs[PORTB].ppar = (IMMR->io_regs[PORTB].ppar & ~0x0fffc000) | 0x0fffc000;
        IMMR->io_regs[PORTB].pdir = (IMMR->io_regs[PORTB].pdir & ~0x0fffc000) | 0x0f030000;
        IMMR->io_regs[PORTB].psor = (IMMR->io_regs[PORTB].psor & ~0x0fffc000) | 0x00000000;
        IMMR->io_regs[PORTB].podr = (IMMR->io_regs[PORTB].podr & ~0x0fffc000) | 0x00000000;

        /* Enable CLK 15 & 16 pins */
        IMMR->io_regs[PORTC].ppar |= 0x0000c000;
        IMMR->io_regs[PORTC].pdir &= ~0x0000c000;
        IMMR->io_regs[PORTC].psor &= ~0x0000c000;
        IMMR->io_regs[PORTC].podr &= ~0x0000c000;

        /* Connect FCC3's Tx and Rx clocks to CLK 15 & 16 */
        IMMR->cpm_mux_cmxfcr = (IMMR->cpm_mux_cmxfcr & 0xffff80ff) |0x00003700;
        break; 
    }
       
    /* Common Parameter RAM Area Initialization */
    IMMR->pram.serials.fcc_pram[eth->num].riptr = 0x3000 + 0x40*eth->num;
    IMMR->pram.serials.fcc_pram[eth->num].tiptr = 0x3020 + 0x40*eth->num;

    /* Let FCC2 know where base of Rx & Tx BDs are    */
    IMMR->pram.serials.fcc_pram[eth->num].rbase = (UWORD)eth->RxBD;
    IMMR->pram.serials.fcc_pram[eth->num].tbase = (UWORD)eth->TxBD;     

    /* Mot byte ordering, Buffers and BDs on 60x bus, enable snooping */
    IMMR->pram.serials.fcc_pram[eth->num].rstate &= 0x00FFFFFF;
    IMMR->pram.serials.fcc_pram[eth->num].rstate |= 0x30000000;  

    /* Mot byte ordering, Buffers and BDs on 60x bus, enable snooping */
    IMMR->pram.serials.fcc_pram[eth->num].tstate &= 0x00FFFFFF;
    IMMR->pram.serials.fcc_pram[eth->num].tstate |= 0x30000000; 

    /* Set MRBLR -- Max. Receive Buffer Length       */
    /* Must be divisible by 32 and not less than 64. */
    IMMR->pram.serials.fcc_pram[eth->num].mrblr = ENET_MRBLR;
      
    
    /* Ethernet Specific Parameter RAM Initialization */

    /* set up abbreviated pointer */
    fcc = (t_EnetFcc_Pram * ) &(IMMR->pram.serials.fcc_pram[eth->num].SpecificProtocol.e);
    fcc->c_mask   = ENET_C_MASK;   /* Constant MASK for CRC */
    fcc->c_pres   = ENET_C_PRES;   /* CRC Preset */
    fcc->crcec    = 0x00000000;     /* CRC Error Counter */
    fcc->alec     = 0x00000000;     /* Align. Error Counter */
    fcc->disfc    = 0x00000000;     /* Discard Frame Counter */
    fcc->ret_lim  = ENET_RET_LIM;  /* Retry Limit Threshold */
    fcc->p_per    = 0x00000000;     /* Persistence */
    fcc->gaddr_h  = 0x00000000;     /* Group Addr. Filter 1 */
    fcc->gaddr_l  = 0x00000000;     /* Group Addr. Filter 2 */
    fcc->tfcstat  = 0x00000000;     /* temp BD holder */
    fcc->mflr     = ENET_MFLR;     /* Max Frame Length Reg. */
    for(eap=(unsigned char *)&(fcc->paddr1_h),i=0;i<6;i++)
        eap[i]=eth->phyaddr[5-i];
    fcc->iaddr_h  = 0x00000000;     /* Ind. Addr. Filter 1 */
    fcc->iaddr_l  = 0x00000000;     /* Ind. Addr. Filter 2 */
    fcc->minflr   = ENET_MINFLR;   /* Min Frame Length Reg. */
    fcc->taddr_h  = 0x00000000;     /* Temp Address (MSB) */
    fcc->taddr_m  = 0x00000000;     /* Temp Address */
    fcc->taddr_l  = 0x00000000;     /* Temp Address (LSB) */
    fcc->pad_ptr  = IMMR->pram.serials.fcc_pram[eth->num].tiptr;/* internal pad pointer */
    fcc->cf_type = 0x00000000;      /* reserved, cleared */
    fcc->maxd1   = ENET_MDMA;      /* Max DMA1 Length Reg. */
    fcc->maxd2   = ENET_MDMA;      /* Max DMA2 Length Reg. */
    
 
    /* Register Initializations */

    IMMR->fcc_regs[eth->num].gfmr = 0x0000000C;                                                              
    IMMR->fcc_regs[eth->num].dsr = ENET_DSR;

    /*  CCITT-CRC | half duplex */
    IMMR->fcc_regs[eth->num].psmr = 0x00000080;    

    /* Clear FCCE Register by writing all 1's. */
    IMMR->fcc_regs[eth->num].fcce = 0xFFFF0000;

   /* Issue Init RX & TX Parameters Command for FCC2. This command to */
   /* the CR lets it know to reinitialize FCC2 with the new parameter */
   /* RAM values. When the ENT/ENR bits are set below, Hunt Mode will */
   /* begin automatically.                                            */
    while ((IMMR->cpm_cpcr & CPCR_FLG) != READY_TO_RX_CMD); 

    switch(eth->num) {
    case 0:
        IMMR->cpm_cpcr = CPCR_INIT_TX_RX_PARAMS | CPCR_FCC1_CH | CPCR_MCN_FEC | CPCR_FLG;
        break;
    case 1:
        IMMR->cpm_cpcr = CPCR_INIT_TX_RX_PARAMS | CPCR_FCC2_CH | CPCR_MCN_FEC | CPCR_FLG;
        break;
    case 2:
        IMMR->cpm_cpcr = CPCR_INIT_TX_RX_PARAMS | CPCR_FCC3_CH | CPCR_MCN_FEC | CPCR_FLG;
        break;
    }

    while ((IMMR->cpm_cpcr & CPCR_FLG) != READY_TO_RX_CMD); 

    /* Set the ENT/ENR bits in the GFMR -- Enable Transmit/Receive */
    IMMR->fcc_regs[eth->num].gfmr |= GFMR_ENT | GFMR_ENR;

    eth->status=ready;
}










