/*******************************************************************************

    Sipo Flash interface
    --------------------

  Author: Karri Kaksonen <karrik@ssh.com>

  Modifications: Arto Vuori <avuori@ssh.com>

  Copyright (C) 2000 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved. 

  Flash burning utility

*******************************************************************************/

#include <string.h>
#include "sipo.h"

#define MADE_BY_INTEL 0x89
#define STRATA_FLASH_16MB 0x18
#define STRATA_FLASH_8MB 0x17
#define FLASH1_OR_16M 0xff0008f6        
#define FLASH1_OR_8M 0xff8008f6 
#define FLASH2_OR_16M 0xff0008f6        
#define FLASH2_OR_8M 0xff8008f6 
#define FLASH2_BR_HI 0xff001001
#define FLASH2_BR_LO 0xfe801001

#define FLASHWORD(index) (*(volatile unsigned short *)(FLASH_START_ADDRESS+(index)))

static unsigned short flash_lanes_swapped=0;
static unsigned short flash_A7_missing=0;

/*
 * flash_sts() reads flash status from IO pins. 
 * Returns zero if either flash is busy
 */

#define PIN_STS_LO (0x80000000>>5)
#define PIN_STS_HI (0x80000000>>4)

static inline int flash_sts(void) 
{
        if((IMMR->io_regs[2].pdat&PIN_STS_HI) && (IMMR->io_regs[2].pdat&PIN_STS_LO))
                return 1; /* Both are ready */
        else
                return 0; /* One or both flash is not ready */
}

/*
  Erase single flash block
*/
int
flash_erase_block(unsigned int blocknum)
{
    int timeout;
    unsigned short dummy;
    unsigned int blockaddr;

    udelay(1000);
    blockaddr = blocknum*FLASH_BLOCKSIZE;

    FLASHWORD(blockaddr) = 0x2020;
    mb();
    FLASHWORD(blockaddr) = 0xd0d0;
    mb();
    udelay(100000);
    dummy = FLASHWORD(blockaddr);
    mb();
    timeout = 50;
    while (timeout-- && !flash_sts()) {
        udelay(100000);
        mb();
    }
    if (timeout == 0)
      printf("flash_erase_block - timeout\n");
    FLASHWORD(blockaddr) = 0xffff;
    mb();
    dummy = FLASHWORD(blockaddr);
    mb();

    return 0;
}

/*
 * Write single buffer. (32 bytes)
 */
void
flash_write_buffer(char *addr, unsigned int i) 
{
        unsigned int k, blockaddr;
        unsigned short dummy;
        int timeout;

        blockaddr = (i&~(FLASH_BLOCKSIZE-1));

        /* Buffer Write command */
        FLASHWORD(blockaddr) = 0xe8e8;
        mb();
        for(k=100;k;k--) {
                udelay(10);
                if((FLASHWORD(blockaddr)&0x8080)!=0) break;
        }
        if(!k)
                return; /* Buffer not available */

        /* Write count & address registers*/
        mb();
        if (flash_lanes_swapped)
            FLASHWORD(blockaddr) = 0x0f00;
        else
            FLASHWORD(blockaddr) = 0x000f;
        mb();

        for(k=0;k<32;k+=2*4) /* Write to buffer (with some loop unrolling) */
        {
                FLASHWORD(i+k+2*0) = *(unsigned short *)(addr+k+2*0); mb();
                FLASHWORD(i+k+2*1) = *(unsigned short *)(addr+k+2*1); mb();
                FLASHWORD(i+k+2*2) = *(unsigned short *)(addr+k+2*2); mb();
                FLASHWORD(i+k+2*3) = *(unsigned short *)(addr+k+2*3); mb();
        }

        /* confirm write */
        FLASHWORD(blockaddr) = 0xd0d0;
        mb();
                        
        /* Wait until status clear */
        dummy=FLASHWORD(blockaddr);
        mb();
        timeout = 100000;
        while (timeout-- && !flash_sts())
          udelay(10);
        if (timeout == 0)
          printf("flash_write_buffer - timeout\n");
}

/*
  Program single flash block 
*/

int
flash_program_block(unsigned int blocknum, void *addr)
{
    int timeout;
    unsigned short dummy;
    unsigned int blockaddr;
    unsigned int i;

    udelay(1000);
    blockaddr = blocknum*FLASH_BLOCKSIZE;

    for(i=0;i<FLASH_BLOCKSIZE;i+=32)
        flash_write_buffer(addr+i, blockaddr+i);

    udelay(210);
    FLASHWORD(blockaddr) = 0xffff;
    mb();
    dummy=FLASHWORD(blockaddr);
    mb();

    return 0;
}


/*
  Write JFFS2 block header to the beginning of the (previously erased) block
*/
static unsigned char jffs2_block_hdr[32] = {
  0x19, 0x85, 0x20, 0x03, 0x00, 0x00, 0x00, 0x0c,
  0xf0, 0x60, 0xdc, 0x98, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

int
flash_program_block_jffs2_hdr(unsigned int blocknum)
{
  unsigned int blockaddr;
  unsigned short dummy;

  udelay(1000);
  blockaddr = blocknum * FLASH_BLOCKSIZE;

  flash_write_buffer(jffs2_block_hdr, blockaddr);

  udelay(210);
  FLASHWORD(blockaddr) = 0xffff;
  mb();
  dummy = FLASHWORD(blockaddr);
  mb();

  return 0;
}


void
flash_clear_lock_bits()
{
    unsigned short dummy;
    int timeout;

    FLASHWORD(0)=0x6060;
    mb();
    FLASHWORD(0)=0xd0d0;
    mb();
    udelay(500000);
    dummy = FLASHWORD(0);
    mb();
    timeout = 3;
    while (timeout-- && !(FLASHWORD(0) & 0x8080)) {
        udelay(100000);
        mb();
    }
    FLASHWORD(0) = 0xffff;
    mb();
}

/*
  Read serial number of the box
  If the code should work on early boards also (below serial no 11),
  you have to call this routine at least once befor doing a write.
  This will check for how the chip was attached to the bus
  and set the static flash_lanes_swapped variable.
  As the signal A7 was also missing on boards up to board 70 we have to test
  for this as well.
  Most of the commands don't care if the lanes are swapped or not.
  But the burst-write mode has to know it.
*/
long long
flash_read_serial_number(unsigned char *Manuf_ID, unsigned char *Device_Type, unsigned char *Lock, char *SerialLock)
{
    long long SerialNr;
    volatile unsigned short *flash;
    int lock_offset;

    flash = (unsigned short *)(FLASH_START_ADDRESS);
    *flash = 0x9090;
    mb();
    flash_lanes_swapped = ((*flash & 0xff00) != 0);
    flash_A7_missing = (flash[64] != 0);
    if (flash_A7_missing)
        lock_offset = 64;
    else
        lock_offset = 128;
    if (flash_lanes_swapped) {
        *Manuf_ID = (flash[0]) >> 8;
        *Device_Type = (flash[1]) >> 8;
        *Lock = (flash[2]) >> 8;
        *SerialLock = (flash[lock_offset]) >> 8;
    } else {
        *Manuf_ID = flash[0];
        *Device_Type = flash[1];
        *Lock = flash[2];
        *SerialLock = flash[lock_offset];
    }
    ((unsigned short *)&SerialNr)[0] = flash[lock_offset+5];
    ((unsigned short *)&SerialNr)[1] = flash[lock_offset+6];
    ((unsigned short *)&SerialNr)[2] = flash[lock_offset+7];
    ((unsigned short *)&SerialNr)[3] = flash[lock_offset+8];
    mb();
    *flash = 0xffff;
    mb();
    return SerialNr;
}

/*
  Write serial number of the box. This will also lock the
  memory location.
*/
void
flash_write_serial_number(long long SerialNr)
{
    volatile unsigned short *flash;
    unsigned short dummy;
    int lock_offset;

    if (flash_A7_missing)
        lock_offset = 64;
    else
        lock_offset = 128;
    flash = (short *)(FLASH_START_ADDRESS);
    *flash=0xc0c0;
    mb();
    flash[lock_offset+5] = ((unsigned short *)&SerialNr)[0];
    mb();
    udelay(630);
    dummy = *flash;
    mb();
    while (!(*flash & 0x8080))
        mb();
    *flash=0xc0c0;
    mb();
    flash[lock_offset+6] = ((unsigned short *)&SerialNr)[1];
    mb();
    udelay(630);
    dummy = *flash;
    mb();
    while (!(*flash & 0x8080))
        mb();
    *flash=0xc0c0;
    mb();
    flash[lock_offset+7] = ((unsigned short *)&SerialNr)[2];
    mb();
    udelay(630);
    dummy = *flash;
    mb();
    while (!(*flash & 0x8080))
        mb();
    *flash=0xc0c0;
    mb();
    flash[lock_offset+8] = ((unsigned short *)&SerialNr)[3];
    mb();
    udelay(630);
    dummy = *flash;
    mb();
    while (!(*flash & 0x8080))
        mb();
    if (SerialNr != -1) {
        *flash=0xc0c0;
        mb();
        if (flash_lanes_swapped)
            flash[lock_offset] = 0xfdff;
        else
            flash[lock_offset] = 0xfffd;
        mb();
        udelay(630);
        dummy = *flash;
        mb();
        while (!(*flash & 0x8080))
            mb();
    }
    *flash = 0xffff;
    mb();
}

int memscmp(const void * cs,const void * ct,size_t count)
{
        const unsigned short *su1, *su2;
        signed char res = 0;

        for( su1 = cs, su2 = ct; 0 < count; ++su1, ++su2, count--)
                if ((res = *su1 - *su2) != 0)
                        break;
        return res;
}

#ifdef DEBUG
void dump_diffs(const void *cs, const void *ct, size_t count)
{
        const unsigned short *su1, *su2;
        int err_count = 0, offset = 0;

        for (su1 = cs, su2 = ct; 0 < count; offset++, ++su1, ++su2, count--)
                if (*su1 - *su2 != 0)
                  {
                    if (err_count < 10)
                      {
                        printf("DIFF<@%d: %d != %d>", offset, *su1, *su2);
                      }
                    err_count++;
                  }

        printf("<TOTAL DIFFS: %d>", err_count);
}
#else /* DEBUG */
inline void dump_diffs(const void *cs, const void *ct, size_t count) {}
#endif /* DEBUG */


/* 
   Erase, program and verify multiple flash blocks
 */
int flash_write(int blocknum, void *addr, unsigned int count)
{
    unsigned char Manuf_ID;
    unsigned char Device_Type;
    unsigned char Lock;
    char SerialLock;
    long long SerialNr;
    unsigned int i,j;

    IMMR->io_regs[2].ppar&=~PIN_STS_LO;
    IMMR->io_regs[2].pdir&=~PIN_STS_LO;
    IMMR->io_regs[2].podr|= PIN_STS_LO;
    IMMR->io_regs[2].psor&=~PIN_STS_LO;

    IMMR->io_regs[2].ppar&=~PIN_STS_HI;
    IMMR->io_regs[2].pdir&=~PIN_STS_HI;
    IMMR->io_regs[2].podr|= PIN_STS_HI;
    IMMR->io_regs[2].psor&= PIN_STS_HI;

    SerialNr = flash_read_serial_number(&Manuf_ID, &Device_Type, &Lock, &SerialLock);
    if ((Device_Type != STRATA_FLASH_16MB) && (Device_Type != STRATA_FLASH_8MB)) 
    {
        printf("Flash chip is unknown\n");
        return -1;
    }

    if(Lock & 0x01) 
    {
        printf("Requires unlocking\n");
        /* If first Block is locked, clear lock bits */
        flash_clear_lock_bits();
        SerialNr = flash_read_serial_number(&Manuf_ID, &Device_Type, &Lock, &SerialLock);
    }

    for(i=0;i<count;i++)
    {
        printf(".");
        for(j=0;j<FLASH_MAX_RETRY;j++) 
        {
            set_leds(SIPO_LED_POWER|SIPO_LED_STATUS_RED);
            if(flash_erase_block(blocknum+i))
                continue;

            set_leds(SIPO_LED_POWER|SIPO_LED_STATUS_WEIRD);
            if(flash_program_block(blocknum+i, addr+i*FLASH_BLOCKSIZE))
                continue;
            if(!memscmp((void *)(FLASH_START_ADDRESS+(blocknum+i)*FLASH_BLOCKSIZE), addr+i*FLASH_BLOCKSIZE, FLASH_BLOCKSIZE/2))
                break;
            else
              {
                printf("-");
                dump_diffs((void *)(FLASH_START_ADDRESS +
                                    (blocknum + i) * FLASH_BLOCKSIZE),
                           addr + i * FLASH_BLOCKSIZE, FLASH_BLOCKSIZE / 2);
              }
        }

        if(j==FLASH_MAX_RETRY) {
            set_leds(SIPO_LED_STATUS_RED);
            printf("flash_write: max retry count reached - abort\n");
            return -1;
        }
    }

    set_leds(SIPO_LED_STATUS_GREEN);
    return 0;
}

/* 
   Write non-aligned and wrong sized memory areas. 
   Cut & Copy & Paste original data to the areas not to be programmed.
 */
int flash_cutcopypaste(unsigned int offset, void *addr, unsigned int count)
{
    static      char *temp=NULL;
    unsigned    int i,j;

    if(!temp)
        temp=malloc(FLASH_BLOCKSIZE);

    i=(offset&(FLASH_BLOCKSIZE-1));
    j=FLASH_BLOCKSIZE-i;

    if(j>=count)
    {
        /* Write single block */
        memcpy(temp, (void *)(FLASH_START_ADDRESS+offset-i), i);
        memcpy(temp+i, addr, count);
        memcpy(temp+i+count, (void *)(FLASH_START_ADDRESS+offset+count), FLASH_BLOCKSIZE-i-count);
        if(flash_write(offset/FLASH_BLOCKSIZE, temp, 1)) {
            printf("Writing single block failed\n");
            return -1;
        }
    } 
    else
    {
        if(i) {
            memcpy(temp, (void *)(FLASH_START_ADDRESS+offset-i), i);
            memcpy(temp+i, addr, j);
            /* Write first block */
            if(flash_write(offset/FLASH_BLOCKSIZE, temp, 1)) {
                printf("Writing first block failed\n");
                return -1;
            }
            addr+=j; count-=j; offset+=j;
        }

        /* Writing middle blocks */
        if(flash_write(offset/FLASH_BLOCKSIZE, addr, count/FLASH_BLOCKSIZE)) {
            printf("Writing middle blocks failed\n");
            return -1;
        }

        addr+=count/FLASH_BLOCKSIZE*FLASH_BLOCKSIZE;
        offset+=count/FLASH_BLOCKSIZE*FLASH_BLOCKSIZE;
        count-=count/FLASH_BLOCKSIZE*FLASH_BLOCKSIZE;

        if(count) {
            memcpy(temp, addr, count);
            memcpy(temp+count, (void *)(FLASH_START_ADDRESS+offset+count), FLASH_BLOCKSIZE-count);
        
            /* Write last block */
            if(flash_write(offset/FLASH_BLOCKSIZE, temp, 1)) {
                printf("Writing last block failed\n");
                return -1;
            }
        }
    }
    printf("\n");

    return 0;
}

flashmain()
{
    serial_init();

    printf("\nProgramming FLASH device\n");
    set_leds(SIPO_LED_POWER);
    while (flash_write(0, (unsigned short *)(0x00000000), ((unsigned long)(_end)+FLASH_BLOCKSIZE-1)/FLASH_BLOCKSIZE))
    printf("\nDone\n");
    while (1) ;
}

/*
  Burn pattern to flash and verify result
*/
static long prandom(long seed)
{
    seed += 0x12348765;
    seed ^= 0x5555aaaa;
    return seed;
}

void
flash_test()
{
    long *ptr;
    long step;
    long val;
    long *base_mem=(long *)KERNEL_LOAD_ADDRESS;
    long blknr1, blknr2, blknr;
    int status;
    int status0;
    volatile unsigned short *flash2;
    volatile unsigned short *flash;
    unsigned short fsize;

    flash = (short *)(FLASH_START_ADDRESS);
    /*
      Save bootloader into ram while we mess up the flash
    */
    memcpy((unsigned short *)(KERNEL_LOAD_ADDRESS + FLASH_BLOCKSIZE), (unsigned short *)flash, FLASH_BD_BLOCK * FLASH_BLOCKSIZE);
    /*
      Taste the flash ID-blocks to find out what we have on the board.
    */
    *flash = 0x9090;
    fsize = flash[1];
    *flash = 0xffff;
    if ((fsize == 0x17) || (fsize == 0x1700)) {
        flash2 = &flash[4*1024*1024];
        blknr1 = 64;
        *flash2 = 0x9090;
        fsize = flash2[1];
        *flash2 = 0xffff;
    } else {
        flash2 = &flash[8*1024*1024];
        blknr1 = 128;
        *flash2 = 0x9090;
        fsize = flash2[1];
        *flash2 = 0xffff;
    }
    blknr2 = 0;
    if ((fsize == 0x17) || (fsize == 0x1700))
        blknr2 = 64;
    if ((fsize == 0x18) || (fsize == 0x1800))
        blknr2 = 128;

    /*
      Write pseudorandom data for every address line in the first chip.
    */
    status = 0;
    blknr = 0;
    val = blknr;
    while (blknr < blknr1) {
        step = 0;
        ptr = base_mem;
        while (step < FLASH_BLOCKSIZE/sizeof(long)) {
            val = prandom(val);
            ptr[step] = val;
            step++;
        }
        flash_write(blknr, (unsigned short *)base_mem, 1);
        if (!blknr)
            blknr = 1;
        else
            blknr *= 2;
    }
    /*
      Write pseudorandom data for every address line in the second chip.
    */
    blknr = 0;
    while (blknr < blknr2) {
        step = 0;
        ptr = base_mem;
        while (step < FLASH_BLOCKSIZE/sizeof(long)) {
            val = prandom(val);
            ptr[step] = val;
            step++;
        }
        flash_write(blknr1 + blknr, (unsigned short *)base_mem, 1);
        if (!blknr)
            blknr = 1;
        else
            blknr *= 2;
    }
    /*
      Check the data in the first chip.
    */
    blknr = 0;
    val = blknr;
    while (blknr < blknr1) {
        step = 0;
        ptr = (long *)flash;
        ptr += (blknr * FLASH_BLOCKSIZE/sizeof(long));
        status0 = 0;
        if (status) {
            if (ptr[0] == *(long *)flash) {
                printf("   %X %X\n", &ptr[0], ptr[0]);
            }
        }
        while (step < FLASH_BLOCKSIZE/sizeof(long)) {
            val = prandom(val);
            if ((ptr[step] != val) && !status0) {
                printf("At %X %X\n", &ptr[step], ptr[step]);
                status = 1;
                status0 = 1;
            }
            step++;
        }
        if (!blknr)
            blknr = 1;
        else
            blknr *= 2;
    }
    if (status)
        printf("flash test failed for flash 1 (%d MB device)\n", blknr1/8);
    else
        printf("flash test passed for flash 1 (%d MB device)\n", blknr1/8);
    /*
      Check the data in the second chip.
    */
    status = 0;
    blknr = 0;
    while (blknr < blknr2) {
        step = 0;
        ptr = (long *)flash2;
        ptr += (blknr * FLASH_BLOCKSIZE/sizeof(long));
        status0 = 0;
        if (status) {
            if (ptr[0] == *(long *)flash2) {
                printf("   %X %X\n", &ptr[0], ptr[0]);
            }
        }
        while (step < FLASH_BLOCKSIZE/sizeof(long)) {
            val = prandom(val);
            if ((ptr[step] != val) && !status0) {
                printf("At %X %X\n", &ptr[step], ptr[step]);
                status0 = 1;
                status = 1;
            }
            step++;
        }
        if (!blknr)
            blknr = 1;
        else
            blknr *= 2;
    }
    if (blknr2) {
        if (status)
            printf("\nflash test failed for flash 2 (%d MB device)\n",
                   blknr2/8);
        else
            printf("\nflash test passed for flash 2 (%d MB device)\n",
                   blknr2/8);
    }
    /*
      Return the bootloader
    */
    flash_write(0, (unsigned short *)(KERNEL_LOAD_ADDRESS + FLASH_BLOCKSIZE), FLASH_BD_BLOCK);
    printf("\n");
}

/* 
 * Flash memory mapping.
 *
 * Read data from the flash chips.
 * Set the memory areas so that the flashes provide a continuous memory area.
 */
int
init_flash()
{
    volatile unsigned short *flash;
    unsigned short fsize;
    int total_flash;

    flash = (short *)(FLASH_START_ADDRESS);
    *flash = 0x9090;
    fsize = flash[1];
    *flash = 0xffff;
    if ((fsize == 0x17) || (fsize == 0x1700)) {
        IMMR->memc_regs[0].or = FLASH1_OR_8M;
        IMMR->memc_regs[7].br = FLASH2_BR_LO;
        IMMR->memc_regs[7].or = FLASH2_OR_16M;
        flash[4*1024*1024] = 0x9090;
        fsize = flash[4*1024*1024+1];
        flash[4*1024*1024] = 0xffff;
        total_flash = 8 * 1024 * 1024;
    } else {
        IMMR->memc_regs[0].or = FLASH1_OR_16M;
        IMMR->memc_regs[7].br = FLASH2_BR_HI;
        IMMR->memc_regs[7].or = FLASH2_OR_16M;
        flash[8*1024*1024] = 0x9090;
        fsize = flash[8*1024*1024+1];
        flash[8*1024*1024] = 0xffff;
        total_flash = 16 * 1024 * 1024;
    }
    if ((fsize == 0x17) || (fsize == 0x1700)) {
        IMMR->memc_regs[7].or = FLASH2_OR_8M;
        total_flash += 8 * 1024 * 1024;
    }
    if ((fsize == 0x18) || (fsize == 0x1800))
        total_flash += 16 * 1024 * 1024;
    return total_flash;
}

