/*
 *      SIPO flash partition management tool
 *
 *      Copyright 2001-2002 SSH Communications Security Corp,
 *      http://www.ssh.com. All Rights Reserved.          
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See COPYING for more
 * details.
 *
 *      8.2.2001 Arto Vuori <avuori@ssh.com>
 *              Initial version
 *
 *      This program can be used to edit SIPO flash partition table and
 *      load images to partitions.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "board.h"

#define IOCTL_FLASH_BLOCK_ERASE 6789

typedef struct {
  unsigned long state[5];
  unsigned long count[2];
  unsigned char buffer[64];
} SHA1_CTX;

extern void SHA1Init(SHA1_CTX* context);
extern void SHA1Update(SHA1_CTX* context, unsigned char* data,
                       unsigned long len);
extern void SHA1Final(unsigned char *digest, SHA1_CTX* context);
extern void sha1(void *buffer, int len, SHA1Digest digest);
extern int verify(void *buffer, int len);

bd_t    bd;
bd_t    *__bd = 0;
int     dev = -1;


int read_bd(void)
{
  SHA1Digest digest;

  /* Do not read twice */
  if (__bd != 0)
    return 0;

  if (dev < 0)
    {
      dev = open("/dev/flash", O_RDWR);
      if (dev < 0)
        {
          fprintf(stderr, "Unable to open /dev/flash\n");
          return -1;
        }
    }

  if (lseek(dev, BD_OFFSET, SEEK_SET) != BD_OFFSET)
    {
      fprintf(stderr, "Unable to seek on device\n");
      return -1;
    }

  if (read(dev, &bd, sizeof(bd)) != sizeof(bd)) {
    fprintf(stderr, "Unable to read BD from device\n");
    return -1;
  }

  sha1((void *) &bd, sizeof(sipo_board_info) - sizeof(bd.verify), digest);
  if (memcmp(bd.verify, digest, sizeof(digest)))
    {
      fprintf(stderr, "BD digest verification failure\n");
      return -1;
    }

  __bd = &bd;

  return 0;
}

int write_bd(void)
{
  SHA1Digest digest;

  /* Write only if BD is read earlier */
  if (__bd == 0)
    return -1;

  if (dev < 0)
    return -1;

  sha1((void *) __bd, sizeof(sipo_board_info) - sizeof(__bd->verify), digest);
  memcpy(__bd->verify, digest, sizeof(__bd->verify));

  if (lseek(dev, BD_OFFSET, SEEK_SET) != BD_OFFSET)
    {
      fprintf(stderr, "Unable to seek on device\n");
      return -1;
    }

  if (write(dev, &bd, sizeof(bd)) != sizeof(bd))
    {
      fprintf(stderr, "Unable to write BD back to device\n");
      return -1;
    }

  return 0;
}



/*
  Modify flash partition table.

  Partition table is array of sipo_flash_partition entries and is located at
  fixed offset in flash device. First partition_table entry is always pt_magic
  type, which does not define any real partition. pt_bootloader should be
  second entry and after that there should be partition table entry.
*/

int partition_set(int index, sipo_partition_type type, int offset, int size)
{
  int i;

  if (__bd == 0)
    return -1;

  if (__bd->pt[0].type != pt_magic)
    {
      printf("Partition table invalid. Initializing table ... \n");
      __bd->pt[0].type = pt_magic;
      __bd->pt[0].size = 0;
      __bd->pt[0].offset = 0;
      __bd->pt[1].type = pt_bootloader;
      __bd->pt[1].offset = 0;
      __bd->pt[1].size = BD_OFFSET;
      __bd->pt[2].type = pt_bd;
      __bd->pt[2].offset = BD_OFFSET;
      __bd->pt[2].size = sizeof(sipo_board_info);
      for (i = 3; i < pt_size; i++)
        __bd->pt[i].type = pt_unused;
    }

  if (offset < 0 || offset + size > __bd->flash_size)
    {
      printf("Partition does not fit in FLASH\n");
      return -1;
    }

  if (index < pt_size && index > 0)
    {
      if (type != pt_unused) {
        for (i = index - 1; i > 0; i--)
                if (__bd->pt[i].type != pt_unused)
                  {
                    if (__bd->pt[i].offset + __bd->pt[i].size > offset)
                      {
                        printf("Overlap with previous partition. No overlap "
                               "allowed\n");
                        return -1;
                      }
                    break;
                  }
        for (i = index + 11; i < pt_size; i++)
          {
            if (__bd->pt[i].type != pt_unused)
              {
                if (__bd->pt[i].offset < size + offset)
                  {
                    printf("Overlap with next partition. No overlap "
                           "allowed\n");
                    return -1;
                  }
                break;
              }
          }
      }
        
      __bd->pt[index].type = type;
      __bd->pt[index].offset = offset;
      __bd->pt[index].size = size;
    }

  return 0;
}


int partition_print(void)
{
  int i;

  if (__bd == 0)
    return -1;

  if (__bd->pt[0].type != pt_magic)
    {
      fprintf(stderr, "Invalid partition table\n");
      return -1;
    }

  printf("Flash partition table (index,type,offset,size)\n");
  for (i = 0; i < pt_size; i++)
    {
      if (__bd->pt[i].type != pt_unused)
        printf("%2d, 0x%-8x, 0x%-8x, 0x%-8x : ", i, __bd->pt[i].type,
               __bd->pt[i].offset, __bd->pt[i].size);
      switch (__bd->pt[i].type)
        {
        case pt_magic:
          printf("Magic type number\n");
          break;
        case pt_bootloader:
          printf("Bootloader\n");
          break;
        case pt_bd:
          printf("Partition table\n");
          break;
        case pt_kernel:
          printf("kernel\n");
          break;
        case pt_rootdisk:
          printf("rootdisk\n");
          break;
        case pt_fs:
          printf("filesystem (JFFS)\n");
          break;
        case pt_jffs2:
          printf("filesystem (JFFS2)\n");
          break;
        case pt_unused:
          break;
        default:
          printf("Unknown\n");
        }
    }

  return 0;
}



#define BUFLEN 0x10000

int check_signature(int fd, int len)
{
  unsigned char *buffer;
  unsigned char sipo_secret[] = SIPO_SECRET;
  int left, i;
  SHA1_CTX ctx;
  SHA1Digest dgst, orig;

  if (len < sizeof(SHA1Digest))
    {
      fprintf(stderr, "Too short file\n");
      return -1;
    }

  left = len - sizeof(SHA1Digest);

  buffer = malloc(BUFLEN);
  if (buffer == NULL)
    {
      fprintf(stderr, "Malloc failed\n");
      return -1;
    }


  SHA1Init(&ctx);
  while (left > 0)
    {
      i = read(fd, buffer, BUFLEN < left ? BUFLEN : left);
      if (i < 0)
        {
          fprintf(stderr, "Read failed\n");
          free(buffer);
          return -1;
        }
      SHA1Update(&ctx, buffer, i);
      left -= i;
    }
  SHA1Update(&ctx, sipo_secret, sizeof(sipo_secret));
  SHA1Final(dgst, &ctx);

  free(buffer);

  if (read(fd, orig, sizeof(SHA1Digest)) != sizeof(SHA1Digest))
    {
      fprintf(stderr, "Could not read checksum\n");
      return -1;
    }

  i = memcmp(orig, dgst, sizeof(SHA1Digest));

  if (i)
    printf("Verification failed!\n");
  else
    printf("Verification OK\n");

  return i;
}

int check_file_signature(const char *fname)
{
  int fin, len, ret;

  fin = open(fname, O_RDONLY);
  if (fin < 0)
    {
      fprintf(stderr, "Unable to open input file '%s'\n", fname);
      return -1;
    }

  len = lseek(fin, 0, SEEK_END);
  lseek(fin, 0, SEEK_SET);

  ret = check_signature(fin, len);

  close(fin);

  return ret;
}


void help(void)
{
  printf("sIPo flash partion management tool v0.2\n");
  printf("Copyright 2001 SSH Communications Security Corp, "
         "All Rights Reserved.\n");
  printf("Usage: flashtool <commands...> <options...>\n");
  printf("Where commands are:\n");
  printf("    -i<file>       Load partition image from the input file\n");
  printf("    -o<file>       Save partition image to the output file\n");
  printf("    -c<dev>        Copy partition from another device\n");
  printf("    -e             Erase partition.\n");
  printf("    -s<index>,<type>,<offset>,<size>\n");
  printf("                   Set partition table entry.\n");
  printf("    -m<file>       Check whether image is modified or identical\n");
  printf("    -p             print Board descriptor & partition table "
         "contents\n");
  printf("    -v<file>       Verify file signature\n");
  printf("    -V             Verify partition signature (use -x, -l, and "
         "possibly -d)\n");
  printf("    -h,H,?         print this help\n");

  printf("\nOptions:\n");
  printf("    -x<num>        Partition index number\n");
  printf("    -d<dev>        device filename for flash device\n");
  printf("    -l             Length of area to copy for -c command\n");
  printf("\n");
}


int main(int argc, char **argv)
{
  int index = -1, base = -1, size = -1, type = -1, copylen = -1;
  int fin = -1, fout = -1, indev = -1, m_in = -1;
  int i, j, k;
  char buffer[FLASH_BLOCKSIZE], buffer2[FLASH_BLOCKSIZE];
  typedef enum {         
    CMD_NONE, CMD_LOAD, CMD_SAVE, CMD_COPY, CMD_ERASE, 
    CMD_PRINT, CMD_VERIFY, CMD_SET, CMD_CMP
  } cmd_type;
  cmd_type command=CMD_NONE;

  if (argc < 2)
    {
      help();
      return 0;
    }

  for (i = 1; i < argc; i++)
    {
      if (argv[i][0] == '-')
        {
          switch (argv[i][1])
            {
            case '?':
            case 'H':
            case 'h':
              help();
              return 0;
            
            case 'V':
              if(command != CMD_NONE)
                {
                  fprintf(stderr, "Invalid arguments. Too many commands.\n");
                  exit(1);
                }
              command = CMD_VERIFY;
              break;

            case 'x':
              index = strtol(&argv[i][2], NULL, 10);
              break;

            case 'l':
              copylen = strtol(&argv[i][2], NULL, 10);
              break;

            case 'd':
              dev = open(&argv[i][2], O_RDWR);
              if (dev < 0)
                {
                  fprintf(stderr, "Unable to open device '%s'\n", &argv[i][2]);
                  return -1;
                }
              break;

            case 's':
              if(command != CMD_NONE)
                {
                  fprintf(stderr, "Invalid arguments. Too many commands.\n");
                  exit(1);
                }
              command = CMD_SET;

              if (read_bd())
                return -1;
              if (sscanf(&argv[i][2], "%d,%d,0x%x,0x%x", &k, &type, &base,
                         &size)!=4) {
                fprintf(stderr, "Invalid arguments. Use '-s<index>,<type>,"
                        "0x<offset>,0x<size>'\n");
                return -1;
              }
              if (partition_set(k, type, base, size))
                return -1;

              break;

            case 'p':
              if(command != CMD_NONE)
                {
                  fprintf(stderr, "Invalid arguments. Too many commands.\n");
                  exit(1);
                }
              command = CMD_PRINT;
              break;

            case 'e':
              if(command != CMD_NONE)
                {
                  fprintf(stderr, "Invalid arguments. Too many commands.\n");
                  exit(1);
                }
              command = CMD_ERASE;
              break;

            case 'i':
              if(command != CMD_NONE)
                {
                  fprintf(stderr, "Invalid arguments. Too many commands.\n");
                  exit(1);
                }
              command = CMD_LOAD;

              fin = open(&argv[i][2], O_RDONLY);
              if (fin < 0)
                {
                  fprintf(stderr, "Unable to open input file '%s'\n",
                          &argv[i][2]);
                  return -1;
                }
              break;

            case 'o':
              if(command != CMD_NONE)
                {
                  fprintf(stderr, "Invalid arguments. Too many commands.\n");
                  exit(1);
                }
              command = CMD_SAVE;

              fout = open(&argv[i][2], O_WRONLY | O_CREAT, 0664);
              if (fout < 0)
                {
                  fprintf(stderr, "Unable to open output file '%s'\n",
                          &argv[i][2]);
                  return -1;
                }
              break;

            case 'm':
              if(command != CMD_NONE)
                {
                  fprintf(stderr, "Invalid arguments. Too many commands.\n");
                  exit(1);
                }
              command = CMD_CMP;

              m_in = open(&argv[i][2], O_RDONLY);
              if (m_in < 0)
                {
                  fprintf(stderr, "Unable to open input file '%s'\n",
                          &argv[i][2]);
                  return -1;
                }
              break;

            case 'c':
              if(command != CMD_NONE)
                {
                  fprintf(stderr, "Invalid arguments. Too many commands.\n");
                  exit(1);
                }
              command = CMD_COPY;

              indev = open(&argv[i][2], O_RDONLY);
              if (indev < 0)
                {
                  fprintf(stderr, "Unable to open input device '%s'\n",
                          &argv[i][2]);
                  return -1;
                }
              break;

            case 'v':
              return check_file_signature(&argv[i][2]);

            default:
              fprintf(stderr, "Unknown option %d '%c'. Use -h for help\n",
                      i, argv[i][1]);
              exit(1);
            }
        }
      else
        {
          fprintf(stderr, "Unknown argument %d '%s'. Use -h for help\n",
                  i, argv[i]);
          exit(1);
        }
    }

  if (read_bd())
    return -1;

  switch(command) 
    {
    case CMD_PRINT:
      fprintf(stderr, "BD digest verification successful\n");
      partition_print();
      exit(0);

    case CMD_SET:
      write_bd();
      exit(0);

    default:
    }
  
  /* Check index, and seek flash to beginning of the partition. */

  if (index == -1)
    {
      fprintf(stderr, "Partition index (-x) required\n");
      return -1;
    }

  size = bd.pt[index].size;
  base = bd.pt[index].offset;
  type = bd.pt[index].type;

  if (lseek(dev, base, SEEK_SET) != base)
    {
      fprintf(stderr, "Seek failed\n");
      return -1;
    }


  switch(command) 
    {
    case CMD_VERIFY:
      if (copylen < 0)
        {
          fprintf(stderr, "Length not specified\n");
          return -1;
        }

      i = check_signature(dev, copylen);
      close(dev);
      return i;

    case CMD_SAVE:
      while (size)
        {
          if (size < FLASH_BLOCKSIZE)
            i = size;
          else
            i = FLASH_BLOCKSIZE;

          j = read(dev, buffer, i);
          if (j <= 0)
            {
              fprintf(stderr, "Error while reading from device\n");
              return -1;
            }

          k = write(fout, buffer, j);
          if (k != j)
            {
              fprintf(stderr, "Error while writing output\n");
              return -1;
            }

          size -= j;
        }
      break;

    case CMD_LOAD:
      i = lseek(fin, 0, SEEK_END);
      if (i < 0)
        {
          fprintf(stderr, "Unable to determine input file length\n");
          return -1;
        }
      else if (i > size)
        {
          fprintf(stderr, "Input file larger than partition\n");
          return -1;
        }
      else
        size = i;
      lseek(fin, 0, SEEK_SET);

      /*
       * write first partial block
       */
      if (lseek(dev, base, SEEK_SET) != base)
        {
          fprintf(stderr, "Seek before first write failed\n");
          return -1;
        }
      i = base & ~(FLASH_BLOCKSIZE - 1);
      if (i != base)
        {
          if (base - i + size > FLASH_BLOCKSIZE)
            j = FLASH_BLOCKSIZE - (base - i);
          else
            j = size;

          read(fin, buffer, j);

          if (write(dev, buffer, j) != j)
            {
              fprintf(stderr, "Unable to write first buffer\n");
              return -1;
            }

          size -= j;
        }

      /* Remaining blocks are full flash blocks, except the last one. */
      while (size > 0)
        {
          j = read(fin, buffer, FLASH_BLOCKSIZE);
          if (j <= 0)
            {
              fprintf(stderr, "Error while reading from input file\n");
              return -1;
            }

          k = write(dev, buffer, j);
          if (k != j)
            {
              fprintf(stderr, "Error while writing to device\n");
              return -1;
            }

          size -= j;
        }
      break;

    case CMD_COPY:
      /* copy the same partition from another device */

      if (index == -1)
        {
          fprintf(stderr, "Partition index (-x) required\n");
          return -1;
        }

      if (copylen < 0)
        {
          printf("Copying the whole partition (%d bytes)\n", size);
          copylen = size;
        }

      if (copylen > size)
        {
          fprintf(stderr, "Copy length larger than partition\n");
          return -1;
        }

      size = copylen;

      if (lseek(indev, base, SEEK_SET) != base)
        {
          fprintf(stderr, "Input device seek before first write failed\n");
          return -1;
        }

      if (lseek(dev, base, SEEK_SET) != base)
        {
          fprintf(stderr, "Seek before first write failed\n");
          return -1;
        }

      i = base & ~(FLASH_BLOCKSIZE - 1);
      if (i != base)
        {
          /* write the first partial block */
          if (base - i + size > FLASH_BLOCKSIZE)
            j = FLASH_BLOCKSIZE - (base - i);
          else
            j = size;

          read(indev, buffer, j);

          if (write(dev, buffer, j) != j)
            {
              fprintf(stderr, "Unable to write first buffer\n");
              return -1;
            }

          size -= j;
        }

      /* Remaining blocks are full flash blocks, except possibly the last one.
       */
      while (size > 0)
        {
          j = read(indev, buffer, FLASH_BLOCKSIZE);
          if (j <= 0)
            {
              fprintf(stderr, "Error while reading from input device\n");
              return -1;
            }

          k = write(dev, buffer, j);
          if (k != j)
            {
              fprintf(stderr, "Error while writing to device\n");
              return -1;
            }

          size -= j;
        }
      break;

    case CMD_CMP:
      /* Compare two images */

      while ((j = read(m_in, buffer, FLASH_BLOCKSIZE)))
        {
          if (j < 0)
            {
              fprintf(stderr, "Error while reading from file\n");
              return -1;
            }

          k = read(dev, buffer2, j);
          if (k != j)
            {
              fprintf(stderr, "Error while reading from device\n");
              return -1;
            }

          if (memcmp(buffer, buffer2, k))
            {
              printf("Images differ\n");
              return 1;
            }
        }

      printf("Identical images\n");
      break;      

    case CMD_ERASE:
      /* Erase partition */
      if(base & (FLASH_BLOCKSIZE-1) || size & (FLASH_BLOCKSIZE-1))
        {
          fprintf(stderr, "Erase can only be used to block aligned partitions\n");
          return 1;
        }

      for(i = base; i < base + size; i += FLASH_BLOCKSIZE)
        {
          if (lseek(dev, i, SEEK_SET) != i)
            {
              fprintf(stderr, "Seek during erase failed\n");
              return -1;
            }
          if((k=ioctl(dev, IOCTL_FLASH_BLOCK_ERASE)))
            {
              fprintf(stderr, "Erase command failed (%d)\n", k);
              return -1;            
            }
        }
      break;

    default:
      fprintf(stderr,"Invalid command\n");
      return 1;
     }

  return 0;
}
