/******************************************************************************

    Sipo Bootloader main
    --------------------

  Author: Arto Vuori <avuori@ssh.com>

  Copyright (C) 2000-2002 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation. See COPYING for more
  details.

  Initialization, main, and some misc routines

******************************************************************************/

#include "sipo.h"
#include "stdio.h"

hardreset() 
{
    /* Disable machine check exception */
    mtmsr(mfmsr()&~MSR_ME);
    mb();

    /* And... This should force us to enter checkstop state */
    *((char *)0x80000000)=0; 
    mb();

    /* ... And reset should happen before exiting this routine */
}

/*
  We should never enter here... 
*/

void fatal_error_exception(unsigned int excnro)
{
    serial_init();
    printf("FATAL ERROR EXCEPTION ");
    printf("NRO %d !!!\n", excnro);
}

/* Watchdog timer decrements wdt_countdown once every second until it reaches zero.
   If wdt_countdown is not zero, hardware watchdog will be pinged to avoid reset.
   To keep system alive, wdt_countdown must be periodically set to nonzero value. It
   is initially set to 30 seconds to allow enough time to uncompress and start kernel.

   NOTE! after 
 */

unsigned int wdt_countdown=SIPO_COUNTDOWN;

wdt_ping() {
    if(wdt_countdown) 
    {
        mb();
        IMMR->siu_swsr = 0x556c;
        mb();
        IMMR->siu_swsr = 0xaa39;
        mb();

        if(!RUNNING_IN_FLASH)
            wdt_countdown--;
    }
}

wdt_init() {
    if(bd_valid() && !__bd->enable_wdt)
        IMMR->siu_sypcr=0xffffffc0; /* Only bus monitors enabled */
    else
        IMMR->siu_sypcr=0xffffffc7; /* Watchdog enabled by default*/
}

/* Length of the decrementer interrupt period */
unsigned int decrementer_period=66000000/4/SIPO_HZ;
/* Software timer interrupt counter */
unsigned int time_counter=0;

/* Decrement interrupt handler. This routine will be called from low-level
   decrementer exception vector. It is triggered when processor decrementer 
   timer expires and the MSR_EE flag is enabled.
*/

void timer_interrupt(void) 
{
    wdt_ping();

    if(!RUNNING_IN_FLASH)
        time_counter++;

    set_dec(decrementer_period);
    
    return;
}

/* 
   Stop all asyncronous activity (communications & interrupts) 
*/

void reset_cpm() 
{
    /* Reset CPM */
    IMMR->cpm_cpcr = (CPCR_RST | CPCR_FLG);
    while (IMMR->cpm_cpcr & CPCR_FLG);

    /* Disable interrupts */
    mtmsr(mfmsr()&~MSR_EE);
}

/*
  Password checks can be disabled by pulling FPGA INIT pin down. State of the pin 
  can be read from MPC8260 IO pin.
*/

hw_passwd_disable() 
{

    IMMR->io_regs[FPGA_INIT_PORT].pdir &= ~ FPGA_INIT_PIN;
    IMMR->io_regs[FPGA_INIT_PORT].ppar &= ~ FPGA_INIT_PIN;
    IMMR->io_regs[FPGA_INIT_PORT].psor &= ~ FPGA_INIT_PIN;
    IMMR->io_regs[FPGA_INIT_PORT].podr &= ~ FPGA_INIT_PIN;

    udelay(1000);

    return ((IMMR->io_regs[FPGA_INIT_PORT].pdat & FPGA_INIT_PIN)==0);
}

/*
  Infinite memory allocator
*/

void *malloc(unsigned size) 
{
    static void *curr=0;
    static void *ptr;

    if(curr==0)
        curr=__HEAP_BEGIN;

    ptr=curr;

    curr+=((size+15)&~15);
    
    return ptr; 
}

/* No need to free anything if memory is infinite */

void free(void *ptr) 
{
}

/*
  Copy board descriptor & command line and then jump to kernel.
*/

void start_kernel(char *cmdline)
{
    bd_t        *bd=(bd_t *)(__bd->bi_memstart+__bd->bi_memsize-0x10000);
    char        *cmd=(char *)(__bd->bi_memstart+__bd->bi_memsize-0x10000+sizeof(bd_t));

    if(hw_passwd_disable()) {
        printf("Kernel booting not allowed if HW password disable is connected\n");
        return;
    }

    memcpy(bd, __bd, sizeof(bd_t));

    strncpy(cmd, cmdline, 0x100);

    /* Stop io and interrupts */
    reset_cpm();

    /* Last ping before jumping to kernel (timer interrupts are now disabled */                
    wdt_ping();

    /* Jump to kernel */
    (*(void (*)(void *,char *))KERNEL_LOAD_ADDRESS)(bd, cmd);         

    /* Never returns */
    while(1);
}

/*
  Boot from flash 
*/

int boot(char *cmd)
{
    int i;

    /* Check for kernel partition */
    if(__bd->pt[0].type == pt_magic) {
        for(i=1;i<pt_size;i++)
            if(__bd->pt[i].type == pt_kernel)
            {
                memcpy((void *)KERNEL_LOAD_ADDRESS, 
                       (void *)(__bd->pt[i].offset+__bd->flash_base),
                       __bd->pt[i].size);

                start_kernel(cmd);
            }
    }

    return -1;
}

/*
  Boot fron net
*/

int net_boot(char *cmd) 
{
    char buffer[200];
    int i,j;

    net_init();

    /* Loop until we get file from server */
    for(j=0;j<NET_BOOT_MAX_REPEAT;j++) {
        if(bootp()) 
            continue;

        i=tftp_get((char *)KERNEL_LOAD_ADDRESS, bootfile, SERVER_IPADDR);

        if(i>0)
            break;
    }
    if(j==NET_BOOT_MAX_REPEAT) {
        printf("Unable to download image. ");
        return -1;
    }

    printf("Booting kernel ... ");

    start_kernel(cmd);
}

/*
  bootswitch() tries to figure out what to do next. Normally it just
  resets current state. If boot_state stored in nv_ram is set to try or
  prepare for flash image testing, it modifies memory/flash size and
  location before starting kernel.
 */

void bootswitch() 
{
    sipo_nv_info        nv;
    unsigned int len, old_state;

    /* Read, save & reset current state */
    nv_read(0, sizeof(nv), (void *)&nv);
    if(nv.magic!=SIPO_MAGIC) {
        printf("NVRAM corrupted or uninitialized ...\n");
        nv.magic=SIPO_MAGIC;
        __bd->boot_state=nv.state=nv_normal;
    } else {
        old_state=nv.state;
        nv.state=nv_normal;
    }
    nv_write(0, sizeof(nv), (void *)&nv);

    switch(old_state) 
    {
    default:
#if 0
        printf("Error: Unknown NVRAM state (%d)... \n", old_state);
        old_state=nv_error;
#else
        printf("Boot state %d\n", old_state);
#endif
        break;

    case nv_normal:
        break;

    case nv_prepare:
        printf("FLASH Upgrade preparation\n");
        __bd->bi_memsize-=__bd->flash_size;
        wdt_countdown=SIPO_CLI_COUNTDOWN;       /* Copying might take long */
        /* Copy only the first blocks (bootloader, partition table, board info
         * because most of the data is overriden anyway..*/
        len = FLASH_BD_BLOCK * FLASH_BLOCKSIZE;
        if (__bd->flash_size < len)
          len = __bd->flash_size;
        memcpy((void *)(__bd->bi_memsize), (void *)(__bd->flash_base), len);
        break;

    case nv_test:
        /* Emulate flash in ram */
        printf("FLASH Upgrade test\n");
        if(!boottest_bd()) 
        {
            __bd->bi_memsize -= __bd->flash_size;
            __bd->flash_base = __bd->bi_memsize;
        } else {
            printf("Board descriptor is not valid. Aborting boot test procedure\n");
            old_state=nv_error;
        }
        break;

    case nv_finalize:
        printf("FLASH Upgrade finalization\n");
        if(!boottest_bd()) 
        {
            __bd->bi_memsize-=__bd->flash_size;
        } else {
            printf("Board descriptor is not valid. Aborting boot test procedure\n");
            old_state=nv_error;
        }
        break;

    case nv_minimal:
        printf("Minimal configuration state\n");
        break;

    case nv_recovery:
        printf("Recovery from startup failure\n");
        break;
    }
    __bd->boot_state=old_state;
}

/* 
   _init() is called from assembler startup code, before SDRAM & CPU stack is correctly initialized.
 */

void _init(void) 
{
    int i;

    set_leds(SIPO_LED_STATUS_RED|SIPO_LED_POWER);

    wdt_init();

    serial_init();

    printf("\n\n\n%s\n%s\n",SIPO_VERSION_STRING,SIPO_COPYRIGHT_STRING);

    rtc_init();

    /* Checkstop should cause reset */
    IMMR->clocks_rmr = 1;

    /* Enable interrupts */
    mtmsr(mfmsr()|MSR_EE);

    set_dec(decrementer_period);

    /* First chance to get in CLI when we are still running from flash */
    cli();

    /* Disable ints while initializing SDRAM */
    mtmsr(mfmsr()&~MSR_EE);

    if(sdram()) {
        printf("SDRAM initialization failure!\n");
    }
    else
        set_leds(SIPO_LED_STATUS_WEIRD|SIPO_LED_POWER);
}

/*
  Everything starts here (after startup initialization).
 */

main() 
{
    init_bd();

    cli();

    bootswitch();

    if(__bd->enable_netboot)
        net_boot(__bd->cmdline);
    else
        boot(__bd->cmdline);

    printf("Halting...\n");
    while(1);
}
