/*******************************************************************************

  MII Station Management interface
  --------------------------------

  Author: Arto Vuori <avuori@ssh.com>

  Copyright (C) 2000 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved. 

  Management frame format is:

   PREAMBLE[32] | ST[2] | OP[2] | PHYAD[5] | REGAD[5] | TA[2] | DATA[16] | idle

 where: 
   PREAMBLE     32 bits of '1'.
   ST           Start of Frame '01'
   OP           Operatio code ('10' for read, '01' for write).
   PHYAD        physical address (5 bits)
   REGAD        register address (5 bits)
   TA           turnaround  (10 for write, z0 for read ( phy writes second bit))
   DATA         16bits of data (for reads phy sends data)   
   idle         idlebits

*******************************************************************************/

#include <string.h>
#include "sipo.h"


#ifdef EVALBOARD
#define MDIO_DATA       0x00400000                      /* Data bit */
#define MDIO_CLK        0x00200000                      /* Clock bit */
#define MDIO_DIR        IMMR->io_regs[PORTC].pdir        /* PIN Direction register */
#define MDIO_IO         IMMR->io_regs[PORTC].pdat        /* PIN Data (Read/Write) register */
#else /* SIPO */
#define MDIO_DATA       0x00400000                      /* Data bit */
#define MDIO_CLK        0x00200000                      /* Clock bit */
#define MDIO_DIR        IMMR->io_regs[PORTC].pdir        /* PIN Direction register */
#define MDIO_IO         IMMR->io_regs[PORTC].pdat        /* PIN Data (Read/Write) register */
#endif

static volatile int foobar;

void mdio_delay(void) {
    for(foobar=0;foobar<1;foobar++);
}


/* SYNC with 32bits of '1' */

void mdio_sync(void) {
    int i;

    MDIO_DIR |= MDIO_DATA;    
    mdio_delay();
    for(i=0;i<32;i++) {
        mdio_delay();
        MDIO_IO = (MDIO_IO & ~MDIO_CLK) | MDIO_DATA;
        mdio_delay();
        MDIO_IO = MDIO_IO | MDIO_CLK | MDIO_DATA;
    }
    mdio_delay();
}



void mdio_write(int phy_id, int location, int value)
{
        int mii_cmd = (0x5002 << 16) | (phy_id << 23) | (location<<18) | value;
        int i;

        mdio_sync();

        /* Shift the command bits out. */
        for (i = 31; i >= 0; i--) {
            int dataval = (mii_cmd & (1 << i)) ? MDIO_DATA : 0;
            MDIO_IO = (MDIO_IO & ~(MDIO_DATA | MDIO_CLK)) | dataval;
            mdio_delay();
            MDIO_IO |= MDIO_CLK;
            mdio_delay();               
        }

        /* Switch iopin direction */
        MDIO_DIR &= ~MDIO_DATA;    

        /* Clear out extra bits. */
        for (i = 2; i > 0; i--) {
            MDIO_IO &= ~(MDIO_DATA | MDIO_CLK);
            mdio_delay();
            MDIO_IO |= MDIO_CLK;
            mdio_delay();       
        }
        MDIO_IO &= ~(MDIO_DATA | MDIO_CLK);

        return;

}

int mdio_read(int phy_id, int location)
{
        int mii_cmd = (0x6000 << 16) | (phy_id << 23) | (location<<18);
        int retval = 0;
        int i;

        mdio_sync();

        /* Shift the command bits out. */
        for (i = 31; i > 17; i--) {
            int dataval = (mii_cmd & (1 << i)) ? MDIO_DATA : 0;
            MDIO_IO = (MDIO_IO & ~(MDIO_DATA | MDIO_CLK)) | dataval;
            mdio_delay();
            MDIO_IO |= MDIO_CLK;
            mdio_delay();               
        }

        /* Switch iopin direction */
        MDIO_DIR &= ~MDIO_DATA;    

        /* Read the two transition, 16 data, and wire-idle bits. */
        for (i = 19; i > 0; i--) {
            MDIO_IO &= ~(MDIO_CLK | MDIO_DATA); 
            mdio_delay();
            retval = (retval << 1) | ((MDIO_IO & MDIO_DATA) ? 1 : 0);
            MDIO_IO |= MDIO_CLK;
            mdio_delay();
        }
        MDIO_IO &= ~(MDIO_DATA | MDIO_CLK);

        return (retval>>1) & 0xffff;
}

void mdio_puke(int phyid) {
    int i,j;

    printf("MII Station management registers 0-7:\n");
    for(i=0;i<8;i++)
    {   
      j=mdio_read(phyid,i);
      printf("%04x ", j);
    }
    printf("\n");
}

void mdio_init(void) {
    int i,j;

#ifndef EVALBOARD
    IMMR->io_regs[PORTC].ppar &= ~(MDIO_DATA | MDIO_CLK);
    IMMR->io_regs[PORTC].podr &= ~(MDIO_DATA | MDIO_CLK);
    IMMR->io_regs[PORTC].pdir |=  (MDIO_DATA | MDIO_CLK);
    IMMR->io_regs[PORTC].pdat &= ~(MDIO_DATA | MDIO_CLK);
#else
    IMMR->io_regs[PORTC].ppar &= ~(MDIO_DATA | MDIO_CLK);
    IMMR->io_regs[PORTC].podr &= ~(MDIO_DATA | MDIO_CLK);
    IMMR->io_regs[PORTC].pdir |=  (MDIO_DATA | MDIO_CLK);
    IMMR->io_regs[PORTC].pdat &= ~(MDIO_DATA | MDIO_CLK);
#endif
}
    
