/*--------------------------------------------------------------------------    
*
* File:  MPC8260.H
*
*
* Description:  
*
* Main Internal Memory Maps for the MPC8260 PowerQUICC 2.
*
* Note that different structures can be overlaid at the same offsets for the 
* different modes of operation.
*
*
* History:
*
* 13 JUL 98    ggh    initial version
*
*----------------------------------------------------------------------------*/

#ifndef _MPC8260_H 
#define _MPC8260_H


#ifndef PACKED
#define PACKED __attribute__((packed))
#endif

/******************************************************************************
*
*  PARAMETER RAM (PRAM) FOR EACH PERIPHERAL
*  
*  Each subsection contains protocol-specific PRAM for each peripheral,
*  followed by the PRAM common to all protocols for that peripheral.  These 
*  structs are used as needed in the main MPC8260 memory map structure.  Note 
*  that different modes of operation will require the use of different PRAM 
*  structs, and that certain structs may overlay and conflict with the use of 
*  other PRAM areas.  Consult the MPC8260 VUser Manual for details as to what
*  is unavailable when certain protocols are run on certain peripherals.
*
******************************************************************************/
                                  

/*---------------------------------------------------------------------------*/
/*                   SERIAL COMMUNICATION CONTROLLER (SCC)                   */
/*---------------------------------------------------------------------------*/

/*----------*/
/* SCC HDLC */
/*----------*/

typedef PACKED struct 
{
    VUBYTE  reserved1[4]; /* Reserved area */
    VUWORD  c_mask;       /* CRC constant */
    VUWORD  c_pres;       /* CRC preset */
    VUHWORD disfc;        /* discarded frame counter */
    VUHWORD crcec;        /* CRC error counter */
    VUHWORD abtsc;        /* abort sequence counter */
    VUHWORD nmarc;        /* nonmatching address rx cnt */
    VUHWORD retrc;        /* frame transmission counter. */
                          /* For FCC this area is reserved.*/
    VUHWORD mflr;         /* maximum frame length reg */
    VUHWORD max_cnt;      /* maximum length counter */
    VUHWORD rfthr;        /* received frames threshold */
    VUHWORD rfcnt;        /* received frames count */
    VUHWORD hmask;        /* user defined frm addr mask */
    VUHWORD haddr1;       /* user defined frm address 1 */
    VUHWORD haddr2;       /* user defined frm address 2 */
    VUHWORD haddr3;       /* user defined frm address 3 */
    VUHWORD haddr4;       /* user defined frm address 4 */
    VUHWORD tmp;          /* temp */
    VUHWORD tmp_mb;       /* temp */
} t_HdlcScc_Pram;

 
/*--------------*/
/* SCC Ethernet */
/*--------------*/

typedef PACKED struct 
{
    VUWORD  c_pres;      /* CRC preset */
    VUWORD  c_mask;      /* CRC constant mask*/
    VUWORD  crcec;       /* CRC error counter */
    VUWORD  alec;        /* alignment error counter */
    VUWORD  disfc;       /* discarded frame counter */
    VUHWORD pads;        /* Short frame pad character. */
    VUHWORD ret_lim;     /* Retry limit threshold. */
    VUHWORD ret_cnt;     /* Retry limit counter. */
    VUHWORD mflr;        /* maximum frame length reg */
    VUHWORD minflr;      /* minimum frame length reg */
    VUHWORD maxd1;       /* max DMA1 length register. */
    VUHWORD maxd2;       /* max DMA2 length register. */
    VUHWORD maxd;        /* Rx max DMA. */
    VUHWORD dma_cnt;     /* Rx DMA counter. */
    VUHWORD max_b;       /* max buffer descriptor byte count. */
    VUHWORD gaddr1;      /* group address filter */
    VUHWORD gaddr2;      /* group address filter */
    VUHWORD gaddr3;      /* group address filter */
    VUHWORD gaddr4;      /* group address filter */
    VUWORD  tbuf0_data0; /* Saved area 0, current frame. */
    VUWORD  tbuf0_data1; /* Saved area 1, current frame. */
    VUWORD  tbuf0_rba0;
    VUWORD  tbuf0_crc;
    VUHWORD tbuf0_bcnt;
    VUHWORD paddr1_h;    /* physical address (MSB) */
    VUHWORD paddr1_m;    /* physical address */
    VUHWORD paddr1_l;    /* physical address (LSB) */
    VUHWORD p_per;       /* persistence */
    VUHWORD rfbd_ptr;    /* Rx first BD pointer. */
    VUHWORD tfbd_ptr;    /* Tx first BD pointer. */
    VUHWORD tlbd_ptr;    /* Tx last BD pointer. */
    VUWORD  tbuf1_data0; /* Saved area 0, next frame. */
    VUWORD  tbuf1_data1; /* Saved area 1, next frame. */
    VUWORD  tbuf1_rba0;
    VUWORD  tbuf1_crc;
    VUHWORD tbuf1_bcnt;
    VUHWORD tx_len;      /* tx frame length counter */
    VUHWORD iaddr1;      /* individual address filter. */
    VUHWORD iaddr2;      /* individual address filter.  */
    VUHWORD iaddr3;      /* individual address filter. */
    VUHWORD iaddr4;      /* individual address filter.  */
    VUHWORD boff_cnt;    /* back-off counter */
    VUHWORD taddr_h;     /* temp address (MSB) */
    VUHWORD taddr_m;     /* temp address */
    VUHWORD taddr_l;     /* temp address (LSB) */
}  t_EnetScc_Pram;

/*----------*/
/* SCC VUART */
/*----------*/

typedef PACKED struct 
{
    VUBYTE  reserved1[8];   /* Reserved area */
    VUHWORD max_idl;        /* maximum idle characters */
    VUHWORD idlc;           /* rx idle counter (internal) */
    VUHWORD brkcr;          /* break count register */
    VUHWORD parec;          /* Rx parity error counter */
    VUHWORD frmec;          /* Rx framing error counter */
    VUHWORD nosec;          /* Rx noise counter */
    VUHWORD brkec;          /* Rx break character counter */
    VUHWORD brkln;          /* Receive break length */
    VUHWORD uaddr1;         /* address character 1 */
    VUHWORD uaddr2;         /* address character 2 */
    VUHWORD rtemp;          /* temp storage */
    VUHWORD toseq;          /* Tx out of sequence char */
    VUHWORD cc[8];          /* Rx control characters */
    VUHWORD rccm;           /* Rx control char mask */
    VUHWORD rccr;           /* Rx control char register */
    VUHWORD rlbc;           /* Receive last break char */
}  t_UartScc_Pram;


/*-----------------*/
/* SCC Transparent */
/*-----------------*/

typedef PACKED struct  
{
    VUWORD  c_mask;         /* CRC constant */
    VUWORD  c_pres;         /* CRC preset */
}  t_TransScc_Pram;


/*------------*/
/* SCC Bisync */
/*------------*/

typedef PACKED struct  
{
    VUBYTE  reserved1[4];   /* Reserved area */
    VUWORD  crcc;           /* CRC Constant Temp Value */
    VUHWORD prcrc;          /* Preset Receiver CRC-16/LRC */
    VUHWORD ptcrc;          /* Preset Transmitter CRC-16/LRC */
    VUHWORD parec;          /* Receive Parity Error Counter */
    VUHWORD bsync;          /* BISYNC SYNC Character */
    VUHWORD bdle;           /* BISYNC DLE Character */
    VUHWORD cc[8];          /* Rx control characters */
    VUHWORD rccm;           /* Receive Control Character Mask */
}  t_BisyncScc_Pram;


/*-----------------*/
/* SCC Common PRAM */
/*-----------------*/

typedef PACKED struct 
{
    VUHWORD rbase;      /* RX BD base address */
    VUHWORD tbase;      /* TX BD base address */
    VUBYTE  rfcr;       /* Rx function code */
    VUBYTE  tfcr;       /* Tx function code */
    VUHWORD mrblr;      /* Rx buffer length */
    VUWORD  rstate;     /* Rx internal state */
    VUWORD  rptr;       /* Rx internal data pointer */
    VUHWORD rbptr;      /* rb BD Pointer */
    VUHWORD rcount;     /* Rx internal byte count */
    VUWORD  rtemp;      /* Rx temp */
    VUWORD  tstate;     /* Tx internal state */
    VUWORD  tptr;       /* Tx internal data pointer */
    VUHWORD tbptr;      /* Tx BD pointer */
    VUHWORD tcount;     /* Tx byte count */
    VUWORD  ttemp;      /* Tx temp */
    VUWORD  rcrc;       /* temp receive CRC */
    VUWORD  tcrc;       /* temp transmit CRC */
    union 
    {
        t_HdlcScc_Pram    h;
        t_EnetScc_Pram    e;
        t_UartScc_Pram    u;
        t_TransScc_Pram   t;
        t_BisyncScc_Pram  b;
    } SpecificProtocol;
    VUBYTE COMPLETE_SIZE_OF_DPRAM_PAGE[0x5c];
}  t_Scc_Pram;



/*---------------------------------------------------------------------------*/
/*                      FAST COMMUNICATION CONTROLLER (FCC)                  */
/*---------------------------------------------------------------------------*/

/*----------*/
/* FCC HDLC */
/*----------*/

typedef PACKED struct 
{
    VUBYTE  reserved1[8];   /* Reserved area */
    VUWORD  c_mask;         /* CRC constant */
    VUWORD  c_pres;         /* CRC preset */
    VUHWORD disfc;          /* discarded frame counter */
    VUHWORD crcec;          /* CRC error counter */
    VUHWORD abtsc;          /* abort sequence counter */
    VUHWORD nmarc;          /* nonmatching address rx cnt */
    VUWORD  max_cnt;        /* maximum length counter */
    VUHWORD mflr;           /* maximum frame length reg */
    VUHWORD rfthr;          /* received frames threshold */
    VUHWORD rfcnt;          /* received frames count */
    VUHWORD hmask;          /* user defined frm addr mask */
    VUHWORD haddr1;         /* user defined frm address 1 */
    VUHWORD haddr2;         /* user defined frm address 2 */
    VUHWORD haddr3;         /* user defined frm address 3 */
    VUHWORD haddr4;         /* user defined frm address 4 */
    VUHWORD tmp;            /* temp */
    VUHWORD tmp_mb;         /* temp */
}  t_HdlcFcc_Pram;


/*--------------*/
/* FCC Ethernet */
/*--------------*/

typedef PACKED struct 
{
    VUWORD  stat_bus;       /* Internal use buffer. */
    VUWORD  cam_ptr;        /* CAM address. */
    VUWORD  c_mask;         /* CRC constant mask*/
    VUWORD  c_pres;         /* CRC preset */
    VUWORD  crcec;          /* CRC error counter */
    VUWORD  alec;           /* alignment error counter */
    VUWORD  disfc;          /* discarded frame counter */
    VUHWORD ret_lim;        /* Retry limit threshold. */
    VUHWORD ret_cnt;        /* Retry limit counter. */
    VUHWORD p_per;          /* persistence */
    VUHWORD boff_cnt;       /* back-off counter */
    VUWORD  gaddr_h;        /* group address filter, high */
    VUWORD  gaddr_l;        /* group address filter, low */
    VUHWORD tfcstat;        /* out of sequece Tx BD staus. */
    VUHWORD tfclen;         /* out of sequece Tx BD length. */
    VUWORD  tfcptr;         /* out of sequece Tx BD data pointer. */
    VUHWORD mflr;           /* maximum frame length reg */
    VUHWORD paddr1_h;       /* physical address (MSB) */
    VUHWORD paddr1_m;       /* physical address */
    VUHWORD paddr1_l;       /* physical address (LSB) */
    VUHWORD ibd_cnt;        /* internal BD counter. */
    VUHWORD ibd_start;      /* internal BD start pointer. */
    VUHWORD ibd_end;        /* internal BD end pointer. */
    VUHWORD tx_len;         /* tx frame length counter */
    VUBYTE  ibd_base[0x20]; /* internal micro code usage. */
    VUWORD  iaddr_h;        /* individual address filter, high */
    VUWORD  iaddr_l;        /* individual address filter, low */
    VUHWORD minflr;         /* minimum frame length reg */
    VUHWORD taddr_h;        /* temp address (MSB) */
    VUHWORD taddr_m;        /* temp address */
    VUHWORD taddr_l;        /* temp address (LSB) */
    VUHWORD pad_ptr;        /* pad_ptr. */
    VUHWORD cf_type;        /* RESERVED (flow control frame type coding) */
    VUHWORD cf_range;       /* flow control frame range. */
    VUHWORD max_b;          /* max buffer descriptor byte count. */
    VUHWORD maxd1;          /* max DMA1 length register. */
    VUHWORD maxd2;          /* max DMA2 length register. */
    VUHWORD maxd;           /* Rx max DMA. */
    VUHWORD dma_cnt;        /* Rx DMA counter. */
    
    /* counter: */
    VUWORD  octc;           /* received octets counter. */
    VUWORD  colc;           /* estimated number of collisions */
    VUWORD  broc;           /* received good packets of broadcast address */
    VUWORD  mulc;           /* received good packets of multicast address */
    VUWORD  uspc;           /* received packets shorter then 64 octets. */
    VUWORD  frgc;           /* as uspc + bad packets */
    VUWORD  ospc;           /* received packets longer then 1518 octets. */
    VUWORD  jbrc;           /* as ospc + bad packets  */
    VUWORD  p64c;           /* received packets of 64 octets. */
    VUWORD  p65c;           /* received packets of 65-128 octets. */
    VUWORD  p128c;          /* received packets of 128-255 octets. */
    VUWORD  p256c;          /* received packets of 256-511 octets. */
    VUWORD  p512c;          /* received packets of 512-1023 octets. */
    VUWORD  p1024c;         /* received packets of 1024-1518 octets. */
    VUWORD  cam_buf;        /* cam respond internal buffer. */
    VUHWORD rfthr;          /* received frames threshold */
    VUHWORD rfcnt;          /* received frames count */
}  t_EnetFcc_Pram;


/*-----------------*/
/* FCC Common PRAM */
/*-----------------*/

typedef PACKED struct 
{
    VUHWORD riptr;          /* Rx internal temporary data pointer. */
    VUHWORD tiptr;          /* Tx internal temporary data pointer. */
    VUHWORD reserved0;      /* Reserved */
    VUHWORD mrblr;          /* Rx buffer length */
    VUWORD  rstate;         /* Rx internal state */
    VUWORD  rbase;          /* RX BD base address */
    VUHWORD rbdstat;        /* Rx BD status and control */
    VUHWORD rbdlen;         /* Rx BD data length */
    VUWORD  rdptr;          /* rx BD data pointer */
    VUWORD  tstate;         /* Tx internal state */
    VUWORD  tbase;          /* TX BD base address */
    VUHWORD tbdstat;        /* Tx BD status and control */
    VUHWORD tbdlen;         /* Tx BD data length */
    VUWORD  tdptr;          /* Tx  data pointer */
    VUWORD  rbptr;          /* rx BD pointer */
    VUWORD  tbptr;          /* Tx BD pointer */
    VUWORD  rcrc;           /* Temp receive CRC */
    VUWORD  reserved_1[0x1];
    VUWORD  tcrc;           /* Temp transmit CRC */
    union                   /* Protocol-Specific parameter ram */
    {
        t_HdlcFcc_Pram    h;
        t_EnetFcc_Pram    e;
    } SpecificProtocol;      
}  t_Fcc_Pram;



/*---------------------------------------------------------------------------*/
/*                  MULTICHANNEL COMMUNICATION CONTROLLER (MCC)              */
/*---------------------------------------------------------------------------*/

/******************************************************************************
* Note that each MCC uses multiple logical channels.  We first define the     *
* PRAM for a logical channel (which can be used in either HDLC or Transparent *
* mode;  wherever there are differences, it is specified), followed by the    *
* PRAM for an MCC itself.                                                     *
******************************************************************************/

/*---------------------*/
/* MCC Logical Channel */
/*---------------------*/

typedef PACKED struct
{
    VUWORD  tstate;         /* Tx internal state. */ 
    VUWORD  zistate;        /* Zero insertion machine state. */
    VUWORD  zidata0;        /* Zero insertion high VUHWORD. */
    VUWORD  zidata1;        /* Zero insertion low VUHWORD. */
    VUHWORD tbdflags;       /* Tx internal BD flags. */
    VUHWORD tbdcnt;         /* Tx internal byte count . */
    VUWORD  tbdptr;         /* Tx internal data pointer. */
    VUHWORD intmask;        /* Interrupt mask flags. */
    VUHWORD chamr;          /* channel mode register. */
    VUWORD  tcrc;           /* Transparent: reserved. */
                            /* Hdlc: Temp receive CRC.*/
    VUWORD  rstate;         /* Rx internal state. */ 
    VUWORD  zdstate;        /* Zero deletion machine state. */
    VUWORD  zddata0;        /* Zero deletion  high VUHWORD. */
    VUWORD  zddata1;        /* Zero deletion  low VUHWORD. */
    VUHWORD rbdflags;       /* Rx internal BD flags. */
    VUHWORD rbdcnt;         /* Rx internal byte count . */
    VUWORD  rbdptr;         /* Rx internal data pointer. */
    VUHWORD maxrlen;        /* Transparent: Max receive buffer length. */
                            /* Hdlc: Max receive frame length. */
    VUHWORD sync_maxcnt;    /* Transparent: Receive synchronization pattern*/
                            /* Hdlc: Max length counter. */
    VUWORD  rcrc;           /* Transparent: reserved. */
                            /* Hdlc: Temp receive CRC.*/
}  t_Mch_Pram;


/*----------*/
/* MCC PRAM */
/*----------*/

typedef PACKED struct
{
    VUWORD  mccbase;        /* A pointer to starting address of BD rings. */
    VUHWORD mccstate;       /* Controller state. */
    VUHWORD mrblr;          /* Maximum receive buffer length. */
    VUHWORD grfthr;         /* Global receive frame threshold. */
    VUHWORD grfcnt;         /* Global receive frame counter. */
    VUWORD  rinttmp;        /* Temp location for rx interrupt table entry. */                            
    VUWORD  data0;          /* Temporary location for holding data. */
    VUWORD  data1;          /* Temporary location for holding data. */
    VUWORD  tintbase;       /* Transmit interrupt table base address. */
    VUWORD  tintptr;        /* Transmit interrupt table pointer. */ 
    VUWORD  tinttmp;        /* Temp location for rx interrupt table entry. */
    VUHWORD sctpbase;       /* Pointer to the super channel transmit table*/
    VUBYTE  res0[0x2];      /* Reserved area */
    VUWORD  c_mask32;       /* CRC constant. */
    VUHWORD xtrabase;       /* Pointer to the beginning of extra parameters */                            
    VUHWORD c_mask16;       /* CRC constant. */
    VUWORD  rinttmp0;       /* Temp location for rx interrupt table entry. */                            
    VUWORD  rinttmp1;       /* Temp location for rx interrupt table entry. */
    VUWORD  rinttmp2;       /* Temp location for rx interrupt table entry. */
    VUWORD  rinttmp3;       /* Temp location for rx interrupt table entry. */
    VUWORD  rintbase0;      /* Rx interrupt table base address. */
    VUWORD  rintptr0;       /* Rx interrupt table pointer. */
    VUWORD  rintbase1;      /* Rx interrupt table base address. */
    VUWORD  rintptr1;       /* Rx interrupt table pointer. */
    VUWORD  rintbase2;      /* Rx interrupt table base address. */
    VUWORD  rintptr2;       /* Rx interrupt table pointer. */
    VUWORD  rintbase3;      /* Rx interrupt table base address. */
    VUWORD  rintptr3;       /* Rx interrupt table pointer. */
    VUBYTE  pad[0xa0];
}  t_Mcc_Pram;



/*---------------------------------------------------------------------------*/
/*                              ATM PARAMETER RAM                            */
/*---------------------------------------------------------------------------*/


/*--------------------------------------*/
/* Address Compression parameters table */
/*--------------------------------------*/

PACKED struct AddressCompressionPram 
{
    VUWORD  VptBase;        /* VP-level addressing table base address */
    VUWORD  VctBase;        /* VC-level addressing table base address */
    VUWORD  Vpt1Base;       /* VP1-level addressing table base address */ 
    VUWORD  Vct1Base;       /* VC1-level addressing table base address */
    VUHWORD VpMask;         /* VP mask for address compression look-up */
} ;


/*-------------------------------*/
/* External CAM parameters table */
/*-------------------------------*/

PACKED struct ExtCamPram 
{
    VUWORD  ExtCamBase;     /* Base address of the external CAM */
    VUBYTE  reserved00[4];  /* Reserved */
    VUWORD  ExtCam1Base;    /* Base address of the external CAM1 */
    VUBYTE  reserved01[6];  /* Reserved */
} ;


/*---------------------------*/
/* ATM mode parameters table */
/*---------------------------*/

typedef PACKED struct AtmPram 
{
    VUBYTE  reserved0[64];  /* Reserved */
    VUHWORD RxCellTmpBase;  /* Rx cell temporary base address */
    VUHWORD TxCellTmpBase;  /* Tx cell temporary base address */
    VUHWORD VUdcTmpBase;    /* VUDC temp base address (in VUDC mode only) */
    VUHWORD IntRctBase;     /* Internal RTC base address */
    VUHWORD IntTctBase;     /* Internal TCT base address */
    VUHWORD IntTcteBase;    /* Internal ACT base address */
    VUBYTE  reserved1[4];   /* reserved four bytes */
    VUWORD  ExtRctBase;     /* Extrnal RTC base address */
    VUWORD  ExtTctBase;     /* Extrnal TCT base address */
    VUWORD  ExtTcteBase;    /* Extrnal ACT base address */
    VUHWORD VUeadOffset;    /* The offset in half-wordunits of the VUEAD
                               entry in the VUDC extra header. Should be
                               even address. If little-endian format is
                               used, the VUeadOffset is of the little-endian
                               format. */
    VUBYTE  reserved2[2];   /* Reserved */
    VUHWORD PmtBase;        /* Performance monitoring table base address */
    VUHWORD ApcParamBase;   /* APC Parameters table base address */
    VUHWORD FbpParamBase;   /* Free buffer pool parameters base address */
    VUHWORD IntQParamBase;  /* Interrupt queue parameters table base */
    VUBYTE  reserved3[2];
    VUHWORD VUniStatTableBase; /* VUNI statistics table base */
    VUWORD  BdBaseExt;      /* BD ring base address extension */
    union 
    {
        struct AddressCompressionPram   AddrCompression;
        struct ExtCamPram               ExtCam;
    } AddrMapping;          /* Address look-up mechanism */
    VUHWORD VciFiltering;   /* VCI filtering enable bits. If bit i is set,
                               the cell with VCI=i will be sent to the
                               raw cell queue. The bits 0-2 and 5 should
                               be zero. */
    VUHWORD Gmode;          /* Global mode */
    VUHWORD CommInfo1;      /* The information field associated with the */
    VUWORD  CommInfo2;      /* last host command */
    VUBYTE  reserved4[4];   /* Reserved */
    VUWORD  CRC32Preset;    /* Preset for CRC32 */
    VUWORD  CRC32Mask;      /* Constant mask for CRC32 */
    VUHWORD AAL1SnpTableBase; /* AAl1 SNP protection look-up table base */
    VUHWORD reserved5;      /* Reserved */
    VUWORD  SrtsBase;       /* External SRTS logic base address. For AAL1
                               only. Should be 16 bytes aligned */
    VUHWORD IdleBase;       /* Idle cell base address */
    VUHWORD IdleSize;       /* Idle cell size: 52, 56, 60, 64 */
    VUWORD  EmptyCellPayload; /* Empty cell payload (little-indian) */
    
    /* ABR specific only */
    VUWORD  Trm; /* VUpper bound on time between F-RM cells for active source */                                 
    VUHWORD Nrm; /* Controls the maximum data cells sent for each F-RM cell. */                           
    VUHWORD Mrm; /* Controls bandwidth between F-RM, B-RM and user data cell */
    VUHWORD Tcr;            /* Tag cell rate */
    VUHWORD AbrRxTcte;      /* ABR reserved area address (2-UHWORD aligned)*/
    VUBYTE  reserved7[76];  /* Reserved */
}  t_Atm_Pram;



/*---------------------------------------------------------------------------*/
/*                      SERIAL MANAGEMENT CHANNEL  (SMC)                     */
/*---------------------------------------------------------------------------*/

typedef PACKED struct  
{
    VUHWORD rbase;          /* Rx BD Base Address */
    VUHWORD tbase;          /* Tx BD Base Address */
    VUBYTE  rfcr;           /* Rx function code */
    VUBYTE  tfcr;           /* Tx function code */
    VUHWORD mrblr;          /* Rx buffer length */
    VUWORD  rstate;         /* Rx internal state */
    VUWORD  rptr;           /* Rx internal data pointer */
    VUHWORD rbptr;          /* rb BD Pointer */
    VUHWORD rcount;         /* Rx internal byte count */
    VUWORD  rtemp;          /* Rx temp */
    VUWORD  tstate;         /* Tx internal state */
    VUWORD  tptr;           /* Tx internal data pointer */
    VUHWORD tbptr;          /* Tx BD pointer */
    VUHWORD tcount;         /* Tx byte count */
    VUWORD  ttemp;          /* Tx temp */
    
    /* SMC VUART-specific PRAM */
    VUHWORD max_idl;        /* Maximum IDLE Characters */
    VUHWORD idlc;           /* Temporary IDLE Counter */
    VUHWORD brkln;          /* Last Rx Break Length */
    VUHWORD brkec;          /* Rx Break Condition Counter */
    VUHWORD brkcr;          /* Break Count Register (Tx) */
    VUHWORD r_mask;         /* Temporary bit mask */  
}  t_Smc_Pram;



/*---------------------------------------------------------------------------*/
/*                            IDMA PARAMETER RAM                             */
/*---------------------------------------------------------------------------*/

typedef PACKED struct  
{
    VUHWORD ibase;          /* IDMA BD Base Address */
    VUHWORD dcm;            /* DMA channel mode register */
    VUHWORD ibdptr;         /* next bd ptr */
    VUHWORD DPR_buf;        /* ptr to internal 64 byte buffer */
    VUHWORD BUF_inv;        /* The quantity of data in DPR_buf */
    VUHWORD SS_max;         /* Steady State Max. transfer size */
    VUHWORD DPR_in_ptr;     /* write ptr for the internal buffer */
    VUHWORD sts;            /* Source Transfer Size */
    VUHWORD DPR_out_ptr;    /* read ptr for the internal buffer */
    VUHWORD seob;           /* Source end of burst */
    VUHWORD deob;           /* Destination end of burst */
    VUHWORD dts;            /* Destination Transfer Size */
    VUHWORD RetAdd;         /* return address when ERM==1 */
    VUHWORD Reserved;       /* reserved */
    VUWORD  BD_cnt;         /* Internal byte count */
    VUWORD  S_ptr;          /* source internal data ptr */
    VUWORD  D_ptr;          /* destination internal data ptr */
    VUWORD  istate;         /* Internal state */
}  t_Idma_Pram;



/*-------------------------------------------------------------------*/
/*                    INTER-INTEGRATED CIRCUIT  (I2C)                */
/*-------------------------------------------------------------------*/

typedef PACKED struct 
{
    VUHWORD rbase;          /* RX BD base address */
    VUHWORD tbase;          /* TX BD base address */
    VUBYTE  rfcr;           /* Rx function code */
    VUBYTE  tfcr;           /* Tx function code */
    VUHWORD mrblr;          /* Rx buffer length */
    VUWORD  rstate;         /* Rx internal state */
    VUWORD  rptr;           /* Rx internal data pointer */
    VUHWORD rbptr;          /* rb BD Pointer */
    VUHWORD rcount;         /* Rx internal byte count */
    VUWORD  rtemp;          /* Rx temp */
    VUWORD  tstate;         /* Tx internal state */
    VUWORD  tptr;           /* Tx internal data pointer */
    VUHWORD tbptr;          /* Tx BD pointer */
    VUHWORD tcount;         /* Tx byte count */
    VUWORD  ttemp;          /* Tx temp */
}  t_I2c_Pram;



/*---------------------------------------------------------------------------*/
/*                       SERIAL PERIPHERAL INTERFACE  (SPI)                  */
/*---------------------------------------------------------------------------*/

typedef PACKED struct  
{
    VUHWORD rbase;          /* Rx BD Base Address */
    VUHWORD tbase;          /* Tx BD Base Address */
    VUBYTE  rfcr;           /* Rx function code */
    VUBYTE  tfcr;           /* Tx function code */
    VUHWORD mrblr;          /* Rx buffer length */
    VUWORD  rstate;         /* Rx internal state */
    VUWORD  rptr;           /* Rx internal data pointer */
    VUHWORD rbptr;          /* Rx BD Pointer */
    VUHWORD rcount;         /* Rx internal byte count */
    VUWORD  rtemp;          /* Rx temp */
    VUWORD  tstate;         /* Tx internal state */
    VUWORD  tptr;           /* Tx internal data pointer */
    VUHWORD tbptr;          /* Tx BD pointer */
    VUHWORD tcount;         /* Tx byte count */
    VUWORD  ttemp;          /* Tx temp */
    VUBYTE  reserved[8];
}  t_Spi_Pram;



/*---------------------------------------------------------------------------*/
/*                      RISC TIMER PARAMETER RAM                             */
/*---------------------------------------------------------------------------*/

typedef PACKED struct  
{
 
    VUHWORD tm_base;        /* RISC timer table base adr */
    VUHWORD tm_ptr;         /* RISC timer table pointer */
    VUHWORD r_tmr;          /* RISC timer mode register */
    VUHWORD r_tmv;          /* RISC timer valid register */
    VUWORD  tm_cmd;         /* RISC timer cmd register */
    VUWORD  tm_cnt;         /* RISC timer internal cnt */
}  t_timer_pram;



/*--------------------------------------------------------------------------*/
/*                  ROM MICROCODE PARAMETER RAM AREA                        */
/*--------------------------------------------------------------------------*/

typedef PACKED struct  
{
    VUHWORD rev_num;        /* VUcode Revision Number */
    VUHWORD d_ptr;          /* MISC Dump area pointer */
}  t_ucode_pram;








/*--------------------------------------------------------------------------*/
/*                MAIN DEFINITION OF MPC8260 INTERNAL MEMORY MAP            */
/*--------------------------------------------------------------------------*/

typedef PACKED struct 
{

/* cpm_ram */
#if 0
    t_Mch_Pram  mch_pram[256];      /* MCC logical channels parameter ram */
#else
    VUBYTE      dpram1[0x4000];
#endif
    VUBYTE      reserved0[0x4000];  /* Reserved area */
    
/* DPR_BASE+0x8000*/

/*---------------------------------------------------------------------*/
/* A note about the pram union:                                        */
/* The pram area has been broken out three ways for clean access into  */
/* certain peripherals' spaces.  This arrangement allows programmers   */
/* flexibility of usage in terms of being able to change which         */
/* peripheral is being accessed by simply changing an array value.     */
/* Given the interweaving of certain peripherals' pram areas, this     */
/* would not be possible with only one large pram structure.           */
/*                                                                     */
/* SERIALS  - For accessing SCC, non-ATM FCC, and MCC pram             */
/* ATM      - For accessing ATM FCC pram                               */
/* STANDARD - For accessing timers, revnum, d_ptr, RAND, and the pram  */
/*            base pointers of the SMCs, IDMAs, SPI, and I2C           */
/*---------------------------------------------------------------------*/
           
    union 
    {
    
    /*for access to the PRAM structs for SCCs, FCCs, and MCCs */ 
        struct serials 
        {
            t_Scc_Pram scc_pram[4];
            t_Fcc_Pram fcc_pram[3];
            t_Mcc_Pram mcc_pram[2];
            VUBYTE reserved1[0x700];
        } serials;
#if 0 /* size too large. Maybe we need to add PACKED specifier somewhere...  */       
    /* for access to ATM PRAM structs */
        struct atm
        {
            VUBYTE reserved2[0x400];
            t_Atm_Pram atm_pram[2];
            VUBYTE reserved3[0xa00];
        } atm;
#endif        
    /* for access to the memory locations holding user-defined 
       base addresses of PRAM for SMCs, IDMA, SPI, and I2C. */     
       struct standard
       {
            VUBYTE scc1[0x100];
            VUBYTE scc2[0x100];
            VUBYTE scc3[0x100];
            VUBYTE scc4[0x100];
            VUBYTE fcc1[0x100];
            VUBYTE fcc2[0x100];
            VUBYTE fcc3[0x100];
            VUBYTE mcc1[0x80];
            VUBYTE reserved_0[0x7c];
            VUBYTE smc1[0x2];
            VUBYTE idma1[0x2];
            VUBYTE mcc2[0x80];
            VUBYTE reserved_1[0x7c];
            VUBYTE smc2[0x2];
            VUBYTE idma2[0x2];
            VUBYTE reserved_2[0xfc];
            VUBYTE spi[0x2];
            VUBYTE idma3[0x2];
            VUBYTE reserved_3[0xe0];
            VUBYTE timers[0x10];
            VUBYTE Rev_num[0x2];
            VUBYTE D_ptr[0x2];
            VUBYTE reserved_4[0x4];
            VUBYTE rand[0x4];
            VUBYTE i2c[0x2];
            VUBYTE idma4[0x2];
            VUBYTE reserved_5[0x500];
        } standard;
        
    } pram;
    
    VUBYTE  reserved11[0x2000];      /* Reserved area */
    VUBYTE  cpm_ram_dpram_2[0x1000]; /* Internal RAM */
    VUBYTE  reserved12[0x4000];      /* Reserved area */

/* siu */
    VUWORD  siu_siumcr;         /* SIU Module Configuration Register */
    VUWORD  siu_sypcr;          /* System Protection Control Register */
    VUBYTE  reserved13[0x6];    /* Reserved area */
    VUHWORD siu_swsr;           /* Software Service Register */

/* buses */
    VUBYTE  reserved14[0x14];   /* Reserved area */
    VUWORD  bcr;                /* Bus Configuration Register */
    VUBYTE  ppc_acr;            /* Arbiter Configuration Register */
    VUBYTE  reserved15[0x3];    /* Reserved area */
    VUWORD  ppc_alrh;           /* Arbitration Level Reg. (First clients)*/
    VUWORD  ppc_alrl;           /* Arbitration Level Reg. (Next clients) */
    VUBYTE  lcl_acr;            /* LCL Arbiter Configuration Register */
    VUBYTE  reserved16[0x3];    /* Reserved area */
    VUWORD  lcl_alrh;      /* LCL Arbitration Level Reg. (First clients) */
    VUWORD  lcl_alrl;      /* LCL Arbitration Level Register (Next clients) */
    VUWORD  tescr1;        /* PPC bus transfer error status control reg. 1 */
    VUWORD  tescr2;        /* PPC bus transfer error status control reg. 2 */
    VUWORD  ltescr1;       /* Local bus transfer error status control reg. 1 */
    VUWORD  ltescr2;       /* Local bus transfer error status control reg. 2 */
    VUWORD  pdtea;              /* PPC bus DMA Transfer Error Address */
    VUBYTE  pdtem;              /* PPC bus DMA Transfer Error MSNUM  */
    VUBYTE  reserved17[0x3];    /* Reserved area */
    VUWORD  ldtea;              /* PPC bus DMA Transfer Error Address */
    VUBYTE  ldtem;              /* PPC bus DMA Transfer Error MSNUM  */
    VUBYTE  reserved18[0xa3];   /* Reserved area */

/* memc */
    struct memc_regs 
    {
        VUWORD br;              /* Base Register */
        VUWORD or;              /* Option Register */
    } memc_regs[12];
    VUBYTE  reserved19[0x8];    /* Reserved area */
    VUWORD  memc_mar;           /* Memory Address Register */
    VUBYTE  reserved20[0x4];    /* Reserved area */
    VUWORD  memc_mamr;          /* Machine A Mode Register */
    VUWORD  memc_mbmr;          /* Machine B Mode Register */
    VUWORD  memc_mcmr;          /* Machine C Mode Register */
    VUWORD  memc_mdmr;          /* Machine D Mode Register */
    VUBYTE  reserved21[0x4];    /* Reserved area */
    VUHWORD memc_mptpr;         /* Memory Periodic Timer Prescaler */
    VUBYTE  reserved22[0x2];    /* Reserved area */
    VUWORD  memc_mdr;           /* Memory Data Register */
    VUBYTE  reserved23[0x4];    /* Reserved area */
    VUWORD  memc_psdmr;         /* PowerPC Bus SDRAM machine Mode Register */
    VUWORD  memc_lsdmr;         /* Local Bus SDRAM machine Mode Register */
    VUBYTE  memc_purt;          /* PowerPC Bus assigned VUPM Refresh Timer */
    VUBYTE  reserved24[0x3];    /* Reserved area */
    VUBYTE  memc_psrt;          /* PowerPC Bus assigned SDRAM Refresh Timer */
    VUBYTE  reserved25[0x3];    /* Reserved area */
    VUBYTE  memc_lurt;          /* Local Bus assigned VUPM Refresh Timer */
    VUBYTE  reserved26[0x3];    /* Reserved area */
    VUBYTE  memc_lsrt;          /* Local Bus assigned SDRAM Refresh Timer */
    VUBYTE  reserved27[0x3];    /* Reserved area */
    VUWORD  memc_immr;          /* Internal Memory Map Register */

/* pci */
    VUWORD  pcibr0;             /* Base address+valid for PCI window 1 */
    VUWORD  pcibr1;             /* Base address+valid for PCI window 2 */
    VUBYTE  reserved28[0x10];   /* Reserved area */
    VUWORD  pcimsk0;            /* Mask for PCI window 1 */
    VUWORD  pcimsk1;            /* Mask for PCI window 2 */
    VUBYTE  reserved29[0x54];   /* Reserved area */

/* si_timers */
    VUHWORD si_timers_tmcntsc;  /* Time Counter Status and Control Register */
    VUBYTE  reserved30[0x2];    /* Reserved area */
    VUWORD  si_timers_tmcnt;    /* Time Counter Register */
    VUWORD  si_timers_tmcntsec; /* Time Counter Seconds*/
    VUWORD  si_timers_tmcntal;  /* Time Counter Alarm Register */
    VUBYTE  reserved31[0x10];   /* Reserved area */
    VUHWORD si_timers_piscr;    /* Periodic Interrupt Status and Control Reg. */
    VUBYTE  reserved32[0x2];    /* Reserved area */
    VUWORD  si_timers_pitc;     /* Periodic Interrupt Count Register */
    VUWORD  si_timers_pitr;     /* Periodic Interrupt Timer Register */
    VUBYTE  reserved33[0x54];   /* Reserved area */

/* test module registers */
    VUWORD  tstmhr;             
    VUWORD  tstmlr;
    VUHWORD tster;
    VUBYTE  reserved34[0x156];  /* Reserved area */
    
/* pci, part 2 */
    VUWORD  pci_pci;            /* PCI Configuration space */
    VUBYTE  reserved35[0x7fc];  /* Reserved area */
    
/* ic */
    VUHWORD ic_sicr;            /* Interrupt Configuration Register */
    VUBYTE  reserved36[0x2];    /* Reserved area */
    VUWORD  ic_sivec;           /* CP Interrupt Vector Register */
    VUWORD  ic_sipnr_h;         /* Interrupt Pending Register (HIGH) */
    VUWORD  ic_sipnr_l;         /* Interrupt Pending Register (LOW) */
    VUWORD  ic_siprr;           /* SIU Interrupt Priority Register */
    VUWORD  ic_scprr_h;         /* Interrupt Priority Register (HIGH) */
    VUWORD  ic_scprr_l;         /* Interrupt Priority Register (LOW) */
    VUWORD  ic_simr_h;          /* Interrupt Mask Register (HIGH) */
    VUWORD  ic_simr_l;          /* Interrupt Mask Register (LOW) */
    VUWORD  ic_siexr;           /* External Interrupt Control Register */
    VUBYTE  reserved37[0x58];   /* Reserved area */

/* clocks */
    VUWORD  clocks_sccr;        /* System Clock Control Register */
    VUBYTE  reserved38[0x4];    /* Reserved area */
    VUWORD  clocks_scmr;        /* System Clock Mode Register */
    VUBYTE  reserved39[0x4];    /* Reserved area */
    VUWORD  clocks_rsr;         /* Reset Status Register */
    VUWORD  clocks_rmr;         /* Reset Moode Register  */
    VUBYTE  reserved40[0x68];   /* Reserved area */

/* io_ports */
    struct io_regs 
    {
        VUWORD  pdir;           /* Port A-D Data Direction Register */
        VUWORD  ppar;           /* Port A-D Pin Assignment Register */
        VUWORD  psor;           /* Port A-D Special Operation Register */
        VUWORD  podr;           /* Port A-D Open Drain Register */
        VUWORD  pdat;           /* Port A-D Data Register */
        VUBYTE reserved41[0xc]; /* Reserved area */
    } io_regs[4];

/* cpm_timers */
    VUBYTE  cpm_timers_tgcr1;   /* Timer Global Configuration Register */
    VUBYTE  reserved42[0x3];    /* Reserved area */
    VUBYTE  cpm_timers_tgcr2;   /* Timer Global Configuration Register */
    VUBYTE  reserved43[0xb];    /* Reserved area */
    VUHWORD cpm_timers_tmr1;    /* Timer Mode Register */
    VUHWORD cpm_timers_tmr2;    /* Timer Mode Register */
    VUHWORD cpm_timers_trr1;    /* Timer Reference Register */
    VUHWORD cpm_timers_trr2;    /* Timer Reference Register */
    VUHWORD cpm_timers_tcr1;    /* Timer Capture Register */
    VUHWORD cpm_timers_tcr2;    /* Timer Capture Register */
    VUHWORD cpm_timers_tcn1;    /* Timer Counter */
    VUHWORD cpm_timers_tcn2;    /* Timer Counter */
    VUHWORD cpm_timers_tmr3;    /* Timer Mode Register */
    VUHWORD cpm_timers_tmr4;    /* Timer Mode Register */
    VUHWORD cpm_timers_trr3;    /* Timer Reference Register */
    VUHWORD cpm_timers_trr4;    /* Timer Reference Register */
    VUHWORD cpm_timers_tcr3;    /* Timer Capture Register */
    VUHWORD cpm_timers_tcr4;    /* Timer Capture Register */
    VUHWORD cpm_timers_tcn3;    /* Timer Counter */
    VUHWORD cpm_timers_tcn4;    /* Timer Counter */
    VUHWORD cpm_timers_ter[4];  /* Timer Event Register */
    VUBYTE reserved44[0x260];   /* Reserved area */

/* sdma general */
    VUBYTE  sdma_sdsr;          /* SDMA Status Register */
    VUBYTE  reserved45[0x3];    /* Reserved area */
    VUBYTE  sdma_sdmr;          /* SDMA Mask Register */
    VUBYTE  reserved46[0x3];    /* Reserved area */

/* idma */
    VUBYTE  idma_idsr1;         /* IDMA Status Register */
    VUBYTE  reserved47[0x3];    /* Reserved area */
    VUBYTE  idma_idmr1;         /* IDMA Mask Register */
    VUBYTE  reserved48[0x3];    /* Reserved area */
    VUBYTE  idma_idsr2;         /* IDMA Status Register */
    VUBYTE  reserved49[0x3];    /* Reserved area */
    VUBYTE  idma_idmr2;         /* IDMA Mask Register */
    VUBYTE  reserved50[0x3];    /* Reserved area */
    VUBYTE  idma_idsr3;         /* IDMA Status Register */
    VUBYTE  reserved51[0x3];    /* Reserved area */
    VUBYTE  idma_idmr3;         /* IDMA Mask Register */
    VUBYTE  reserved52[0x3];    /* Reserved area */
    VUBYTE  idma_idsr4;         /* IDMA Status Register */
    VUBYTE  reserved53[0x3];    /* Reserved area */
    VUBYTE  idma_idmr4;         /* IDMA Mask Register */
    VUBYTE  reserved54[0x2c3];  /* Reserved area */
    
/* fcc */
    struct fcc_regs 
    {
        VUWORD  gfmr;           /* FCC General Mode Register */
        VUWORD  psmr;           /* FCC Protocol Specific Mode Register */
        VUHWORD todr;           /* FCC Transmit On Demand Register */
        VUBYTE  reserved55[0x2];/* Reserved area */
        VUHWORD dsr;            /* FCC Data Sync. Register */
        VUBYTE  reserved56[0x2];/* Reserved area */
        VUWORD  fcce;           /* FCC Event Register */
        VUWORD  fccm;           /* FCC Mask Register */
        VUBYTE  fccs;           /* FCC Status Register */
        VUBYTE  reserved57[0x3];/* Reserved area */
        VUWORD  ftprr;          /* FCC Transmit Partial Rate Register */
    } fcc_regs[3];
    VUBYTE reserved58[0x290];   /* Reserved area */
    
/* brgs 5-8 */
    VUWORD  brgs_brgc5;         /* BRG Configuration Register */
    VUWORD  brgs_brgc6;         /* BRG Configuration Register */
    VUWORD  brgs_brgc7;         /* BRG Configuration Register */
    VUWORD  brgs_brgc8;         /* BRG Configuration Register */
    VUBYTE  reserved59[0x260];  /* Reserved area */
    
/* i2c */
    VUBYTE  i2c_i2mod;          /* IC Mode Register */
    VUBYTE  reserved60[0x3];    /* Reserved area */
    VUBYTE  i2c_i2add;          /* IC Address Register */
    VUBYTE  reserved61[0x3];    /* Reserved area */
    VUBYTE  i2c_i2brg;          /* IC BRG Register */
    VUBYTE  reserved62[0x3];    /* Reserved area */
    VUBYTE  i2c_i2com;          /* IC Command Register */
    VUBYTE  reserved63[0x3];    /* Reserved area */
    VUBYTE  i2c_i2cer;          /* IC Event Register */
    VUBYTE  reserved64[0x3];    /* Reserved area */
    VUBYTE  i2c_i2cmr;          /* IC Mask Register */
    VUBYTE  reserved65[0x14b];  /* Reserved area */
    
/* cpm */
    VUWORD  cpm_cpcr;           /* Communication Processor Command Register */
    VUWORD  cpm_rccr;           /* RISC Configuration Register */
    VUWORD  cpm_rmdr;           /* RISC Microcode Dev. Support Control Reg. */
    VUHWORD cpm_rctr1;          /* RISC Controller Trap Register */
    VUHWORD cpm_rctr2;          /* RISC Controller Trap Register */
    VUHWORD cpm_rctr3;          /* RISC Controller Trap Register */
    VUHWORD cpm_rctr4;          /* RISC Controller Trap Register */
    VUBYTE  reserved66[0x2];    /* Reserved area */
    VUHWORD cpm_rter;           /* RISC Timers Event Register */
    VUBYTE  reserved67[0x2];    /* Reserved area */
    VUHWORD cpm_rtmr;           /* RISC Timers Mask Register */
    VUHWORD cpm_rtscr;          /* RISC Time-Stamp Timer Control Register */
    VUHWORD cpm_rmds;           /* RISC Development Support Status Register */
    VUWORD  cpm_rtsr;           /* RISC Time-Stamp Register */
    VUBYTE  reserved68[0xc];    /* Reserved area */

/* brgs 1-4 */
    VUWORD  brgs_brgc1;         /* BRG Configuration Register */
    VUWORD  brgs_brgc2;         /* BRG Configuration Register */
    VUWORD  brgs_brgc3;         /* BRG Configuration Register */
    VUWORD  brgs_brgc4;         /* BRG Configuration Register */
    
/* scc */
    struct scc_regs_8260
    {
        VUWORD  gsmr_l;         /* SCC General Mode Register */
        VUWORD  gsmr_h;         /* SCC General Mode Register */
        VUHWORD psmr;           /* SCC Protocol Specific Mode Register */
        VUBYTE  reserved69[0x2];/* Reserved area */
        VUHWORD todr;           /* SCC Transmit-On-Demand Register */
        VUHWORD dsr;            /* SCC Data Synchronization Register */
        VUHWORD scce;           /* SCC Event Register */
        VUBYTE  reserved70[0x2];/* Reserved area */
        VUHWORD sccm;           /* SCC Mask Register */
        VUBYTE  reserved71;     /* Reserved area */
        VUBYTE  sccs;           /* SCC Status Register */
        VUBYTE  reserved72[0x8];/* Reserved area */
    } scc_regs[4];
    
/* smc */
    struct smc_regs_8260
    {
        VUBYTE  reserved73[0x2];/* Reserved area */
        VUHWORD smcmr;          /* SMC Mode Register */
        VUBYTE  reserved74[0x2];/* Reserved area */
        VUBYTE  smce;           /* SMC Event Register */
        VUBYTE  reserved75[0x3];/* Reserved area */
        VUBYTE  smcm;           /* SMC Mask Register */
        VUBYTE  reserved76[0x5];/* Reserved area */
    } smc_regs[2];
    
/* spi */
    VUHWORD spi_spmode;         /* SPI Mode Register */
    VUBYTE  reserved77[0x4];    /* Reserved area */
    VUBYTE  spi_spie;           /* SPI Event Register */
    VUBYTE  reserved78[0x3];    /* Reserved area */
    VUBYTE  spi_spim;           /* SPI Mask Register */
    VUBYTE  reserved79[0x2];    /* Reserved area */
    VUBYTE  spi_spcom;          /* SPI Command Register */
    VUBYTE  reserved80[0x52];   /* Reserved area */
    
/* cpm_mux */
    VUBYTE  cpm_mux_cmxsi1cr;   /* CPM MUX SI Clock Route Register */
    VUBYTE  reserved81;         /* Reserved area */
    VUBYTE  cpm_mux_cmxsi2cr;   /* CPM MUX SI Clock Route Register */
    VUBYTE  reserved82;         /* Reserved area */
    VUWORD  cpm_mux_cmxfcr;     /* CPM MUX FCC Clock Route Register */
    VUWORD  cpm_mux_cmxscr;     /* CPM MUX SCC Clock Route Register */
    VUBYTE  cpm_mux_cmxsmr;     /* CPM MUX SMC Clock Route Register */
    VUBYTE  reserved83;         /* Reserved area */
    VUHWORD cpm_mux_cmxuar;     /* CPM MUX VUTOPIA Address Register */
    VUBYTE  reserved84[0x10];   /* Reserved area */
    
/* si */
    struct si_regs 
    {
        VUHWORD sixmr[4];       /* SI TDM Mode Registers A-D */
        VUBYTE  sigmr;          /* SI Global Mode Register */
        VUBYTE  reserved85;     /* Reserved area */
        VUBYTE  sicmdr;         /* SI Command Register */
        VUBYTE  reserved86;     /* Reserved area */
        VUBYTE  sistr;          /* SI Status Register */
        VUBYTE  reserved87;     /* Reserved area */
        VUHWORD sirsr;          /* SI RAM Shadow Address Register */
        VUHWORD mcce;           /* MCC Event Register */
        VUBYTE  reserved88[0x2];/* Reserved area */
        VUHWORD mccm;           /* MCC Mask Register */
        VUBYTE  reserved89[0x2];/* Reserved area */
        VUBYTE  mccr;           /* MCC Configuration Register */
        VUBYTE  reserved90[0x7];/* Reserved area */
    } si_regs[2];
    VUBYTE reserved91[0x4a0];   /* Reserved area */
    
/* si_ram */
    struct si_ram 
    {
        VUHWORD tx_siram[0x100];/* SI Transmit Routing RAM */
        VUBYTE  reserved92[0x200];/* Reserved area */
        VUHWORD rx_siram[0x100];/* SI Receive Routing RAM */
        VUBYTE  reserved93[0x200];/* Reserved area */
    } si_ram[2];
    VUBYTE  reserved94[0x1000]; /* Reserved area */

}  t_PQ2IMM;





/***************************************************************************/
/*                   General Global Definitions                            */
/***************************************************************************/

#define PORTA   0           /* Index into Parallel I/O Regs Array */              
#define PORTB   1           /* Index into Parallel I/O Regs Array */              
#define PORTC   2           /* Index into Parallel I/O Regs Array */              
#define PORTD   3           /* Index into Parallel I/O Regs Array */
                                 
#define PAGE1   0           /* Index 1 into PRAM Array */
#define PAGE2   1           /* Index 2 into PRAM Array */
#define PAGE3   2           /* Index 3 into PRAM Array */
#define PAGE4   3           /* Index 4 into PRAM Array */

#define FCC1    0           /* FCC1 Index into FCC PRAM Array */
#define FCC2    1           /* FCC2 Index into FCC PRAM Array */
#define FCC3    2           /* FCC3 Index into FCC PRAM Array */

#define MCC1    0           /* MCC1 Index into MCC PRAM Array */
#define MCC2    1           /* MCC2 Index into MCC PRAM Array */

#define SCC1    0           /* SCC1 Index into SCC PRAM Array */             
#define SCC2    1           /* SCC2 Index into SCC PRAM Array */             
#define SCC3    2           /* SCC3 Index into SCC PRAM Array */             
#define SCC4    3           /* SCC4 Index into SCC PRAM Array */

#define TDMA    0           /* Index into TDM Mode Register Array */
#define TDMB    1           /* Index into TDM Mode Register Array */
#define TDMC    2           /* Index into TDM Mode Register Array */
#define TDMD    3           /* Index into TDM Mode Register Array */

#define IMMR ((t_PQ2IMM *)0xf0000000)

/* 603e CPU registers */

#define MSR_VEC         (1<<25)         /* Enable AltiVec */
#define MSR_POW         (1<<18)         /* Enable Power Management */
#define MSR_WE          (1<<18)         /* Wait State Enable */
#define MSR_TGPR        (1<<17)         /* TLB Update registers in use */
#define MSR_CE          (1<<17)         /* Critical Interrupt Enable */
#define MSR_ILE         (1<<16)         /* Interrupt Little Endian */
#define MSR_EE          (1<<15)         /* External Interrupt Enable */
#define MSR_PR          (1<<14)         /* Problem State / Privilege Level */
#define MSR_FP          (1<<13)         /* Floating Point enable */
#define MSR_ME          (1<<12)         /* Machine Check Enable */
#define MSR_FE0         (1<<11)         /* Floating Exception mode 0 */
#define MSR_SE          (1<<10)         /* Single Step */
#define MSR_BE          (1<<9)          /* Branch Trace */
#define MSR_DE          (1<<9)          /* Debug Exception Enable */
#define MSR_FE1         (1<<8)          /* Floating Exception mode 1 */
#define MSR_IP          (1<<6)          /* Exception prefix 0x000/0xFFF */
#define MSR_IR          (1<<5)          /* Instruction Relocate */
#define MSR_DR          (1<<4)          /* Data Relocate */
#define MSR_PE          (1<<3)          /* Protection Enable */
#define MSR_PX          (1<<2)          /* Protection Exclusive Mode */
#define MSR_RI          (1<<1)          /* Recoverable Exception */
#define MSR_LE          (1<<0)          /* Little Endian */

#define MSR_            MSR_ME|MSR_RI
#define MSR_KERNEL      MSR_|MSR_IR|MSR_DR
#define MSR_USER        MSR_KERNEL|MSR_PR|MSR_EE

/* Floating Point Status and Control Register (FPSCR) Fields */

#define FPSCR_FX        0x80000000      /* FPU exception summary */
#define FPSCR_FEX       0x40000000      /* FPU enabled exception summary */
#define FPSCR_VX        0x20000000      /* Invalid operation summary */
#define FPSCR_OX        0x10000000      /* Overflow exception summary */
#define FPSCR_UX        0x08000000      /* Underflow exception summary */
#define FPSCR_ZX        0x04000000      /* Zero-devide exception summary */
#define FPSCR_XX        0x02000000      /* Inexact exception summary */
#define FPSCR_VXSNAN    0x01000000      /* Invalid op for SNaN */
#define FPSCR_VXISI     0x00800000      /* Invalid op for Inv - Inv */
#define FPSCR_VXIDI     0x00400000      /* Invalid op for Inv / Inv */
#define FPSCR_VXZDZ     0x00200000      /* Invalid op for Zero / Zero */
#define FPSCR_VXIMZ     0x00100000      /* Invalid op for Inv * Zero */
#define FPSCR_VXVC      0x00080000      /* Invalid op for Compare */
#define FPSCR_FR        0x00040000      /* Fraction rounded */
#define FPSCR_FI        0x00020000      /* Fraction inexact */
#define FPSCR_FPRF      0x0001f000      /* FPU Result Flags */
#define FPSCR_FPCC      0x0000f000      /* FPU Condition Codes */
#define FPSCR_VXSOFT    0x00000400      /* Invalid op for software request */
#define FPSCR_VXSQRT    0x00000200      /* Invalid op for square root */
#define FPSCR_VXCVI     0x00000100      /* Invalid op for integer convert */
#define FPSCR_VE        0x00000080      /* Invalid op exception enable */
#define FPSCR_OE        0x00000040      /* IEEE overflow exception enable */
#define FPSCR_UE        0x00000020      /* IEEE underflow exception enable */
#define FPSCR_ZE        0x00000010      /* IEEE zero divide exception enable */
#define FPSCR_XE        0x00000008      /* FP inexact exception enable */
#define FPSCR_NI        0x00000004      /* FPU non IEEE-Mode */
#define FPSCR_RN        0x00000003      /* FPU rounding control */

/* Special Purpose Registers (SPRNs)*/

#define SPRN_CDBCR      0x3D7   /* Cache Debug Control Register */
#define SPRN_CTR        0x009   /* Count Register */
#define SPRN_DABR       0x3F5   /* Data Address Breakpoint Register */
#define SPRN_DAC1       0x3F6   /* Data Address Compare 1 */
#define SPRN_DAC2       0x3F7   /* Data Address Compare 2 */
#define SPRN_DAR        0x013   /* Data Address Register */
#define SPRN_DBAT0L     0x219   /* Data BAT 0 Lower Register */
#define SPRN_DBAT0U     0x218   /* Data BAT 0 Upper Register */
#define SPRN_DBAT1L     0x21B   /* Data BAT 1 Lower Register */
#define SPRN_DBAT1U     0x21A   /* Data BAT 1 Upper Register */
#define SPRN_DBAT2L     0x21D   /* Data BAT 2 Lower Register */
#define SPRN_DBAT2U     0x21C   /* Data BAT 2 Upper Register */
#define SPRN_DBAT3L     0x21F   /* Data BAT 3 Lower Register */
#define SPRN_DBAT3U     0x21E   /* Data BAT 3 Upper Register */
#define SPRN_DBCR       0x3F2   /* Debug Control Regsiter */
#define   DBCR_EDM      0x80000000
#define   DBCR_IDM      0x40000000
#define   DBCR_RST(x)   (((x) & 0x3) << 28)
#define     DBCR_RST_NONE               0
#define     DBCR_RST_CORE               1
#define     DBCR_RST_CHIP               2
#define     DBCR_RST_SYSTEM             3
#define   DBCR_IC       0x08000000      /* Instruction Completion Debug Evnt */
#define   DBCR_BT       0x04000000      /* Branch Taken Debug Event */
#define   DBCR_EDE      0x02000000      /* Exception Debug Event */
#define   DBCR_TDE      0x01000000      /* TRAP Debug Event */
#define   DBCR_FER      0x00F80000      /* First Events Remaining Mask */
#define   DBCR_FT       0x00040000      /* Freeze Timers on Debug Event */
#define   DBCR_IA1      0x00020000      /* Instr. Addr. Compare 1 Enable */
#define   DBCR_IA2      0x00010000      /* Instr. Addr. Compare 2 Enable */
#define   DBCR_D1R      0x00008000      /* Data Addr. Compare 1 Read Enable */
#define   DBCR_D1W      0x00004000      /* Data Addr. Compare 1 Write Enable */
#define   DBCR_D1S(x)   (((x) & 0x3) << 12)     /* Data Adrr. Compare 1 Size */
#define     DAC_BYTE    0
#define     DAC_HALF    1
#define     DAC_WORD    2
#define     DAC_QUAD    3
#define   DBCR_D2R      0x00000800      /* Data Addr. Compare 2 Read Enable */
#define   DBCR_D2W      0x00000400      /* Data Addr. Compare 2 Write Enable */
#define   DBCR_D2S(x)   (((x) & 0x3) << 8)      /* Data Addr. Compare 2 Size */
#define   DBCR_SBT      0x00000040      /* Second Branch Taken Debug Event */
#define   DBCR_SED      0x00000020      /* Second Exception Debug Event */
#define   DBCR_STD      0x00000010      /* Second Trap Debug Event */
#define   DBCR_SIA      0x00000008      /* Second IAC Enable */
#define   DBCR_SDA      0x00000004      /* Second DAC Enable */
#define   DBCR_JOI      0x00000002      /* JTAG Serial Outbound Int. Enable */
#define   DBCR_JII      0x00000001      /* JTAG Serial Inbound Int. Enable */
#define SPRN_DBCR0      0x3F2   /* Debug Control Register 0 */
#define SPRN_DBCR1      0x3BD   /* Debug Control Register 1 */
#define SPRN_DBSR       0x3F0   /* Debug Status Register */
#define SPRN_DCCR       0x3FA   /* Data Cache Cacheability Register */
#define   DCCR_NOCACHE          0       /* Noncacheable */
#define   DCCR_CACHE            1       /* Cacheable */
#define SPRN_DCMP       0x3D1   /* Data TLB Compare Register */
#define SPRN_DCWR       0x3BA   /* Data Cache Write-thru Register */
#define   DCWR_COPY             0       /* Copy-back */
#define   DCWR_WRITE            1       /* Write-through */
#define SPRN_DEAR       0x3D5   /* Data Error Address Register */
#define SPRN_DEC        0x016   /* Decrement Register */
#define SPRN_DMISS      0x3D0   /* Data TLB Miss Register */
#define SPRN_DSISR      0x012   /* Data Storage Interrupt Status Register */
#define SPRN_EAR        0x11A   /* External Address Register */
#define SPRN_ESR        0x3D4   /* Exception Syndrome Register */
#define   ESR_IMCP      0x80000000      /* Instr. Machine Check - Protection */
#define   ESR_IMCN      0x40000000      /* Instr. Machine Check - Non-config */
#define   ESR_IMCB      0x20000000      /* Instr. Machine Check - Bus error */
#define   ESR_IMCT      0x10000000      /* Instr. Machine Check - Timeout */
#define   ESR_PIL       0x08000000      /* Program Exception - Illegal */
#define   ESR_PPR       0x04000000      /* Program Exception - Priveleged */
#define   ESR_PTR       0x02000000      /* Program Exception - Trap */
#define   ESR_DST       0x00800000      /* Storage Exception - Data miss */
#define   ESR_DIZ       0x00400000      /* Storage Exception - Zone fault */
#define SPRN_EVPR       0x3D6   /* Exception Vector Prefix Register */
#define SPRN_HASH1      0x3D2   /* Primary Hash Address Register */
#define SPRN_HASH2      0x3D3   /* Secondary Hash Address Resgister */
#define SPRN_HID0       0x3F0   /* Hardware Implementation Register 0 */
#define   HID0_EMCP     (1<<31)         /* Enable Machine Check pin */
#define   HID0_EBA      (1<<29)         /* Enable Bus Address Parity */
#define   HID0_EBD      (1<<28)         /* Enable Bus Data Parity */
#define   HID0_SBCLK    (1<<27)
#define   HID0_EICE     (1<<26)
#define   HID0_ECLK     (1<<25)
#define   HID0_PAR      (1<<24)
#define   HID0_DOZE     (1<<23)
#define   HID0_NAP      (1<<22)
#define   HID0_SLEEP    (1<<21)
#define   HID0_DPM      (1<<20)
#define   HID0_ICE      (1<<15)         /* Instruction Cache Enable */
#define   HID0_DCE      (1<<14)         /* Data Cache Enable */
#define   HID0_ILOCK    (1<<13)         /* Instruction Cache Lock */
#define   HID0_DLOCK    (1<<12)         /* Data Cache Lock */
#define   HID0_ICFI     (1<<11)         /* Instr. Cache Flash Invalidate */
#define   HID0_DCI      (1<<10)         /* Data Cache Invalidate */
#define   HID0_SPD      (1<<9)          /* Speculative disable */
#define   HID0_SGE      (1<<7)          /* Store Gathering Enable */
#define   HID0_SIED     (1<<7)          /* Serial Instr. Execution [Disable] */
#define   HID0_BTIC     (1<<5)          /* Branch Target Instruction Cache Enable */
#define   HID0_ABE      (1<<3)          /* Address Broadcast Enable */
#define   HID0_BHTE     (1<<2)          /* Branch History Table Enable */
#define   HID0_BTCD     (1<<1)          /* Branch target cache disable */
#define SPRN_HID1       0x3F1   /* Hardware Implementation Register 1 */
#define SPRN_IABR       0x3F2   /* Instruction Address Breakpoint Register */
#define SPRN_IAC1       0x3F4   /* Instruction Address Compare 1 */
#define SPRN_IAC2       0x3F5   /* Instruction Address Compare 2 */
#define SPRN_IBAT0L     0x211   /* Instruction BAT 0 Lower Register */
#define SPRN_IBAT0U     0x210   /* Instruction BAT 0 Upper Register */
#define SPRN_IBAT1L     0x213   /* Instruction BAT 1 Lower Register */
#define SPRN_IBAT1U     0x212   /* Instruction BAT 1 Upper Register */
#define SPRN_IBAT2L     0x215   /* Instruction BAT 2 Lower Register */
#define SPRN_IBAT2U     0x214   /* Instruction BAT 2 Upper Register */
#define SPRN_IBAT3L     0x217   /* Instruction BAT 3 Lower Register */
#define SPRN_IBAT3U     0x216   /* Instruction BAT 3 Upper Register */
#define SPRN_ICCR       0x3FB   /* Instruction Cache Cacheability Register */
#define   ICCR_NOCACHE          0       /* Noncacheable */
#define   ICCR_CACHE            1       /* Cacheable */
#define SPRN_ICDBDR     0x3D3   /* Instruction Cache Debug Data Register */
#define SPRN_ICMP       0x3D5   /* Instruction TLB Compare Register */
#define SPRN_ICTC       0x3FB   /* Instruction Cache Throttling Control Reg */
#define SPRN_IMISS      0x3D4   /* Instruction TLB Miss Register */
#define SPRN_IMMR       0x27E   /* Internal Memory Map Register */
#define SPRN_L2CR       0x3F9   /* Level 2 Cache Control Regsiter */
#define   L2CR_PIPE_LATEWR   (0x01800000)   /* late-write SRAM */
#define   L2CR_L2CTL         (0x00100000)   /* RAM control */
#define   L2CR_INST_DISABLE  (0x00400000)   /* disable for insn's */
#define   L2CR_L2I           (0x00200000)   /* global invalidate */
#define   L2CR_L2E           (0x80000000)   /* enable */
#define   L2CR_L2WT          (0x00080000)   /* write-through */
#define SPRN_LR         0x008   /* Link Register */
#define SPRN_MMCR0      0x3B8   /* Monitor Mode Control Register 0 */
#define SPRN_MMCR1      0x3BC   /* Monitor Mode Control Register 1 */
#define SPRN_PBL1       0x3FC   /* Protection Bound Lower 1 */
#define SPRN_PBL2       0x3FE   /* Protection Bound Lower 2 */
#define SPRN_PBU1       0x3FD   /* Protection Bound Upper 1 */
#define SPRN_PBU2       0x3FF   /* Protection Bound Upper 2 */
#define SPRN_PID        0x3B1   /* Process ID */
#define SPRN_PIR        0x3FF   /* Processor Identification Register */
#define SPRN_PIT        0x3DB   /* Programmable Interval Timer */
#define SPRN_PMC1       0x3B9   /* Performance Counter Register 1 */
#define SPRN_PMC2       0x3BA   /* Performance Counter Register 2 */
#define SPRN_PMC3       0x3BD   /* Performance Counter Register 3 */
#define SPRN_PMC4       0x3BE   /* Performance Counter Register 4 */
#define SPRN_PVR        0x11F   /* Processor Version Register */
#define SPRN_RPA        0x3D6   /* Required Physical Address Register */
#define SPRN_SDA        0x3BF   /* Sampled Data Address Register */
#define SPRN_SDR1       0x019   /* MMU Hash Base Register */
#define SPRN_SGR        0x3B9   /* Storage Guarded Register */
#define   SGR_NORMAL            0
#define   SGR_GUARDED           1
#define SPRN_SIA        0x3BB   /* Sampled Instruction Address Register */
#define SPRN_SPRG0      0x110   /* Special Purpose Register General 0 */
#define SPRN_SPRG1      0x111   /* Special Purpose Register General 1 */
#define SPRN_SPRG2      0x112   /* Special Purpose Register General 2 */
#define SPRN_SPRG3      0x113   /* Special Purpose Register General 3 */
#define SPRN_SRR0       0x01A   /* Save/Restore Register 0 */
#define SPRN_SRR1       0x01B   /* Save/Restore Register 1 */
#define SPRN_SRR2       0x3DE   /* Save/Restore Register 2 */
#define SPRN_SRR3       0x3DF   /* Save/Restore Register 3 */
#define SPRN_TBRL       0x10C   /* Time Base Read Lower Register (user, R/O) */
#define SPRN_TBRU       0x10D   /* Time Base Read Upper Register (user, R/O) */
#define SPRN_TBWL       0x11C   /* Time Base Lower Register (supervisor, R/W) */
#define SPRN_TBWU       0x11D   /* Time Base Upper Register (supervisor, R/W) */

#define stringify(s)    tostring(s)
#define tostring(s)     #s

#define mfmsr()         ({unsigned int rval; \
                        asm volatile("mfmsr %0" : "=r" (rval)); rval;})
#define mtmsr(v)        asm volatile("mtmsr %0" : : "r" (v))

#define mfspr(rn)       ({unsigned int rval; \
                        asm volatile("mfspr %0," stringify(rn) \
                                     : "=r" (rval)); rval;})
#define mtspr(rn, v)    asm volatile("mtspr " stringify(rn) ",%0" : : "r" (v))


static __inline__ unsigned int get_dec(void)
{
        return (mfspr(SPRN_DEC));
}

static __inline__ void set_dec(unsigned int val)
{
        mtspr(SPRN_DEC, val);
}


             
#endif
