/*******************************************************************************

    Minimal ARP/IP/UDP stack for polling ethernet interface
    -------------------------------------------------------

  Author: Arto Vuori <avuori@ssh.com>

  Copyright (C) 2000 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved. 

*******************************************************************************/

#include <string.h>
#include "sipo.h"

/* Wait for t microseconds */
extern void udelay(int t);

#define NAME_LENGTH 64
in_addr MY_IPADDR={0x00000000};
in_addr MY_NETMASK={0x00000000};
in_addr MY_GATEWAY={0x00000000};
in_addr MY_DNSSERV={0x00000000};
char    MY_NAME[NAME_LENGTH];
in_addr SERVER_IPADDR={0x00000000};
char    SERVER_NAME[NAME_LENGTH];
in_addr BROADCAST_IPADDR=0xffffffff;
ethaddr BROADCAST_ETHADDR={0xff,0xff,0xff,0xff,0xff,0xff};
static char dhcp_request[]={RFC1533_COOKIE, RFC1533_END};
char    rootpath[NAME_LENGTH];
char    bootfile[NAME_LENGTH];


#define MAX_ARP 16
struct arptable_t arptable[MAX_ARP];

#define MAX_SOCKET 16
udp_socket socket_table[MAX_SOCKET];

/* initialize network code */

net_init() {
    int i;

    memcpy(eth_interface[0].phyaddr, __bd->bi_enetaddr, sizeof(ethaddr));
    eth_init(&eth_interface[0]);
    udelay(100000);

    for(i=0;i<MAX_SOCKET;i++)
        socket_table[i].state=SOCK_FREE;

    for(i=0;i<MAX_ARP;i++)
        arptable[i].ipaddr=0;
}

/* Find hwaddr for IP addr by using ARP protocol. */

static ethaddr *arp_lookup(in_addr ip_addr) {
    char        buffer[256];
    ethhdr      *eth;
    ethaddr     *ret;
    arprequest *arp;
    static int next_i=0;
    int repeat;
    int i,j;

    if(ip_addr==BROADCAST_IPADDR) 
        return &BROADCAST_ETHADDR;

    if((ip_addr & MY_NETMASK) != (MY_IPADDR & MY_NETMASK))
      ip_addr = MY_GATEWAY;

    eth=(ethhdr *)buffer;
    arp=(arprequest *)(buffer+sizeof(ethhdr));

    for(i=0;i<MAX_ARP;i++)
        if(!memcmp(&arptable[i].ipaddr, &ip_addr,sizeof(in_addr)))
            return &(arptable[i].ethaddr);

    for(repeat=0;repeat<ARP_MAX_REPEAT;repeat++)
    {
        memcpy(eth->dstaddr, BROADCAST_ETHADDR, sizeof(ethaddr));
        memcpy(eth->srcaddr, eth_interface[0].phyaddr, sizeof(ethaddr));
        eth->type=PROT_ARP;

        /* Fill arp packet */
        arp->hwtype=1;
        arp->protocol=PROT_IP;
        arp->hwlen=6;
        arp->protolen=4;
        arp->opcode=ARP_REQUEST;
        memcpy(&arp->shwaddr, eth_interface[0].phyaddr, sizeof(ethaddr));
        memcpy(&arp->sipaddr, &MY_IPADDR, sizeof(in_addr));
        memcpy(&arp->thwaddr, &BROADCAST_ETHADDR, sizeof(ethaddr));
        memcpy(&arp->tipaddr, &ip_addr, sizeof(in_addr));

        /* And send it */
        eth_send(&eth_interface[0], buffer, sizeof(ethhdr)+sizeof(arprequest));

        /* Wait for reply */
        for(i=0;i<1000;i++) {
            /* Receive packets */
            j=eth_recv(&eth_interface[0], buffer,256);

            /* If no packets received, wait for 1ms */
            if(j==0)
                udelay(1000);

            /* Drop invalid packets */
            if(j<=0)
                continue;

            /* Drop unwanted packets */
            if(eth->type!=PROT_ARP 
               || arp->opcode!=ARP_REPLY
/*               || memcmp(&eth->dstaddr, eth_interface[0].phyaddr, sizeof(ethaddr))  */
               || memcmp(&arp->sipaddr, &ip_addr, sizeof(in_addr)))

                continue;

            /* Save result to ARP table */
            arptable[next_i].ipaddr=ip_addr;
            memcpy(&arptable[next_i].ethaddr, &arp->shwaddr, sizeof(ethaddr));
            ret=&(arptable[next_i].ethaddr);

            /* Advance ARP cache update pointer */
            next_i=(next_i+1>=MAX_ARP)?0:next_i+1;

            return ret;
        }
    }
}

/* create udp socket */

udp_socket *socket(int proto, in_addr src_ip, int src_port, in_addr dst_ip,
                   int dst_port, unsigned int flags)
{
    int i;
    udp_socket *sock;

    /* Find free socket */
    for(i=0;i<MAX_SOCKET && socket_table[i].state!=SOCK_FREE;i++);
    if(i==MAX_SOCKET) return NULL;
    sock=&socket_table[i];
    sock->state=SOCK_READY;
    sock->flags = flags;

    sock->src_ip=src_ip;
    sock->src_port=src_port;
    sock->dst_ip=dst_ip;
    sock->dst_port=dst_port;
    sock->data=NULL;

    return sock;
}

static unsigned short ipchksum(unsigned short *ip, int len)
{
    unsigned short c;
    unsigned long sum = 0;

    len >>= 1;
    while (len--) {
        c=*(ip++);
        sum += c;
        if (sum > 0xFFFF)
            sum -= 0xFFFF;
    }
    return((~sum) & 0x0000FFFF);
}

/* XXXX Uhh.. this doesn't work */
static unsigned short udpchksum(iphdr *ip, udphdr *udp)
{
    int len;
    unsigned short c, *p;
    unsigned long sum = 0;
    struct pseudohdr {
        in_addr src,dest;
        char    zero;
        char    proto;
        unsigned short len;
    } pseudo;

    pseudo.src=ip->src;
    pseudo.dest=ip->dest;
    pseudo.zero=0;
    pseudo.proto=ip->protocol;
    pseudo.len=udp->len;

    p=(unsigned short *)&pseudo;
    len = sizeof(struct pseudohdr)>>1; 
    while (len--) {
        c=*(p++);
        sum += c;
        if (sum > 0xFFFF)
            sum -= 0xFFFF;
    }

    p = (unsigned short *)udp;
    len = udp->len >> 1;
    while (len--) {
        c=*(p++);
        sum += c;
        if (sum > 0xFFFF)
            sum -= 0xFFFF;
    }

    if (udp->len & 1)
      {
        c = *(((unsigned char *) udp) + udp->len - 1) << 8;
        sum += c;
        if (sum > 0xFFFF)
            sum -= 0xFFFF;
      }

    sum=((~sum) & 0x0000FFFF);
    return sum;
}



/* send udp packet */

int udp_send(udp_socket *sock, void *data, int len)
{
    static int  ident=0;
    char        buffer[2048];
    ethhdr      *eth;
    iphdr               *ip;
    udphdr      *udp;
    void                *payload;

    eth=(ethhdr *)buffer;
    ip=(iphdr *)((void *)eth + sizeof(ethhdr));
    udp=(udphdr *)((void *)ip + sizeof(iphdr));
    payload=(void *)udp + sizeof(udphdr);

    /* Build ethernet header */
    memcpy(eth->dstaddr, arp_lookup(sock->dst_ip), sizeof(ethaddr));
    memcpy(eth->srcaddr, eth_interface[0].phyaddr, sizeof(ethaddr));
    eth->type=PROT_IP;

    /* Build IP header */
    ip->verhdrlen=0x45;                 /* ver=4, len=0x20 */
    ip->service=0x00;
    ip->len=20+sizeof(udphdr)+len;      /* IPhdr len + udphdr + data length */
    ip->ident=ident++;
    ip->frags=0x4000;                   /* Don't fragment */
    ip->ttl=0x40;
    ip->protocol=PROT_IP_UDP;
    ip->chksum=0;
    ip->src=sock->src_ip;
    ip->dest=sock->dst_ip;
    ip->chksum = ipchksum((unsigned short *)ip, sizeof(iphdr));

    /* Build UDP header */
    udp->src=sock->src_port;
    udp->dest=sock->dst_port;
    udp->len=len+sizeof(udphdr);
    udp->chksum=0;

    /* copy payload */
    memcpy(payload, data, len);

    if (sock->flags & UDP_SOCK_CHECKSUM)
      {
        udp->chksum = udpchksum(ip, udp);
        if (udp->chksum == 0)
          udp->chksum = 0xffff;
      }
    else
      udp->chksum = 0; 

    /* Pass frame to ethernet driver */
    eth_send(&eth_interface[0], eth, sizeof(ethhdr)+sizeof(iphdr)+sizeof(udphdr)+len);

    return 0;
}

/* Receive one udp packet matching socket.
   Return > 0 if packet destinated to socket received. Returns zero otherwise. 
   Also sends an ARP_REPLY if anyone is asking for our ethaddr.
*/

int udp_recv(udp_socket *sock, void *data, int maxlen)
{
    char        buffer[2048];
    ethhdr      *eth;
    arprequest  *arp;
    iphdr       *ip;
    udphdr      *udp;
    void        *payload;
    int         j;

    eth=(ethhdr *)buffer;
    arp=(arprequest *)(buffer+sizeof(ethhdr));
    ip=(iphdr *)((void *)eth + sizeof(ethhdr));
    udp=(udphdr *)((void *)ip + sizeof(iphdr));
    payload=(void *)udp + sizeof(udphdr);
    
    while(1) {
        /* Receive packet */
        j=eth_recv(&eth_interface[0], buffer,2048);

        /* return 0, if no packets available */
        if(j==0)
            return 0;

        /* Drop invalid packets */
        if(j<0)
            continue;

        /* Check if ARP request for us */
        if(eth->type==PROT_ARP
           && arp->opcode==ARP_REQUEST
           && !memcmp(&arp->tipaddr, &MY_IPADDR, sizeof(in_addr)))
        {
            /* Send ARP reply */
            memcpy(&eth->dstaddr, &eth->srcaddr, sizeof(ethaddr));
            memcpy(&eth->srcaddr, eth_interface[0].phyaddr, sizeof(ethaddr));

            memcpy(&arp->tipaddr, &arp->sipaddr, sizeof(in_addr));
            memcpy(&arp->thwaddr, &arp->shwaddr, sizeof(ethaddr));
            memcpy(&arp->sipaddr, &MY_IPADDR, sizeof(in_addr));
            memcpy(&arp->shwaddr, eth_interface[0].phyaddr, sizeof(ethaddr));
            arp->opcode=ARP_REPLY;
            eth_send(&eth_interface[0], buffer, j);
            continue;
        }

        /* Drop unwanted packets */
        if(eth->type!=PROT_IP
           || j<sizeof(iphdr)+sizeof(ethhdr)+sizeof(udphdr)
           || ip->verhdrlen != 0x45
           || ip->protocol!=PROT_IP_UDP
           || ipchksum((unsigned short *)ip, sizeof(iphdr))
           || (sock->flags & UDP_SOCK_CHECKSUM && udpchksum(ip,udp))
            ) 
            continue;

        /* Drop packets that are sent to wrong socket */
        if(udp->dest != sock->src_port
#if 0
           || udp->src != sock->dst_port
#endif
           || memcmp(&ip->dest, &sock->src_ip, sizeof(in_addr))
           || memcmp(&ip->src, &sock->dst_ip, sizeof(in_addr))
            )
        {
#if 0
            printf("UDP: Wrong port, packet dropped\n");
#endif
            continue;
        }
        
        /* Drop too long packets */
        if(udp->len-8>maxlen) {
            printf("UDP: Too long, packet dropped\n");
            return 0;
        }

        /* Take copy of UDP packet source address */
        sock->dst_port = udp->src;

        memcpy(data, payload, udp->len-8);

        return udp->len-8;
    }
}

/*
  Broadcast an BOOTP/DHCP request to find out our IP addr + rootpath + rootfile ... etc
*/

int bootp() {
    char        buffer[2048];
    ethhdr      *eth;
    iphdr       *ip;
    udphdr      *udp;
    bootp_t     bp, *bpr;
    static int  xid=0xca63e752;
    int         i,j,repeat;
    udp_socket  *sock;
    unsigned char *s,c,l, str[1000];

    /* Don't do bootp, if we already have an ip address. */
    if(MY_IPADDR!=0)
        return 0;

    eth=(ethhdr *)buffer;
    ip=(iphdr *)((void *)eth + sizeof(ethhdr));
    udp=(udphdr *)((void *)ip + sizeof(iphdr));
    bpr=(void *)udp + sizeof(udphdr);

    sock=socket(PROT_IP_UDP, 0x0, BOOTP_CLIENT, BROADCAST_IPADDR, BOOTP_SERVER, UDP_SOCK_CHECKSUM);

    memset(&bp,0,sizeof(bootp_t));

    bp.op=BOOTP_REQUEST;
    bp.htype=1;
    bp.hlen=6;
    bp.hops=0;
    bp.xid=xid;
    bp.secs=0;
    bp.unused=0;
    bp.ciaddr=0;
    bp.yiaddr=0;
    bp.siaddr=0;
    bp.giaddr=0;
    memcpy(&bp.hwaddr, eth_interface[0].phyaddr, sizeof(ethaddr));
    bp.sname[0]=0;
    memcpy(&bp.vend, &dhcp_request, sizeof(dhcp_request));

    for(repeat=0;repeat<BOOTP_MAX_REPEAT;repeat++)
    {
        printf("Sending BOOTP request ... \n");
        /* Send bootp request */
        udp_send(sock, &bp, sizeof(bootp_t));

        for(i=0;i<1000;i++)
        {
            /* Receive packet */
            j=eth_recv(&eth_interface[0], buffer,2048);

            /* wait while, if no packets available */
            if(j==0)
                udelay(1000);

            /* Drop invalid packets */
            if(j<=0)
                continue;

            /* Drop unwanted packets */
            if(eth->type!=PROT_IP
               || j<sizeof(iphdr)+sizeof(ethhdr)+sizeof(udphdr)
               || ip->verhdrlen != 0x45
               || ip->protocol!=PROT_IP_UDP
               || ipchksum((unsigned short *)ip, sizeof(iphdr)))
                continue;

            /* Drop packets that are sent to wrong socket */
            if(udp->dest != sock->src_port
               || udp->src != sock->dst_port)
            {
#if 0
                printf("Unknown UDP packet dropped\n");
#endif
                continue;
            }
        
            if(bpr->op!=BOOTP_REPLY || bpr->xid != xid)
            {
                printf("Not a BOOTP reply\n");
                continue;
            }

            printf("BOOTP Reply received (");

            memcpy(&MY_IPADDR, &bpr->yiaddr, sizeof(in_addr));
            printf("my IP addr = %d.%d.%d.%d)\n", 
                   ((char *)&MY_IPADDR)[0],
                   ((char *)&MY_IPADDR)[1],
                   ((char *)&MY_IPADDR)[2],
                   ((char *)&MY_IPADDR)[3]);

            memcpy(&SERVER_IPADDR, &bpr->siaddr, sizeof(in_addr));
            memcpy(SERVER_NAME, bpr->sname, 64);
            SERVER_NAME[64]=0;
#ifdef DEBUG
            printf("  Server IP: %d.%d.%d.%d (%s)\n", 
                   ((char *)&SERVER_IPADDR)[0],
                   ((char *)&SERVER_IPADDR)[1],
                   ((char *)&SERVER_IPADDR)[2],
                   ((char *)&SERVER_IPADDR)[3], bpr->sname);
#endif

            memcpy(bootfile, bpr->file, 64);
            bootfile[64]=0;
#ifdef DEBUG
            printf("   Bootfile: %s\n", bpr->file);
#endif


            s=(char *)&bpr->vend;
            /* compare DHCP magic. Return if not correct */
            if(memcmp(s,dhcp_request,4)) 
                return -1;
            s+=4;
            /* Parse DHCP vendor information block */
            while(1) {
                c=*s++;
                l=*s++;
                if(c==RFC1533_END)
                    break;
                switch(c) {
                case RFC1533_NETMASK:
#ifdef DEBUG
                    printf("    Netmask: %d.%d.%d.%d\n", s[0], s[1], s[2], s[3]);
#endif
                    memcpy(&MY_NETMASK,s,4);
                    break;                  
                case RFC1533_GATEWAY:
#ifdef DEBUG
                    printf("    Gateway: %d.%d.%d.%d\n", s[0], s[1], s[2], s[3]);
#endif
                    memcpy(&MY_GATEWAY,s,4);
                    break;
                case RFC1533_DNS:      
#ifdef DEBUG
                    printf("   Nameserv: %d.%d.%d.%d\n", s[0], s[1], s[2], s[3]);
#endif
                    memcpy(&MY_DNSSERV,s,4);
                    break;
                case RFC1533_HOSTNAME:
                    memcpy(MY_NAME,s,(l>NAME_LENGTH)?NAME_LENGTH:l); 
                    MY_NAME[(l>64)?64:l]=0;
#ifdef DEBUG
                    printf("   hostname: %s\n", MY_NAME);
#endif
                    break;
                case RFC1533_ROOTPATH:
                    memcpy(rootpath,s,(l>NAME_LENGTH)?NAME_LENGTH:l); 
                    rootpath[(l>64)?64:l]=0;
#ifdef DEBUG
                    printf("   rootpath: %s\n", rootpath);
#endif
                    break;
                case RFC1533_BOOTFILESIZE:
#ifdef DEBUG
                    printf(" rootfilesz: %d\n", (int)(*(unsigned short *)s)*512);
#endif
                    break;
                }
                s+=l;
            }

            return 0;
        }
    }
    return -2;
}

/* Download file using TFTP */

int tftp_get(char *buffer, char *filename, in_addr serverip)
{    
    tftp_t      tftp;
    udp_socket  *sock;
    int         i,j, k, l, block=1, last_ack=0;

    printf("TFTP get \"%s\" from %d.%d.%d.%d.\n", filename, 
           ((unsigned char *)&serverip)[0],
           ((unsigned char *)&serverip)[1],
           ((unsigned char *)&serverip)[2],
           ((unsigned char *)&serverip)[3]);

    sock=socket(PROT_IP_UDP, MY_IPADDR, 12345, serverip,  TFTP_PORT, UDP_SOCK_CHECKSUM);
    
    memset(&tftp, 0, sizeof(tftp));
    tftp.opcode=TFTP_RRQ;
    strcpy(tftp.u.rrq, filename);
    strcpy(tftp.u.rrq+strlen(filename)+1, "octet");
    udp_send(sock, &tftp, 2 + strlen(filename) + 1 + 5 + 1);

    while(1) {
        for(l=0;l<3;l++) /* Try with three exp backoff sequences */
        {
            for(i=0, k=100; i<30000;i++) /* Do 30sec exponential backoff sequence */
            {
                j=udp_recv(sock, &tftp, sizeof(tftp));
                if(!j)
                    udelay(1000);
                if(j>0)
                    break;
                if(i>k && last_ack) {
                    k=k*2;
                    tftp.u.data.block=last_ack;
                    tftp.opcode=TFTP_ACK;
                    udp_send(sock, &tftp, 4);
                }
            }
            if(i!=30000) break;
        }
        if(l==3) return -1;

        switch(tftp.opcode) {
        case TFTP_DATA:
            if(tftp.u.data.block>block) {
                printf("?");
                break;
            }

            /* Send ack */
            tftp.opcode=TFTP_ACK;
            udp_send(sock, &tftp, 4);

            if(tftp.u.data.block<block)
                break;

            last_ack=block;

            /* copy data */
            memcpy(buffer+(block-1)*512, tftp.u.data.download, j-4);

            /* Print '.' at every 10k */
            if(block%20==0)  
                printf(".");

            /* stop if partial block is received */
            if(j-4!=512) {
                printf("\ndone (%d bytes read)\n", (block-1)*512+j-4);
                return (block-1)*512+j-4;
            }
            /* advance block number */
            block++;

            break;

        case TFTP_ERROR:
            printf("TFTP Error %d: %s\n", tftp.u.err.errcode, tftp.u.err.errmsg);
            return -1;
            break;
        }
    }
    
}


int tftp_put(char *buffer, int len, char *filename, in_addr serverip)
{    
  tftp_t tftp;
  udp_socket *sock;
  int i,j, k, l, block_count, last_len, block = 0;

  printf("TFTP put \"%s\" from %d.%d.%d.%d.\n", filename, 
         ((unsigned char *)&serverip)[0],
         ((unsigned char *)&serverip)[1],
         ((unsigned char *)&serverip)[2],
         ((unsigned char *)&serverip)[3]);

  block_count = len / 512;
  last_len = len % 512;
  if (last_len == 0)
    block_count++;

  sock = socket(PROT_IP_UDP, MY_IPADDR, 12345, serverip, TFTP_PORT,
                UDP_SOCK_CHECKSUM);

  memset(&tftp, 0, sizeof(tftp));
  tftp.opcode = TFTP_WRQ;
  strcpy(tftp.u.wrq, filename);
  strcpy(tftp.u.wrq + strlen(filename) + 1, "octet");
  udp_send(sock, &tftp, 2 + strlen(filename) + 1 + 5 + 1);

  while (block <= block_count)
    {
      /* Try with three exp backoff sequences */
      for (l = 0; l < 3; l++)
        {
          /* Do 30sec exponential backoff sequence */
          for (i = 0, k = 0; i < 30000; i++)
            {
              if (i > k && block)
                {
                  int blen;

                  if (k == 0)
                    k = 100;
                  else
                    k *= 2;
                  blen = block == block_count ? last_len : 512;

                  tftp.opcode = TFTP_DATA;
                  tftp.u.data.block = block;
                  memcpy(tftp.u.data.download, buffer + (block - 1) * 512,
                         blen);
                  udp_send(sock, &tftp, 4 + blen);
                }

              j = udp_recv(sock, &tftp, sizeof(tftp));
              if (!j)
                udelay(1000);
              if (j > 0)
                break;
            }
          if (i != 30000)
            break;
        }

      if (l == 3)
        return -1;

      switch (tftp.opcode) {
      case TFTP_ACK:
        if (tftp.u.ack.block != block)
          {
            printf("?");
            break;
          }

        /* Print '.' at every 10k */
        if (block % 20 == 0)
          printf(".");

        block++;
        break;

      case TFTP_ERROR:
        printf("TFTP Error %d: %s\n", tftp.u.err.errcode, tftp.u.err.errmsg);
        return -1;
        break;
      }
    }

  printf("\ndone (%d bytes written)\n", len);

  return len;
}
