/*******************************************************************************

    Sipo Partition managemen
    --------------------

  Author: Arto Vuori <avuori@ssh.com>

  Copyright (C) 2001 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved. 

  Board info block & Partition table management

*******************************************************************************/


#include "sipo.h"
#include "stdio.h"

sipo_board_info *__bd=(sipo_board_info *)BD_OFFSET; /* Points to BD when, running from flash */

/* Return nonzero if board descriptor is valid */
int bd_valid() 
{
    SHA1Digest digest;

    sha1((void *)__bd, sizeof(sipo_board_info)-sizeof(__bd->verify), digest);
    if(memcmp(__bd->verify, digest, sizeof(digest)) ||__bd->bd_version != __BD_bd_version)
        return 0;

    return 1;
}

void write_bd() {
    SHA1Digest digest;

    sha1((void *)__bd, sizeof(sipo_board_info)-sizeof(__bd->verify), digest);
    memcpy(__bd->verify, digest,sizeof(__bd->verify));
    flash_cutcopypaste(BD_OFFSET, __bd, sizeof(sipo_board_info));
}

int boottest_bd() 
{
    sipo_board_info *oldbd=__bd;
    
    __bd=(sipo_board_info *)malloc(sizeof(sipo_board_info));
    memcpy(__bd, (void *)(oldbd->bi_memsize-oldbd->flash_size+BD_OFFSET), sizeof(sipo_board_info));
    if(bd_valid()) {
        return 0;       /* All ok */
    } else {            
        free(__bd);     /* Abort test */
        __bd=oldbd;
        return 1;
    }
}

init_bd() 
{
    SHA1Digest digest;

    __bd=(sipo_board_info *)malloc(sizeof(sipo_board_info));
    memcpy(__bd, (void *)BD_START_ADDRESS, sizeof(sipo_board_info));

    /* Initialize board descriptor if not yet done */
    if(!bd_valid())
    {
        char    eth[6]=__BD_enetaddr;
        int     i;

        printf("Warning: No valid setting found. Using default setting... \n");

        memset(__bd,0,sizeof(sipo_board_info));

        __bd->bi_memstart=__BD_memstart;
        __bd->bi_memsize=__BD_memsize;
        __bd->bi_memtotal=__BD_memtotal;
        __bd->bi_intfreq=__BD_intfreq;
        __bd->bi_busfreq=__BD_busfreq;
        __bd->bi_cpmfreq=__BD_cpmfreq;
        __bd->bi_brgfreq=__BD_brgfreq;
        __bd->bi_vco=__BD_vco;
        __bd->bi_baudrate=__BD_baudrate;
        __bd->bi_immr=(unsigned)IMMR;
        for(i=0;i<6;i++) __bd->bi_enetaddr[i]=eth[i];
        __bd->flash_base=__BD_flash_base;
        __bd->true_flash_base=__BD_flash_base;
        __bd->bd_version=__BD_bd_version;
        __bd->hw_version=__BD_hw_version;
        __bd->sw_version=__BD_sw_version;
        __bd->boot_state=__BD_boot_state;
        __bd->enable_wdt=__BD_enable_wdt;
        __bd->enable_netboot=__BD_enable_netboot;
        __bd->enable_debugport=__BD_enable_debugport;
        strcpy(__bd->cmdline, __BD_cmdline);
    }
    __bd->flash_size=init_flash();
}
void print_bd(void) 
{
        printf("memstart = 0x%08x\n",__bd->bi_memstart);    /* Memory start address */
        printf("memsize  = 0x%08x\n",__bd->bi_memsize);     /* Usable memory size in bytes */
        printf("memtotal = 0x%08x\n",__bd->bi_memtotal);    /* Physical memory size in bytes */
        printf("intfreq  = %d\n",__bd->bi_intfreq);     /* Internal Freq, in Hz */
        printf("busfreq  = %d\n",__bd->bi_busfreq);     /* Bus Freq, in MHz */
        printf("cpmfreq  = %d\n",__bd->bi_cpmfreq);     /* CPM Freq, in MHz */
        printf("brgfreq  = %d\n",__bd->bi_brgfreq);     /* BRG Freq, in MHz */
        printf("vco      = %d\n",__bd->bi_vco);         /* VCO Out from PLL */
        printf("baudrate = %d\n",__bd->bi_baudrate);    /* Default console baud rate */
        printf("immr     = 0x%08x\n",__bd->bi_immr);        /* IMMR when called from boot rom */
        printf("enetaddr = 0x%02x%02x%02x%02x%02x%02x\n",
               __bd->bi_enetaddr[0], __bd->bi_enetaddr[1], __bd->bi_enetaddr[2],
               __bd->bi_enetaddr[3], __bd->bi_enetaddr[4], __bd->bi_enetaddr[5]);

        printf("flashbase= 0x%08x\n", __bd->flash_base);
        printf("flashsize= 0x%08x\n", __bd->flash_size);
        printf("bd ver   = %d\n", __bd->bd_version);             /* Board version descriptor */
        printf("hw ver   = %d\n",__bd->hw_version);             /* hardware version */
        printf("sw ver   = %d\n",__bd->sw_version);             /* software version */
        printf("bootstate= %d\n", __bd->boot_state);             /* Boot state copied from NVRAM at initialization */
        printf("netboot  = %s\n",__bd->enable_netboot?"enabled":"disabled");        
        printf("debugport= %s\n", __bd->enable_debugport?"enabled":"disabled");
        printf("passwd   = %s\n", __bd->enable_passwd?"enabled":"disabled");
}


/*
  Modify flash partition table.

  Partition table is array of sipo_flash_partition entries and is located at fixed offset 
  in flash device. First partition_table entry is always pt_magic type, which doesn't define
  any real partition. pt_bootloader should be second entry and after that there should be 
  partition table entry.
*/

int partition_set(int index, sipo_partition_type type, int offset, int size) 
{
    int i;

    if(__bd->pt[0].type!=pt_magic) {
        printf("Partition table invalid. Initializing table ... \n");
        __bd->pt[0].type=pt_magic;
        __bd->pt[0].size=0;
        __bd->pt[0].offset=0;
        __bd->pt[1].type=pt_bootloader;
        __bd->pt[1].offset=0;
        __bd->pt[1].size=BD_OFFSET;
        __bd->pt[2].type=pt_bd;
        __bd->pt[2].offset=BD_OFFSET;
        __bd->pt[2].size=sizeof(sipo_board_info);
        for(i=3;i<pt_size;i++)
            __bd->pt[i].type=pt_unused;
    }

    if(offset < 0 || offset + size > __bd->flash_size)
    {
        printf("Partition does not fit in FLASH\n");
        return -1;
    }

    if(index<pt_size && index>0) {
        if(type!=pt_unused) {
            for(i=index-1;i>0;i--)
                if(__bd->pt[i].type!=pt_unused) {
                    if(__bd->pt[i].offset+__bd->pt[i].size>offset) {
                        printf("Overlap with previous partition. No overlap allowed\n");
                        return -1;
                    }
                    break;
                }
            for(i=index+11;i<pt_size;i++)
                if(__bd->pt[i].type!=pt_unused) {
                    if(__bd->pt[i].offset<size+offset) {
                        printf("Overlap with next partition. No overlap allowed\n");
                        return -1;
                    }
                    break;
                }
        }

        __bd->pt[index].type=type;
        __bd->pt[index].offset=offset;
        __bd->pt[index].size=size;
    }

    return 0;
}


void partition_print(void) {
    int i;
    
    if(__bd->pt[0].type!=pt_magic) {
        partition_set(0,0,0,0);
    }

    printf("Flash partition table (index,type,offset,size)\n");
    for(i=0;i<pt_size;i++) {

        if(__bd->pt[i].type!=pt_unused)
            printf("%2d, 0x%-8x, 0x%-8x, 0x%-8x : ", i, __bd->pt[i].type, __bd->pt[i].offset, __bd->pt[i].size);
        switch(__bd->pt[i].type) {
        case pt_magic:
            printf("Magic type number\n");
            break;
        case pt_bootloader:
            printf("Bootloader\n");
            break;
        case pt_bd:
            printf("Partition table\n");
            break;
        case pt_kernel:
            printf("kernel\n");
            break;
        case pt_rootdisk:
            printf("rootdisk\n");
            break;
        case pt_fs:
            printf("filesystem\n");
            break;
        case pt_jffs2:
            printf("filesystem (JFFS2)\n");
            break;
        case pt_unused:
            break;
        default:
            printf("Unknown\n");            
        }
        
    }

}

/* Load flash partition by using tftp.
   Uses bootp to find own IP address & TFTP Server.
 */

int
flash_tftp(char *fname, int pt, int skip_verify)
{
    int i,j,k,offset;

#if 0
    if(__bd->pt[pt].type!=pt && pt!=pt_bootloader) {
        printf("Partition invalid\n");
        return -1;
    }
#endif

    if(pt!=pt_bootloader) 
        offset=__bd->pt[pt].offset;
    else
        offset=0;

    net_init();

    bootp();

    /* Load file and verify digest */
    i=tftp_get((char *)KERNEL_LOAD_ADDRESS, fname, SERVER_IPADDR);
    if(i<=0) return -1;

    printf("Verifying image checksum\n");
    if (skip_verify)
      {
	printf("Skipping image checksum verification.\n");
      }
    else if(verify((void *)KERNEL_LOAD_ADDRESS,i)) {
        printf("Warning: Image digest verification failed\n");
        if(!hw_passwd_disable()) 
            return -1;
        else
            printf("Passwords & verification disabled with HW switch... will continue\n");
    } else
        printf("Image digest verification successful - writing to flash\n");

    if(i>__bd->pt[pt].size && pt!=pt_bootloader) {
        printf("File too large to fit partition\n");
        return -1;
    }

    if(flash_cutcopypaste(offset, (void *)KERNEL_LOAD_ADDRESS, i)) {
        printf("Flash write failed\n");
        return -1;
    }

#if 0 /* double-check if image is written correctly */
    if(verify((void *)FLASH_START_ADDRESS+offset,i)) 
    {
        printf("Warning: Flash partition digest verification failed after write\n");
        for(j=0;j<i;j++) 
        {
            if(*(char *)(FLASH_START_ADDRESS+offset+j)!=*(char *)(KERNEL_LOAD_ADDRESS+j)) 
            {
                printf("Compare failed at flash_base + 0x%x:\n", j);
                printf("image:\n");                 
                for(k=0;k<16;k++)
                    printf("%02x ",*(char *)(KERNEL_LOAD_ADDRESS+j+k));

                printf("\nflash:");
                for(k=0;k<16;k++)
                    printf("%02x ",*(char *)(FLASH_START_ADDRESS+offset+j+k+i*16));
                printf("\n");
            }
        }

        return -1;
    }
#endif
    return 0;
}

/* Upload flash partition to TFTP server.
   Uses bootp to find own IP address & TFTP Server.
 */

int
flash_to_tftp(char *fname, int pt)
{
  int offset, i;

  if (pt != pt_bootloader) 
    offset = __bd->pt[pt].offset;
  else
    offset = 0;

  net_init();

  bootp();

  i = tftp_put((char *) FLASH_START_ADDRESS + offset, __bd->pt[pt].size,
               fname, SERVER_IPADDR);
  if (i <= 0)
    return -1;

  return 0;
}

int partition_erase(int pt)
{
    int offset, size, type, i, start, end, prev_proc, proc;

    if(pt!=pt_bootloader) 
        offset=__bd->pt[pt].offset;
    else
        offset=0;

    if(offset&(FLASH_BLOCKSIZE-1)) {
        printf("Error: Partition offset must be aligned to flash block size\n");
        return -1;
    }

    size = __bd->pt[pt].size;
    if(size&(FLASH_BLOCKSIZE-1)) {
        printf("Error: Partition size must be aligned to flash block size\n");
        return -1;
    }

    type = __bd->pt[pt].type;

    start = offset / FLASH_BLOCKSIZE;
    end = (offset + size) / FLASH_BLOCKSIZE;
    prev_proc = 0;
    for(i = start; i < end; i++) {
        if(i&1)
            set_leds(SIPO_LED_POWER|SIPO_LED_STATUS_RED);
        else
            set_leds(SIPO_LED_POWER|SIPO_LED_STATUS_WEIRD);

        if ((i - start) % 3 == 0)
          printf(".");
        if (end - start < 1)
          proc = 0;
        else
          proc = (i - start) * 100 / (end - start);
        if (proc >= prev_proc + 10)
          {
            prev_proc = (proc / 10) * 10;
            printf(".%d%%", prev_proc);
          }

        flash_erase_block(i);
	if (type == pt_jffs2)
	  flash_program_block_jffs2_hdr(i);
    }
    printf(".100%\n");

    set_leds(SIPO_LED_POWER|SIPO_LED_STATUS_GREEN);    

    return 0;
}
