/*******************************************************************************

  Dallas 1305 RTC/NVRAM interface
  -------------------------------

  Copyright (C) 2001 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved. 

  Routines to read/write data to the Dallas RTC/NVRAM chip

  Original Linux driver by Marko Teiste <mte@ssh.com>.
  Modified to work in ssh_boot by Arto Vuori <avuori@ssh.com>

*******************************************************************************/

#include "sipo.h"

/*
 *  OK. Where are trying to emulate the SPI by wiggling the pins directly
 *  Why? Because this code is only used to set/get data from rtc and going
 *  thru the interrupt hassle does not seem to be worth the trouble.
 *
 *  The thing is wired to port D pin
 *
 *      16      SPIMOSI (output to RTC)
 *      17      SPIMOSO (input from RTC)
 *      18      SPICLK
 *      19      SPISEL
 */

#define SIPO_RTC_DIR IMMR->io_regs[3].pdir
#define SIPO_RTC_DAT IMMR->io_regs[3].pdat

#define SIPO_RTC_OUT (0x80000000 >> 16)
#define SIPO_RTC_IN  (0x80000000 >> 17)
#define SIPO_RTC_CLK (0x80000000 >> 18)
#define SIPO_RTC_SEL (0x80000000 >> 19)
#define SIPO_RTC_ALL (SIPO_RTC_IN | SIPO_RTC_OUT | SIPO_RTC_CLK | SIPO_RTC_SEL)

/*  Help functions for toggling the I/O pins
 */
static int
pin_state(unsigned int pin)
{
    return (0 != (SIPO_RTC_DAT & pin));
}

static void
pin_up(unsigned int pin)
{
    SIPO_RTC_DAT |= pin;
}

static void
pin_down(unsigned int pin)
{
    SIPO_RTC_DAT &= ~pin;
}

static void
pin_flip(unsigned int pin)
{
    SIPO_RTC_DAT ^= pin;
}


/*  Initialize the pins we use to communicate with the rtc 
 *  e.g. mark them general and set the directions
 */
static void
pins_init(void)
{
    IMMR->io_regs[3].ppar &= ~SIPO_RTC_ALL;
    IMMR->io_regs[3].psor &= ~SIPO_RTC_ALL;
    IMMR->io_regs[3].podr &= ~SIPO_RTC_ALL;
    IMMR->io_regs[3].pdat &= ~SIPO_RTC_ALL;
    IMMR->io_regs[3].pdir |= SIPO_RTC_OUT | SIPO_RTC_CLK | SIPO_RTC_SEL;
    IMMR->io_regs[3].pdir &= ~SIPO_RTC_IN;
}

/* Output one byte to rtc
 */
static void
rtc_out(unsigned char c)
{
        unsigned char mask = 0x80;
        int i;

        for(i = 0; i < 8; i++, mask = (mask >> 1)) {
                pin_flip(SIPO_RTC_CLK);
                if (c & mask) 
                        pin_up(SIPO_RTC_OUT);
                else 
                        pin_down(SIPO_RTC_OUT);
                udelay(1);
                pin_flip(SIPO_RTC_CLK);
                udelay(1);
        }
}

/* Get one byte from rtc
 */
static unsigned char
rtc_in(void)
{
        unsigned char mask = 0x80;
        unsigned char c = 0;
        int i;

        for(i = 0; i < 8; i++, mask = (mask >> 1)) {
                pin_flip(SIPO_RTC_CLK);
                udelay(1);
                pin_flip(SIPO_RTC_CLK);
                udelay(1);
                if (pin_state(SIPO_RTC_IN))
                        c |= mask;
        }

        return c;
}

/* Write range of bytes to the RTC chip
 */
void 
rtc_write(unsigned int address, unsigned int len, unsigned char *buffer)
{
    int i;

    address+=0x80;              /* Write area offset */

    pin_down(SIPO_RTC_CLK);
    pin_up(SIPO_RTC_SEL);
    udelay(1);
    rtc_out (address & 0xff);
    for(i = 0; i < len; i++) {
        rtc_out(buffer[i]);
    }
    pin_down(SIPO_RTC_SEL);
}

/* Read range of bytes from the RTC chip
 */
void 
rtc_read(unsigned int address, unsigned int len, unsigned char *buffer)
{
    int i;

    pin_down(SIPO_RTC_CLK);
    pin_up(SIPO_RTC_SEL);
    udelay(1);
    rtc_out(address & 0xff);
    for(i = 0; i < len; i++) {
        buffer[i] = rtc_in();
    }
    pin_down(SIPO_RTC_SEL);
}

/* External interface functions
 */
void
rtc_init()
{
        char buffer[4];

        pins_init();
        pin_down(SIPO_RTC_SEL);

        /* Enable oscilator, Enable write */
        buffer[0] = 0x00;
        rtc_write(0x0f, 1, buffer);

        /* Tickle charge */
        buffer[0] = 0xa5;
        rtc_write(0x11, 1, buffer);

}

/*  Dallas 1305 stores the time as two BCD digits in 8bit registers 0 to 6.
 *      0       seconds         (00-59)
 *      1       minutes         (00-59)
 *      2       hours           (00-23)
 *      3       day             (01-07)   ignored
 *      4       day of month    (01-31)
 *      5       month           (01-12)
 *      6       year            (00-99)
 */
void
print_rtc_time()
{
        unsigned int year, mon, day, hour, min, sec;
        unsigned char buffer[7];

        rtc_read(0x00, 7, buffer);

        sec =  (((buffer[0] >> 4) & 0x0f) * 10) + (buffer[0] & 0x0f);
        min =  (((buffer[1] >> 4) & 0x0f) * 10) + (buffer[1] & 0x0f);
        hour = (((buffer[2] >> 4) & 0x0f) * 10) + (buffer[2] & 0x0f);
        day =  (((buffer[4] >> 4) & 0x0f) * 10) + (buffer[4] & 0x0f);
        mon =  (((buffer[5] >> 4) & 0x0f) * 10) + (buffer[5] & 0x0f);
        year = (((buffer[6] >> 4) & 0x0f) * 10) + (buffer[6] & 0x0f) + 2000;

        printf("date = %d/%d/%d %02d:%02d:%02d\n", year,mon,day,hour,min,sec);
}

int
set_rtc_time(int sec, int min, int hour, int day, int mon, int year)
{
        unsigned char buffer[18];

        if(sec<0 || sec>=60 ||
           min<0 || min>=60 ||
           hour<0 || hour>=24 ||
           day<0 || day>31 ||
           mon<0 || mon>12 ||
           year<2000 || year>2100) 
        {
            printf("set_rtc_time: Date is out of range\n");
            return -1;
        }

        /* Create and write the time on chip */
        buffer[0] = (((sec / 10) << 4) & 0xf0) 
                    | ((sec % 10) & 0x0f);
        buffer[1] = (((min / 10) << 4) & 0xf0) 
                    | ((min % 10) & 0x0f);
        buffer[2] = (((hour / 10) << 4) & 0xf0) 
                    | ((hour % 10) & 0x0f);
        buffer[3] = 1;
        buffer[4] = (((day / 10) << 4) & 0xf0) 
                    | ((day % 10) & 0x0f);
        buffer[5] = ((((mon) / 10) << 4) & 0xf0) 
                    | (((mon) % 10) & 0x0f);
        buffer[6] = ((((year % 100) / 10) << 4) & 0xf0) 
                    | (((year % 100) % 10) & 0x0f);

        rtc_write(0x00, 7, buffer);

        return 0;
}

/* Write to NVRAM */
int nv_write(unsigned int offset, unsigned int len, unsigned char *buffer)
{
    if(offset>0x60) {
        printf("nv_write: Offset is out of range\n");
        return 0;
    }

    if(len>0x60) {
        printf("nv_write: length is out of range\n");
        return 0;
    }

    rtc_write(offset+0x20,len,buffer);

    return len;
}

/* Read from NVRAM */
int nv_read(unsigned int offset, unsigned int len, unsigned char *buffer)
{
    if(offset>0x60) {
        printf("nv_read: Offset is out of range\n");
        return 0;
    }

    if(len>0x60) {
        printf("nv_read: length is out of range\n");
        return 0;
    }

    rtc_read(offset+0x20,len,buffer);
    
    return len;
}

