/*******************************************************************************

    Sipo SDRAM initialization
    --------------------

  Author: Arto Vuori <avuori@ssh.com>

  Copyright (C) 2000 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved. 

  Initialize and Test SDRAM. Read SPD PROM.

*******************************************************************************/


#include "sipo.h"

/* 
 * SPD PROM structure 
 */

typedef unsigned char byte;

typedef struct {
    byte        num_bytes;
    byte        num_size;
    byte        type;
    byte        rows;
    byte        cols;
    byte        mod_rows;
    byte        width;
    byte        width_hi;
    byte        signal_levels;
    byte        cycle_time;
    byte        cas_max;
    byte        ecc_type;
    byte        refresh_type;
    byte        chip_width;
    byte        ecc_width;
    byte        col_addr_clocks;
    byte        burst_types;
    byte        banks;
    byte        cas_latency;
    byte        cs_latency;
    byte        we_latency;
    byte        module_attributes;
    byte        chip_attributes;
    byte        cycle_time2;
    byte        cas_max2;
    byte        cycle_time3;
    byte        cas_max3;
    byte        min_pre_to_act;
    byte        min_row_to_row;
    byte        min_ras_to_cas;
    byte        min_ras_width;
    byte        row_density;
    byte        _res[255];                      /* byte 32 ... */
} t_spd;

#define SPD_DIR IMMR->io_regs[3].pdir
#define SPD_DAT IMMR->io_regs[3].pdat
#define SPD_SCL (0x80000000>>14)
#define SPD_SDA (0x80000000>>15)

#define SPD_DELAY udelay(10);

/*
 * Send one bit 
 */
void spd_bitw(int b) 
{
    SPD_DELAY;
    if(b) 
        SPD_DAT|= SPD_SDA;
    else
        SPD_DAT&=~SPD_SDA;
    SPD_DELAY;
    SPD_DAT|= SPD_SCL;
    SPD_DELAY;
    SPD_DAT&=~SPD_SCL;
}

/*
 * Read one bit 
 */
int spd_bitr() {
    int b;
    
    SPD_DIR&=~SPD_SDA;
    SPD_DELAY;
    SPD_DAT|= SPD_SCL;
    SPD_DELAY;
    if(SPD_DAT&SPD_SDA)
        b=1;
    else 
        b=0;
    SPD_DAT&=~SPD_SCL;
    SPD_DIR|= SPD_SDA;

    return b;
}

/*
 * Read one byte from any address 
 */
int spd_read(int addr) {
    int i,j=-1;

    /* Start condition */
    SPD_DELAY;    
    SPD_DAT&=~SPD_SDA;
    SPD_DELAY;
    SPD_DAT&=~SPD_SCL;

    spd_bitw(1);        /* Preamble */
    spd_bitw(0);
    spd_bitw(1);
    spd_bitw(0);

    spd_bitw(1);        /* Addr */
    spd_bitw(0);
    spd_bitw(0);

    spd_bitw(0);        /* Write Addr */

    if(spd_bitr())
        goto error;;

    for(i=0;i<8;i++)
        spd_bitw((addr>>(7-i))&1);

    if(spd_bitr())
        goto error;

    /* Start condition again*/
    SPD_DELAY;    
    SPD_DAT|= SPD_SDA;
    SPD_DELAY
    SPD_DAT|= SPD_SCL;
    SPD_DELAY;
    SPD_DAT&=~SPD_SDA;
    SPD_DELAY;
    SPD_DAT&=~SPD_SCL;

    spd_bitw(1);        /* Preamble */
    spd_bitw(0);
    spd_bitw(1);
    spd_bitw(0);

    spd_bitw(1);        /* Addr */
    spd_bitw(0);
    spd_bitw(0);

    spd_bitw(1);        /* Read */

    if(spd_bitr())
        goto error;

    for(i=j=0;i<8;i++)
        j=(j<<1)+spd_bitr();
error:
    /* Stop condition */
    SPD_DELAY;    
    SPD_DAT&=~SPD_SDA;
    SPD_DELAY;    
    SPD_DAT|= SPD_SCL;
    SPD_DELAY;
    SPD_DAT|= SPD_SDA;

    return j;
}

/*
 * Initialize IO pins 
 */
int spd_init(void) {
    int i,c;
    
    /* clock OUT */
    IMMR->io_regs[3].pdat |= (0x80000000>>14);
    IMMR->io_regs[3].ppar &=~(0x80000000>>14);
    IMMR->io_regs[3].pdir |= (0x80000000>>14);
    IMMR->io_regs[3].podr &=~(0x80000000>>14);
    IMMR->io_regs[3].psor &=~(0x80000000>>14);

    /* Data IN/OUT */
    IMMR->io_regs[3].pdat |= (0x80000000>>15);
    IMMR->io_regs[3].ppar &=~(0x80000000>>15);
    IMMR->io_regs[3].pdir |= (0x80000000>>15);
    IMMR->io_regs[3].podr &=~(0x80000000>>15);
    IMMR->io_regs[3].psor &=~(0x80000000>>15);

    /* Try to read the first SPD byte maximum of 3 times */
    for(i=c=0;i<3 && c>=0;i++)
        c=spd_read(0);  

    return c;
}

void parse(t_spd *spd) 
{
    printf("%d bytes in SPD structure\n", spd->num_bytes);
    printf("%d PROM total size\n", 1<<spd->num_size);

    switch(spd->type) {
    case 0x02: printf("memory type is EDO\n"); break;
    case 0x04: printf("memory type is SDRAM\n"); break;
    default: printf("Unknown type\n");
    }

    printf("%d row bits\n", spd->rows);
    printf("%d column bits\n", spd->cols);
    printf("%d module rows\n", spd->mod_rows);
    printf("%d bits width\n", spd->width);

    switch(spd->signal_levels) {
    case 0x00: printf("5v TTL\n"); break;
    case 0x01: printf("LVTTL\n"); break;
    case 0x02: printf("HSTL 1.5\n"); break;
    case 0x03: printf("SSTL 3.3\n"); break;
    case 0x04: printf("SSTL 2.5\n"); break;
    default: printf("Unknown signal levels\n");
    }

    printf("%d.%dns min cycle time\n", spd->cycle_time>>4, spd->cycle_time&0x0f);
    printf("%d.%dns max clock to data\n", spd->cas_max>>4, spd->cas_max&0x0f);

    switch(spd->ecc_type) {
    case 0x00: printf("No error codes\n"); break;
    case 0x01: printf("Parity\n"); break;
    case 0x02: printf("ECC\n"); break;
    default: printf("Unknown error correction\n");
    }

    switch(spd->refresh_type&0x7f) {
    case 0x00: printf("Normal refresh\n"); break;
    case 0x01: printf("Reduced 0.25x refresh\n"); break;
    case 0x02: printf("Reduced 0.5x refresh\n"); break;
    case 0x03: printf("Extended 2x refresh\n"); break;
    case 0x04: printf("Extended 4x refresh\n"); break;
    case 0x05: printf("Extended 8x refresh\n"); break;
    default: printf("Unknown refresh\n");
    }

    if(spd->refresh_type&0x80)
        printf("Self refesh\n");


    printf("%d bits primary components width\n", spd->chip_width);
    printf("%d bits ecc width\n", spd->ecc_width);

    printf("%d clocks delay for random column address\n");

    printf("{ ");
    if(spd->burst_types&0x80) printf("Page ");
    if(spd->burst_types&0x08) printf("8 ");
    if(spd->burst_types&0x04) printf("4 ");
    if(spd->burst_types&0x02) printf("2 ");
    if(spd->burst_types&0x02) printf("1 ");
    printf("} are supported burst types\n");

    printf("%d banks on device\n", spd->banks);

    printf("{ ");
    if(spd->cas_latency&0x80) printf("TBD ");
    if(spd->cas_latency&0x40) printf("7 ");
    if(spd->cas_latency&0x20) printf("6 ");
    if(spd->cas_latency&0x10) printf("5 ");
    if(spd->cas_latency&0x08) printf("4 ");
    if(spd->cas_latency&0x04) printf("3 ");
    if(spd->cas_latency&0x02) printf("2 ");
    if(spd->cas_latency&0x02) printf("1 ");
    printf("} are supported CAS latencies\n");

    printf("{ ");
    if(spd->cs_latency&0x80) printf("TBD ");
    if(spd->cs_latency&0x40) printf("7 ");
    if(spd->cs_latency&0x20) printf("6 ");
    if(spd->cs_latency&0x10) printf("5 ");
    if(spd->cs_latency&0x08) printf("4 ");
    if(spd->cs_latency&0x04) printf("3 ");
    if(spd->cs_latency&0x02) printf("2 ");
    if(spd->cs_latency&0x02) printf("1 ");
    printf("} are supported CS latencies\n");

    printf("{ ");
    if(spd->we_latency&0x80) printf("TBD ");
    if(spd->we_latency&0x40) printf("7 ");
    if(spd->we_latency&0x20) printf("6 ");
    if(spd->we_latency&0x10) printf("5 ");
    if(spd->we_latency&0x08) printf("4 ");
    if(spd->we_latency&0x04) printf("3 ");
    if(spd->we_latency&0x02) printf("2 ");
    if(spd->we_latency&0x02) printf("1 ");
    printf("} are supported WE latencies\n");

    /* XXX module & device attributes */

    printf("%d.%dns min 2nd cycle time\n", spd->cycle_time2>>4, spd->cycle_time2&0x0f);
    printf("%d.%dns max 2nd clock to data\n", spd->cas_max2>>4, spd->cas_max2&0x0f);
    printf("%d.%dns min 3rd cycle time\n", spd->cycle_time3>>4, spd->cycle_time3&0x0f);
    printf("%d.%dns max 3rd clock to data\n", spd->cas_max3>>4, spd->cas_max3&0x0f);

    printf("%dns Mininmum row precharge time\n", spd->min_pre_to_act);
    printf("%dns Minimum row to row time\n", spd->min_row_to_row);
    printf("%dns Mininmum RAS to CAS\n", spd->min_ras_to_cas);
    printf("%dns Mininmum RAS pulse width\n", spd->min_ras_width);

    /* ROW density etc... */
}

/* 
 * Test SPD interface 
 */
int spd_test(char *cmd) 
{
    unsigned char buf[0x100];
    t_spd       *spd=(t_spd *)buf;
    int         i,j,c;

    if((c=spd_init())<=0) {
        printf("spd_test: Unable to communicate with SPD\n");
        return -1;
    }

    printf("SPD dump:\n");

    for(i=0;i<c && i<0x100;i++) {
        buf[i]=j=spd_read(i);
        
        if((i&0xf)==0) 
            printf("0x%02x: ", i);

        printf("%02x ", j);

        if((i&0xf)==0x07) 
            printf("  ");
        else if((i&0xf)==0x0f || i+1==c) 
            printf("\n");
    }
    printf("done. (%d bytes read)\n",c);

    parse(spd);

    return 0;
}

/* a really lousy (and slow) pseudo random hash function 
 */
unsigned pattern(unsigned int s) {
    unsigned j=0xf00ba37e,i;

    j+=s;

    for(i=0;i<64;i++)
    {
        if((s>>(i&31))&1) 
            j=(j<<1)|(((j>>1)^(j>>7)^(j>>11)^(j>>17))&1);       /* LSFR1 */
        else  
            j=(j<<1)|(((j>>3)^(j>>5)^(j>>13)^(j>>21))&1);       /* LSFR2 */
    }

    return j;
}

/* Read & Write test pattern to exponential addresses 
*/
int memtest(volatile unsigned *base, int size) 
{
    int i,j,t;

    for(i=0;i<size/sizeof(unsigned);i=(i==0)?1:i*2)
        base[i]=pattern(i+j);

    for(i=0;i<size/sizeof(unsigned);i=(i==0)?1:i*2)
        if(base[i]!=pattern(i+j))
            return 1;                           /* Error */

    for(i=0;i<size/sizeof(unsigned);i=(i==0)?1:i*2)
        base[i]=~pattern(i+j);

    for(i=0;i<size/sizeof(unsigned);i=(i==0)?1:i*2)
        if(base[i]!=~pattern(i+j))
            return 1;                           /* Error */

    return 0;

}

/* SDRAM controller (default) settings */

#define SDRAM_OR (0xFC000000+(SDRAM_BPD<<13)+(SDRAM_ROWST<<9)+(SDRAM_NUMR<<6)+(SDRAM_IBID<<4))
#define SDRAM_BR 0x00000041     
#define SDRAM_BR_BOOT 0x10000041

#define SDRAM_OP_NORMAL 0x40000000
#define SDRAM_OP_PRECHARGE 0x28000000
#define SDRAM_OP_CBRREFR 0x08000000
#define SDRAM_OP_MODEWR 0x18000000
        
/* BPD (Banks Per Device) = 00 = 2 internal banks per device. */
#define SDRAM_BPD 1
/* ROWST (Row Start Address Bit) = 011 = A9 being the row start address bit */
#define SDRAM_ROWST 6
/* NUMR (Number of Row Address Lines) = 010 = 11 row address lines. */
#define SDRAM_NUMR 3
/* IBID (Internal bank interleaving within same device disable) */
#define SDRAM_IBID 1

#define SDRAM_PBI 1
#define SDRAM_SDAM 3    
#define SDRAM_RFRC 6
#define SDRAM_BSMA 2
#define SDRAM_SDA10 4
#define SDRAM_PRETOACT 1
#define SDRAM_ACTTORW 1
#define SDRAM_LDOTOPRE 2
#define SDRAM_WRC 1
#define SDRAM_EAMUX 1
#define SDRAM_CL 2
#define SD_PSDMR_HI (SDRAM_PBI<<15)+(SDRAM_SDAM<<8)+(SDRAM_BSMA<<5)+(SDRAM_SDA10<<2)+(SDRAM_RFRC>>1)
#define SD_PSDMR_LO ((SDRAM_RFRC & 1)<<15)+(SDRAM_PRETOACT<<12)+(SDRAM_LDOTOPRE<<6)+(SDRAM_WRC<<4)+(SDRAM_EAMUX<<3)+(SDRAM_ACTTORW<<9)+SDRAM_CL
#define SD_PSDMR ((SD_PSDMR_HI<<16)|SD_PSDMR_LO)

#define SDRAM_BURSTLEN 2

#define BUS_CYCLE_TIME  15      /* ns */

/* 
 * SDRAM controller initialization.
 *
 * Read data from SPD PROM. Parse information and check for DIMM compatibility.
 * information stored in SPD is used to determine some timings for memory 
 * controller. Critical settings are checked to be in correct range.
 * If SPD is not found, default settings will be tried. Also if memory test
 * fails after 'intelligent' configuration, default settings will be tried.
 * If memory test passes, return value will be zero.
 */
int sdram(void)
{
    unsigned char buf[0x100];
    t_spd       *spd=(t_spd *)buf;
    int         i,j,c,defaults=0;
    unsigned    int sd_psdmr=(((SDRAM_PBI<<15)+(SDRAM_SDAM<<8)+(SDRAM_BSMA<<5)+(SDRAM_SDA10<<2)+(SDRAM_RFRC>>1) << 16) \
                              +((SDRAM_RFRC & 1)<<15)+(SDRAM_EAMUX<<3));

    unsigned    int sd_pretoact=SDRAM_PRETOACT;
    unsigned    int sd_ldotopre=SDRAM_LDOTOPRE;
    unsigned    int sd_wrc=SDRAM_WRC;
    unsigned    int sd_acttorw=SDRAM_ACTTORW;
    unsigned    int sd_cl=SDRAM_CL;

    if((c=spd_init())<=0) 
    {
        printf("ERROR: Unable to communicate with SPD. No SDRAM present or type unknown.\n");

        defaults=1;
    }
    else
    {
        for(i=0;i<c && i<0x100;i++)
            buf[i]=j=spd_read(i);

        if(spd->type!=0x04) {
            printf("Memory type must be SDRAM\n"); defaults=1; }
        if(spd->rows!=0x0c) {
            printf("Only 12 row bits configuration supported"); defaults=1; }
        if(spd->cols!=9) {
            printf("Only 9 column bits configuration supported"); defaults=1; }
        if(spd->banks!=4) {
            printf("Only 4 banks configuration supported"); defaults=1; }
        if(spd->width!=64) {
            printf("Only 64 data bits configuration supported\n"); defaults=1; }
        if(spd->signal_levels!=0x01) {
            printf("Only LVTTL signal levels supported\n"); defaults=1; }
        if(spd->cycle_time>0xa0) {
            printf("Cycle time should be maximum of 10ns\n"); defaults=1; }
        if(spd->cas_max>0x60) {
            printf("Clock to data should be maximum of 6ns\n"); defaults=1; }
        if(!(spd->cas_latency&0x06)) {
            printf("Only CAS latency of 2 or 3 cycles is supported\n"); defaults=1; }
        if(!(spd->burst_types&0x04)) {
            printf("Only burst lenght of 4 is supported\n"); defaults=1; }

        if(spd->cas_latency&0x02)
            sd_cl=2;
        else if(spd->cas_latency&0x04)
            sd_cl=3;

        sd_ldotopre=4-sd_cl;            

        sd_pretoact=(spd->min_pre_to_act+BUS_CYCLE_TIME-1)/BUS_CYCLE_TIME-1;
        if(sd_pretoact<1) 
            sd_pretoact=1;
        else if(sd_pretoact>7) {
            printf("Min clocks from precharge to activete is out of range\n"); defaults=1; }

        sd_acttorw=(spd->min_ras_to_cas+BUS_CYCLE_TIME-1)/BUS_CYCLE_TIME-1;
        if(sd_acttorw<1) 
            sd_acttorw=1;
        else if(sd_acttorw>7) {
            printf("Min clocks from active to read/write is out of range\n"); defaults=1; }

        sd_wrc=(spd->min_row_to_row+BUS_CYCLE_TIME-1)/BUS_CYCLE_TIME-1;
        if(sd_wrc<1) 
            sd_wrc=1;
        else if(sd_wrc>3) {
            printf("Min clocks for write recovery is out of range\n"); defaults=1; }
    }

    if(!defaults) 
    {
        sd_psdmr |= (sd_pretoact<<12)+(sd_ldotopre<<6)+(sd_wrc<<4)+(sd_acttorw<<9)+sd_cl;

        IMMR->memc_regs[1].or=SDRAM_OR;
        IMMR->memc_regs[1].br=SDRAM_BR_BOOT;

        IMMR->memc_mptpr = 0x3200;
        IMMR->memc_psrt = 0x0e;
        IMMR->memc_psdmr = sd_psdmr + SDRAM_OP_PRECHARGE;

        ((char *)(SDRAM_BR_BOOT&0xffff0000))[0]=0;

        IMMR->memc_psdmr = sd_psdmr + SDRAM_OP_CBRREFR;

        ((char *)(SDRAM_BR_BOOT&0xffff0000))[0]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[1]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[2]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[3]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[4]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[5]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[6]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[7]=0;

        IMMR->memc_psdmr = sd_psdmr + SDRAM_OP_MODEWR;

        ((char *)(SDRAM_BR_BOOT&0xffff0000))[((sd_cl<<4)+ SDRAM_BURSTLEN)<<3]=0xff;
        
        IMMR->memc_psdmr = sd_psdmr + SDRAM_OP_NORMAL;

        /* Test if these settiongs work... use default values if test fails */
        if(memtest((unsigned *)(SDRAM_BR_BOOT&0xffff0000), 0x400000)) {
            printf("Automagic SDRAM initialization failed... falling back to default values\n");
            defaults = 1;
        }
        else
            return 0;
    }

    if(defaults) 
    {
        /* Use default values */

        IMMR->memc_regs[1].or=SDRAM_OR;
        IMMR->memc_regs[1].br=SDRAM_BR_BOOT;

        IMMR->memc_mptpr = 0x3200;
        IMMR->memc_psrt = 0x0e;
        IMMR->memc_psdmr = SD_PSDMR + SDRAM_OP_PRECHARGE;

        ((char *)(SDRAM_BR_BOOT&0xffff0000))[0]=0;

        IMMR->memc_psdmr = SD_PSDMR + SDRAM_OP_CBRREFR;

        ((char *)(SDRAM_BR_BOOT&0xffff0000))[0]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[1]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[2]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[3]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[4]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[5]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[6]=0;
        ((char *)(SDRAM_BR_BOOT&0xffff0000))[7]=0;

        IMMR->memc_psdmr = SD_PSDMR + SDRAM_OP_MODEWR;

        ((char *)(SDRAM_BR_BOOT&0xffff0000))[((SDRAM_CL<<4)+ SDRAM_BURSTLEN)<<3]=0xff;
        
        IMMR->memc_psdmr = SD_PSDMR + SDRAM_OP_NORMAL;
    }

    return memtest((unsigned *)(SDRAM_BR_BOOT&0xffff0000), 0x400000);

}

