/*******************************************************************************

  SMC/SCC Serial Console interface
  ----------------------------

  Author: Arto Vuori <avuori@ssh.com>

  Copyright (C) 2000 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved. 

  Simple polling interface for MPC8260 SMC/SCC serial controllers.
  Adapted from SMC example from motorola.

*******************************************************************************/

#include "sipo.h"

#define PTR2IMMROFFSET(x) ((unsigned int)(x) - (unsigned int)(IMMR))

/* Single buffer component */
typedef struct BufferPool
{
   UBYTE  RxBuffer;
   UBYTE  TxBuffer;
} LB;

typedef bufferdesc BD;

/* Buffer Descriptor Ring format */
typedef struct BufferDescRings 
{
    BD RxBD;    /* Rx BD ring */
    BD TxBD;    /* Tx BD ring */

} BDRINGS;

/* Serial interface structure */

typedef enum  { SMC, SCC } serial_type;

typedef struct {
    serial_type type;   /* controller type */
    int num;            /* controller numberk SMC1 = 0, SCC1=0*/
    char *name;

    void *pram;
    
    BDRINGS *RxTxBD;    /* buffer descriptors base pointer */
    LB *Buffers;        /* data Buffers */
} serial_if;

serial_if serial_interfaces[]={
    { SMC, 0, "ttyS0 @ SMC1", (void *)IMMR+0x0000, (BDRINGS *)(((BYTE *)IMMR)+0xc0), (LB *)(((BYTE *)IMMR)+0xd0)},
    { SCC, 1, "ttyS1 @ SCC2", (void *)IMMR+0x0000, (BDRINGS *)(((BYTE *)IMMR)+0x40), (LB *)(((BYTE *)IMMR)+0x50)},
    { SCC, 2, "ttyS2 @ SCC3", (void *)IMMR+0x0000, (BDRINGS *)(((BYTE *)IMMR)+0x40), (LB *)(((BYTE *)IMMR)+0x50)},
    { -1}
};

/* Default serial console */
static serial_if *tty=&serial_interfaces[1];

/*
 * Set Baudrate generator
 */

#define CPM_BRG_EN      ((UWORD)0x00010000)

void
setbrg(int brg, int rate)
{
        VUWORD   *bp;

        if(brg<5) 
          bp = (&IMMR->brgs_brgc1)+brg-1;
        else
          bp = (&IMMR->brgs_brgc5)+brg-5;

#if 0
        if(__bd->magic==SIPO_MAGIC)
            *bp = ((__bd->est.bi_brgfreq/16/ __bd->est.bi_baudrate) << 1) | CPM_BRG_EN;
        else
#endif
            *bp = ((__BD_brgfreq/16/ __BD_baudrate) << 1) | CPM_BRG_EN;
}

/* Initialize serial port on mpc8260  */

void serial_init(void)
{   
    int i;
    t_Smc_Pram *smcp;
    t_Scc_Pram *sccp;
    t_UartScc_Pram *sccu;

    if(tty->type==SMC) 
    {
        switch(tty->num) 
        {
        case 0: /* SMC1 */
            /* Initialize IO pins. Allow SMC TX, RX out */
            IMMR->io_regs[3].ppar |= 0x00C00000;     
            IMMR->io_regs[3].pdir |= 0x00400000;
            IMMR->io_regs[3].pdir &= 0xFF7FFFFF;
            IMMR->io_regs[3].psor &= 0xFF3FFFFF;
   
            /* SMC1 uses BRG1 */
            setbrg(1,9600);
            IMMR->cpm_mux_cmxsmr = (IMMR->cpm_mux_cmxsmr & ~0xb0) | 0x00;

            /* set SMC Pram start */
            *(VUHWORD *)(&(IMMR->pram.standard.smc1)) = PTR2IMMROFFSET(tty->pram);
        }
        smcp=tty->pram;

        /* Set RXBD table start */
        smcp->rbase = PTR2IMMROFFSET(&tty->RxTxBD->RxBD);

        /* Set TXBD table start */
        smcp->tbase = PTR2IMMROFFSET(&tty->RxTxBD->TxBD);

        /* Initialize Rx and Tx Params for SMC: */
        /* Spin until cpcr flag is cleared             */
        for(IMMR->cpm_cpcr = 0x1d010000; IMMR->cpm_cpcr & 0x00010000;) ;
   
        /* Set RFCR,TFCR -- Rx,Tx Function Code */
        /* Normal Operation and Motorola byte   */
        /* ordering                             */
        smcp->rfcr = 0x10;
        smcp->tfcr = 0x10;

        /* MRBLR = MAX buffer length */
        smcp->mrblr = 1;

        /* MAX_IDL = Disable MAX Idle Feature */
        smcp->max_idl = 0;  

        /* BRKEC = No break condition occurred */
        smcp->brkec = 0;    

        /* BRKCR = 1 break char sent on top XMIT */
        smcp->brkcr = 1;    

        /* Initialize the BDs */

        /* Setup Receiver Buffer Descriptors */
        tty->RxTxBD->RxBD.bd_cstatus = 0xA000;            /* Enable, Last BD */
        tty->RxTxBD->RxBD.bd_length = 1;
        tty->RxTxBD->RxBD.bd_addr = &(tty->Buffers->RxBuffer);

        /* Setup Transmitter Buffer Descriptors */
        tty->RxTxBD->TxBD.bd_cstatus = 0x2000;      /* Buffer not yet ready; Last BD */
        tty->RxTxBD->TxBD.bd_length = 1;
        tty->RxTxBD->TxBD.bd_addr = &(tty->Buffers->TxBuffer);

        /* Clear any pending events */
        IMMR->smc_regs[tty->num].smce = 0xFF; 

        /* SMC_SMCM = Mask all interrupts, use polling mode */
        IMMR->smc_regs[tty->num].smcm = 0;     

        /* 8-bit mode,  no parity, 1 stop-bit */
        /* UART SMC Mode                                                       */
        /* Normal operation (no loopback),     */
        /* SMC Transmitter/Receiver Enabled    */
        IMMR->smc_regs[tty->num].smcmr = 0x4820;

        /* Enable receive and transmit */
        IMMR->smc_regs[tty->num].smcmr = 0x4823;
    } 
    else if(tty->type==SCC)
    {
        switch(tty->num) 
        {
        case 1: /* SCC2 */
#if 0 /* Flow control */
            /* Initialize IO pins. If you also need DCD then use 0x000c0000.*/
            IMMR->io_regs[2].ppar = (IMMR->io_regs[2].ppar & ~0x00040000) | 0x00040000;
            IMMR->io_regs[2].psor = (IMMR->io_regs[2].psor & ~0x00040000) | 0x00000000;
            IMMR->io_regs[2].pdir = (IMMR->io_regs[2].pdir & ~0x00040000) | 0x00000000;
#endif

            IMMR->io_regs[3].ppar = (IMMR->io_regs[3].ppar & ~0x00000018) | 0x00000018;
            IMMR->io_regs[3].psor = (IMMR->io_regs[3].psor & ~0x00000018) | 0x00000000;
            IMMR->io_regs[3].pdir = (IMMR->io_regs[3].pdir & ~0x00000038) | 0x00000030;
            IMMR->io_regs[3].pdat = (IMMR->io_regs[3].pdat & ~0x00000020); /* set RTS */
            /* SCC2 uses BRG 3 */
#if 0
            setbrg(3,9600);
            IMMR->cpm_mux_cmxscr = (IMMR->cpm_mux_cmxscr & ~0x00ff0000) | 0x00120000;
#else
            setbrg(1,9600);
            IMMR->cpm_mux_cmxscr = (IMMR->cpm_mux_cmxscr & ~0x0000ff00) | 0x00000000;
#endif
            break;

        case 2: /* SCC3 */
            IMMR->io_regs[3].ppar = (IMMR->io_regs[3].ppar & ~0x000000c0) | 0x000000c0;
            IMMR->io_regs[3].psor = (IMMR->io_regs[3].psor & ~0x000000c0) | 0x00000000;
            IMMR->io_regs[3].pdir = (IMMR->io_regs[3].pdir & ~0x000000c0) | 0x00000080;
   
            /* SCC3 uses BRG 4 */
            setbrg(4,9600);
            IMMR->cpm_mux_cmxscr = (IMMR->cpm_mux_cmxscr & ~0x0000ff00) | 0x00001b00;
            break;
        }

        tty->pram = sccp = &IMMR->pram.serials.scc_pram[tty->num];

        /* Set RXBD table start */
        sccp->rbase = PTR2IMMROFFSET(&tty->RxTxBD->RxBD);

        /* Set TXBD table start */
        sccp->tbase = PTR2IMMROFFSET(&tty->RxTxBD->TxBD);

        /* Initialize Rx and Tx Params for SCC: Spin until cpcr flag is cleared   */
        switch(tty->num) {
        case 1: /* SCC2 */
            for(IMMR->cpm_cpcr = 0x04a10000; IMMR->cpm_cpcr & 0x00010000;) ;
            break;
        case 2: /* SCC3 */
            for(IMMR->cpm_cpcr = 0x08c10000; IMMR->cpm_cpcr & 0x00010000;) ;
            break;
        }
   
        /* Set RFCR,TFCR -- Rx,Tx Function Code */
        /* Normal Operation and Motorola byte   */
        /* ordering & Snooping                            */
        sccp->rfcr = 0x10;
        sccp->tfcr = 0x10;

        /* MRBLR = MAX buffer length */
        sccp->mrblr = 1;

        sccu=&sccp->SpecificProtocol.u;

        /* MAX_IDL = Disable MAX Idle Feature */
        sccu->max_idl = 0;  

        /* BRKCR = 1 break char sent on top XMIT */
        sccu->brkcr = 1;    

        sccu->parec = 0;
        sccu->frmec = 0;
        sccu->nosec = 0;
        sccu->brkec = 0;
        sccu->uaddr1 = 0;
        sccu->uaddr2 = 0;
        sccu->toseq = 0;
        for(i=0;i<8;i++)
            sccu->cc[i] = 0x8000;
        sccu->rccm = 0xc0ff;

        /* Initialize the BDs */

        /* Setup Receiver Buffer Descriptors */
        tty->RxTxBD->RxBD.bd_cstatus = 0xa000;            /* Enable, Last BD */
        tty->RxTxBD->RxBD.bd_length = 1;
        tty->RxTxBD->RxBD.bd_addr = &(tty->Buffers->RxBuffer);

        /* Setup Transmitter Buffer Descriptors */
        tty->RxTxBD->TxBD.bd_cstatus = 0x2000;      /* Buffer not yet ready; Last BD */
        tty->RxTxBD->TxBD.bd_length = 1;
        tty->RxTxBD->TxBD.bd_addr = &(tty->Buffers->TxBuffer);

        /* Clear any pending events */
        IMMR->scc_regs[tty->num].scce = 0xFFFF; 

        /* SMC_SMCM = Mask all interrupts, use polling mode */
        IMMR->scc_regs[tty->num].sccm = 0; 

        /* 8-bit mode,  no parity, 1 stop-bit */
        /* UART Mode                                                       */
        /* Normal operation (no loopback),     */
        IMMR->scc_regs[tty->num].gsmr_h = 0x00000060; 
        IMMR->scc_regs[tty->num].gsmr_l = 0x00028004; 
        IMMR->scc_regs[tty->num].psmr = 0x3000;

        /* Transmitter/Receiver Enabled    */
        IMMR->scc_regs[tty->num].gsmr_l = 0x00028034;   
    }

#if 0
    printf("%s Initialized\n", tty->name);
#endif
        
}


/* Write single character. */

void serial_putc(UBYTE ch)
{
    /* Loop until transmission completed */
    while (tty->RxTxBD->TxBD.bd_cstatus  &  0x8000);    
   
    /* Store data */
    *(tty->RxTxBD->TxBD.bd_addr) = ch;                  
    tty->RxTxBD->TxBD.bd_length = 1;

    /* Set Ready bit */
    tty->RxTxBD->TxBD.bd_cstatus |= 0x8000;              
}


/* Read single character. Wait if no characters is available. */

UBYTE serial_getc(void)
{
    UBYTE ch;
   
    /* Loop if RxBD empty */
    while (tty->RxTxBD->RxBD.bd_cstatus  &  0x8000);      

    /* Receive data */
    ch = *(tty->RxTxBD->RxBD.bd_addr);   
                   
    /* Set Buffer Empty bit */
    tty->RxTxBD->RxBD.bd_cstatus |= 0x8000;   

    return ch;
}

/* Return next character. Do not remove it from buffer */

UBYTE serial_peekc(void)
{
    UBYTE ch;
   
    /* Loop if RxBD empty */
    while (tty->RxTxBD->RxBD.bd_cstatus  &  0x8000);      

    /* Receive data */
    ch = *(tty->RxTxBD->RxBD.bd_addr);   
                   
    return ch;
}



/* Test if new character is available.
 *
 * returns: 0, if not available. 
 *          1, if is available
 */

int serial_tstc(void)
{      
   if(tty->RxTxBD->RxBD.bd_cstatus & 0x8000) 
       return 0;  /*  character NOT available */
   else
       return 1;  /*  character IS available */
} 


/* Print string and newline.*/

void serial_puts(char *s) 
{
    while(*s!=0)
        serial_putc(*s++);
    serial_putc(0x0a);
    serial_putc(0x0d);
}
