/*
 * ifdhandler for SIPO smart card reader (driver for pcsc-lite)
 * Copyright (C) 2002, Jouni Malinen <jkmaline@cc.hut.fi>
 *
 * This program is free software; you can redistribute them and/or modify
 * it under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation. See COPYING for more details.
 *
 * Note:
 * SIPO has only one smart card slot, so Lun and Channel arguments are ignored.
 */

#include <pcsclite.h>
#include <ifdhandler.h>
#include <fcntl.h>

#if 0
#define DPRINT(...) do { } while (0)
#else
#define DPRINT(args...) printf(args)
#endif


static int sipo_scard_present(void)
{
	int fd;
	char buf;
	fd = open("/proc/sipo/scard", 0);
	if (fd < 0)
		return -1;
	if (read(fd, &buf, 1) != 1) {
		close(fd);
		return -2;
	}
	close(fd);
	if (buf == '1')
		return 1;
	return 0;
}


static struct sipo_rdr_data {
	int fd;
	unsigned char atr[MAX_ATR_SIZE];
	int atr_len;
} sipo_rdr;


static int sipo_scard_open(struct sipo_rdr_data *rdr)
{
	int ret;

	rdr->atr_len = -1;

	if (sipo_scard_present() != 1)
		return -1;

	rdr->fd = open("/dev/misc/siposcard", O_NONBLOCK);
	if (rdr->fd < 0) {
		perror("open");
		return -1;
	}

	ret = read(rdr->fd, rdr->atr, sizeof(rdr->atr));
	if (ret < 0) {
		perror("read");
		close(rdr->fd);
		rdr->fd = -1;
		return -1;
	}

	if (rdr->atr[0] != 0x3b && rdr->atr[0] != 0x3f) {
		DPRINT("ATR TS invalid (0x%02x)\n", rdr->atr[0]);
		close(rdr->fd);
		rdr->fd = -1;
		return -1;
	}

	rdr->atr_len = ret;

	/* FIX: should probably parse ATR (see scard.c for example code) and
	 * could change baudrate and clk_freq */

	return 0;
}


static int sipo_scard_close(struct sipo_rdr_data *rdr)
{
	if (rdr->fd < 0)
		return -1;
	close(rdr->fd);
	rdr->fd = -1;
	return 0;
}


RESPONSECODE IFDHCreateChannel(DWORD Lun, DWORD Channel)
{
 	DPRINT("IFDHCreateChannel(Lun=0x%x, Channel=0x%x)\n", Lun, Channel);

	return IFD_SUCCESS;
}


RESPONSECODE IFDHCloseChannel(DWORD Lun)
{
 	DPRINT("IFDHCloseChannel(Lun=0x%x)\n", Lun);

	if (sipo_scard_close(&sipo_rdr) < 0)
		return IFD_COMMUNICATION_ERROR;

	return IFD_SUCCESS;
}

RESPONSECODE IFDHGetCapabilities(DWORD Lun, DWORD Tag,
				 PDWORD Length, PUCHAR Value)
{
	RESPONSECODE ret;

	DPRINT("IFDHGetCapabilities(Lun=0x%x, Tag=0x%x)\n", Lun, Tag);

	switch (Tag) {
	case TAG_IFD_SLOTS_NUMBER:
		DPRINT("  TAG_IFD_SLOTS_NUMBER\n");
		*Length = 1;
		Value[0] = 1;
		ret = IFD_SUCCESS;
		break;

	default:
		DPRINT("  unsupported Tag\n");
		ret = IFD_NOT_SUPPORTED;
		break;
	}

	return ret;
}

RESPONSECODE IFDHSetCapabilities(DWORD Lun, DWORD Tag,
				 DWORD Length, PUCHAR Value)
{
	DPRINT("IFDHSetCapabilities(Lun=0x%x, Tag=0x%x)\n", Lun, Tag);

	return IFD_NOT_SUPPORTED;
}

RESPONSECODE IFDHSetProtocolParameters(DWORD Lun, DWORD Protocol, UCHAR Flags,
				       UCHAR PTS1, UCHAR PTS2, UCHAR PTS3)
{
	DPRINT("IFDHSetProtocolParameters(Lun=0x%x, Protocol=%d, Flags=0x%x, "
	       "PTS1=%u, PTS2=%u, PTS3=%u\n", Lun, Protocol, Flags, PTS1,
	       PTS2, PTS3);

	return IFD_NOT_SUPPORTED;
}

RESPONSECODE IFDHPowerICC(DWORD Lun, DWORD Action,
			  PUCHAR Atr, PDWORD AtrLength)
{
	RESPONSECODE ret = IFD_SUCCESS;

	DPRINT("IFDHPowerICC(Lun=0x%x, Action=0x%x)\n", Lun, Action);

	switch (Action) {
	case IFD_RESET:
		(void) sipo_scard_close(&sipo_rdr);
		DPRINT("  IFD_RESET\n");
		/* fall through to power UP */

	case IFD_POWER_UP:
		DPRINT("  IFD_POWER_UP\n");
		if (sipo_scard_open(&sipo_rdr) < 0) {
			ret = IFD_COMMUNICATION_ERROR;
			goto done;
		}
		memcpy(Atr, sipo_rdr.atr, sipo_rdr.atr_len);
		*AtrLength = sipo_rdr.atr_len;
		break;

	case IFD_POWER_DOWN:
		DPRINT("  IFD_POWER_DOWN\n");
		(void) sipo_scard_close(&sipo_rdr);
		memset(Atr, 0, MAX_ATR_SIZE);
		*AtrLength = 0;
		break;
	}

 done:
	return ret;
}

RESPONSECODE IFDHTransmitToICC(DWORD Lun, SCARD_IO_HEADER SendPci,
			       PUCHAR TxBuffer, DWORD TxLength,
			       PUCHAR RxBuffer, PDWORD RxLength,
			       PSCARD_IO_HEADER RecvPci)
{
	RESPONSECODE ret = IFD_SUCCESS;
	int len;
	DPRINT("IFDHTransmitToICC(Lun=0x%x, SendPci.Protocol=%d, "
	       "TxLength=%u, RxLength=%u)\n", Lun, SendPci.Protocol, TxLength,
	       *RxLength);

	/* Check that Prot matches with the card */

	if (sipo_rdr.fd < 0) {
		ret = IFD_COMMUNICATION_ERROR;
		goto done;
	}

	if (TxLength < 1) {
		ret = IFD_COMMUNICATION_ERROR;
		goto done;
	}

	len = write(sipo_rdr.fd, TxBuffer, TxLength);
	if (len < 0) {
		ret = IFD_COMMUNICATION_ERROR;
		DPRINT("  write failed; res=%d\n", len);
		goto done;
	}

	if (len != TxLength) {
		ret = IFD_COMMUNICATION_ERROR;
		DPRINT("  partial write; %d/%d\n", len, TxLength);
		/* FIX: try again? non-blocking fd.. */
		goto done;
	}

	len = read(sipo_rdr.fd, RxBuffer, *RxLength);
	if (len < 0) {
		ret = IFD_COMMUNICATION_ERROR;
		DPRINT("  read failed; res=%d\n", len);
		goto done;
	}
	*RxLength = len;

 done:
	if (ret != IFD_SUCCESS)
		*RxLength = 0;

	return ret;
}

RESPONSECODE IFDHControl(DWORD Lun, PUCHAR TxBuffer, DWORD TxLength,
			 PUCHAR RxBuffer, PDWORD RxLength)
{
	/* This function is for data exchange with the reader (not the card).
	 * Not currently used in SIPO scard driver. */

	DPRINT("IFDHControl(Lun=0x%x, TxLength=%u)\n", Lun, TxLength);

	return IFD_NOT_SUPPORTED;
}

RESPONSECODE IFDHICCPresence(DWORD Lun)
{
	int ret;

	/* DPRINT("IFDHICCPresence(Lun=0x%x)\n", Lun); */

	ret = sipo_scard_present();

	if (ret != 1 && sipo_rdr.fd >= 0) {
		(void) sipo_scard_close(&sipo_rdr);
	}

	if (ret == 1)
		return IFD_ICC_PRESENT;
	if (ret == 0)
		return IFD_ICC_NOT_PRESENT;
	return IFD_COMMUNICATION_ERROR;
}
