/* Example code for using SIPO smart card reader */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>


int sipo_scard_present(void)
{
	int fd;
	char buf;
	fd = open("/proc/sipo/scard", 0);
	if (fd < 0) {
		fprintf(stderr, "Could not read /proc/sipo/scard\n");
		return 0;
	}
	if (read(fd, &buf, 1) != 1) {
		fprintf(stderr, "Could not read data from /proc/sipo/scard\n");
		close(fd);
		return 0;
	}
	close(fd);
	if (buf == '1')
		return 1;
	return 0;
}


void dump_atr(char *atr, int len)
{
	int i;
	int pos = 0;
	unsigned char ts, t0, td1 = 0, t = 0;
	int fi = -1, di = -1, ii = -1, n = -1, pi1 = -1, pi2 = -1;

	printf("ATR:\n");
	if (len < 2) {
		printf("  Too short (%i bytes) ATR.\n", len);
		return;
	}
	if (len > 33) {
		printf("  Too long (%i bytes) ATR.\n", len);
		return;
	}

	ts = atr[pos++];
	printf("  TS=%02x (", ts);
	if (ts == 0x3b)
		printf("direct convention)\n");
	else if (ts == 0x3f)
		printf("inverse convention)\n");
	else {
		printf("unknown)!\n");
		return;
	}

	t0 = atr[pos++];

	printf("  T0=%02x\n", t0);

#define BIT(x) (1 << (x))
#define TA_PRESENT BIT(7)
#define TB_PRESENT BIT(6)
#define TC_PRESENT BIT(5)
#define TD_PRESENT BIT(4)

	printf("    Y1:");
	if (t0 & TA_PRESENT)
		printf(" TA1");
	if (t0 & TB_PRESENT)
		printf(" TB1");
	if (t0 & TC_PRESENT)
		printf(" TC1");
	if (t0 & TD_PRESENT)
		printf(" TD1");
	printf("\n");

	printf("    K: %u (number of historical characters)\n", t0 & 0x0f);

	if (t0 & TA_PRESENT) {
		if (pos >= len) goto overflow;
		fi = (atr[pos] & 0xf0) >> 4;
		di = atr[pos] & 0x0f;
		printf("  TA1=%02x (FI=%u DI=%u)\n", atr[pos++], fi, di);
	}

	if (t0 & TB_PRESENT) {
		if (pos >= len) goto overflow;
		ii = (atr[pos] & (BIT(6) | BIT(5))) >> 5;
		pi1 = atr[pos] & (BIT(4) | BIT(3) | BIT(2) | BIT(1) | BIT(0));
		printf("  TB1=%02x (II=%u PI2=%u)\n", atr[pos++], ii, pi1);
	}

	if (t0 & TC_PRESENT) {
		if (pos >= len) goto overflow;
		n = atr[pos];
		printf("  TC1=%02x (N=%u)\n", atr[pos++], n);
	}

	if (t0 & TD_PRESENT) {
		if (pos >= len) goto overflow;
		td1 = atr[pos++];
		printf("  TD1=%02x\n", td1);
		printf("    Y2:");
		if (td1 & TA_PRESENT)
			printf(" TA2");
		if (td1 & TB_PRESENT)
			printf(" TB2");
		if (td1 & TC_PRESENT)
			printf(" TC2");
		if (td1 & TD_PRESENT)
			printf(" TD2");
		printf("\n");

		t = td1 & 0x0f;
		printf("    T: %u (protocol type)\n", t);

		if (td1 & TA_PRESENT) {
			if (pos >= len) goto overflow;
			printf("  TA2=%02x\n", atr[pos++]);
		}

		if (td1 & TB_PRESENT) {
			if (pos >= len) goto overflow;
			pi2 = atr[pos];
			printf("  TB2=%02x (PI2=%u)\n", atr[pos++], pi2);
		}

		if (td1 & TC_PRESENT) {
			if (pos >= len) goto overflow;
			printf("  TC2=%02x\n", atr[pos++]);
		}

		if (td1 & TC_PRESENT) {
			if (pos >= len) goto overflow;
			printf("  TD2=%02x\n", atr[pos++]);
			/* FIX: should really do this recursively.. this TD2
			 * indicates which T?3 are present and so on */
			printf("NOT SUPPORTED.. rest of the ATR parse might "
			       "be invalid\n");
		}
	}


	if ((t0 & 0x0f) > (len - pos)) goto overflow;
	printf("  historical characters:");
	for (i = 0; i < (t0 & 0x0f); i++)
		printf(" %02x", atr[pos + i]);
	printf("\n    (");
	for (i = 0; i < (t0 & 0x0f); i++) {
		if (isprint(atr[pos + i]))
			printf("%c", atr[pos + i]);
		else
			printf("<%02x>", atr[pos + i]);
	}
	printf(")\n");
	pos += t0 & 0x0f;


	if (t != 0) {
		unsigned char csum = 0;

		if (pos >= len) goto overflow;
		printf("  TCK=%02x\n", atr[pos++]);
		for (i = 1; i < pos; i++)
			csum ^= atr[i];
		if (csum != 0)
			printf("    Invalid TCK check character\n");
	}

	if (pos < len)
		printf("  %u extra characters\n", len - pos);

	return;

 overflow:
	printf("ATR overflow\n");
}


void sipo_init_scard(void)
{
	int fd, ret, i;
	unsigned char buf[128];

	fd = open("/dev/misc/siposcard", O_NONBLOCK);
	if (fd < 0) {
		perror("open");
		sleep(5);
	}

	for (;;) {
		ret = read(fd, buf, sizeof(buf));
		if (ret < 0) {
			perror("read");
			close(fd);
			sleep(1);
			return;
		}

		printf("read %i bytes:", ret);
		for (i = 0; i < ret; i++)
			printf(" %02x", buf[i]);
		printf("\n");
		dump_atr(buf, ret);
	}

	close(fd);
}


int main(int argc, char *argv[])
{
	for (;;) {
		int count = 0;
		while (!sipo_scard_present()) {
			if (count == 0)
				printf("Waiting for smart card..\n");
			else if (count > 30) {
				printf("No smart scard found.\n");
				exit(1);
			}
			sleep(1);
		}

		sipo_init_scard();
	}

	return 0;
}


/*

Setec:
read 20 bytes: 3b 9f 94 40 1e 00 67 11 43 46 49 53 45 10 52 66 ff 81 90 00
ATR:
  TS=3b (direct convention)
  T0=9f
    Y1: TA1 TD1
    K: 15 (number of historical characters)
  TA1=94
  TD1=40
    Y2: TB2
    T: 0 (protocol type)
  TB2=1e
  historical characters: 00 67 11 43 46 49 53 45 10 52 66 ff 81 90 00
    (<00>g<11>CFISE<10>Rf<ff><81><90><00>)


Master Card:

read 18 bytes: 3b 6e 00 00 00 62 80 42 57 41 56 41 4e 54 10 81 90 00
ATR:
  TS=3b (direct convention)
  T0=6e
    Y1: TB1 TC1
    K: 14 (number of historical characters)
  TB1=00
  TC1=00
  historical characters: 00 62 80 42 57 41 56 41 4e 54 10 81 90 00
    (<00>b<80>BWAVANT<10><81><90><00>)



TODO:

Could use IOCTL_SCARD_BAUDRATE (0x8800) to setup faster baudrate and/or
clk_freq after parsing ATR. Defaults:
    info->scard_baudrate = 9600;
    info->scard_clk_freq = 372 * 9600;

*/
