/* 
   sipoclock.c -- read and write the SIPO rtc clock via /dev/nvram
   Usage: sipoclock [-r|w]
   Time-stamp: <20010512 12:53:32 janski>

  Copyright (C) 2001-2002 SSH Communications Security Corp, Helsinki, Finland.
  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation. See COPYING for more
  details.
*/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

time_t 
read_hwclock(void)
{
  time_t time;
  int f, r;
  unsigned char buffer[7];
  struct tm tm;

  f = open("/dev/nvram", O_RDONLY);
  if (f < 0)
    exit(1);

  r = read(f, buffer, 7);
  if (r < 0)
    exit(2);

  r = close(f);
  if (r < 0)
      exit(3);

/*  Dallas 1305 stores the time as two BCD digits in 8bit registers 0 to 6.
 *      0       seconds         (00-59)
 *      1       minutes         (00-59)
 *      2       hours           (00-23)
 *      3       day             (01-07)   ignored
 *      4       day of month    (01-31)
 *      5       month           (01-12)
 *      6       year            (00-99)
 * 
 *  Months are off by one compared to system clock.
 *  System clock starts from year 1900.
 */

  tm.tm_sec =  (((buffer[0] >> 4) & 0x0f) * 10) + (buffer[0] & 0x0f);
  tm.tm_min =  (((buffer[1] >> 4) & 0x0f) * 10) + (buffer[1] & 0x0f);
  tm.tm_hour = (((buffer[2] >> 4) & 0x0f) * 10) + (buffer[2] & 0x0f);
  tm.tm_mday = (((buffer[4] >> 4) & 0x0f) * 10) + (buffer[4] & 0x0f);
  tm.tm_mon =  (((buffer[5] >> 4) & 0x0f) * 10) + (buffer[5] & 0x0f) - 1;
  tm.tm_year = (((buffer[6] >> 4) & 0x0f) * 10) + (buffer[6] & 0x0f) + 100;
  
  time = mktime(&tm);

  return time;
}

void 
hwclock_to_sysclock(void)
{
  struct timeval tv = {0, 0};
  struct timezone tz = {0, 0};
  time_t time;
  int r;

  time = read_hwclock();
  tv.tv_sec = time;

  r = settimeofday(&tv, &tz);

  if (r < 0)
    exit(4);

  return;
}

void 
sysclock_to_hwclock(void)
{
  struct timeval tv;
  struct timezone tz = {0, 0};
  int f, r;
  unsigned char buffer[7];
  struct tm *tm_p;

  r = gettimeofday(&tv, &tz);
  if (r < 0)
    exit(5);

  tm_p = gmtime(&tv.tv_sec);
  if (tm_p < 0)
    exit(6);
  
  /* Encode into BCD, two digits per byte.
     Months are off by one, years are modulo 100 */

  buffer[0] = (((tm_p->tm_sec / 10) << 4) & 0xf0) 
    | ((tm_p->tm_sec % 10) & 0x0f);
  buffer[1] = (((tm_p->tm_min / 10) << 4) & 0xf0) 
    | ((tm_p->tm_min % 10) & 0x0f);
  buffer[2] = (((tm_p->tm_hour / 10) << 4) & 0xf0) 
    | ((tm_p->tm_hour % 10) & 0x0f);
  buffer[3] = 1;
  buffer[4] = (((tm_p->tm_mday / 10) << 4) & 0xf0) 
    | ((tm_p->tm_mday % 10) & 0x0f);
  buffer[5] = ((((tm_p->tm_mon + 1) / 10) << 4) & 0xf0) 
    | (((tm_p->tm_mon + 1) % 10) & 0x0f);
  buffer[6] = ((((tm_p->tm_year % 100) / 10) << 4) & 0xf0) 
    | (((tm_p->tm_year % 100) % 10) & 0x0f);

  f = open("/dev/nvram", O_RDWR);
  if (f < 0)
    exit(7);

  r = write(f, buffer, 7);
  if (r < 0)
    exit(8);
  
  r = close(f);
  if (r < 0)
    exit(9);
  
  return;
}

int 
main(int argc, char *argv[])
{
  int c;
  char *options = "rw";
  time_t time;

  time = read_hwclock();

  c = getopt(argc, argv, options);

  switch (c)
    {
    case 'r':
      hwclock_to_sysclock();
      break;
    case 'w':
      sysclock_to_hwclock();
      break;
    default:
      printf("The HW clock time is: %s", ctime(&time));
      break;
    }

  exit(0);
}
